#ifndef MPC_PASSAGEWAY_HPP
#define MPC_PASSAGEWAY_HPP

#include "common/inc/PositionTime.hpp"
#include "common/inc/StateTime.hpp"

namespace mpc
{
  typedef struct WAYPOINT
  {
    XYThS_Config cfg;               /**< configuration (x,y,th) of the waypoint */
    float dist;                     /**< minimal distance to obstacles */
  } WayPoint;


                                    /**************************************************/
                                    /**************** Class Passageway ****************/
                                    /**************************************************/

  /** \class Passageway
   *  \brief free passageway of a reference corridor
   */
  class Passageway
  {
  public : 
    float s;                                           /**< curve abscica */
    float minWidthB;                                   /**< lateral minimal bound */
    float maxWidthB;                                   /**< lateral maximal bound */
    PositionXY maxDistPoint;                           /**< position of the interval maximizing its distance to obstacles */
    float dist;                                        /**< distance to the closest obstacle */
    PositionXY nearObs;                                /**< position of the closest obstacle */

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    Passageway();

    /**
     *  \brief copy constructor
     */
    Passageway(const Passageway& p);

    /**
     *  \brief init constructor
     *  \param s : curve abscica
     *  \param minWidthB : lateral minimal bound
     *  \param maxWidthB : lateral maximal bound
     *  \param maxDistPoint : position of the interval maximizing its distance to obstacles
     *  \param dist : distance to the closest obstacle
     *  \param nearObs : position of the closest obstacle
     */
    Passageway(float s, float minWidthB, float maxWidthB, const PositionXY& maxDistPoint, float dist, const PositionXY& nearObs);

    /**
     *  \brief destructor
     */
    virtual ~Passageway();

    /******************* Accessors ********************/

    /**
     *  \brief return the average lateral shift of the passageway wrt. the reference corridor
     *  \param tildY_origin : origin of the lateral shift on the corridor
     *  \return the average lateral shift of the passageway wrt. the reference corridor
     */
    float latShift(float tildY_origin) const;

    /**************** Display functions ***************/

  public : 
    /**
     *  \brief display the passageway
     */
    void display() const;
  };


                                    /**************************************************/
                                    /*************** Class PassagewayWP ***************/
                                    /**************************************************/

  /** \class PassagewayWP
   *  \brief free passageway of a reference corridor (including the waypoint that must be reached)
   */
  class PassagewayWP : public Passageway
  {
  public : 
    XYThS_Config cfg;                                   /**< configuration (x,y,th) of the waypoint */
    double v;                                           /**< linear velocity at which the waypoint must be reached */
    double t;                                           /**< time at which the waypoint must be reached */
    
    /********** Constructors and destructors **********/
    
  protected : 
    /**
     *  \brief default constructor
     */
    PassagewayWP();
    
  public : 
    /**
     *  \brief copy constructor
     */
    PassagewayWP(const PassagewayWP& p);

    /**
     *  \brief init constructor
     *  \param s : curve abscica
     *  \param minWidthB : lateral minimal bound
     *  \param maxWidthB : lateral maximal bound
     *  \param maxDistPoint : position of the interval maximizing its distance to obstacles
     *  \param dist : distance to the closest obstacle
     *  \param nearObs : position of the closest obstacle
     *  \param cfg : the configuration that must be reach
     *  \param v : linear velocity at which it must be reached
     *  \param t : time at which it must be reached
     */
    PassagewayWP(float s, float minWidthB, float maxWidthB, const PositionXY& maxDistPoint, float dist, const PositionXY& nearObs, XYThS_Config cfg, double v, double t);
      
    /**
     *  \brief destructor
     */
    virtual ~PassagewayWP();

    /**
     *  \brief return a copy of the current PassagewayST instance
     *  \return a copy of the current PassagewayST instance
     */
    PassagewayWP* copy() const;

    /******************* Accessors ********************/

    /**
     *  \brief return the differential drive state-time corresponding to the current PassagewayWP
     *  \return the differential drive state-time corresponding to the current PassagewayWP
     */
    DD2StateT toDD2StateT() const;

    /**
     *  \brief return the waypoint corresponding to the current PassagewayWP
     *  \return the waypoint corresponding to the current PassagewayWP
     */
    WayPoint toWayPoint() const;
  };

};

#endif // MPC_PASSAGEWAY_HPP
