#include "collisionCheck/inc/CollisionCheck.hpp"
using namespace mpc;

#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /************** Class CollisionCheck **************/
                                    /**************************************************/

double CollisionCheck::dt_traj_sampling_ = 0.1;
bool CollisionCheck::unknown_is_obs_ = false;

/**
 *  \brief check for collision between a trajectory and an obstacle map
 *  \return INFINITY if no collisions exists between both trajectory and obstacle map
 *  \return the minimal time of collision if a collision may occur
 */
double CollisionCheck::checkCol(const BoolObsMap& map, const PSTraj& pstraj)
{
  STraj* straj = pstraj.sample(CollisionCheck::dt_traj_sampling_);

  // check the sampling process succeeded
  if(straj == NULL) throw string("CollisionCheck::checkCol : the trajectory sampling failed.");

  double timeCol = CollisionCheck::checkCol(map, *straj);
  delete straj;

  return timeCol;
}

/**
 *  \brief check for collision between a trajectory and an obstacle map
 *  \param map : the obstacle map
 *  \param traj : the followed state trajectory
 *  \return INFINITY if no collisions exists between both trajectory and obstacle map
 *  \return the minimal time of collision if a collision may occur
 */
double CollisionCheck::checkCol(const BoolObsMap& map, const STraj& straj)
{
  int N = straj.size();
  bool colFound = false;
  double colTime = -1;

  int i=0;
  while(i<N && !colFound)
    {
      // get back the current position time of the robot
      PositionXYT post = straj.at(i).posXYT();

      // check if the current position is in collision on the obstacle map
      obs2D_cell_t cell_t = map.getCellType((float) post.x(),(float) post.x());

      if(cell_t == CELL_OBS) { colTime = post.t(); colFound = true; }
      else if(cell_t == CELL_UNKWN && unknown_is_obs_) { colTime = post.t(); colFound = true; }
      else i++;
    }

  if(colFound) // should be always true
    return colTime;
  else
    return INFINITY;
}

/**
 *  \brief check for collision all along a trajectory
 *  \param map : the obstacle map
 *  \param traj : the followed state trajectory
 *  \param vec_col : resulting vector containing the status (collision=true or free=false) of each configuration along the discrete path
 */
void CollisionCheck::checkCol(const BoolObsMap& map, const STraj& straj, vector<bool>& vec_col)
{
  int N = straj.size();

  cout << "{ ";

  for(int i=0; i<N; i++)
    {
      // get back the current position time of the robot
      PositionXYT post = straj.at(i).posXYT();

      // check if the current position is in collision on the obstacle map
      obs2D_cell_t cell_t = map.getCellType((float) post.x(),(float) post.x());

//       if(cell_t == CELL_OBS) cout << "OBS";
//       else if(cell_t == CELL_FREE) cout << "FREE";
//       else if(cell_t == CELL_UNKWN) cout << "UNKWN";
//       else cout << "???";

      if(cell_t == CELL_OBS || (cell_t == CELL_UNKWN && unknown_is_obs_)) 
	{ 
	  vec_col.push_back(true); 
// 	  cout << "(true); "; 
	}
      else 
	{ 
	  vec_col.push_back(false); 
// 	  cout << "(false); "; 
	}
    }

//   cout << "}" << endl;
}
