#ifndef MPC_CONFIG_HPP
#define MPC_CONFIG_HPP

#include "common/inc/Vector.hpp"
#include "common/inc/Position.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /******************* Class Config *****************/
                                    /**************************************************/

  /** \class Config
   *  \brief Configuration of a robotic system
   */
  class Config : public Vector
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    Config();

    /**
     *  \brief copy constructor
     *  \param p : the control to copy
     */
    Config(const Config& c);

    /**
     *  \brief init constructor
     *  \param n : dimension of the configuration
     */
    Config(int n);

    /**
     *  \brief destructor
     */
    virtual ~Config();

    /******************** Accessors *******************/

    /**
     *  \brief return the position of the workspace (in W=XxY)
     *  \return the position of the workspace (in W=XxY)
     */
    virtual PositionXY posXY() const = 0;

  };

};
#endif // MPC_CONFIG_HPP


#ifndef MPC_PMCONFIG_HPP
#define MPC_PMCONFIG_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class PMConfig *****************/
                                    /**************************************************/

  /** \class PMConfig
   *  \brief Config of a point mass system
   */
  class PMConfig : public Config
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    PMConfig();

    /**
     *  \brief copy constructor
     *  \param p : the control to copy
     */
    PMConfig(const PMConfig& c);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     */
    PMConfig(double x,double y);

    /**
     *  \brief destructor
     */
    virtual ~PMConfig();

  protected : 
    /**
     *  \brief init constructor
     *  \param n : dimension of the configuration
     *
     *  Only used by subclasses
     */
    PMConfig(int n);

    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double& x() { return (*this)[0]; }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double& y() { return (*this)[1]; }

    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double x() const { return this->at(0); }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double y() const { return this->at(1); }

    /**
     *  \brief return the position of the workspace (in W=XxY)
     *  \return the position of the workspace (in W=XxY)
     */
    inline virtual PositionXY posXY() const { return PositionXY(this->x(),this->y()); }
  };

};
#endif // MPC_PMCONFIG_HPP


#ifndef MPC_DPMCONFIG_HPP
#define MPC_DPMCONFIG_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class DPMConfig ****************/
                                    /**************************************************/

  /** \class DPMConfig
   *  \brief Config of a directed point mass system
   */
  class DPMConfig : public PMConfig
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    DPMConfig();

    /**
     *  \brief copy constructor
     *  \param p : the configuration to copy
     */
    DPMConfig(const DPMConfig& c);

    /**
     *  \brief init constructor
     *  \param x  : position x
     *  \param y  : position y
     *  \param th : heading th
     */
    DPMConfig(double x,double y,double th);

    /**
     *  \brief destructor
     */
    virtual ~DPMConfig();

  protected : 
    /**
     *  \brief init constructor
     *  \param n : dimension of the configuration
     *
     *  Only used by subclasses
     */
    DPMConfig(int n);

    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double& th() { return (*this)[2]; }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double th() const { return this->at(2); }
  };

};
#endif // MPC_DPMCONFIG_HPP


#ifndef MPC_XYTHS_CONFIG_HPP
#define MPC_XYTHS_CONFIG_HPP

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class XYThS_Config ***************/
                                    /**************************************************/

  /** \class XYThS_Config
   *  \brief Config of a directed planar system
   */
  class XYThS_Config : public DPMConfig
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    XYThS_Config();

    /**
     *  \brief copy constructor
     *  \param p : the configuration to copy
     */
    XYThS_Config(const XYThS_Config& c);

    /**
     *  \brief init constructor
     *  \param x  : position x
     *  \param y  : position y
     *  \param th : heading th
     */
    XYThS_Config(double x,double y,double th,double s);

    /**
     *  \brief destructor
     */
    virtual ~XYThS_Config();

    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double& s() { return (*this)[3]; }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double s() const { return this->at(3); }

    /******************* Operators ********************/

    /**
     *  \brief operator = : copy the input configuration
     *  \param s2 : the configuration to copy
     */
    void operator=(const XYThS_Config& s2);

    /**
     *  \brief operator += : add the input configuration
     *  \param s2 : the configuration to add
     */
    void operator+=(const XYThS_Config& s2);

    /**
     *  \brief operator -= : substract the input configuration
     *  \param s2 : the configuration to substract
     */
    void operator-=(const XYThS_Config& s2);

    /**
     *  \brief operator += : multiply with the input coefficient
     *  \param d : the coefficient to mutilpy with
     */
    void operator*=(double d);

    /**
     *  \brief operator /= : divide with the input coefficient
     *  \param d : the coefficient to divide with
     */
    void operator/=(double d);

    /**
     *  \brief operator + : sumation
     *  \param s2 : the second vector to add
     *  \return the sum of both vectors
     */
    XYThS_Config operator+(const XYThS_Config& s2) const;

    /**
     *  \brief operator + : substraction
     *  \param s2 : the second vector to substract
     *  \return the substraction of both vectors
     */
    XYThS_Config operator-(const XYThS_Config& s2) const;

    /**
     *  \brief operator * : scalar multiplication
     *  \param d : the scalar to multiply with
     *  \return the product of the current vector and the scalar
     */
    XYThS_Config operator*(double d) const;

    /**
     *  \brief operator / : division by a scalar
     *  \param d : the scalar to divide with
     *  \return the division result
     */
    XYThS_Config operator/(double d) const;
  };

};
#endif // MPC_XYTHS_CONFIG_HPP


#ifndef MPC_CLCONFIG_HPP
#define MPC_CLCONFIG_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class CLConfig *****************/
                                    /**************************************************/

  /** \class CLConfig
   *  \brief Config of a car-like system
   */
  class CLConfig : public DPMConfig
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    CLConfig();

    /**
     *  \brief copy constructor
     *  \param p : the configuration to copy
     */
    CLConfig(const CLConfig& c);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param th : heading th
     */
    CLConfig(double x,double y,double th);

    /**
     *  \brief destructor
     */
    virtual ~CLConfig();
  };

};
#endif // MPC_CLCONFIG_HPP


#ifndef MPC_SSCONFIG_HPP
#define MPC_SSCONFIG_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class SSConfig *****************/
                                    /**************************************************/

  /** \class SSConfig
   *  \brief Config of a space ship system
   */
  class SSConfig : public DPMConfig
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    SSConfig();

    /**
     *  \brief copy constructor
     *  \param p : the configuration to copy
     */
    SSConfig(const SSConfig& c);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param th : heading th
     */
    SSConfig(double x,double y,double th);

    /**
     *  \brief destructor
     */
    virtual ~SSConfig();
  };

};
#endif // MPC_SSCONFIG_HPP


#ifndef MPC_DDCONFIG_HPP
#define MPC_DDCONFIG_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class DDConfig *****************/
                                    /**************************************************/

  /** \class DDConfig
   *  \brief Config of a differential drive system
   */
  class DDConfig : public DPMConfig
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    DDConfig();

    /**
     *  \brief copy constructor
     *  \param p : the configuration to copy
     */
    DDConfig(const DDConfig& c);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param th : heading th
     */
    DDConfig(double x,double y,double th);

    /**
     *  \brief destructor
     */
    virtual ~DDConfig();
  };

};
#endif // MPC_DDCONFIG_HPP


#ifndef MPC_DD2CONFIG_HPP
#define MPC_DD2CONFIG_HPP

#define DD2Config DPMConfig

#endif // MPC_DD2CONFIG_HPP


#ifndef MPC_WCCONFIG_HPP
#define MPC_WCCONFIG_HPP

#define WCConfig DPMConfig

#endif // MPC_WCCONFIG_HPP
