#ifndef CONTROL_HPP
#define CONTROL_HPP

#include "common/inc/Vector.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class Control *****************/
                                    /**************************************************/

  /** \class Control
   *  \brief Control applied on a robotic system
   */
  class Control : public Vector
  {
  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    Control();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    Control(const Control& c);                          // copy constructor

    /**
     *  \brief init construtor (init a control of n parameters)
     *  \param n : the number of parameters of the control
     */
    Control(int n);

    /**
     *  \brief destructor
     */
    virtual ~Control();
  };


                                    /**************************************************/
                                    /**************** Class PM3DControl ***************/
                                    /**************************************************/

  /** \class PM3DControl
   *  \brief Control applied on a 3D point mass system
   */
  class PM3DControl : public Control
  {
  public : 
    static double maxNoiseAcc_;                         /*< maximal noise by second over the acceleration */

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    PM3DControl();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    PM3DControl(const PM3DControl& c);

    /**
     *  \brief init constructor
     *  \param acc : the acceleration control
     *  \param ss : the steering speed control
     */
    PM3DControl(double ax,double ay,double az);

    /**
     *  \brief destructor
     */
    virtual ~PM3DControl();

  public : 
    /******************* Accessors ********************/

    /**
     *  \brief getter : return a reference to the acceleration on X control
     *  \return reference to the acceleration on X control
     */
    inline double& ax() {return (*this)[0];}

    /**
     *  \brief getter : return the acceleration on X control
     *  \return the acceleration on X control
     */
    inline double ax() const {return this->at(0);}

    /**
     *  \brief getter : return a reference to the acceleration on Y control
     *  \return reference to the acceleration on Y control
     */
    inline double& ay() {return (*this)[1];}

    /**
     *  \brief getter : return the acceleration on Y control
     *  \return the acceleration on Y control
     */
    inline double ay() const {return this->at(1);}

    /**
     *  \brief getter : return a reference to the acceleration on Z control
     *  \return reference to the acceleration on Z control
     */
    inline double& az() {return (*this)[2];}

    /**
     *  \brief getter : return the acceleration on Z control
     *  \return the acceleration on Z control
     */
    inline double az() const {return this->at(2);}


    /**************** Noise functions *****************/

    /**
     *  \brief return a randomly noised control from the current one
     *  \param t : the application time of the control
     *  \return noised control
     *
     *  The longer the control is applied, the more it will be noised
     */
    PM3DControl noisedCtrl(double t);
  };

                                    /**************************************************/
                                    /***************** Class CLControl ****************/
                                    /**************************************************/

  /** \class CLControl
   *  \brief Control applied on a car-like robotic system
   */
  class CLControl : public Control
  {
  public : 
    static double maxNoiseAcc_;                         /*< maximal noise by second over the acceleration */
    static double maxNoiseSS_;                          /*< maximal noise by second over the steering angle */

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    CLControl();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    CLControl(const CLControl& c);

    /**
     *  \brief init constructor
     *  \param acc : the acceleration control
     *  \param ss : the steering speed control
     */
    CLControl(double acc,double ss);

    /**
     *  \brief destructor
     */
    virtual ~CLControl();

  public : 
    /******************* Accessors ********************/

    /**
     *  \brief getter : return a reference to the acceleration control
     *  \return reference to the acceleration control
     */
    inline double& acc() {return (*this)[0];}

    /**
     *  \brief getter : return the acceleration control
     *  \return the acceleration control
     */
    inline double acc() const {return this->at(0);}

    /**
     *  \brief getter : return a reference to the steering speed control
     *  \return reference to the steering speed control
     */
    inline double& ss() {return (*this)[1];}

    /**
     *  \brief getter : return the steering speed control
     *  \return the steering speed control
     */
    inline double ss() const {return this->at(1);}

    /**************** Noise functions *****************/

    /**
     *  \brief return a randomly noised control from the current one
     *  \param maxNAcc : the maximal noise applied on the acceleration
     *  \param maxNSS : the maximal noise applied on the steering speed
     *  \return noised control
     */
    CLControl noisedCtrl(double maxNAcc,double maxNSS);

    /**
     *  \brief return a randomly noised control from the current one
     *  \param t : the application time of the control
     *  \return noised control
     *
     *  The longer the control is applied, the more it will be noised
     */
    CLControl noisedCtrl(double t);
  };


                                    /**************************************************/
                                    /***************** Class SSControl ****************/
                                    /**************************************************/

  /** \class SSControl
   *  \brief Control applied on a spaceship robotic system
   */
  class SSControl : public Control
  {
  public : 
    static double maxNoiseAcc_;                          /*< maximal noise by second over the acceleration */
    static double maxNoiseAAcc_;                         /*< maximal noise by second over the angular acceleration */

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    SSControl();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    SSControl(const SSControl& c);

    /**
     *  \brief init constructor
     *  \param acc : the acceleration control
     *  \param aacc : the angular acceleration control
     */
    SSControl(double acc,double aacc);

    /**
     *  \brief destructor
     */
    virtual ~SSControl();

  public : 
    /******************* Accessors ********************/

    /**
     *  \brief getter : return a reference to the acceleration control
     *  \return reference to the acceleration control
     */
    inline double& acc() {return (*this)[0];}

    /**
     *  \brief getter : return the acceleration control
     *  \return the acceleration control
     */
    inline double acc() const {return this->at(0);}

    /**
     *  \brief getter : return a reference to the angular acceleration control
     *  \return reference to the angular acceleration control
     */
    inline double& aacc() {return (*this)[1];}

    /**
     *  \brief getter : return the angular acceleration control
     *  \return the angular acceleration control
     */
    inline double aacc() const {return this->at(1);}
  };


                                    /**************************************************/
                                    /***************** Class DDControl ****************/
                                    /**************************************************/

  /** \class DDControl
   *  \brief Control applied on a differential robotic system
   */
  class DDControl : public Control
  {
  public : 
    static double maxNoiseWAcc_;                         /*< maximal noise by second over the wheels acceleration */

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    DDControl();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    DDControl(const DDControl& c);

    /**
     *  \brief init constructor
     *  \param wracc : the right wheel acceleration control
     *  \param wlacc : the left wheel acceleration control
     */
    DDControl(double wracc,double wlacc);

    /**
     *  \brief destructor
     */
    virtual ~DDControl();

  public : 
    /******************* Accessors ********************/

    /**
     *  \brief getter : return a reference to the rigth wheel acceleration control
     *  \return reference to the rigth wheel acceleration control
     */
    inline double& wracc() {return (*this)[0];}

    /**
     *  \brief getter : return the rigth wheel acceleration control
     *  \return the rigth wheel acceleration control
     */
    inline double wracc() const {return this->at(0);}

    /**
     *  \brief getter : return a reference to the rigth wheel acceleration control
     *  \return reference to the rigth wheel acceleration control
     */
    inline double& wlacc() {return (*this)[1];}

    /**
     *  \brief getter : return the rigth wheel acceleration control
     *  \return the rigth wheel acceleration control
     */
    inline double wlacc() const {return this->at(1);}

    /**************** Noise functions *****************/

    /**
     *  \brief return a randomly noised control from the current one
     *  \param t : the application time of the control
     *  \return noised control
     *
     *  The longer the control is applied, the more it will be noised
     */
    DDControl noisedCtrl(double t);
  };


                                    /**************************************************/
                                    /***************** Class DD2Control ****************/
                                    /**************************************************/

  /** \class DD2Control
   *  \brief Control applied on a wheelchair robotic system
   */
  class DD2Control : public Control
  {
  public : 
    static double maxNoiseV_;                         /*< maximal noise by second over the linear acceleration */
    static double maxNoiseAV_;                        /*< maximal noise by second over the angular acceleration */

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    DD2Control();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    DD2Control(const DD2Control& c);

    /**
     *  \brief init constructor
     *  \param v : the linear velocity control
     *  \param av : the angular velocity control
     */
    DD2Control(double v,double av);

    /**
     *  \brief destructor
     */
    virtual ~DD2Control();

  public : 
    /******************* Accessors ********************/

    /**
     *  \brief getter : return a reference to the linear speed control
     *  \return reference to the linear speed control
     */
    inline double& v() {return (*this)[0];}

    /**
     *  \brief getter : return the linear speed control
     *  \return the linear speed control
     */
    inline double v() const {return this->at(0);}

    /**
     *  \brief getter : return a reference to the angular speed control
     *  \return reference to the angular speed control
     */
    inline double& av() {return (*this)[1];}

    /**
     *  \brief getter : return the angular speed control
     *  \return the angular speed control
     */
    inline double av() const {return this->at(1);}

    /**
     *  \brief getter : return a reference to the angular speed control
     *  \return reference to the angular speed control
     */
    inline double& w() {return (*this)[1];}

    /**
     *  \brief getter : return the angular speed control
     *  \return the angular speed control
     */
    inline double w() const {return this->at(1);}

    /**************** Noise functions *****************/

    /**
     *  \brief return a randomly noised control from the current one
     *  \param t : the application time of the control
     *  \return noised control
     *
     *  The longer the control is applied, the more it will be noised
     */
    DD2Control noisedCtrl(double t);
  };

#define WCControl DD2Control
#define ManaControl DD2Control


                                    /**************************************************/
                                    /*************** Class DD2AccControl **************/
                                    /**************************************************/

  /** \class DD2AccControl
   *  \brief Acceleration control applied on a Differential drive robotic system
   */
  class DD2AccControl : public Control
  {
  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    DD2AccControl();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    DD2AccControl(const DD2AccControl& c);

    /**
     *  \brief init constructor
     *  \param acc : the linear acceleration control
     *  \param aacc : the angular acceleration control
     */
    DD2AccControl(double acc,double aacc);

    /**
     *  \brief destructor
     */
    virtual ~DD2AccControl();

    /******************* Accessors ********************/

  public : 

    /**
     *  \brief getter : return a reference to the linear acceleration control
     *  \return reference to the linear acceleration control
     */
    inline double& acc() {return (*this)[0];}

    /**
     *  \brief getter : return the linear acceleration control
     *  \return the linear acceleration control
     */
    inline double acc() const {return this->at(0);}

    /**
     *  \brief getter : return a reference to the angular acceleration control
     *  \return reference to the angular acceleration control
     */
    inline double& aacc() {return (*this)[1];}

    /**
     *  \brief getter : return the angular acceleration control
     *  \return the angular acceleration control
     */
    inline double aacc() const {return this->at(1);}
  };

#define ManaAccControl DD2AccControl


                                    /**************************************************/
                                    /**************** Class AASControl ****************/
                                    /**************************************************/

  /** \class AASControl
   *  \brief Tricky control in acceleration and angular speed
   */
  class AASControl : public Control
  {
  public : 
    //   static double maxNoiseV_;                         /*< maximal noise by second over the linear acceleration */
    //   static double maxNoiseAV_;                        /*< maximal noise by second over the angular acceleration */

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    AASControl();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    AASControl(const AASControl& c);

    /**
     *  \brief init constructor
     *  \param acc : the linear acceleration control
     *  \param av : the angular speed control
     */
    AASControl(double acc,double as);

    /**
     *  \brief destructor
     */
    virtual ~AASControl();

  public : 
    /******************* Accessors ********************/

    /**
     *  \brief getter : return a reference to the linear acceleration control
     *  \return reference to the linear acceleration control
     */
    inline double& acc() {return (*this)[0];}

    /**
     *  \brief getter : return the linear acceleration control
     *  \return the linear acceleration control
     */
    inline double acc() const {return this->at(0);}

    /**
     *  \brief getter : return a reference to the angular speed control
     *  \return reference to the angular speed control
     */
    inline double& as() {return (*this)[1];}

    /**
     *  \brief getter : return the angular speed control
     *  \return the angular speed control
     */
    inline double as() const {return this->at(1);}

    //   /**************** Noise functions *****************/

    //   /**
    //    *  \brief return a randomly noised control from the current one
    //    *  \param t : the application time of the control
    //    *  \return noised control
    //    *
    //    *  The longer the control is applied, the more it will be noised
    //    */
    //   WCControl noisedCtrl(double t);
  };


                                    /**************************************************/
                                    /***************** Class CCControl ****************/
                                    /**************************************************/

  /** \class CCControl
   *  \brief Control applied on a cycab robotic system
   */
  class CCControl : public Control
  {
  public : 
    static double maxNoiseV_;                          /*< maximal noise by second over the velocity */
    static double maxNoisePh_;                         /*< maximal noise by second over the steering angle */

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    CCControl();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    CCControl(const CCControl& c);

    /**
     *  \brief init constructor
     *  \param v : the velocity control
     *  \param ph : the steering angle control
     */
    CCControl(double v,double ph);

    /**
     *  \brief destructor
     */
    virtual ~CCControl();

  public : 
    /******************* Accessors ********************/

    /**
     *  \brief getter : return a reference to the velocity control
     *  \return reference to the velocity control
     */
    inline double& v() {return (*this)[0];}

    /**
     *  \brief getter : return the velocity control
     *  \return the velocity control
     */
    inline double v() const {return this->at(0);}

    /**
     *  \brief getter : return a reference to the steering speed control
     *  \return reference to the steering speed control
     */
    inline double& ph() {return (*this)[1];}

    /**
     *  \brief getter : return the steering speed control
     *  \return the steering speed control
     */
    inline double ph() const {return this->at(1);}

    /**************** Noise functions *****************/

    /**
     *  \brief return a randomly noised control from the current one
     *  \param maxNV : the maximal noise applied on the velocity
     *  \param maxNPh : the maximal noise applied on the steering angle
     *  \return noised control
     */
    CCControl noisedCtrl(double maxNV,double maxNPh);

    /**
     *  \brief return a randomly noised control from the current one
     *  \param t : the application time of the control
     *  \return noised control
     *
     *  The longer the control is applied, the more it will be noised
     */
    CCControl noisedCtrl(double t);
  };

};
#endif // CONTROL_HPP
