#ifndef MPC_CONTROLTIME_HPP
#define MPC_CONTROLTIME_HPP

#include "common/inc/Control.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class ControlTime ***************/
                                    /**************************************************/

  /** \class ControlTime
   *  \brief Control-time applied on a robotic system
   */
  class ControlTime : public Control
  {
  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    ControlTime();

    /**
     *  \brief copy constructor
     *  \param c : the control-time to copy
     */
    ControlTime(const ControlTime& c);

    /**
     *  \brief init construtor (init a control of n parameters)
     *  \param n : the number of parameters of the control
     */
    ControlTime(int n);

    /**
     *  \brief destructor
     */
    virtual ~ControlTime();

    /******************** Accessors *******************/
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    virtual double& t() = 0;

    /**
     *  \brief getter : time T
     *  \return time T
     */
    virtual double t() const = 0;
  };


                                    /**************************************************/
                                    /*************** Class PM3DControlT ***************/
                                    /**************************************************/

  /** \class PM3DControlT
   *  \brief Control-time applied on a 3D point mass system
   */
  class PM3DControlT : public ControlTime, public PM3DControl
  {
  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    PM3DControlT();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    PM3DControlT(const PM3DControlT& c);

    /**
     *  \brief init constructor
     *  \param ax : the acceleration on X control
     *  \param ay : the acceleration on Y control
     *  \param az : the acceleration on Z control
     *  \param t : the time of application of the control
     */
    PM3DControlT(double ax,double ay,double az,double t);

    /**
     *  \brief destructor
     */
    virtual ~PM3DControlT();

    /******************** Accessors *******************/
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double& t() { return (*((ControlTime*)this))[3]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double t() const { return this->ControlTime::at(3); }

    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;
  };


                                    /**************************************************/
                                    /**************** Class CLControlT ****************/
                                    /**************************************************/

  /** \class CLControlT
   *  \brief Control-time applied on a car-like robotic system
   */
  class CLControlT : public ControlTime, public CLControl
  {
  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    CLControlT();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    CLControlT(const CLControlT& c);

    /**
     *  \brief init constructor
     *  \param acc : the acceleration control
     *  \param ss : the steering speed control
     *  \param t : the time of application of the control
     */
    CLControlT(double acc,double ss,double t);

    /**
     *  \brief destructor
     */
    virtual ~CLControlT();

    /******************** Accessors *******************/
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double& t() { return (*((ControlTime*)this))[2]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double t() const { return this->ControlTime::at(2); }

    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;
  };


                                    /**************************************************/
                                    /**************** Class SSControlT ****************/
                                    /**************************************************/

  /** \class SSControlT
   *  \brief Control-time applied on a spaceship robotic system
   */
  class SSControlT : public ControlTime, public SSControl
  {
  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    SSControlT();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    SSControlT(const SSControlT& c);

    /**
     *  \brief init constructor
     *  \param accx : the acceleration control on X
     *  \param accy : the acceleration control on Y
     *  \param aacc : the angular acceleration control
     *  \param t : the time of application of the control
     */
    SSControlT(double acc,double aacc,double t);

    /**
     *  \brief destructor
     */
    virtual ~SSControlT();

    /******************** Accessors *******************/
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double& t() { return (*((ControlTime*)this))[2]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double t() const { return this->ControlTime::at(2); }

    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;
  };

                                    /**************************************************/
                                    /**************** Class DDControlT ****************/
                                    /**************************************************/

  /** \class DDControlT
   *  \brief Control-time applied on a differential drive robotic system
   */
  class DDControlT : public ControlTime, public DDControl
  {
  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    DDControlT();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    DDControlT(const DDControlT& c);

    /**
     *  \brief init constructor
     *  \param wracc : the right wheel acceleration control
     *  \param wlacc : the left wheel acceleration control
     *  \param t : the time of application of the control
     */
    DDControlT(double acc,double ss,double t);

    /**
     *  \brief destructor
     */
    virtual ~DDControlT();

    /******************** Accessors *******************/
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double& t() { return (*((ControlTime*)this))[2]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double t() const { return this->ControlTime::at(2); }

    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;
  };


                                    /**************************************************/
                                    /**************** Class DD2ControlT ****************/
                                    /**************************************************/

  /** \class DD2ControlT
   *  \brief Control-time applied on a wheelchair robotic system
   */
  class DD2ControlT : public ControlTime, public DD2Control
  {
  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    DD2ControlT();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    DD2ControlT(const DD2ControlT& c);

    /**
     *  \brief init constructor
     *  \param v : the linear velocity control
     *  \param av : the angular velocity control
     *  \param t : the time of application of the control
     */
    DD2ControlT(double v,double av,double t);

    /**
     *  \brief destructor
     */
    virtual ~DD2ControlT();

    /******************** Accessors *******************/
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double& t() { return (*((ControlTime*)this))[2]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double t() const { return this->ControlTime::at(2); }


    /******************* Operators ********************/

    /**
     *  \brief operator = : copy the input state
     *  \param s2 : the state to copy
     */
    void operator=(const DD2ControlT& c2);

    /**
     *  \brief operator + : sumation
     *  \param s2 : the second vector to add
     *  \return the sum of both vectors
     */
    DD2ControlT operator+(const DD2ControlT& c2) const;

    /**
     *  \brief operator + : substraction
     *  \param s2 : the second vector to substract
     *  \return the substraction of both vectors
     */
    DD2ControlT operator-(const DD2ControlT& c2) const;

    /**
     *  \brief operator * : scalar multiplication
     *  \param d : the scalar to multiply with
     *  \return the product of the current vector and the scalar
     */
    DD2ControlT operator*(double d) const;


    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;
  };

#define WCControlT DD2ControlT
#define ManaControlT DD2ControlT


                                    /**************************************************/
                                    /************* Class DD2AccControlT **************/
                                    /**************************************************/

  /** \class DD2AccControlT
   *  \brief Acceleration control-time applied on a Differential drive robotic system
   */
  class DD2AccControlT : public ControlTime, public DD2AccControl
  {
  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    DD2AccControlT();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    DD2AccControlT(const DD2AccControlT& c);

    /**
     *  \brief init constructor
     *  \param acc : the linear acceleration control
     *  \param aacc : the angular acceleration control
     *  \param t : the time of application of the control
     */
    DD2AccControlT(double acc,double aacc,double t);

    /**
     *  \brief destructor
     */
    virtual ~DD2AccControlT();

    /******************** Accessors *******************/
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double& t() { return (*((ControlTime*)this))[2]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double t() const { return this->ControlTime::at(2); }

    /******************* Operators ********************/

    /**
     *  \brief operator = : copy the input state
     *  \param s2 : the state to copy
     */
    void operator=(const DD2AccControlT& c2);

    /**
     *  \brief operator + : sumation
     *  \param s2 : the second vector to add
     *  \return the sum of both vectors
     */
    DD2AccControlT operator+(const DD2AccControlT& c2) const;

    /**
     *  \brief operator + : substraction
     *  \param s2 : the second vector to substract
     *  \return the substraction of both vectors
     */
    DD2AccControlT operator-(const DD2AccControlT& c2) const;

    /**
     *  \brief operator * : scalar multiplication
     *  \param d : the scalar to multiply with
     *  \return the product of the current vector and the scalar
     */
    DD2AccControlT operator*(double d) const;


    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;
  };

#define ManaAccControlT DD2AccControlT


                                    /**************************************************/
                                    /**************** Class AASControlT ***************/
                                    /**************************************************/

  /** \class AASControlT
   *  \brief Tricky control-time in acceleration and angular speed
   */
  class AASControlT : public ControlTime, public AASControl
  {
  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    AASControlT();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    AASControlT(const AASControlT& c);

    /**
     *  \brief init constructor
     *  \param acc : the linear acceleration control
     *  \param as : the angular speed control
     *  \param t : the time of application of the control
     */
    AASControlT(double acc,double as,double t);

    /**
     *  \brief destructor
     */
    virtual ~AASControlT();

    /******************** Accessors *******************/
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double& t() { return (*((ControlTime*)this))[2]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double t() const { return this->ControlTime::at(2); }

    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;
  };


                                    /**************************************************/
                                    /**************** Class CCControlT ****************/
                                    /**************************************************/

  /** \class CCControlT
   *  \brief Control-time applied on a cycab robotic system
   */
  class CCControlT : public ControlTime, public CCControl
  {
  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    CCControlT();

    /**
     *  \brief copy constructor
     *  \param c : the control to copy
     */
    CCControlT(const CCControlT& c);

    /**
     *  \brief init constructor
     *  \param v : the velocity control
     *  \param ss : the steering angle control
     *  \param t : the time of application of the control
     */
    CCControlT(double v,double ph,double t);

    /**
     *  \brief destructor
     */
    virtual ~CCControlT();

    /******************** Accessors *******************/
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double& t() { return (*((ControlTime*)this))[2]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline virtual double t() const { return this->ControlTime::at(2); }

    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;
  };

};
#endif // MPC_CONTROLTIME_HPP
