#ifndef MPC_STATETIME_HPP
#define MPC_STATETIME_HPP

#include "common/inc/State.hpp"
#include "common/inc/PositionTime.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class StateTime ****************/
                                    /**************************************************/

  /** \class StateTime
   *  \brief State of a robotic system increased by the time dimension (S*T)
   */
  class StateTime : public State
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    StateTime();

    /**
     *  \brief copy constructor
     *  \param s : the state to copy
     */
    StateTime(const StateTime& s);

    /**
     *  \brief init constructor
     *  \param n : dimension of the state
     */
    StateTime(int n);

    /**
     *  \brief destructor
     */
    virtual ~StateTime();

    /**
     *  \brief return a copy of the current state-time
     *  \return a copy of the current state-time
     */
    virtual StateTime* copy() const = 0;


    /******************** Accessors *******************/
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    virtual double& t() = 0;

    /**
     *  \brief getter : time T
     *  \return time T
     */
    virtual double t() const = 0;

    /**
     *  \brief return the position-time of the space (in XxYxT)
     *  \return the position-time of the space (in XxYxT)
     */
    virtual PositionXYT posXYT() const = 0;
  
//     /**
//      *  \brief accessor of the i-th coordinate
//      *  \param i : index of the coordinate to return
//      *  \return the i-th coordinate of the vector
//      */
//     virtual double& operator[](int i) = 0;

//     /**
//      *  \brief accessor of the i-th coordinate
//      *  \param i : index of the coordinate to return
//      *  \return the i-th coordinate of the vector
//      */
//     virtual double at(int i) const = 0;

//     /**
//      *  \brief size of the vector
//      *  \return the size of the vector
//      */
//     virtual int size() const = 0;

    /*************** Display functions ****************/

    /**
     *  \brief display the state-time in an openGL viewer
     */
    virtual void toGL() const;

    /**
     *  \brief display the state-time in an openGL viewer
     *  \param R : the red color
     *  \param G : the green color
     *  \param B : the blue color
     *  \param size : the size of the object displayed   
     */
    virtual void toGL(double R,double G,double B,double size) const;
  
//     /**
//      *  \brief display the state's coordinates
//      */
//     virtual void display() const = 0;
  };

};
#endif // MPC_STATETIME_HPP


#ifndef MPC_PMSTATET_HPP
#define MPC_PMSTATET_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class PMStateT *****************/
                                    /**************************************************/

  /** \class PMStateT
   *  \brief State-time of a point mass system
   */
  class PMStateT : public StateTime // , public PMState
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    PMStateT();

    /**
     *  \brief copy constructor
     *  \param p : the state-time to copy
     */
    PMStateT(const PMStateT& s);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param sx : speed sx
     *  \param sy : speed sy
     *  \param t : time t
     */
    PMStateT(double x,double y,double sx,double sy,double t);

    /**
     *  \brief destructor
     */
    virtual ~PMStateT();

    /**
     *  \brief return a state from the current state-time
     *  \return a state from the current state-time
     */
    inline PMState toState() const { return PMState(x(),y(),sx(),sy()); }

    /**
     *  \brief return a copy of the current state-time
     *  \return a copy of the current state-time
     */
    inline virtual PMStateT* copy() const { return new PMStateT(*this); }


    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : position X \return position X
     */
    inline double& x() { return (*this)[0]; }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double& y() { return (*this)[1]; }

    /**
     *  \brief getter : speed SX
     *  \return speed SX
     */
    inline double& sx() { return (*this)[2]; }

    /**
     *  \brief getter : speed SY
     *  \return speed SY
     */
    inline double& sy() { return (*this)[3]; }

    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double x() const { return this->at(0); }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double y() const { return this->at(1); }

    /**
     *  \brief getter : speed SX
     *  \return speed SX
     */
    inline double sx() const { return this->at(2); }

    /**
     *  \brief getter : speed SY
     *  \return speed SY
     */
    inline double sy() const { return this->at(3); }

    /**
     *  \brief retrieve a configuration from the current state
     *  \return the corresponding configuration
     */
    inline virtual PMConfig cfg() const { return PMConfig(x(),y()); }
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline double& t() { return (*this)[4]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline double t() const { return this->at(4); }

    /**
     *  \brief return the position-time of the space (in XxYxT)
     *  \return the position-time of the space (in XxYxT)
     */
    inline virtual PositionXYT posXYT() const { return PositionXYT(this->x(),this->y(),this->t()); }

    /**
     *  \brief return the position of the space (in W = R^2 x S)
     *  \return the position-time of the space (in W = R^2 x S)
     */
    inline virtual PositionXYTh posXYTh() const { return PositionXYTh(x(),y(),0.0); }

//     /**
//      *  \brief accessor of the i-th coordinate
//      *  \param i : index of the coordinate to return
//      *  \return the i-th coordinate of the vector
//      */
//     inline virtual double& operator[](int i) { return coord_.at(i); }

//     /**
//      *  \brief accessor of the i-th coordinate
//      *  \param i : index of the coordinate to return
//      *  \return the i-th coordinate of the vector
//      */
//     inline virtual double at(int i) const { return this->at(i); }

//     /**
//      *  \brief size of the vector
//      *  \return the size of the vector
//      */
//     inline virtual int size() const { return this->PMState::size(); }

    /******************* Operators ********************/

    /**
     *  \brief operator = : copy the input state
     *  \param s2 : the state to copy
     */
    void operator=(const PMStateT& s2);

    /**
     *  \brief operator += : add the input state
     *  \param s2 : the state to add
     */
    void operator+=(const PMStateT& s2);

    /**
     *  \brief operator -= : substract the input state
     *  \param s2 : the state to substract
     */
    void operator-=(const PMStateT& s2);

    /**
     *  \brief operator += : multiply with the input coefficient
     *  \param d : the coefficient to mutilpy with
     */
    void operator*=(double d);

    /**
     *  \brief operator /= : divide with the input coefficient
     *  \param d : the coefficient to divide with
     */
    void operator/=(double d);

    /**
     *  \brief operator + : sumation
     *  \param s2 : the second vector to add
     *  \return the sum of both vectors
     */
    PMStateT operator+(const PMStateT& s2) const;

    /**
     *  \brief operator + : substraction
     *  \param s2 : the second vector to substract
     *  \return the substraction of both vectors
     */
    PMStateT operator-(const PMStateT& s2) const;

    /**
     *  \brief operator * : scalar multiplication
     *  \param d : the scalar to multiply with
     *  \return the product of the current vector and the scalar
     */
    PMStateT operator*(double d) const;

    /**
     *  \brief operator / : division by a scalar
     *  \param d : the scalar to divide with
     *  \return the division result
     */
    PMStateT operator/(double d) const;

    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;
  };

};
#endif // MPC_PMSTATET_HPP


#ifndef MPC_PM3DSTATET_HPP
#define MPC_PM3DSTATET_HPP

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class PM3DStateT ****************/
                                    /**************************************************/

  /** \class PM3DStateT
   *  \brief State-time of a 3D point mass system
   */
  class PM3DStateT : public StateTime // , public PM3DState
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    PM3DStateT();

    /**
     *  \brief copy constructor
     *  \param p : the state-time to copy
     */
    PM3DStateT(const PM3DStateT& s);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param z : position y
     *  \param sx : speed sx
     *  \param sy : speed sy
     *  \param sz : speed sz
     *  \param t : time t
     */
    PM3DStateT(double x,double y,double z,double sx,double sy,double sz,double t);

    /**
     *  \brief destructor
     */
    virtual ~PM3DStateT();

    /**
     *  \brief return a state from the current state-time
     *  \return a state from the current state-time
     */
    inline PM3DState toState() const { return PM3DState(x(),y(),z(),sx(),sy(),sz()); }

    /**
     *  \brief return a copy of the current state-time
     *  \return a copy of the current state-time
     */
    inline virtual PM3DStateT* copy() const { return new PM3DStateT(*this); }


    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : position X \return position X
     */
    inline double& x() { return (*this)[0]; }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double& y() { return (*this)[1]; }

    /**
     *  \brief getter : position Z
     *  \return position Z
     */
    inline double& z() { return (*this)[2]; }

    /**
     *  \brief getter : speed SX
     *  \return speed SX
     */
    inline double& sx() { return (*this)[3]; }

    /**
     *  \brief getter : speed SY
     *  \return speed SY
     */
    inline double& sy() { return (*this)[4]; }

    /**
     *  \brief getter : speed SZ
     *  \return speed SZ
     */
    inline double& sz() { return (*this)[5]; }

    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double x() const { return this->at(0); }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double y() const { return this->at(1); }

    /**
     *  \brief getter : position Z
     *  \return position Z
     */
    inline double z() const { return this->at(2); }

    /**
     *  \brief getter : speed SX
     *  \return speed SX
     */
    inline double sx() const { return this->at(3); }

    /**
     *  \brief getter : speed SY
     *  \return speed SY
     */
    inline double sy() const { return this->at(4); }

    /**
     *  \brief getter : speed SZ
     *  \return speed SZ
     */
    inline double sz() const { return this->at(5); }

//     /**
//      *  \brief retrieve a configuration from the current state
//      *  \return the corresponding configuration
//      */
//     inline virtual PMConfig cfg() const { return PMConfig(x(),y()); }
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline double& t() { return (*this)[6]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline double t() const { return this->at(6); }

    /**
     *  \brief return the position of the space (in W = R^2 x S)
     *  \return the position-time of the space (in W = R^2 x S)
     */
    inline virtual PositionXYTh posXYTh() const { cerr << "PM3DStateT::posXYTh cannot be implemented!" << endl; exit(1); }

    /**
     *  \brief return the position-time of the space (in R^3xT)
     *  \return the position-time of the space (in R^3xT)
     */
    virtual PositionXYT posXYThT() const { cerr << "PM3DStateT::posXYThT cannot be implemented!" << endl; exit(1); }

//     /**
//      *  \brief return the position-time of the space (in R^3xT)
//      *  \return the position-time of the space (in R^3xT)
//      */
//     inline virtual PositionXYZT posXYZT() const { return PositionXYZT(this->x(),this->y(),this->z(),this->t()); }

    /**
     *  \brief return the position of the space (in W = R^3)
     *  \return the position-time of the space (in W = R^3)
     */
    inline virtual PositionXYT posXYT() const { return PositionXYT(x(),y(),t()); }

    /******************* Operators ********************/

    /**
     *  \brief operator = : copy the input state
     *  \param s2 : the state to copy
     */
    void operator=(const PM3DStateT& s2);

    /**
     *  \brief operator += : add the input state
     *  \param s2 : the state to add
     */
    void operator+=(const PM3DStateT& s2);

    /**
     *  \brief operator -= : substract the input state
     *  \param s2 : the state to substract
     */
    void operator-=(const PM3DStateT& s2);

    /**
     *  \brief operator += : multiply with the input coefficient
     *  \param d : the coefficient to mutilpy with
     */
    void operator*=(double d);

    /**
     *  \brief operator /= : divide with the input coefficient
     *  \param d : the coefficient to divide with
     */
    void operator/=(double d);

    /**
     *  \brief operator + : sumation
     *  \param s2 : the second vector to add
     *  \return the sum of both vectors
     */
    PM3DStateT operator+(const PM3DStateT& s2) const;

    /**
     *  \brief operator + : substraction
     *  \param s2 : the second vector to substract
     *  \return the substraction of both vectors
     */
    PM3DStateT operator-(const PM3DStateT& s2) const;

    /**
     *  \brief operator * : scalar multiplication
     *  \param d : the scalar to multiply with
     *  \return the product of the current vector and the scalar
     */
    PM3DStateT operator*(double d) const;

    /**
     *  \brief operator / : division by a scalar
     *  \param d : the scalar to divide with
     *  \return the division result
     */
    PM3DStateT operator/(double d) const;

    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;
  };

};
#endif // MPC_PM3DSTATET_HPP


#ifndef MPC_CLSTATET_HPP
#define MPC_CLSTATET_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class CLStateT *****************/
                                    /**************************************************/

  /** \class CLStateT
   *  \brief State-time of a car-like system
   */
  class CLStateT : public StateTime // , public CLState
  {
  protected : 

  public : 
    static double maxNoiseXY_;                          /*< maximal noise by second over the position (XY) */
    static double maxNoiseTH_;                          /*< maximal noise by second over the heading (TH) */
    static double maxNoiseS_;                           /*< maximal noise by second over the linear velocity (S) */
    static double maxNoisePH_;                          /*< maximal noise by second over the steering angle (PH) */

    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    CLStateT();

    /**
     *  \brief copy constructor
     *  \param p : the state-time to copy
     */
    CLStateT(const CLStateT& s);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param th : heading th
     *  \param ph : steering angle ph
     *  \param s : linear speed s
     *  \param t : time t
     */
    CLStateT(double x,double y,double th,double ph,double s,double t);

    /**
     *  \brief destructor
     */
    virtual ~CLStateT();

    /**
     *  \brief return a state from the current state-time
     *  \return a state from the current state-time
     */
    inline CLState toState() const { return CLState(x(),y(),th(),ph(),s()); }

    /**
     *  \brief return a copy of the current state-time
     *  \return a copy of the current state-time
     */
    inline virtual CLStateT* copy() const { return new CLStateT(*this); }


    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double& x() { return (*this)[0]; }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double& y() { return (*this)[1]; }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double& th() { return (*this)[2]; }

    /**
     *  \brief getter : steering angle PH
     *  \return heading PH
     */
    inline double& ph() { return (*this)[3]; }

    /**
     *  \brief getter : linear speed S
     *  \return linear speed S
     */
    inline double& s() { return (*this)[4]; }

    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double x() const { return this->at(0); }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double y() const { return this->at(1); }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double th() const { return this->at(2); }

    /**
     *  \brief getter : steering angle PH
     *  \return heading PH
     */
    inline double ph() const { return this->at(3); }

    /**
     *  \brief getter : linear speed S
     *  \return linear speed S
     */
    inline double s() const { return this->at(4); }

    /**
     *  \brief retrieve a configuration from the current state
     *  \return the corresponding configuration
     */
    inline virtual CLConfig cfg() const { return CLConfig(x(),y(),th()); }
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline double& t() { return (*this)[5]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline double t() const { return this->at(5); }

//     /**
//      *  \brief accessor of the i-th coordinate
//      *  \param i : index of the coordinate to return
//      *  \return the i-th coordinate of the vector
//      */
//     inline virtual double at(int i) const { return this->CLState::coord_.at(i); }
    
    /**
     *  \brief return the position-time of the space (in XxYxT)
     *  \return the position-time of the space (in XxYxT)
     */
    inline virtual PositionXYT posXYT() const { return PositionXYT(this->x(),this->y(),this->t()); }

    /**
     *  \brief return the position of the space (in W = R^2 x S)
     *  \return the position-time of the space (in W = R^2 x S)
     */
    inline virtual PositionXYTh posXYTh() const { return PositionXYTh(x(),y(),th()); }

//     /**
//      *  \brief accessor of the i-th coordinate
//      *  \param i : index of the coordinate to return
//      *  \return the i-th coordinate of the vector
//      */
//     inline virtual double& operator[](int i) { return coord_.at(i); }

//     /**
//      *  \brief accessor of the i-th coordinate
//      *  \param i : index of the coordinate to return
//      *  \return the i-th coordinate of the vector
//      */
//     inline virtual double at(int i) const { return this->CLState::at(i); }

//     /**
//      *  \brief size of the vector
//      *  \return the size of the vector
//      */
//     inline virtual int size() const { return this->CLState::size(); }

    /******************* Operators ********************/

    /**
     *  \brief operator = : copy the input state
     *  \param s2 : the state to copy
     */
    void operator=(const CLStateT& s2);

    /**
     *  \brief operator += : add the input state
     *  \param s2 : the state to add
     */
    void operator+=(const CLStateT& s2);

    /**
     *  \brief operator -= : substract the input state
     *  \param s2 : the state to substract
     */
    void operator-=(const CLStateT& s2);

    /**
     *  \brief operator += : multiply with the input coefficient
     *  \param d : the coefficient to mutilpy with
     */
    void operator*=(double d);

    /**
     *  \brief operator /= : divide with the input coefficient
     *  \param d : the coefficient to divide with
     */
    void operator/=(double d);

    /**
     *  \brief operator + : sumation
     *  \param s2 : the second vector to add
     *  \return the sum of both vectors
     */
    CLStateT operator+(const CLStateT& s2) const;

    /**
     *  \brief operator + : substraction
     *  \param s2 : the second vector to substract
     *  \return the substraction of both vectors
     */
    CLStateT operator-(const CLStateT& s2) const;

    /**
     *  \brief operator * : scalar multiplication
     *  \param d : the scalar to multiply with
     *  \return the product of the current vector and the scalar
     */
    CLStateT operator*(double d) const;

    /**
     *  \brief operator / : division by a scalar
     *  \param d : the scalar to divide with
     *  \return the division result
     */
    CLStateT operator/(double d) const;

    /**************** Noise functions *****************/

    /**
     *  \brief return a randomly noised state from the current one
     *  \param t : the elapsed time from the previous known state
     *  \return noised state
     */
    CLStateT noisedState(double t);

    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;

    /*************** Utility functions ****************/

    /**
     *  \brief compute the relative coordinates between both state-times
     *  \param initST : the initial state-time
     *  \param finalST : the final state-time
     *
     *  initST is center in position and heading (0,0,0)
     *  finalST is adjust in the same frame than initST
     */
    static void relativeStateT(CLStateT& initST,CLStateT& finalST);
  };

};
#endif // MPC_CLSTATET_HPP


#ifndef MPC_SSSTATET_HPP
#define MPC_SSSTATET_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class SSStateT *****************/
                                    /**************************************************/

  /** \class SSStateT
   *  \brief State-time of a space ship system
   */
  class SSStateT : public StateTime // , public SSState
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    SSStateT();

    /**
     *  \brief copy constructor
     *  \param p : the state-time to copy
     */
    SSStateT(const SSStateT& s);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param th : heading th
     *  \param sx : speed sx
     *  \param sy : speed sy
     *  \param om : angular velocity om
     *  \param t : time t
     */
    SSStateT(double x,double y,double th,double sx,double sy,double om,double t);

    /**
     *  \brief destructor
     */
    virtual ~SSStateT();

    /**
     *  \brief return a state from the current state-time
     *  \return a state from the current state-time
     */
    inline SSState toState() const { return SSState(x(),y(),th(),sx(),sy(),om()); }

    /**
     *  \brief return a copy of the current state-time
     *  \return a copy of the current state-time
     */
    inline virtual SSStateT* copy() const { return new SSStateT(*this); }


    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double& x() { return (*this)[0]; }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double& y() { return (*this)[1]; }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double& th() { return (*this)[2]; }

    /**
     *  \brief getter : speed SX
     *  \return speed SX
     */
    inline double& sx() { return (*this)[3]; }

    /**
     *  \brief getter : speed SY
     *  \return speed SY
     */
    inline double& sy() { return (*this)[4]; }

    /**
     *  \brief getter : angular velocity OM
     *  \return angular velocity OM
     */
    inline double& om() { return (*this)[5]; }

    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double x() const { return this->at(0); }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double y() const { return this->at(1); }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double th() const { return this->at(2); }

    /**
     *  \brief getter : speed SX
     *  \return speed SX
     */
    inline double sx() const { return this->at(3); }

    /**
     *  \brief getter : speed SY
     *  \return speed SY
     */
    inline double sy() const { return this->at(4); }

    /**
     *  \brief getter : angular velocity OM
     *  \return angular velocity OM
     */
    inline double om() const { return this->at(5); }

    /**
     *  \brief retrieve a configuration from the current state
     *  \return the corresponding configuration
     */
    inline virtual SSConfig cfg() const { return SSConfig(x(),y(),th()); }
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline double& t() { return (*this)[6]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline double t() const { return this->at(6); }

    /**
     *  \brief return the position-time of the space (in XxYxT)
     *  \return the position-time of the space (in XxYxT)
     */
    inline virtual PositionXYT posXYT() const { return PositionXYT(this->x(),this->y(),this->t()); }

    /**
     *  \brief return the position of the space (in W = R^2 x S)
     *  \return the position-time of the space (in W = R^2 x S)
     */
    inline virtual PositionXYTh posXYTh() const { return PositionXYTh(x(),y(),th()); }

//     /**
//      *  \brief accessor of the i-th coordinate
//      *  \param i : index of the coordinate to return
//      *  \return the i-th coordinate of the vector
//      */
//     inline virtual double& operator[](int i) { return coord_.at(i); }

//     /**
//      *  \brief accessor of the i-th coordinate
//      *  \param i : index of the coordinate to return
//      *  \return the i-th coordinate of the vector
//      */
//     inline virtual double at(int i) const { return this->SSState::at(i); }

//     /**
//      *  \brief size of the vector
//      *  \return the size of the vector
//      */
//     inline virtual int size() const { return this->SSState::size(); }

    /******************* Operators ********************/

    /**
     *  \brief operator = : copy the input state
     *  \param s2 : the state to copy
     */
    void operator=(const SSStateT& s2);

    /**
     *  \brief operator += : add the input state
     *  \param s2 : the state to add
     */
    void operator+=(const SSStateT& s2);

    /**
     *  \brief operator -= : substract the input state
     *  \param s2 : the state to substract
     */
    void operator-=(const SSStateT& s2);

    /**
     *  \brief operator += : multiply with the input coefficient
     *  \param d : the coefficient to mutilpy with
     */
    void operator*=(double d);

    /**
     *  \brief operator /= : divide with the input coefficient
     *  \param d : the coefficient to divide with
     */
    void operator/=(double d);

    /**
     *  \brief operator + : sumation
     *  \param s2 : the second vector to add
     *  \return the sum of both vectors
     */
    SSStateT operator+(const SSStateT& s2) const;

    /**
     *  \brief operator + : substraction
     *  \param s2 : the second vector to substract
     *  \return the substraction of both vectors
     */
    SSStateT operator-(const SSStateT& s2) const;

    /**
     *  \brief operator * : scalar multiplication
     *  \param d : the scalar to multiply with
     *  \return the product of the current vector and the scalar
     */
    SSStateT operator*(double d) const;

    /**
     *  \brief operator / : division by a scalar
     *  \param d : the scalar to divide with
     *  \return the division result
     */
    SSStateT operator/(double d) const;

    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;

    /*************** Utility functions ****************/

    /**
     *  \brief compute the relative coordinates between both state-times
     *  \param initST : the initial state-time
     *  \param finalST : the final state-time
     *
     *  initST is center in position and heading (0,0,0)
     *  finalST is adjust in the same frame than initST
     */
    static void relativeStateT(SSStateT& initST,SSStateT& finalST);
  };

};
#endif // MPC_SSSTATET_HPP


#ifndef MPC_DDSTATET_HPP
#define MPC_DDSTATET_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class DDStateT *****************/
                                    /**************************************************/

  /** \class DDStateT
   *  \brief State-time of a differential drive system
   */
  class DDStateT : public StateTime // , public DDState
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    DDStateT();

    /**
     *  \brief copy constructor
     *  \param p : the state-time to copy
     */
    DDStateT(const DDStateT& s);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param th : heading th
     *  \param omr : right wheel velocity omr
     *  \param oml : left wheel velocity omr
     *  \param t : time t
     */
    DDStateT(double x,double y,double th,double omr,double oml,double t);

    /**
     *  \brief destructor
     */
    virtual ~DDStateT();

    /**
     *  \brief return a state from the current state-time
     *  \return a state from the current state-time
     */
    inline DDState toState() const { return DDState(x(),y(),th(),omr(),oml()); }

    /**
     *  \brief return a copy of the current state-time
     *  \return a copy of the current state-time
     */
    inline virtual DDStateT* copy() const { return new DDStateT(*this); }


    /******************** Accessors *******************/
  
  public : 
    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double& x() { return (*this)[0]; }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double& y() { return (*this)[1]; }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double& th() { return (*this)[2]; }

    /**
     *  \brief getter : right wheel velocity OMR
     *  \return right wheel velocity OMR
     */
    inline double& omr() { return (*this)[3]; }

    /**
     *  \brief getter : left wheel velocity OML
     *  \return left wheel velocity OML
     */
    inline double& oml() { return (*this)[4]; }

    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double x() const { return this->at(0); }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double y() const { return this->at(1); }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double th() const { return this->at(2); }

    /**
     *  \brief getter : right wheel velocity OMR
     *  \return right wheel velocity OMR
     */
    inline double omr() const { return this->at(3); }

    /**
     *  \brief getter : left wheel velocity OML
     *  \return left wheel velocity OML
     */
    inline double oml() const { return this->at(4); }

    /**
     *  \brief retrieve a configuration from the current state
     *  \return the corresponding configuration
     */
    inline virtual DDConfig cfg() const { return DDConfig(x(),y(),th()); }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline double& t() { return (*this)[5]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline double t() const { return this->at(5); }

    /**
     *  \brief return the position-time of the space (in XxYxT)
     *  \return the position-time of the space (in XxYxT)
     */
    inline virtual PositionXYT posXYT() const { return PositionXYT(this->x(),this->y(),this->t()); }

    /**
     *  \brief return the position of the space (in W = R^2 x S)
     *  \return the position-time of the space (in W = R^2 x S)
     */
    inline virtual PositionXYTh posXYTh() const { return PositionXYTh(x(),y(),th()); }

//     /**
//      *  \brief accessor of the i-th coordinate
//      *  \param i : index of the coordinate to return
//      *  \return the i-th coordinate of the vector
//      */
//     inline virtual double& operator[](int i) { return coord_.at(i); }

//     /**
//      *  \brief accessor of the i-th coordinate
//      *  \param i : index of the coordinate to return
//      *  \return the i-th coordinate of the vector
//      */
//     inline virtual double at(int i) const { return this->DDState::at(i); }

//     /**
//      *  \brief size of the vector
//      *  \return the size of the vector
//      */
//     inline virtual int size() const { return this->DDState::size(); }

    /******************* Operators ********************/

    /**
     *  \brief operator = : copy the input state
     *  \param s2 : the state to copy
     */
    void operator=(const DDStateT& s2);

    /**
     *  \brief operator += : add the input state
     *  \param s2 : the state to add
     */
    void operator+=(const DDStateT& s2);

    /**
     *  \brief operator -= : substract the input state
     *  \param s2 : the state to substract
     */
    void operator-=(const DDStateT& s2);

    /**
     *  \brief operator += : multiply with the input coefficient
     *  \param d : the coefficient to mutilpy with
     */
    void operator*=(double d);

    /**
     *  \brief operator /= : divide with the input coefficient
     *  \param d : the coefficient to divide with
     */
    void operator/=(double d);

    /**
     *  \brief operator + : sumation
     *  \param s2 : the second vector to add
     *  \return the sum of both vectors
     */
    DDStateT operator+(const DDStateT& s2) const;

    /**
     *  \brief operator + : substraction
     *  \param s2 : the second vector to substract
     *  \return the substraction of both vectors
     */
    DDStateT operator-(const DDStateT& s2) const;

    /**
     *  \brief operator * : scalar multiplication
     *  \param d : the scalar to multiply with
     *  \return the product of the current vector and the scalar
     */
    DDStateT operator*(double d) const;

    /**
     *  \brief operator / : division by a scalar
     *  \param d : the scalar to divide with
     *  \return the division result
     */
    DDStateT operator/(double d) const;

    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;
  };

};
#endif // MPC_DDSTATET_HPP


#ifndef MPC_DD2STATET_HPP
#define MPC_DD2STATET_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class DD2StateT *****************/
                                    /**************************************************/

  /** \class DD2StateT
   *  \brief State-time of a wheelchair system
   */
  class DD2StateT :public StateTime
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    DD2StateT();

    /**
     *  \brief copy constructor
     *  \param p : the state-time to copy
     */
    DD2StateT(const DD2StateT& s);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param th : heading th
     *  \param s : linear velocity s
     *  \param om : angular velocity om
     *  \param t : time t
     */
    DD2StateT(double x,double y,double th,double s,double om,double t);

    /**
     *  \brief init constructor
     *  \param s : state s
     *  \param t : time t
     */
    DD2StateT(const DD2State& s,double t);

    /**
     *  \brief destructor
     */
    virtual ~DD2StateT();

    /**
     *  \brief return a state from the current state-time
     *  \return a state from the current state-time
     */
    inline DD2State toState() const { return DD2State(x(),y(),th(),s(),om()); }

    /**
     *  \brief return a copy of the current state-time
     *  \return a copy of the current state-time
     */
    inline virtual DD2StateT* copy() const { return new DD2StateT(*this); }


    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double& x() { return (*this)[0]; }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double& y() { return (*this)[1]; }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double& th() { return (*this)[2]; }

    /**
     *  \brief getter : liner velocity S
     *  \return linear velocity S
     */
    inline double& s() { return (*this)[3]; }

    /**
     *  \brief getter : angular velocity OM
     *  \return angular velocity OM
     */
    inline double& om() { return (*this)[4]; }

    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double x() const { return this->at(0); }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double y() const { return this->at(1); }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double th() const { return this->at(2); }

    /**
     *  \brief getter : linear velocity S
     *  \return linear velocity S
     */
    inline double s() const { return this->at(3); }

    /**
     *  \brief getter : angular velocity OM
     *  \return angular velocity OM
     */
    inline double om() const { return this->at(4); }

    /**
     *  \brief getter : liner velocity S
     *  \return linear velocity S
     */
    inline double& v() { return (*this)[3]; }

    /**
     *  \brief getter : angular velocity OM
     *  \return angular velocity OM
     */
    inline double& w() { return (*this)[4]; }

    /**
     *  \brief getter : linear velocity S
     *  \return linear velocity S
     */
    inline double v() const { return this->at(3); }

    /**
     *  \brief getter : angular velocity OM
     *  \return angular velocity OM
     */
    inline double w() const { return this->at(4); }

    /**
     *  \brief retrieve a configuration from the current state
     *  \return the corresponding configuration
     */
    inline virtual DD2Config cfg() const { return DD2Config(x(),y(),th()); }
  
    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline double& t() { return (*this)[5]; }

    /**
     *  \brief getter : time T
     *  \return time T
     */
    inline double t() const { return this->at(5); }

    /**
     *  \brief return the position-time of the space (in XxYxT)
     *  \return the position-time of the space (in XxYxT)
     */
    inline virtual PositionXYT posXYT() const { return PositionXYT(this->x(),this->y(),this->t()); }

    /**
     *  \brief return the position of the space (in W = R^2 x S)
     *  \return the position-time of the space (in W = R^2 x S)
     */
    inline virtual PositionXYTh posXYTh() const { return PositionXYTh(x(),y(),th()); }

//     /**
//      *  \brief accessor of the i-th coordinate
//      *  \param i : index of the coordinate to return
//      *  \return the i-th coordinate of the vector
//      */
//     inline virtual double& operator[](int i) { return coord_.at(i); }

//     /**
//      *  \brief accessor of the i-th coordinate
//      *  \param i : index of the coordinate to return
//      *  \return the i-th coordinate of the vector
//      */
//     inline virtual double at(int i) const { return this->DD2State::at(i); }

//     /**
//      *  \brief size of the vector
//      *  \return the size of the vector
//      */
//     inline virtual int size() const { return this->DD2State::size(); }

    /******************* Operators ********************/

    /**
     *  \brief operator = : copy the input state
     *  \param s2 : the state to copy
     */
    void operator=(const DD2StateT& s2);

    /**
     *  \brief operator += : add the input state
     *  \param s2 : the state to add
     */
    void operator+=(const DD2StateT& s2);

    /**
     *  \brief operator -= : substract the input state
     *  \param s2 : the state to substract
     */
    void operator-=(const DD2StateT& s2);

    /**
     *  \brief operator += : multiply with the input coefficient
     *  \param d : the coefficient to mutilpy with
     */
    void operator*=(double d);

    /**
     *  \brief operator /= : divide with the input coefficient
     *  \param d : the coefficient to divide with
     */
    void operator/=(double d);

    /**
     *  \brief operator + : sumation
     *  \param s2 : the second vector to add
     *  \return the sum of both vectors
     */
    DD2StateT operator+(const DD2StateT& s2) const;

    /**
     *  \brief operator + : substraction
     *  \param s2 : the second vector to substract
     *  \return the substraction of both vectors
     */
    DD2StateT operator-(const DD2StateT& s2) const;

    /**
     *  \brief operator * : scalar multiplication
     *  \param d : the scalar to multiply with
     *  \return the product of the current vector and the scalar
     */
    DD2StateT operator*(double d) const;

    /**
     *  \brief operator / : division by a scalar
     *  \param d : the scalar to divide with
     *  \return the division result
     */
    DD2StateT operator/(double d) const;

    /**************** Noise functions *****************/

    /**
     *  \brief return a randomly noised state from the current one
     *  \param t : the elapsed time from the previous known state
     *  \return noised state
     */
    DD2StateT noisedState(double t);

    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;

    /*************** Utility functions ****************/

    /**
     *  \brief compute the relative coordinates between both state-times
     *  \param initST : the initial state-time
     *  \param finalST : the final state-time
     *
     *  initST is center in position and heading (0,0,0)
     *  finalST is adjust in the same frame than initST
     */
    static void relativeStateT(DD2StateT& initST,DD2StateT& finalST);
  };

};
#endif // MPC_DD2STATET_HPP


#ifndef MPC_WCSTATET_HPP
#define MPC_WCSTATET_HPP

#define WCStateT DD2StateT

#endif // MPC_WCSTATET_HPP


#ifndef MPC_MANASTATET_HPP
#define MPC_MANASTATET_HPP

#define ManaStateT DD2StateT

#endif // MPC_MANASTATET_HPP
