#ifndef MPC_TREE_HPP
#define MPC_TREE_HPP

#include <vector>
#include <list>
#include <stdlib.h>

namespace mpc
{
                                    /**************************************************/
                                    /******************* Class Node *******************/
                                    /**************************************************/

  /*  \class Node
   *  \brief template node class for tree construction
   */
  template <typename T> class Node
  {
  protected : 
    Node<T>* father_;                              /*< the father of the current node */
    std::vector<Node<T>*> sons_;                   /*< the sons of the current node */
    T val_;                                        /*< the value of the node */

    /********** Constructors and destructors **********/

  protected :
    /**
     *  \brief default constructor
     */  
    Node()
    {
      this->father() = NULL;
    }

  public : 
    /**
     *  \brief copy constructor
     *  \param n : the node to copy
     */  
    Node(const Node<T>& n)
    {
      this->father() = n.father();
      int N = n.nbSons();
      for(int i=0; i<N; i++)
	this->addSon(n.son(i));
      this->val() = n.val();
    }

    /**
     *  \brief init constructor
     *  \param val : the value contained in the node
     */  
    Node(T val)
    {
      this->father() = NULL;
      this->val()    = val;
    }
  
    /**
     *  \brief destructor
     */  
    virtual ~Node()
    {
      for(int i=0; i<sons_.size(); i++) delete sons_.at(i);
    }
  
  
    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the father of the current node
     *  \return a reference to the father of the current node
     */
    virtual inline Node<T>*& father() { return this->father_; }

    /**
     *  \brief return the father of the current node
     *  \return a pointer to the father of the current node
     */
    virtual inline Node<T>* father() const { return this->father_; }

    /**
     *  \brief return the number of sons of the current node
     *  \return the number of sons of the current node
     */
    inline int nbSons() const { return this->sons_.size(); }

    /**
     *  \brief return the i-th son of the current node
     *  \param i : the index of the son
     *  \return the i-th son of the current node
     */
    virtual inline Node<T>* son(int i) const { return this->sons_.at(i); }

    /**
     *  \brief return a reference to the i-th son of the current node
     *  \return a reference to the i-th son of the current node
     */
    inline T& val() { return this->val_; }
  
    /**
     *  \brief return the i-th son of the current node
     *  \return the i-th son of the current node
     */
    inline T val() const { return this->val_; }

    /**
     *  \brief add a son to the current node
     *  \param s : a pointer to the new son
     */
    void addSon(Node<T>* s)
    {
      this->sons_.push_back(s); 
      s->father_ = this;
    }

    /**
     *  \brief add a son to the current node
     *  \param s : a pointer to the new son
     */
    void addSon(T val)
    {
      Node<T>* newNode = new Node<T>(val);
      this->addSon(newNode);
    }

    /**
     *  \brief return true if the current node is a leaf
     *  \return true if the current node is a leaf
     */
    bool isLeaf() const
    {
      if(this->sons_.size() == 0) return true;
      else return false;
    }

    /**
     *  \brief return true if the current node is a root
     *  \return true if the current node is a root
     */
    bool isRoot() const
    {
      if(this->father_ == NULL) return true;
      else return false;
    }


    /**************** Utility functions ***************/

  protected : 
        /**
     *  \brief return a vector containing all the nodes corresponding to the leaves ordered from left to right
     *  \return a vector containing all the nodes corresponding to the leaves ordered from left to right
     *
     *  A new instance of std::vector is returned
     */
    void getLeavesNodes(std::vector<Node<T>*>& vecLeaves)
    {
      if(this->isLeaf()) vecLeaves.push_back(this);
      else 
	{
	  for(int i=0; i<this->nbSons(); i++)
	    {
	      this->son(i)->getLeavesNodes(vecLeaves);
	    }
	}
    }

  public : 
    /**
     *  \brief return a vector containing all the values of the leaves ordered from left to right
     *  \return a vector containing all the values of the leaves ordered from left to right
     */
    void getLeavesVal(std::vector<T>& vecLeaves) const
    {
      if(this->isLeaf()) vecLeaves.push_back(val_);
      else 
	{
	  for(int i=0; i<this->nbSons(); i++)
	    {
	      this->son(i)->getLeavesVal(vecLeaves);
	    }
	}
    }

  private : 
    /*
     *  \brief recursive method of the computeAllBranches method
     *  \param vecAllBranches : the vector containing all branches
     *  \param vecCurBranch : the current branch built
     */
    void computeAllBranchesRec(std::vector< std::vector<T>* >* vecAllBranches, std::vector<T>* vecCurBranch) const
    {
      // add the current value to the current branch
      vecCurBranch->push_back(this->val());

      if(this->isLeaf())
	{
	  // add the current branch the the vector of branches
	  vecAllBranches->push_back(vecCurBranch);
	}
      else
	{
	  // call the recursive function on each son of the current node
	  int N = this->nbSons();

	  for(int i=0; i<N; i++)
	    {
	      std::vector<T>* vecNewCurBranch = NULL;

	      if(i == N-1) vecNewCurBranch = vecCurBranch;
	      else vecNewCurBranch = new std::vector<T>(*vecCurBranch);

	      this->son(i)->computeAllBranchesRec(vecAllBranches, vecNewCurBranch);
	    }
	}
    }

  public : 
    /**
     *  \brief compute every possible branch from the root to a leaf of the tree
     *  \return a vector of every possible branches
     */
    std::vector< std::vector<T>* > computeAllBranches() const
    {
      std::vector< std::vector<T>* > vecAllBranches;
      std::vector<T>* vecCurBranch = new std::vector<T>;

      this->computeAllBranchesRec(&vecAllBranches, vecCurBranch);

      return vecAllBranches;
    }

    /**
     *  \brief get back the list of values of the nodes from the root of the tree up to the current node
     *  \param vecRes : the resulting vector of nodes
     */
    void getVecValuesFromRoot(std::list<T>& vec_vals)
    {
      Node<T>* curNode = this;
      while(curNode != NULL)
	{
	  vec_vals.push_front(curNode->val());
	  curNode = curNode->father();
	}
    }
  };


//                                     /**************************************************/
//                                     /******************* Class Tree *******************/
//                                     /**************************************************/

//   /*  \class Tree
//    *  \brief template tree
//    */
//   template <typename T> class Tree : Node<T>
//   {
//     /********** Constructors and destructors **********/
  
//   protected : 
//     /**
//      *  \brief default constructor
//      */  
//     Tree();

//   public : 
//     /**
//      *  \brief init constructor
//      *  \param val : the value contained in the node
//      */  
//     Tree(T val);

//     /**
//      *  \brief init constructor
//      *  \param val : the value contained in the node
//      */  
//     Tree(const Node<T>& n);

//     /**
//      *  \brief destructor
//      */  
//     ~Tree();
//   };

};
#endif // MPC_TREE_HPP
