#include "Control.hpp"
#include "Utils.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /****************** Class Control *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

// default constructor
Control::Control() : Vector()
{
}

// copy constructor
Control::Control(const Control& c) : Vector(c)
{
}

// init construtor (init a control of n parameters)
Control::Control(int n) : Vector(n)
{
}

// destructor
Control::~Control()
{
}


                                    /**************************************************/
                                    /**************** Class PM3DControl ***************/
                                    /**************************************************/

double PM3DControl::maxNoiseAcc_ = 0.2;

/********** Constructors and destructors **********/

// default constructor
PM3DControl::PM3DControl() : Control(3)
{
}

// copy constructor
PM3DControl::PM3DControl(const PM3DControl& c) : Control(3)
{
  this->ax() = c.ax();
  this->ay() = c.ay();
  this->az() = c.az();
}

// init construtor
PM3DControl::PM3DControl(double ax,double ay,double az) : Control(3)
{
  this->ax() = ax;
  this->ay() = ay;
  this->az() = az;
}

// destructor
PM3DControl::~PM3DControl()
{
}


/**************** Noise functions *****************/

/**
 *  \brief return a randomly noised control from the current one
 *  \param t : the application time of the control
 *  \return noised control
 *
 *  The longer the control is applied, the more it will be noised
 */
PM3DControl PM3DControl::noisedCtrl(double t)
{
  double noiseAX = getDoubleRandom(-maxNoiseAcc_*t,maxNoiseAcc_*t);
  double noiseAY = getDoubleRandom(-maxNoiseAcc_*t,maxNoiseAcc_*t);
  double noiseAZ = getDoubleRandom(-maxNoiseAcc_*t,maxNoiseAcc_*t);
 
  return PM3DControl(this->ax() + noiseAX,
		     this->ay() + noiseAY,
		     this->az() + noiseAZ);
}


                                    /**************************************************/
                                    /***************** Class CLControl ****************/
                                    /**************************************************/

double CLControl::maxNoiseAcc_ = 0.2;
double CLControl::maxNoiseSS_ = 0.003*M_PI;

/********** Constructors and destructors **********/

// default constructor
CLControl::CLControl() : Control(2)
{
}

// copy constructor
CLControl::CLControl(const CLControl& c) : Control(2)
{
  this->acc() = c.acc();
  this->ss()  = c.ss();
}

// init construtor
CLControl::CLControl(double acc,double ss) : Control(2)
{
  this->acc() = acc;
  this->ss()  = ss;
}

// destructor
CLControl::~CLControl()
{
}


/**************** Noise functions *****************/

// return a noised control from the current one
CLControl CLControl::noisedCtrl(double maxNAcc,double maxNSS)
{
  double noiseAcc = getDoubleRandom(-maxNAcc,maxNAcc);
  double noiseSS = getDoubleRandom(-maxNSS,maxNSS);
 
  return CLControl(this->acc() + noiseAcc,
		   this->ss() + noiseSS);
}

/**
 *  \brief return a randomly noised control from the current one
 *  \param t : the application time of the control
 *  \return noised control
 *
 *  The longer the control is applied, the more it will be noised
 */
CLControl CLControl::noisedCtrl(double t)
{
  double noiseAcc = getDoubleRandom(-maxNoiseAcc_*t,maxNoiseAcc_*t);
  double noiseSS = getDoubleRandom(-maxNoiseSS_*t,maxNoiseSS_*t);
 
  return CLControl(this->acc() + noiseAcc,
		   this->ss() + noiseSS);  
}


                                    /**************************************************/
                                    /***************** Class SSControl ****************/
                                    /**************************************************/

double maxNoiseAccX_ = 0.1;
double maxNoiseAccY_ = 0.1;
double maxNoiseAAcc_ = 0.1*M_PI;

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
SSControl::SSControl() : Control(2)
{
}

/**
 *  \brief copy constructor
 *  \param c : the control to copy
 */
SSControl::SSControl(const SSControl& c) : Control(2)
{
  this->acc() = c.acc();
  this->aacc() = c.aacc();
}

/**
 *  \brief init constructor
 *  \param accx : the acceleration control on X
 *  \param accy : the acceleration control on Y
 *  \param aacc : the angular acceleration control
 */
SSControl::SSControl(double acc,double aacc) : Control(2)
{
  this->acc() = acc;
  this->aacc() = aacc;
}

/**
 *  \brief destructor
 */
SSControl::~SSControl()
{
}


                                    /**************************************************/
                                    /***************** Class DDControl ****************/
                                    /**************************************************/

double DDControl::maxNoiseWAcc_ = 0.2;

/********** Constructors and destructors **********/

// default constructor
DDControl::DDControl() : Control(2)
{
}

// copy constructor
DDControl::DDControl(const DDControl& c) : Control(2)
{
  this->wracc() = c.wracc();
  this->wlacc() = c.wlacc();
}

// init construtor
DDControl::DDControl(double wracc,double wlacc) : Control(2)
{
  this->wracc() = wracc;
  this->wlacc() = wlacc;
}

// destructor
DDControl::~DDControl()
{
}


/**************** Noise functions *****************/

/**
 *  \brief return a randomly noised control from the current one
 *  \param t : the application time of the control
 *  \return noised control
 *
 *  The longer the control is applied, the more it will be noised
 */
DDControl DDControl::noisedCtrl(double t)
{
  double noiseWRAcc = getDoubleRandom(-maxNoiseWAcc_*t,maxNoiseWAcc_*t);
  double noiseWLAcc = getDoubleRandom(-maxNoiseWAcc_*t,maxNoiseWAcc_*t);
  
  return DDControl(this->wracc() + noiseWRAcc,
		   this->wlacc() + noiseWLAcc);
}


                                    /**************************************************/
                                    /***************** Class DD2Control ****************/
                                    /**************************************************/

// double DD2Control::maxNoiseV_ = 1.5;
// double DD2Control::maxNoiseAV_ = 0.3*M_PI;
double DD2Control::maxNoiseV_ = 0.0;
double DD2Control::maxNoiseAV_ = 0.0;

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DD2Control::DD2Control() : Control(2)
{
}

/**
 *  \brief copy constructor
 *  \param c : the control to copy
 */
DD2Control::DD2Control(const DD2Control& c) : Control(2)
{
  this->v()  = c.v();
  this->av() = c.av();
}

/**
 *  \brief init constructor
 *  \param v : the linear velocity control
 *  \param av : the angular velocity control
 */
DD2Control::DD2Control(double v,double av) : Control(2)
{
  this->v()  = v;
  this->av() = av;
}

/**
 *  \brief destructor
 */
DD2Control::~DD2Control()
{
}


/**************** Noise functions *****************/

/**
 *  \brief return a randomly noised control from the current one
 *  \param t : the application time of the control
 *  \return noised control
 *
 *  The longer the control is applied, the more it will be noised
 */
DD2Control DD2Control::noisedCtrl(double t)
{
  double noiseV = getDoubleRandom(-maxNoiseV_*t,maxNoiseV_*t);
  double noiseAV = getDoubleRandom(-maxNoiseAV_*t,maxNoiseAV_*t);
  
  return DD2Control(this->v() + noiseV,
		   this->av() + noiseAV);
}


//                                     /**************************************************/
//                                     /**************** Class ManaControl ***************/
//                                     /**************************************************/

// /********** Constructors and destructors **********/

// /**
//  *  \brief default constructor
//  */
// ManaControl::ManaControl() : WCControl()
// {
// }

// /**
//  *  \brief copy constructor
//  *  \param c : the control to copy
//  */
// ManaControl::ManaControl(const ManaControl& c) : WCControl(c)
// {
// }

// /**
//  *  \brief init constructor
//  *  \param v : the linear velocity control
//  *  \param w : the angular velocity control
//  */
// ManaControl::ManaControl(double v,double w) : WCControl(v,w)
// {
// }

// /**
//  *  \brief destructor
//  */
// ManaControl::~ManaControl()
// {
// }


                                    /**************************************************/
                                    /*************** Class DD2AccControl *************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
DD2AccControl::DD2AccControl() : Control(2)
{
}

/**
 *  \brief copy constructor
 *  \param c : the control to copy
 */
DD2AccControl::DD2AccControl(const DD2AccControl& c) : Control(2)
{
  this->acc() = c.acc();
  this->aacc() = c.aacc();
}

/**
 *  \brief init constructor
 *  \param acc : the linear acceleration control
 *  \param aacc : the angular acceleration control
 */
DD2AccControl::DD2AccControl(double acc,double aacc) : Control(2)
{
  this->acc() = acc;
  this->aacc() = aacc;
}

/**
 *  \brief destructor
 */
DD2AccControl::~DD2AccControl()
{
}


                                    /**************************************************/
                                    /**************** Class AASControl ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
AASControl::AASControl() : Control(2)
{
}

/**
 *  \brief copy constructor
 *  \param c : the control to copy
 */
AASControl::AASControl(const AASControl& c) : Control(2)
{
  this->acc() = c.acc();
  this->as()  = c.as();
}

/**
 *  \brief init constructor
 *  \param acc : the acceleration control
 *  \param as : the angular speed control
 */
AASControl::AASControl(double acc,double as) : Control(2)
{
  this->acc() = acc;
  this->as()  = as;
}

/**
 *  \brief destructor
 */
AASControl::~AASControl()
{
}


                                    /**************************************************/
                                    /***************** Class CCControl ****************/
                                    /**************************************************/

double CCControl::maxNoiseV_ = 10.0;
double CCControl::maxNoisePh_ = 0.2;
// double CCControl::maxNoiseV_ = 0.4;
// double CCControl::maxNoisePh_ = 0.08;

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
CCControl::CCControl() : Control(2)
{
}

/**
 *  \brief copy constructor
 *  \param c : the control to copy
 */
CCControl::CCControl(const CCControl& c) : Control(2)
{
  this->v()  = c.v();
  this->ph() = c.ph();
}

/**
 *  \brief init constructor
 *  \param acc : the velocity control
 *  \param ph : the steering angle control
 */
CCControl::CCControl(double v,double ph) : Control(2)
{
  this->v()  = v;
  this->ph() = ph;  
}

/**
 *  \brief destructor
 */
CCControl::~CCControl()
{
}


/**************** Noise functions *****************/

/**
 *  \brief return a randomly noised control from the current one
 *  \param maxNV : the maximal noise applied on the velocity
 *  \param maxNPh : the maximal noise applied on the steering angle
 *  \return noised control
 */
CCControl CCControl::noisedCtrl(double maxNV,double maxNPh)
{
  double noiseV = getDoubleRandom(-maxNV,maxNV);
  double noisePh = getDoubleRandom(-maxNPh,maxNPh);
 
  return CCControl(this->v() + noiseV,
		   this->ph() + noisePh);
}

/**
 *  \brief return a randomly noised control from the current one
 *  \param t : the application time of the control
 *  \return noised control
 *
 *  The longer the control is applied, the more it will be noised
 */
CCControl CCControl::noisedCtrl(double t)
{
  double noiseV = getDoubleRandom(-maxNoiseV_*t,maxNoiseV_*t);
  double noisePh = getDoubleRandom(-maxNoisePh_*t,maxNoisePh_*t);
 
  return CCControl(this->v() + noiseV,
		   this->ph() + noisePh);  
}

