#include "Timer.hpp"
using namespace mpc;

#include <cstdlib>
#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /****************** Class Timer *******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
Timer::Timer()
{
  this->timerSt_ = TIMER_NONINIT;
  this->sumTotElapsedTime_ = 0;
  this->sumElapsedTime_ = 0;
  this->sumTotRequestedTime_ = 0;
  this->tic_ = 0;
}

/**
 *  \brief destructor
 */
Timer::~Timer()
{
  // nothing to do
}


/******************** Accessors *******************/

/**
 *  \brief compute and return the last elapsed time
 *  \return the last elapsed time
 */
double Timer::elapsedTime()
{
  double timePast;

  switch(this->timerSt_)
    {
      case TIMER_NONINIT : 
	return -1;
	break;

    case TIMER_STOPPED : 
      // compute the elapsed time from the last requested time
      timePast = this->sumElapsedTime_;

      // save the last requested time
      lastRequestedTime_ = stopTime_;

      // reinit the sum of elapsed time
      sumElapsedTime_ = 0;

      // return the resulting time
      return timePast;      
      break;

    case TIMER_STARTED : 
      // get back the current time
      struct timeval curTime;
      gettimeofday(&curTime,NULL);
      
      // compute the elapsed time from the last requested time
      if(lastTime_ >= lastRequestedTime_) timePast = Timer::computeElapsedTime(lastTime_,curTime);
      else timePast = Timer::computeElapsedTime(lastRequestedTime_,curTime);
      timePast += this->sumElapsedTime_;
      
      // save the last requested time
      lastRequestedTime_ = curTime;

      // reinit the sum of elapsed time
      sumElapsedTime_ = 0;

      // increase the tic value
      this->tic_++;
      this->sumTotRequestedTime_ += timePast;
      
      // return the resulting time
      return timePast;

    default : 
      cout << "Timer::elapsedTime : Unknown timer status" << endl;
      exit(1);
    }
}

/**
 *  \brief compute and return the total elapsed time
 *  \return the total elapsed time
 */
double Timer::totalElapsedTime() const
{
  double timePast;

  switch(this->timerSt_)
    {
    case TIMER_NONINIT : 
      return -1;
      break;

    case TIMER_STOPPED : 
      // compute the elapsed time from the initial time
      timePast = this->sumTotElapsedTime_;

      // return the resulting time
      return timePast;      
      break;

    case TIMER_STARTED :   
      // get back the current time
      struct timeval curTime;
      gettimeofday(&curTime,NULL);
      
      // compute the elapsed time from the initial time
      timePast = Timer::computeElapsedTime(lastTime_,curTime);
      timePast += this->sumTotElapsedTime_;
      
      // return the resulting time
      return timePast;
      break;

    default : 
      cout << "Timer::totalElapsedTime : Unknown timer status" << endl;
      exit(1);
    }
}

/**
 *  \brief return the average period
 *  \return the average period
 */
double Timer::avgPeriod() const
{
  if(tic_ == 0 || sumTotRequestedTime_ == 0) return 0;
  else return sumTotRequestedTime_ / ((double)tic_);
}

/**
 *  \brief return the average frequency
 *  \return the average frequency
 */
double Timer::avgFrequency() const
{
  if(tic_ == 0 || sumTotRequestedTime_ == 0) return 0;
  else return ((double)tic_) / sumTotRequestedTime_;
}


/********************* Actions ********************/

// /**
//  *  \brief init the timer with the current time
//  */
// void Timer::init()
// {
//   gettimeofday(initTime_);
//   lastTime_ = initTime_;
//   stopTime_ = initTime_;
// }

/**
 *  \brief start the timer
 */
void Timer::start()
{
  switch(this->timerSt_)
    {
    case TIMER_NONINIT : 
      this->timerSt_ = TIMER_STARTED;

      gettimeofday(&initTime_,NULL);
      lastTime_ = initTime_;
      stopTime_ = initTime_;
      lastRequestedTime_ = initTime_;
      break;

    case TIMER_STOPPED : 
      this->timerSt_ = TIMER_STARTED;

      // save the time at which we started the timer
      gettimeofday(&lastTime_,NULL);
      break;
      
    case TIMER_STARTED : 
      // nothing to do
      break;
    }
}

/**
 *  \brief stop the timer
 */
void Timer::stop()
{
  switch(this->timerSt_)
    {
    case TIMER_NONINIT : 
      // nothing to do
      break;

    case TIMER_STOPPED : 
      // nothing to do
      break;
      
    case TIMER_STARTED : 
      this->timerSt_ = TIMER_STOPPED;

      // save the time at which we stopped the timer
      gettimeofday(&stopTime_,NULL);
      // compute the elapsed time from the last starting time
      double timePast = Timer::computeElapsedTime(lastTime_,stopTime_);
      this->sumTotElapsedTime_ += timePast;
//       cout << "sumTotElapsedTime_ : " << sumTotElapsedTime_ << endl;
      // compute the elapsed time from the last requested time
      double timePast2;
      if(lastTime_ >= lastRequestedTime_) timePast2 = timePast;
      else timePast2 = Timer::computeElapsedTime(lastRequestedTime_,stopTime_);
      this->sumElapsedTime_ += timePast2;
//       cout << "sumElapsedTime_ : " << sumElapsedTime_ << endl;
      break;
    }
}

/**
 *  \brief reinitialize the timer
 *  \return 0 if the reinitialization succeed, -1 else
 */
int Timer::reinit()
{
  switch(this->timerSt_)
    {  
    case TIMER_NONINIT : 
    case TIMER_STOPPED : 
      this->timerSt_ = TIMER_NONINIT;

      this->sumTotElapsedTime_ = 0;
      this->sumElapsedTime_ = 0;
      this->sumTotRequestedTime_ = 0;
      this->tic_ = 0;
      return 0;
      break;

    case TIMER_STARTED : 
      return -1;
      break;

    default : 
      cout << "Timer::reinit : Unknown timer status" << endl;
      exit(1);
    }
}

/**
 *  \brief return the absolute init time
 *  \param time : the absolute init time
 */
void Timer::absoluteInitTime(struct timeval& time)
{
  time.tv_sec = initTime_.tv_sec;
  time.tv_usec = initTime_.tv_usec;
}

/**
 *  \brief return the absolute init time
 *  \param time : the absolute init time
 */
void Timer::absoluteInitTime(double& time)
{
  time = initTime_.tv_sec + 0.000001*initTime_.tv_usec;
}


/***************** Static functions ***************/

/**
 *  \brief compute the elapsed time between two struct timeval t1 and t2
 *  \param t1 : the start timeval
 *  \param t2 : the end timeval
 *  \return the elapsed time
 */
double Timer::computeElapsedTime(const struct timeval& t1,const struct timeval& t2)
{
  // compute the time past
  time_t elapsedTs = t2.tv_sec - t1.tv_sec;
  suseconds_t elapsedTus = t2.tv_usec - t1.tv_usec;

  double res = ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);
  return res;
}

/**
 *  \brief get back the absolute time
 *  \param time : resulting structure containing the current absolute time
 */
void Timer::absoluteTime(struct timeval& time)
{
  gettimeofday(&time,NULL);
}

/**
 *  \brief get back the absolute time
 *  \param time : resulting current absolute time
 */
void Timer::absoluteTime(double& time)
{
  struct timeval tv;
  gettimeofday(&tv,NULL);

  time = tv.tv_sec + 0.000001*tv.tv_usec;
}


                                    /**************************************************/
                                    /************* struct timeval operators ***********/
                                    /**************************************************/

/**
 *  \brief compare two struct timeval
 *  \param t1 : the first time to evaluate
 *  \param t2 : the second time to evaluate
 *  \return true if t1 < t2
 */
bool mpc::operator<(const struct timeval& t1,const struct timeval& t2)
{
  if(t1.tv_sec < t2.tv_sec || (t1.tv_sec == t2.tv_sec && t1.tv_usec < t2.tv_usec)) return true;
  else return false;
}

/**
 *  \brief compare two struct timeval
 *  \param t1 : the first time to evaluate
 *  \param t2 : the second time to evaluate
 *  \return true if t1 > t2
 */
bool mpc::operator>(const struct timeval& t1,const struct timeval& t2)
{
  if(t1.tv_sec > t2.tv_sec || (t1.tv_sec == t2.tv_sec && t1.tv_usec > t2.tv_usec)) return true;
  else return false;
}

/**
 *  \brief compare two struct timeval
 *  \param t1 : the first time to evaluate
 *  \param t2 : the second time to evaluate
 *  \return true if t1 <= t2
 */
bool mpc::operator<=(const struct timeval& t1,const struct timeval& t2)
{
  if(t1.tv_sec < t2.tv_sec || (t1.tv_sec == t2.tv_sec && t1.tv_usec <= t2.tv_usec)) return true;
  else return false;
}

/**
 *  \brief compare two struct timeval
 *  \param t1 : the first time to evaluate
 *  \param t2 : the second time to evaluate
 *  \return true if t1 >= t2
 */
bool mpc::operator>=(const struct timeval& t1,const struct timeval& t2)
{
  if(t1.tv_sec > t2.tv_sec || (t1.tv_sec == t2.tv_sec && t1.tv_usec >= t2.tv_usec)) return true;
  else return false;
}

/**
 *  \brief compare two struct timeval
 *  \param t1 : the first time to evaluate
 *  \param t2 : the second time to evaluate
 *  \return true if t1 == t2
 */
bool mpc::operator==(const struct timeval& t1,const struct timeval& t2)
{
  if(t1.tv_sec == t2.tv_sec && t1.tv_usec == t2.tv_usec) return true;
  else return false;
}

// /**
//  *  \brief copy a struct timeval
//  *  \param t1 : the resulting time
//  *  \param t2 : the time to copy
//  */
// void mpc::operator=(struct timeval& t1,const struct timeval& t2)
// {
//   t1.tv_sec = t2.tv_sec;
//   t1.tv_usec = t2.tv_usec;
// }
