#ifndef MPC_SIMPLECONTROLLER_HPP
#define MPC_SIMPLECONTROLLER_HPP

#include "system/inc/System.hpp"
#include "trajGen/inc/TrajGen.hpp"
#include "motion/inc/PCTraj.hpp"
#include "motion/inc/PSTraj.hpp"

#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>

namespace mpc
{

                                    /**************************************************/
                                    /************* Class SimpleController *************/
                                    /**************************************************/

  /** \class SimpleController
   *  \brief Simple controller
   */
  class SimpleController
  {
  protected : 
    System* sys_;                                          /**< the system used */
    double dt_;                                            /**< period at which is called the controller */

    TrajGen* trajgen_;                                     /**< trajectory generator tool used to find a control trajectory to the goal */

    int debug_;                                            /**< debug level of the controller */
    PCTraj* ctraj_;                                        /**< control trajectory currently followed */
    PSTraj* straj_;                                        /**< state trajectory currently followed */

    boost::mutex mutexTraj_;                               /**< mutex used to prevent control and state trajectories access conflicts */

    /********** Constructors and destructors **********/
  
  protected : 
    /**
     *  \brief default constructor
     *  Only used by subclasses
     */
    SimpleController();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the robotic system handled
     *  \param dt : the controller period
     *  \param debug : the debugging level
     */
    SimpleController(System* sys,double dt = 0.5, int debug=0);

    /**
     *  \brief destructor
     */
    virtual ~SimpleController();
  

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the system constraints
     *  \return the system constraints
     */
    virtual const System* sys() const = 0;

    /**
     *  \brief return the controller period
     *  \return the controller period
     */
    double dt() const;

    /**
     *  \brief set the controller period
     *  \param dt : the new controller period
     */
    void setDT(double dt);


    /*************** Control Computation **************/

  protected : 
    /**
     *  \brief Initialize the standard parameters of the trajectory generator
     *  \param dbg : the trajectory generator debugging mode
     */
    virtual void initTrajGen(int dbg=0);


    /***************** Debug / Display ****************/

  public : 
    /**
     *  \brief return a copy of the current control trajectory
     *  \return a copy of the current control trajectory
     * 
     *  A new instance of a subclass of PCTraj is returned
     */
    virtual const PCTraj* getPCTraj() = 0;

    /**
     *  \brief return a copy of the current control trajectory
     *  \return a copy of the current control trajectory
     * 
     *  A new instance of a subclass of PSTraj is returned
     */
    virtual const PSTraj* getPSTraj() = 0;
  };

};

#endif // MPC_SIMPLECONTROLLER_HPP
