#ifndef MPC_PMTRAJDEFORM_HPP
#define MPC_PMTRAJDEFORM_HPP

#include "deformation/inc/TrajDeform.hpp"
#include "motion/inc/STrajectory.hpp"
#include "system/inc/PointMassSys.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class PMTrajDeform ***************/
                                    /**************************************************/

  /** \class PMTrajDeform
   *  \brief Trajectory deformer TEDDY for a Point Mass system
   */
  class PMTrajDeform : public TrajDeform
  {
  protected : 
    PMStateT   *tableFInt_;                            /**< table of the internal forces */
    PMStateT   *tableFExt_;                            /**< table of the external forces */

    /********** Constructors and destructors **********/
  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    PMTrajDeform();

  public : 
    /**
     *  \brief init constructor
     *  \param traj : the initial trajectory to deform
     *  \param env : the environment in which the system evolve
     *  \param sys : the features of the robot system used
     */
    PMTrajDeform(PMSTraj* traj,Environment* env,PointMassSys* sys);

    /**
     *  \brief init constructor
     *  \param traj : the initial trajectory to deform
     *  \param env : the environment in which the system evolve
     *  \param sys : the features of the robot system used
     *  \param dbg : the debugging level used
     */
    PMTrajDeform(PMSTraj* traj,Environment* env,PointMassSys* sys,int dbg);

    //   /**
    //    *  \brief init constructor
    //    *  \param traj : the initial trajectory to deform
    //    *  \param env : the environment in which the system evolve
    //    *  \param sys : the features of the robot system used
    //    *  \param typeS : pointer to the table of state types (for simulation display)
    //    *  \param tns : pointer to the table of nearest states (for simulation display)
    //    *  \param tfi : pointer to the table of internal forces (for simulation display)
    //    *  \param tfe : pointer to the table of external forces (for simulation display)
    //    */
    //   PMTrajDeform(PMSTraj* traj,Environment* env,PointMassSys* sys,int** typeS,PositionXYT*** tns,StateTime** tfi,StateTime** tfe);

    /**
     *  \brief destructor
     */
    virtual ~PMTrajDeform();

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    inline PMSTraj* traj() const { return ((PMSTraj*)(this->traj_)); }

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    inline PointMassSys* sys() const { return ((PointMassSys*)(this->sys_)); }

    /**
     *  \brief set the debugging level
     *  \param dbg : the new debugging level
     */
    virtual void setDebug(int dbg);

    /*************** Forces computation ***************/

  protected : 
    /**
     *  \brief computing of the internal forces
     *  \param prev : the previous node of the trajectory
     *  \param cur : the current node of the trajectory
     *  \param next : the next node of the trajectory
     */
    PMStateT computeFInt(const PMStateT& prev, const PMStateT& cur, const PMStateT& next);

    /**
     *  \brief compute the total forces
     *  \param prev : the previous node of the trajectory
     *  \param cur : the current node of the trajectory
     *  \param next : the next node of the trajectory
     *  \param env : the environment of the system
     *  \param time : the current time
     *  \param last : true if the current node is the last node of the trajectory
     */
    PMStateT computeForces(const PMStateT& prev, const PMStateT& cur, const PMStateT& next, Environment* env, double time, bool last);

    /**
     *  \brief apply the internal forces on the current node
     *  \param i : index of the current node on the trajectory
     *  \param prev : the previous node of the trajectory
     *  \param cur : the current node of the trajectory
     *  \param next : the next node of the trajectory
     */
    virtual void applyFInt(int i,const PMStateT& prev, PMStateT& cur, const PMStateT& next);

    /**
     *  \brief compute the total external forces
     *  \param i : index of the current node on the trajectory
     *  \param cur : the current node of the trajectory
     *  \param env : the environment of the system
     *  \param time : the current time
     *  \param last : true if the current node is the last node of the trajectory
     */
    virtual void applyFExt(int i,PMStateT& cur, Environment* env, double time, bool last);


    /******* Connectivity and Collision checking ******/

  protected : 
    /**
     *  \brief check the connectivity between two nodes of the trajectory
     *  \param i : 
     *  \param oldCur : 
     *  \param prev : the previous node of the trajectory
     *  \param cur : the current node of the trajectory
     *  \param force : the resulting force to apply to restore the connectivity
     */
    void checkConnectivity(int i, const PMStateT& oldCur, const PMStateT& prev, const PMStateT& cur, PMStateT& force);


    /******************* Resampling *******************/
  
  protected : 
    /**
     *  \brief resample the trajectory
     *  \param curStep : index of the node already reached
     */
    virtual void resample(int curStep);


    /*************** Deformation process **************/

  protected : 
    /**
     *  \brief init the deformation and check the temporal space between two consecutive states
     *  \param time : the current time
     *  \param curStep : index of the node of the trajectory that has been already reached
     */
    virtual void initDeform(double time,int curStep);

    /**
     *  \brief finish the deformation
     *  \param time : the current time
     *  \param curStep : index of the node of the trajectory that has been already reached
     */
    virtual int finishDeform(double time,int curStep);

  public : 
    /**
     *  \brief deform the current trajectory
     *  \param time : the current time
     *  \param curStep : index of the node of the trajectory that has been already reached
     */
    virtual int deform(double time, int curStep);


    /**************** Display functions ***************/

  public : 
    /**
     *  \brief display the external forces (from dynamic obstacles) in an openGL viewer
     */
    virtual void dispFExt();

    /**
     *  \brief display the external forces (from static obstacles) in an openGL viewer
     */
    virtual void dispStFExt();

    /**
     *  \brief display the internal forces in an openGL viewer
     */
    virtual void dispFInt();
  };

};
#endif // MPC_PMTRAJDEFORM_HPP
