#ifndef STDEFORMATION_HPP
#define STDEFORMATION_HPP

#include "StateR.hpp"
#include "StateTime.hpp"
#include "STrajectory.hpp"
#include "PVDTrajPlanner.hpp"
#include "PFPathPlanner.hpp"
#include "Environment.hpp"
#include "Robot.hpp"
#include "ReachabilityCone.hpp"
#include "NHPath.hpp"
#include "NHPathPlanner.hpp"

// Different deformation modes
typedef enum {
  SPATIOTEMP_DEFORM  = 1, // spatio-temporal deformation
  TEMPONLY_DEFORM    = 2, // temporal deformation only
  SPATONLY_DEFORM    = 3  // spatial deformation only
} modeDeform_t;

#define T_FEA   0
#define T_FAR   1
#define T_UNFEA 2

class STDeformation
{
protected : 
  STrajectory  *traj_;                  // the trajectory which will be deformed
  Environment *env_;                    // the environment by which the trajectory is deformed
  Robot       *rb_;                     // the features of the robot system which is used

  double      *K_INT;                   // internal forces coefficient
  double       K_EXT_SPT;               // external spatial forces coefficient (deprecated)
  double       K_EXT_TMP;               // external temporal forces coefficient (deprecated)
  double      *K_SPT_TMP;               // balance between spatial and temporal forces
  double      *K_EXT;                   // external forces coefficient (dynamic obstacles)
  double      *K_STEXT;                 // external forces coefficient (static obstacles)
//   double      *maxEffectDist;           // maximum distance until which external forces take effects (dynamic obstacles)

  modeDeform_t modeDeform;              // mode of the deformation which will be applied

  double      *alphaMin;                // minimal spatial coefficient (for the metric)
  double      *alphaMax;                // maximal spatial coefficient (for the metric)
  double      *betaMin;                 // minimal temporal coefficient (for the metric)
  double      *betaMax;                 // maximal temporal coefficient (for the metric)

  double      *spMed;                   // spatial maximal effective distance
  double      *tmpMed;                  // temporal maximal effective distance
  double      *maxStEffectDist;         // maximum distance until which external forces take effects (static obstacles)

  double      *reliabilityTime;         // reliability time
  double      *epsilonRT;               // epsilon for reliability time
  double      *predHorizon_;            // prediction horizon during which we consider 
                                        // the previsional model to b completely reliable

  bool        *useTemporalConsistency;  // true if the temporal consistency is used

  int         **typeState;              // table of the type of the states of the trajectory

  StateR      ***tableNearState;        // table of nearest states of the states of the trajectory
  StateR      **tableFInt;              // table of the internal forces
  StateR      **tableFExt;              // table of the external forces

  double      oldTime_;                 // time of the previous deformation which has been done

  double      timeStep_;                // time step of the trajectory computation (used to adjust the number of state-time)
  double      posStep_;                 // pos step of the trajectory computation (used to adjust the number of state-time)

  int         I;

  double      lastDeformDuration_;      // duration of the last deformation process

public : 
  /************** Constructors and destructors *************/

  STDeformation();                                            // default constructor
  STDeformation(STrajectory *traj,Environment *env,Robot *rb,int** typeS); 
                                                              // init constructor
  STDeformation(STrajectory *traj,Environment *env,Robot *rb,int** typeS,StateR*** tns,StateR** tfi,StateR** tfe); 
                                                              // init constructor
  STDeformation(STrajectory *traj,Environment *env,Robot *rb,int** typeS,StateR6D*** tns,StateR6D** tfi,StateR6D** tfe); 
                                                              // init constructor
  virtual ~STDeformation();                                   // destructor

  /*********************** Accessors ***********************/

  STrajectory* getTraj() const;                                // return the current trajectory
  void setTraj(STrajectory* traj);                             // set the current trajectory
  Environment* getEnv() const;                                // return the current environment
  void setEnv(Environment* env);                              // set the current environment
  Robot* getRobot() const;                                    // return the current robot
  void setRobot(Robot* rb);                                   // set the current robot

  void setModeDeform(modeDeform_t mode);                      // set the deformation mode
  void setForcesCoef(double* kint, double* kext, double *kstext, double* bal); 
       // set the coefficients of internal and external forces and the balance between external spatial and temporal forces
  void setForcesCoef(double* kint, double* kext, double *kstext, double* amin, double* amax, double* bmin, double* bmax); 
       // set the coefficients of internal and external forces and the balance between external spatial and temporal forces
//   void setMaxEffectDist(double* mad);                         // set the distance from the robot until which forces take effects (dynamic obstacles)
//   void setMaxStEffectDist(double* mad);                       // set the distance from the robot until which forces take effects (static obstacles)
  void setMaxEffectDist(double* spMed,double* tmpMed, double* staticMed); // set the spatial and temporal distances from the robot until which forces take effects
  void setReliability(double* time, double* eps);             // set the reliability time and the associated epsilon value
  void setReliability(double* predH, double* time, double* eps); 
                                      // set the prediction horizon, the reliability time and the associated epsilon value
  void setTemporalConst(bool *b);                             // set if the temporal consistency is used or not

  void setTimeStep(double ts);                                // set the time step
  void setPosStep(double ps);                                 // set the position step

  // set the duration of the last deformation process
  inline void setLastDeformDuration(double deformDur) { lastDeformDuration_ = deformDur; }

  /*** Computing of the forces applied on the trajectory ***/

  StateR5D computeFInt(const StateR& prev, const StateR& cur, const StateR& next);
    // computing of the internal forces
  StateR5D computeF1Ext(const StateR& prev, const StateR& cur, const StateR& next, Environment* env, bool last);
  StateR5D computeF2Ext(const StateR& cur, Environment* env);
  StateR5D computeF3Ext(const StateR& cur, Environment* env);
    // compute the external spatial force due to the obstacles
  StateR5D computeF4Ext(const StateR& cur, Environment* env);
    // compute the external temporary force due to the obstacles
  StateR5D computeF5Ext(const StateR& cur, Environment* env, double time, bool last);
    // compute the external spatio-temporary force due to the obstacles
  StateR5D computeF6Ext(const StateR& cur, Environment* env, double time, bool last);
    // compute the external spatio-temporary force due to the obstacles
  StateR5D searchNearTempFS(const StateR& cur, Environment* env);
    // search the nearest temporary free space
  StateR5D computeFExt(const StateR& cur, Environment* env, double time, bool last);
    // compute the total external forces
  StateR5D computeStFExt(const StateR& cur, Environment* env, bool last);
    // compute the external forces due to static obstacles
  StateR5D computeForces(const StateR& prev, const StateR& cur, const StateR& next, Environment* env, double time, bool last);
    // compute the total forces

  double getCoefReliability(double dt) const;               // return the reliability coefficient to apply on the forces 
                                                            // knowing the difference dt between the current time and the 
                                                            // time of the node on which we apply the force

  /********** Connectivity and Collision checking **********/

  double adjustTime(STrajectory& oldTraj, STrajectory& newTraj, int i, int j);
    // adjust the time of the trajectory
  virtual void checkConnectivity(int i, const StateR& oldCur, const StateR& prev, const StateR& cur, StateR& force);
    // check the connectivity between two nodes of the trajectory
  virtual double checkCollision(STrajectory* traj,Environment* env,double curTime);
    // check if the still-to-be executed part of the trajectory is in collision with the environment
    // return the first time from the current time where a collision occurs. return -1 if there is no collision

  /************** Spatio-temporal deformation **************/

  virtual void initDeform();                                // init the deformation and check the temporal space between two consecutive states
  virtual int finishDeform(double time);                    // finish the deformation
  virtual int deform(double time, int curStep);             // deform the current trajectory

  /******************* Display functions *******************/
  void displayPredHorizonProfile(double dt) const;          // Display the predictional horizon profile on interval [0;dt]
};

#endif
