#ifndef STNHDEFORMATION_HPP
#define STNHDEFORMATION_HPP

#include "STDeformation.hpp"
#include "NHReachabilityCone.hpp"
#include "NHReachabilitySpace.hpp"
#include "CLTrajGen.hpp"

#include <vector>
#include <list>

// Various speed disconnections types
typedef enum {
  NO_DISCO_SPACC   = 0, // no disconnection over speed or acceleration
  DISCO_MIN_SPACC  = 1, // too much time to travel a so small distance
  DISCO_MAX_SPACC  = 2  // not enough time to travel a so long distance
} disco_spacc_t;

// Various curvature disconnections types
typedef enum {
  NO_DISCO_CURV    = 0, // no disconnection over the curvature
  DISCO_LEFT_CURV  = 1, // too great left turn
  DISCO_RIGHT_CURV = 2  // too great right turn
} disco_curv_t;


class ParamCtrl
{
public : 
  double B,C,D,B2,C2,D2;
  
  /************** Constructors and destructors *************/
  ParamCtrl();                                               // default constructor
  ParamCtrl(const ParamCtrl& p);                             // copy constructor
  ParamCtrl(double b,double c,double d,double b2,double c2,double d2); // init constructor
  ~ParamCtrl();                                              // destructor
};


class STNHDeformation : public STDeformation
{
public : 
  std::vector <Trajectory*> vecTraj_;
  std::list <Vec_DP*>* listCtrl_;
  std::list <int>* feaCtrl_;

  std::vector<StateS6D> itdTraj_;                       // intermediate trajectory (after external forces application)
  std::vector<StateS6D> itdTraj2_;                      // intermediate trajectory (after time adjustment)

  bool parallelExec;
  unsigned int nbThread;

  Vec_DP constraints_;

  bool usePreviousParam_;                               // true if the previous parameters for trajectory generation are used
  bool useResampling_;                                  // true if the resampling process is used

protected : 
  
public : 
  /************** Constructors and destructors *************/
  
  STNHDeformation();                                         // default constructor
  STNHDeformation(Trajectory *traj,Environment *env,Robot *rb,int** typeS,StateR*** tns,StateR** tfi,StateR** tfe); 
                                                             // init constructor
  virtual ~STNHDeformation();                                // destructor

  /****************** Getters and Setters ******************/
  inline bool isParallelExec() const { return parallelExec; }
  inline void setParallelExec(bool p) { parallelExec = p; }
  inline int getNbThread() const { return nbThread; }
  inline void setNbThread(int nbt) { nbThread = nbt; }
  inline void setUsePrevParam(bool u) { usePreviousParam_ = u; }
  inline void setUseResampling(bool u) { useResampling_ = u; }

  /*** Computing of the forces applied on the trajectory ***/

  StateS6D computeNHFInt(const StateR& prev, const StateR& cur, const StateR& next, const StateR& next5, int &seg5, Trajectory6D* traj, int I);
    // computing of the internal forces
  StateS6D computeNHFInt(const StateR& prev, const StateR& cur, const StateR& next, bool last,Vec_DP** param, int *fea);
    // computing of the internal forces
  void computeNHFInt(const StateR& prev, const StateR& cur, const StateR& next, bool last,Vec_DP** param, int* fea, StateS6D& fres);
    // computing of the internal forces
  StateS6D computeNHFIntHe(const StateR& prev, const StateR& cur, const StateR& next, disco_spacc_t& discoSA, disco_curv_t& discoC, Trajectory6D* traj,int I);
    // computing of the internal forces
  StateS6D computeNHFIntRK(const StateR& prev, const StateR& cur, const StateR& next, disco_spacc_t& discoSA, disco_curv_t& discoC);
    // computing of the internal forces
  StateS6D computeNHF6Ext(const StateR& cur, Environment* env, double time, bool last);
    // compute the total external forces
  StateS6D computeNHFExt(const StateR& cur, Environment* env, double time, bool last);
    // compute the total external forces
  StateS6D computeNHStFExt(const StateR& cur, Environment* env, bool last);
    // compute forces exerted by static obstacles
  StateS6D computeNHForces(const StateR& prev, const StateR& cur, const StateR& next, Environment* env, double time, bool last);// compute the total forces

  /********** Connectivity and Collision checking **********/

  double checkConnectivity(Trajectory* traj, double curTime);
    // check the connectivity of the trajectory
  virtual void checkNHConnectivity(int i, const StateR& oldCur, const StateR& prev, const StateR& cur, StateR& force);
    // check the connectivity between two nodes of the trajectory
  void adjustTime(int curStep);
    // adjust the time of the nodes of the trajectory to maintain its consistency
  virtual double checkCollision(STrajectory* traj,Environment* env,double curTime);
    // check if the still-to-be executed part of the trajectory is in collision with the environment
    // return the first time from the current time where a collision occurs. return -1 if there is no collision

  /*********************** Resampling **********************/

  void resample(int curStep);                                // resample the trajectory

  /************** Spatio-temporal deformation **************/
  
  virtual void initDeform(int curStep);                      // init the deformation and check the temporal space between two consecutive states
  virtual int finishDeform(double time, int curStep);        // finish the deformation
  virtual int deform(double time, int curStep);              // deform the current trajectory

  /*************** Feedback control methods ****************/

  void initListCtrl(std::list <Vec_DP*>* listCtrl);       // init the list of feedback controls

  /***************** Trajectory computation ****************/

  int genTraj(const StateS6D& prev,const StateS6D& next,CLTrajGen& carModel,Vec_DP** param,double& dt) const;
    // initialize the trajectory generator and compute a trajectory from prev to next
    // return the trajectory generator return status

  int computeItdStateT_Strict(const StateS6D& prev,const StateS6D& next,StateS6D& res,Vec_DP** param);
    // compute the intermediate node to add between prev and next 
    // strict case : compute the node only if the convergence has been ensured
    // return status : 0 -> convergence has been ensured
    //                 1 -> convergence has not been ensured

  int computeItdStateT_Avg(const StateS6D& prev,const StateS6D& next,StateS6D& res,Vec_DP** param);
    // compute the intermediate node to add between prev and next 
    // average case : compute a intermediate node even if the convergence hasn't been ensured
    // return status : 0 -> near convergence has been ensured
    //                 1 -> far convergence has been ensured
    //                 2 -> convergence hasn't been ensured


  /**************** Trajectory modifications ***************/

  StateS6D getStateTraj(double t) const;                     // return the state of the deformed trajectory at time t
  void displayTraj() const;                                  // display the refined trajectory in an openGL viewer
  
  int locateLoop(int curStep,std::vector<int*> &vecLoop) const; // locate the loops along the trajectory
                                                         // return the number of loops found and fill the vector of loops
  void removeLoop(int curStep);                              // remove the loops from the trajectory

  void insertStateInTraj(const StateS6D& s,int pos,Vec_DP** param,int fea); // insert a state in the deformed trajectory
  void deleteStateFromTraj(int start,int end);               // delete one or several states from the trajectory
                                                             // including start but not including end

  /******************* Utility functions *******************/
  
  void initConstraints();                                    // initialize the vector of constraints

};

#endif // STNHDEFORMATION_HPP
