#ifndef BOOLOBSMAP_HPP
#define BOOLOBSMAP_HPP

#include "environment/inc/Map2D.hpp"
#include "environment/inc/SlopeMap.hpp"
#include "environment/inc/VarSlopeMap.hpp"

#include <cmath>

namespace mpc
{

/**
 * \enum obs2D_cell_t
 * \brief Enumeration of the different obs2D cell types
 *
 * CELL_UNKWN : unknown cell
 * CELL_OBS   : obstructed cell
 * CELL_FREE  : free cell
 */
typedef enum {
  CELL_UNKWN   =  -1,
  CELL_OBS     =   0,
  CELL_FREE    =   1
} obs2D_cell_t;

                                    /**************************************************/
                                    /***************** Class BoolObsMap ***************/
                                    /**************************************************/

  /** \class BoolObsMap
   * \brief map dividing a 2D environment into CObs and CFree
   */
  class BoolObsMap : public Map2D
  {
  protected : 
    obs2D_cell_t** cellState_;    /**< State of the obstacle map cells (CELL_UNKWN, CELL_OBS or CELL_FREE) */

    float thresSlope_;            /**< the maximal slope that the robot can climb */
    float thresVarSlope_;         /**< the maximal slope derivative that the robot can climb */
    float safeDist_;              /**< the minimal distance that the robot may keep to be safe for collisions */
    int maxNbCells_;              /*< Maximal niumber of cells corresponding to maxDist_ */

  protected : 
    static float colDepth_;       /**< depth of the colors used to convert the map into images */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    BoolObsMap();
    
  public : 
    /**
     *  \brief copy constructor
     *  \param map : the map to copy
     */
    BoolObsMap(const BoolObsMap& map);

    /**
     *  \brief init constructor
     *  \param nbCols : number of columns of the map table
     *  \param nbLines : number of lines of the map table
     *  \param xOrigin : X coordinate of the top left corner of the (0,0) pixel in the global frame
     *  \param yOrigin : Z coordinate of the top left corner of the (0,0) pixel in the global frame
     *  \param xScale : X size of a cell in meters
     *  \param yScale : Z size of a cell in meters
     *  \param thresSlope : the maximal slope that the robot can climb
     *  \param thresVarSlope : the maximal slope derivative that the robot can climb
     *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
     */
    BoolObsMap(int nbCols, int nbLines, float xOrigin, float yOrigin, float xScale, float yScale, float thresSlope, float thresVarSlope, float safeDist);

    /**
     *  \brief init constructor (from a data elevation map)
     *  \param dem : the input dem
     *  \param thresSlope : the maximal slope that the robot can climb
     *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
     */
    BoolObsMap(const SlopeMap* slopemap, float thresSlope=0.25*M_PI, float safeDist=0.5);

    /**
     *  \brief init constructor (from a data elevation map)
     *  \param dem : the input dem
     *  \param thresSlope : the maximal slope that the robot can climb
     *  \param thresVarSlope : the maximal slope derivative that the robot can climb
     *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
     */
    BoolObsMap(const VarSlopeMap* varslopemap, float thresSlope=0.25*M_PI, float thresVarSlope=5.0*M_PI, float safeDist=0.5);

    /**
     *  \brief destructor
     */
    virtual ~BoolObsMap();

    /**
     *  \brief return a copy of the current map
     *  \return a pointer to the copy of the current map
     *
     *  A new instance of BoolObsMap is allocated
     */
    virtual BoolObsMap* copy() const;

    /***************** Initialization *****************/

  protected : 
    /*
     *  \brief init tables from their dimesions
     *  \param number of columns of the table
     *  \param number of lines of the table
     */
    void initTables(int nbCols,int nbLines);
    
    /*
     *  \brief delete all the content of the slope map tables
     */
    void clearTables();


    /******************** Accessors *******************/

  protected : 
    /*
     *  \brief set the cell type
     *  \param (i,j) : index of the cell in the grid
     *  \param type : the new cell type
     */
    void setCellType(int i, int j,obs2D_cell_t type);

  public : 
    /**
     *  \brief get the type of cell (i,j)
     *  \param (i,j) : index of the cell in the grid
     *  \return the cell type
     */
    obs2D_cell_t getCellType(int i, int j) const;

    /**
     *  \brief get the type of cell (i,j)
     *  \param (x,y) : position of the cell in the grid
     *  \return the cell type
     */
    obs2D_cell_t getCellType(float x, float y) const;

    /**
     *  \brief return the maximal slope that the robot can climb
     *  \return the maximal slope that the robot can climb
     */
    float thresSlope() const { return this->thresSlope_; }

    /**
     *  \brief return the maximal slope derivative that the robot can climb
     *  \return the maximal slope derivative that the robot can climb
     */
    float thresVarSlope() const { return this->thresVarSlope_; }

    /**
     *  \brief return the minimal distance that the robot may keep to be safe for collisions
     *  \return the minimal distance that the robot may keep to be safe for collisions
     */
    float safeDist() const { return this->safeDist_; }


    /************** Obstacle map building *************/

  protected : 
    /*
     *  \brief thresholding the input slope map to obstain an obstacle map
     *  \param sm : the input slope map
     *  \param thresSlope : the maximal slope that the robot can climb
     */
    void thresSM(const SlopeMap* sm, float thresSlope);

    /*
     *  \brief thresholding the input var slope map to obstain an obstacle map
     *  \param vsm : the input var slope map
     *  \param thresSlope : the maximal slope that the robot can climb
     *  \param thresVarSlope : the maximal slope derivative that the robot can climb
     */
    void thresVSM(const VarSlopeMap* vsm, float thresSlope, float thresVarSlope);

    /*
     *  \brief increase the obstacles with the input safe distance
     *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
     */
    void increaseObs(float safeDist);

    /*
     *  \brief initialize the maximal number of cells considered to increase the obstacles
     *  \param safeDist : the maximal distance considered to increase the obstacles
     */
    void initMaxNbCells(float safeDist);

  public : 
    /**
     *  \brief build the obstacle map from an input slope map
     *  \param sm : the input slope map
     *  \param thresSlope : the maximal slope that the robot can climb
     *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
     *  \return 0 if the process succeed, -1 else
     */
    int buildFromSM(const SlopeMap* sm, float thresSlope=0.25*M_PI, float safeDist=0.5);

    /**
     *  \brief build the obstacle map from an input slope map
     *  \param vsm : the input var slope map
     *  \param thresSlope : the maximal slope that the robot can climb
     *  \param thresVarSlope : the maximal slope derivative that the robot can climb
     *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
     *  \return 0 if the process succeed, -1 else
     */
    int buildFromVSM(const VarSlopeMap* vsm, float thresSlope=0.25*M_PI, float thresVarSlope=5.0*M_PI, float safeDist=0.5);


    /************ Input / Output functions ************/

  protected : 
    /*
     *  \brief init the Z color parameters
     *  \param minColZ : output minimal Z value used to convert the DEM into an image
     *  \param maxColZ : output maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : output Z scale value used to convert the DEM into an image
     */
    virtual void initZ_ColorParams(float &minColZ, float &maxColZ, float &scaleColZ) const;

    /*
     *  \brief return the colour corresponding to a cell of the DEM
     *  \param minColZ : minimal Z value used to convert the DEM into an image
     *  \param maxColZ : maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : Z scale value used to convert the DEM into an image
     *  \return the colour corresponding to a cell of the DEM
     */
    virtual int greyColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ) const;

    /*
     *  \brief return the colour corresponding to a cell of the DEM
     *  \param minColZ : minimal Z value used to convert the DEM into an image
     *  \param maxColZ : maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : Z scale value used to convert the DEM into an image
     *  \param R : the output red colour
     *  \param G : the output green colour
     *  \param B : the output blue colour
     */
    virtual void rgbColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ,int& R,int& G,int& B) const;


    /**************** Utility functions ***************/

  protected : 
    /**
     *  \brief extract a sub-map of the current map
     *  \param iMin : the number of the first column to keep
     *  \param iMax : the number of the last column to keep
     *  \param jMin : the number of the first line to keep
     *  \param jMax : the number of the last line to keep
     *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
     *
     *  A new instance of BoolObsMap is created
     */
    virtual BoolObsMap* subMap(int iMin, int iMax, int jMin, int jMax) const;

  public : 
    /**
     *  \brief display the main features of the DEM
     *  \param dispGrid : display the grid if true
     */
    virtual void display(bool dispGrid) const;

    /**
     *  \brief add a linear obstacle on the map from its end points
     *  \param (x1,y1) : coordinates of the first end point on the map
     *  \param (x2,y2) : coordinates of the second end point on the map
     */
    void addLineObs(float x1, float y1, float x2, float y2);
  };


  

};

#endif // BOOLOBSMAP_HPP
