#ifndef __MAP_HPP
#define __MAP_HPP

#include "motion/inc/Path.hpp"
#include "common/inc/Position.hpp"
#include "environment/inc/Pos.hpp"
#include <cstdlib>

#ifndef _DIR_T_
#define _DIR_T_

// Various directions
typedef enum {
  TMP  = -2, // temporary direction 
  IMP  = -1, // impossible direction
  UND  =  0, // undefined direction
  NW   =  1, // north west
  N    =  2, // north
  NE   =  3, // north east
  E    =  4, // east
  SE   =  5, // south east
  S    =  6, // south
  SW   =  7, // south west
  W    =  8  // west
} dir_t;

#endif

namespace mpc
{
class Pos;

class Map
{
	public:
		// Various map types
		typedef enum {ERR     = -1, // unrecognized type
		              ND      =  0, // undefinied type
		              PGM     =  1, // magic PGM
		              PGM_T   =  2, // thresholded map
		              PGM_TG  =  3, // thresholded and grown map
		              PGM_N   =  4, // standard PGM
			      PGM_M   =  5, // medium map
			      PGM_PFO =  6, // map with obstacles potential fields
			      PGM_PFT =  7, // map with target potential fields
			      PGM_PF  =  8  // map with all potential fields
		} typ_t;

	protected : 
		typ_t typ;
		int w;
		int h;
		int d;
		double* tabC;

	public : 	
		/************* Constructors and destructors ************/
		Map();                                            // default constructor
		Map(const Map&);                                  // copy constructor
		Map(const int wt,const int ht);                   // init constructor
                Map(const int wt,const int ht,const int dph);     // init constructor
		virtual ~Map();                                   // destructor

                /********************** Accessors **********************/
		typ_t getType() const;                            // return map type
		int getWidth() const;                             // return map width
		int getHeight() const;                            // return map height
		int getDepth() const;                             // return color depth of the map
		double getCol(int x,int y) const;                 // return point color at coord (x,y)
                double getCol(Pos p);                             // return point color at position p
		void setType(typ_t f);                            // modify map type
		void setWidth(int wt);                            // modify map width
		void setHeight(int ht);                           // modify map height
                void setCol(int x,int y,double col);              // modify point color at coord (x,y)
		void setCol(Pos p,double col);               // modify point color at position p

                /**************** Map utility functions ****************/
                void initTabC(int lg,int ht);                     // initialize map table
                void display() const;                             // display the table of the map

		/************ Map loading/saving functions *************/
		virtual int load(const char* nomfic);             // Reading function for PGM type map
                                                                  // return 0 if it works, -1 else
		virtual int save(const char* nomfic) const;       // Writting function for PGM type map
                                                                  // return 0 if it works, -1 else

                /*************** Map treatment functions ***************/
		void thresholdPGM(int seuil);                     // Thresholding function for PGM typed map
		void increaseObs(int nbPix);                      // Obstacles growing for thresholded maps
		Map* computeWave(Pos* p);                    // compute a wave on the map from current position
  // Map* applyPFields();                              // Apply potentials fields on the map
  // Map* applyPFields2();                             // Apply potentials fields on the map (2nd version)
		Map* applyPFields3();                             // Apply potentials fields on the map (3rd version)
                static Map* mediumMap(Map* m1,Map* m2);           // Compute medium colour map between two maps
                static Map* addingPFMap(Map* mapTP,Map* mapPFO);  // Compute adding of both potential fields map
                static Map* addingPFMap2(Map* mapTP,Map* mapPFO); // Compute adding of both potential fields map

                /**************** Map merging functions ****************/
                void merge(Map* map2);                            // merge the obstacles of the current map with map2

                /**************** Obstruction functions ****************/
                double isFree(int x,int y) const;                 // return the freedom probability of the position (x,y)
                void setFreedom(int x,int y,double freed);        // set the freedom probability of the position (x,y)
                void setFreedom(Pos p1,Pos p2,double freed);      // set the freedom probability of the line defined between p1 and p2
                void setFreedom(int x1,int y1,int x2,int y2,double freed); // set the freedom probability of the line defined between (x1,y1) and (x2,y2)
                void setBordersFreedom(double freed);             // set the freedom probability of borders of the map
                void setCompletelyFree();                         // set the map completely free
                void setCompletelyObstructed();                   // set the map completely obstructed

                /**
		 *  \brief return true if the map if free at position p
		 *  \return true if the map if free at position p
		 */
		bool isFreeAt(const PositionXY& p) const;

                /**
		 *  \brief return the color of the map at position p
		 *  \return the color of the map at position p
		 */
		double getColAt(const PositionXY& p) const;

                /**************** Path tracing functions ***************/

		/**
		 *  \brief trace a path on the map
		 *  \param p : the input path to trace
		 */
		void tracePath(const CPath& p);

};

};
#endif
