#include "Obs.hpp"
using namespace mpc;

#include <GL/gl.h>
#include <GL/glu.h>

                                    /**************************************************/
                                    /******************* Class Obs ********************/
                                    /**************************************************/

double Obs::thresChange_ = 1.0;                     /**< probability threshold to change its motion after one sec (random motion mode) */
double Obs::maxAngleChange_ = M_PI*0.3;             /**< maximal angle change after one sec (random motion mode) */
double Obs::maxSpdChange_ = 0.05;                   /**< maximal speed change after one sec (random motion mode) */

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
Obs::Obs()
{
  this->moving_ = false;
  this->type_   = UNDEFINED_OBS;
}

/**
 *  \brief copy constructor
 *  \param o : the obstacle to copy
 */
Obs::Obs(const Obs& o)
{
  this->moving_ = o.moving_;
}

/**
 *  \brief destructor
 */
Obs::~Obs()
{
}


                                    /**************************************************/
                                    /****************** Class Obs2D *******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
Obs2D::Obs2D() : Object2D()
{
  this->type_   = UNDEFINED_OBS;
}

/**
 *  \brief init constructor
 *  \param 
 */
Obs2D::Obs2D(const PositionXYThT& post,const SpeedXYTh& spd, const BoundingShape& bs) : Object2D(bs)
{
  this->type_   = UNDEFINED_OBS;
  this->lastPosT_ = post;
  this->lastSpd_ = spd;  
}

/**
 *  \brief copy constructor
 *  \param o : the obstacle to copy
 */
Obs2D::Obs2D(const Obs2D& o) : Obs(o), Object2D(o)
{
  this->type_   = o.type_;
  this->lastPosT_ = o.lastPosT_;
  this->lastSpd_ = o.lastSpd_;
//   this->moving_ = o.moving_;
}

/**
 *  \brief destructor
 */
Obs2D::~Obs2D()
{
}


/******************** Accessors *******************/

/**
 *  \brief return the forecast position of the obstacle at time t
 *  \param t : the input time at which the position of the obstacle is evaluated
 *  \return the position forecast at time t
 */
PositionXYThT Obs2D::forecastPosT(double t) const
{
  double dt = t - this->lastPosT().t();
//   cout << "dt:" << dt << "  t:" << t << "  lastPosT:"; lastPosT().display(); cout << endl;
  if(dt >= 0)
    {
      double x = this->lastPosT().x() + dt * this->lastSpd().x();
      double y = this->lastPosT().y() + dt * this->lastSpd().y();
      double th = normalizedAngle(this->lastPosT().th() + dt * this->lastSpd().th());
//       return PositionXYThT(this->lastPosT().x(),this->lastPosT().y(),th,t);
      return PositionXYThT(x,y,th,t);
    }
  else
    {
      cout << "Obs2D::forecastPosT : Request of a forecast position-time in the past... dt:" << dt << endl; 
      return this->lastPosT();
//       throw string("Obs2D::forecastPosT : Request of a forecast position-time in the past");
    }
}

/**
 *  \brief register a new position-time of the obstacle
 *  \param post : the new position-time perceived
 */
void Obs2D::registerPosT(const PositionXYThT& post,const SpeedXYTh& spd)
{
  this->lastPosT() = post;
  this->lastSpd() = spd;
}


/*************** Environment update ***************/

/**
 *  \brief simulate a random motion of all dynamic obstacles
 *  \param the time at which the environment is evaluated
 */
void Obs2D::simuRandMotion(double t)
{
  // check if the obstacle is moving
  if(moving_)
    {
      // get the last time at which the obstacle has been seen
      double lastT = lastPosT_.t();

      // compute the elapsed time from the previous update
      double elapsedT = t - lastT;

      // check if time are consistent
      if(elapsedT < 0)
	{
	  cerr << "Obs2D::simuRandMotion : unconsistent elapsed time..." << endl;
	  return;
	}

      double rdmChgt = getRandom();

      // compute the obstacle motion change if needed
      if(rdmChgt > 0.5)
      //if(rdmChgt > Obs::thresChange_*elapsedT)
	{
	  // compute the old norm of the speed
	  double oldNormSpd = sqrt(lastSpd_.x()*lastSpd_.x() + lastSpd_.y()*lastSpd_.y());
	  // compute the old angle of the speed
	  double oldAngleSpd = atan2(lastSpd_.y(),lastSpd_.x());

	  // get the norm changement factor
	  double normFactor = getRandomM11();
	  // get the angle changement factor
	  double angleFactor = getRandomM11();

	  // compute the new norm of the speed
	  double newNormSpd = (1.0 + normFactor*Obs::maxSpdChange_*elapsedT) * oldNormSpd;
	    
	  // compute the new angle of the speed
	  double newAngleSpd = oldAngleSpd + normalizedAngle(angleFactor*Obs::maxAngleChange_*elapsedT);

	  //compute the new speed
	  lastSpd_.x() = cos(newAngleSpd)*newNormSpd;
	  lastSpd_.y() =  sin(newAngleSpd)*newNormSpd;
	}

      // compute the new position of the obstacle
      lastPosT_.x() += lastSpd_.x() * elapsedT;
      lastPosT_.y() += lastSpd_.y() * elapsedT;
      lastPosT_.t() += elapsedT;
    }
  else
    {
      // non-moving obstacle : update the last 
      // time at which the obstacle has been seen
      lastPosT_.t() = t;
    }

//   cout << "Obs2D::simuRandMotion -> t : " << t << "  lastPosT_ : "; lastPosT_.display(); cout << endl;
}


/*************** Sampling functions ***************/

/**
 *  \brief sample the obstructed positions of the obstacle of the static environment
 *  \param map : the occupancy grid on which the obstacle must be sampled
 *  \param sampleX,sampleY : sampling scale of the resulting map
 *  \param minY,maxX,minY,maxY : the bounds of the window in which the obstacle must be sampled
 *  \return an occupancy grid of the environment
 *
 *  A new instance of occupancy grid is returned
 */
void Obs2D::sample(Map* map,double minX,double maxX,double minY,double maxY,int sampleX,int sampleY) const
{
//   for(int i=0; i<sampleX; i++)
//     {
//       double x = minX + (i + 0.5) * ((maxX-minX) / ((double)sampleX));
      
//       for(int j=0; j<sampleY; j++)
// 	{
// 	  double y = minY + (j + 0.5) * ((maxY-minY) / ((double)sampleY));

// 	  // create a configuration from the current x and y
// 	  PositionXY curPos(x,y);
// 	  curPos.display(); cout << " ";

// 	  // check if the obstacle obstructs the current position of the map
// 	  if(this->obstruct(curPos))
// 	    {
// 	      cout << "obstruct!!!!!!!!" << endl;
// 	      map->setFreedom(i,j,0.0);
// 	    }
// 	  else
// 	    map->setFreedom(i,j,1.0);
// 	}
//       cout << endl;
//     }
//   return map;
}


                                    /**************************************************/
                                    /**************** Class DiskObs2D *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
DiskObs2D::DiskObs2D() : Obs2D()
{
  this->moving_ = false;
  this->type_ = DISK_OBS2D;
}
  
/**
 *  \brief init constructor
 *  \param post : the initial position-time of the obstacle
 *  \param spd : the initial speed of the obstacle
 *  \param rad : the radius of the disk obstacle
 */
DiskObs2D::DiskObs2D(const PositionXYT& post,const SpeedXY& spd,double rad) : Obs2D()
{
  this->type_ = DISK_OBS2D;
  this->lastPosT() = PositionXYThT(post.x(),post.y(),0,post.t());
  this->lastSpd() = SpeedXYTh(spd.x(),spd.y(),0);
  this->bshape_ = new BDisk(rad);

  if(spd.x() != 0 || spd.y() != 0) this->moving_ = true;
  else this->moving_ = false;
}

/**
 *  \brief copy constructor
 *  \param o : the obstacle to copy
 */
DiskObs2D::DiskObs2D(const DiskObs2D& o) : Obs2D(o)
{
  this->bshape_ = new BDisk(o.bshape());
}

/**
 *  \brief destructor
 */
DiskObs2D::~DiskObs2D()
{
}

/**
 *  \brief return a pointer on a copy of the current obstacle
 *  \return a pointer on a copy of the current obstacle
 */
DiskObs2D* DiskObs2D::copy() const
{
  return new DiskObs2D(*this);
}

/**
 *  \brief return a pointer on a copy of the current obstacle
 *  \param frame : the frame in which we want to express the coordinates of the obstacle
 *  \return a pointer on a copy of the current obstacle
 */
DiskObs2D* DiskObs2D::copyFromFrame(const PositionXYTh& frame) const
{
  // compute the position of the obstacle in the new frame
  PositionXYThT oldPosT = this->lastPosT(); 
  PositionXYTh posFromF = coordFromFrame(oldPosT,frame);
  PositionXYT posTfromF(posFromF.x(),posFromF.y(),oldPosT.t());

  // compute the speed of the obstacle in the new frame
  SpeedXY spd = this->lastSpd();
  spd.rotate(frame.th());

  // copy the radius of the obstacle
  double rad = this->radius();

  // construct the new obstacle
  DiskObs2D* obs = new DiskObs2D(posTfromF,spd,rad);

  // return it
  return obs;
}

/**
 *  \brief return a pointer on a copy of the current obstacle
 *  \param frame : the frame in which we want to express the coordinates of the obstacle
 *  \return a pointer on a copy of the current obstacle
 */
DiskObs2D* DiskObs2D::copyFromFrame(const PositionXYThT& frame) const
{
  // compute the position of the obstacle in the new frame
  PositionXYThT oldPosT = this->lastPosT(); 
  PositionXYTh posFromF = coordFromFrame(oldPosT,frame);
  PositionXYT posTfromF(posFromF.x(),posFromF.y(),frame.t());

  // compute the speed of the obstacle in the new frame
  SpeedXY spd = this->lastSpd();
  spd.rotate(frame.th());

  // copy the radius of the obstacle
  double rad = this->radius();

  // construct the new obstacle
  DiskObs2D* obs = new DiskObs2D(posTfromF,spd,rad);

//   cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
//   cout << "cur : "; this->display();
//   cout << "obs : "; this->display();
//   cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;

  // return it
  return obs;
}

/**
 *  \brief return a pointer on a copy of the current obstacle
 *  \param frame : the frame in which we want to express the coordinates of the obstacle
 *  \return a pointer on a copy of the current obstacle
 */
DiskObs2D* DiskObs2D::copyInFrame(const PositionXYTh& frame) const
{
  // compute the position of the obstacle in the new frame
  PositionXYThT oldPosT = this->lastPosT(); 
  PositionXYTh posInF = coordInFrame(oldPosT,frame);
//   PositionXYT posTinF(posInF.x(),posInF.y(),oldPosT.t() + frame.t());
  PositionXYT posTinF(posInF.x(),posInF.y(),oldPosT.t());

  // compute the speed of the obstacle in the new frame
  SpeedXY spd = this->lastSpd();
  spd.rotate(-frame.th());

  // copy the radius of the obstacle
  double rad = this->radius();

  // construct the new obstacle
  DiskObs2D* obs = new DiskObs2D(posTinF,spd,rad);

  // return it
  return obs;
}

/**
 *  \brief return a pointer on a copy of the current obstacle
 *  \param frame : the frame in which we want to express the coordinates of the obstacle
 *  \return a pointer on a copy of the current obstacle
 */
DiskObs2D* DiskObs2D::copyInFrame(const PositionXYThT& frame) const
{
  // compute the position of the obstacle in the new frame
  PositionXYThT oldPosT = this->lastPosT(); 
  PositionXYTh posInF = coordInFrame(oldPosT,frame);
//   PositionXYT posTinF(posInF.x(),posInF.y(),oldPosT.t() + frame.t());
  PositionXYT posTinF(posInF.x(),posInF.y(),frame.t());

  // compute the speed of the obstacle in the new frame
  SpeedXY spd = this->lastSpd();
  spd.rotate(-frame.th());

  // copy the radius of the obstacle
  double rad = this->radius();

  // construct the new obstacle
  DiskObs2D* obs = new DiskObs2D(posTinF,spd,rad);

  // return it
  return obs;
}


/************* Obstruction functions **************/

/**
 *  \brief test if the environment obstruct the position pos
 *  \param pos : the position to test
 *  \param t : the input time at which to test collisions
 */
bool DiskObs2D::obstruct(const PositionXY& pos,double t) const
{
  double dist;
  PositionXY near;
  PositionXY nearP2;
  bool res = areInCollision(this->forecastPosT(t), this->bshape(), pos, PointShape(), near, nearP2, dist);
  return res;
}

/**
 *  \brief test if the environment obstruct the position pos at time t
 *  \param pos : the position to test
 *  \param t : the input time at which to test collisions
 *  \param near : the nearest position of the obstacle found
 *  \param dist : the minimal distance between pos and the nearest obstacle
 */
bool DiskObs2D::obstruct(const PositionXY& pos,double t,PositionXY& near,double& dist) const
{
  PositionXY nearP2;
  bool res = areInCollision(this->forecastPosT(t), this->bshape(), pos, PointShape(), near, nearP2, dist);
  return res;
}

/**
 *  \brief test if the environment obstruct the segment p1p2
 *  \param p1 : the first end position of the segment
 *  \param p2 : the second end position of the segment
 *  \param t : the input time at which to test collisions
 */
bool DiskObs2D::obstruct(const PositionXY& p1,const PositionXY& p2,double t) const
{
  PositionXY itsc1;
  PositionXY itsc2; 
  double dist;
  int nbItsc = areInCollision(this->forecastPosT(t), this->bshape(), p1, p2, itsc1, itsc2, dist);
  if(nbItsc == 0) return false;
  else return true;
}

/**
 *  \brief test if the environment obstruct the segment p1p2
 *  \param p1 : the first end position of the segment
 *  \param p2 : the second end position of the segment
 *  \param t : the input time at which to test collisions
 *  \param near : the nearest position of the obstacle found
 *  \param dist : the minimal distance between pos and the nearest obstacle
 */
bool DiskObs2D::obstruct(const PositionXY& p1,const PositionXY& p2,double t,PositionXY& near,double& dist) const
{
  cerr << "DiskObs2D::obstruct(segment,near) : Not implemented yet!" << endl;
  exit(1);
}


// /*************** Sampling functions ***************/

// /**
//  *  \brief sample the obstructed positions of the obstacle of the static environment
//  *  \param sampleX,sampleY : sampling scale of the resulting map
//  *  \param minY,maxX,minY,maxY : the bounds of the window in which the obstacle must be sampled
//  *  \return an occupancy grid of the environment
//  *
//  *  A new instance of occupancy grid is returned
//  */
// virtual Map* DiskObs2D::sample(double minX,double maxX,double minY,double maxY,int sampleX,int sampleY) const
// {
//   cerr << "DiskObs2D::sample : Not implemented yet!" << endl;
//   exit(1);
// }


/*************** Display functions ****************/
  
/**
 *  \brief display the object's features
 */
void DiskObs2D::display() const
{
  cout << "Disk Obstacle 2D -> rad : " << this->bshape().radius() << ",  PosT : "; 
  lastPosT().display(); cout << ",  Spd : "; lastSpd().display(); cout << endl;
}

/**
 *  \brief draw the object in an openGL viewer
 *  \param dispFM : true if the forecast model must be displayed
 */
void DiskObs2D::toGL(bool dispFM) const
{
  const GLfloat redMaterial[]={1.,0.,0.,1.};
  const GLfloat greenMaterial[]={0.,1.,0.,1.};
//   const GLfloat greenTsMaterial[]={0.,1.,0.,0.2};

  PositionXYThT posT = this->lastPosT();

  double radObs = this->bshape().radius(); 
  
  if(dispFM)
    {
      int nbSide = 10;      

      PositionXYThT finalPosT = this->forecastPosT(posT.t() + 20.0);

//       cout << "//////////////////////////////////" << endl;
//       cout << "lastPosT : "; posT.display(); 
//       cout << "  speed : "; lastSpd().display();
//       cout << "  finalPosT : "; finalPosT.display(); cout << endl;

      glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);
		  	      
      glPushMatrix();
      glTranslated(posT.x(),posT.y(),posT.t());
      glNormal3f(0,0,-1);
      gluDisk(gluNewQuadric(), 0, radObs, nbSide, 1);
      glPopMatrix();
      
      glPushMatrix();
      glTranslated(finalPosT.x(),finalPosT.y(),finalPosT.t());
      glNormal3f(0,0,1);
      gluDisk(gluNewQuadric(), 0, radObs, nbSide, 1);
      glPopMatrix();
	      
      for(int j=0; j<nbSide; j++)
	{
	  glBegin(GL_QUADS);
	  PositionXYT v1(posT.x() + cos(j*2*M_PI/((double)nbSide)) * radObs,
			 posT.y() + sin(j*2*M_PI/((double)nbSide)) * radObs,
			 posT.t());
	      
	  PositionXYT v2(posT.x() + cos(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
			 posT.y() + sin(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
			 posT.t());
	      
	  PositionXYT v3(finalPosT.x() + cos(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
			 finalPosT.y() + sin(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
			 finalPosT.t());
	      
	  PositionXYT v4(finalPosT.x() + cos(j*2*M_PI/((double)nbSide)) * radObs,
			 finalPosT.y() + sin(j*2*M_PI/((double)nbSide)) * radObs,
			 finalPosT.t());
	      
	  PositionXYT v1v2 = (v2 - v1);
	  double normV1V2 = sqrt(v1v2.x()*v1v2.x() + v1v2.y()*v1v2.y() + v1v2.t()*v1v2.t());
	  PositionXYT v1v2n;
	  if(normV1V2 != 0) v1v2n = PositionXYT(v1v2.x()/normV1V2,v1v2.y()/normV1V2,v1v2.t()/normV1V2);

	  PositionXYT v1v4 = (v4 - v1);
	  double normV1V4 = sqrt(v1v4.x()*v1v4.x() + v1v4.y()*v1v4.y() + v1v4.t()*v1v4.t());
	  PositionXYT v1v4n;
	  if(normV1V4 != 0) v1v4n = PositionXYT(v1v4.x()/normV1V4,v1v4.y()/normV1V4,v1v4.t()/normV1V4);

	  PositionXYT n = v1v2n^v1v4n;
	  
	  glNormal3f(n.x(),n.y(),n.t());
	  glVertex3f(v1.x(), v1.y(), v1.t());
	  glNormal3f(n.x(),n.y(),n.t());
	  glVertex3f(v2.x(), v2.y(), v2.t());
	  glNormal3f(n.x(),n.y(),n.t());
	  glVertex3f(v3.x(), v3.y(), v3.t());
	  glNormal3f(n.x(),n.y(),n.t());
	  glVertex3f(v4.x(), v4.y(), v4.t());
	      
	  glEnd();
	}
    }
  else
    {
      // display obstacles only on the current time
      glMaterialfv(GL_FRONT, GL_AMBIENT, greenMaterial);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, greenMaterial);
      
      glPushMatrix();
      glTranslated(posT.x(),posT.y(),posT.t());
      gluDisk(gluNewQuadric(), 0, radObs, 20, 1);
      gluCylinder(gluNewQuadric(), radObs, radObs, 0.05*radObs, 20, 20);
      glTranslated(0,0,radObs*0.05);
      gluDisk(gluNewQuadric(), 0, radObs, 20, 1);
      glPopMatrix();
    }
}

/**
 *  \brief draw the object in an openGL viewer
 *  \param frame : the frame of the perception system from which is construct the environment model
 *  \param dispFM : true if the forecast model must be displayed
 */
void DiskObs2D::toGL(const PositionXYTh& frame,bool dispFM) const
// void DiskObs2D::toGL(const PositionXYThT& frame,bool dispFM) const
{
  PositionXYThT frameT(frame.x(),frame.y(),frame.th(),0);
  this->toGL(frameT,dispFM);
}

/**
 *  \brief draw the object in an openGL viewer
 *  \param frame : the frame of the perception system from which is construct the environment model
 *  \param dispFM : true if the forecast model must be displayed
 */
void DiskObs2D::toGL(const PositionXYThT& frame,bool dispFM) const
{
  const GLfloat redMaterial[]={1.,0.,0.,1.};
  const GLfloat greenMaterial[]={0.,1.,0.,1.};
//   const GLfloat greenTsMaterial[]={0.,1.,0.,0.2};

  PositionXYThT posT = this->lastPosT();
  double radObs = this->bshape().radius(); 
//   cout << "posT : "; posT.display(); cout << endl;
  
  PositionXYTh posInF = coordFromFrame(posT,frame);
  PositionXYThT posTinF(posInF.x(),posInF.y(),posInF.th(),posT.t());

  if(dispFM)
    {
      int nbSide = 10;
	      
      PositionXYThT finalPosT = this->forecastPosT(posT.t() + 20.0);

      glMaterialfv(GL_FRONT, GL_AMBIENT, redMaterial);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, redMaterial);

      PositionXYTh finalPosInF = coordFromFrame(finalPosT,frame);
      PositionXYThT finalPosTinF(finalPosInF.x(),finalPosInF.y(),finalPosInF.th(),finalPosT.t());

//       cout << "//////////////////////////////////" << endl;
//       cout << "lastPosT : "; posT.display(); 
//       cout << "  speed : "; lastSpd().display();
//       cout << "  finalPosT : "; finalPosT.display(); cout << endl;
//       cout << "posTinF : "; posTinF.display(); 
//       cout << "  finalPosTinF : "; finalPosTinF.display(); cout << endl;
//       cout << "  speed2 : (" << (finalPosT.x() - posT.x())/20.0 << "," << (finalPosT.y() - posT.y())/20.0 << ")" << endl;
//       cout << "  speed2F : (" << (finalPosTinF.x() - posTinF.x())/20.0 << "," << (finalPosTinF.y() - posTinF.y())/20.0 << ")" << endl;

      glPushMatrix();
      glTranslated(posTinF.x(),posTinF.y(),posTinF.t());
      glNormal3f(0,0,-1);
      gluDisk(gluNewQuadric(), 0, radObs, nbSide, 1);
      glPopMatrix();
	      
      glPushMatrix();
      glTranslated(finalPosTinF.x(),finalPosTinF.y(),finalPosTinF.t());
      glNormal3f(0,0,1);
      gluDisk(gluNewQuadric(), 0, radObs, nbSide, 1);
      glPopMatrix();
	      
      for(int j=0; j<nbSide; j++)
	{
	  glBegin(GL_QUADS);
	  PositionXYT v1(posTinF.x() + cos(j*2*M_PI/((double)nbSide)) * radObs,
			 posTinF.y() + sin(j*2*M_PI/((double)nbSide)) * radObs,
			 posTinF.t());
	      
	  PositionXYT v2(posTinF.x() + cos(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
			 posTinF.y() + sin(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
			 posTinF.t());
	      
	  PositionXYT v3(finalPosTinF.x() + cos(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
			 finalPosTinF.y() + sin(((j+1)%nbSide)*2*M_PI/((double)nbSide)) * radObs,
			 finalPosTinF.t());
	      
	  PositionXYT v4(finalPosTinF.x() + cos(j*2*M_PI/((double)nbSide)) * radObs,
			 finalPosTinF.y() + sin(j*2*M_PI/((double)nbSide)) * radObs,
			 finalPosTinF.t());
	      
	  PositionXYT v1v2 = (v2 - v1);
	  double normV1V2 = sqrt(v1v2.x()*v1v2.x() + v1v2.y()*v1v2.y() + v1v2.t()*v1v2.t());
	  PositionXYT v1v2n;
	  if(normV1V2 != 0) v1v2n = PositionXYT(v1v2.x()/normV1V2,v1v2.y()/normV1V2,v1v2.t()/normV1V2);

	  PositionXYT v1v4 = (v4 - v1);
	  double normV1V4 = sqrt(v1v4.x()*v1v4.x() + v1v4.y()*v1v4.y() + v1v4.t()*v1v4.t());
	  PositionXYT v1v4n;
	  if(normV1V4 != 0) v1v4n = PositionXYT(v1v4.x()/normV1V4,v1v4.y()/normV1V4,v1v4.t()/normV1V4);

	  PositionXYT n = v1v2n^v1v4n;
	  
	  glNormal3f(n.x(),n.y(),n.t());
	  glVertex3f(v1.x(), v1.y(), v1.t());
	  glNormal3f(n.x(),n.y(),n.t());
	  glVertex3f(v2.x(), v2.y(), v2.t());
	  glNormal3f(n.x(),n.y(),n.t());
	  glVertex3f(v3.x(), v3.y(), v3.t());
	  glNormal3f(n.x(),n.y(),n.t());
	  glVertex3f(v4.x(), v4.y(), v4.t());
	      
	  glEnd();
	}
    }
  else
    {
      // display obstacles only on the current time
      glMaterialfv(GL_FRONT, GL_AMBIENT, greenMaterial);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, greenMaterial);
      
      glPushMatrix();
      glTranslated(posTinF.x(),posTinF.y(),posTinF.t());
      gluDisk(gluNewQuadric(), 0, radObs, 20, 1);
      gluCylinder(gluNewQuadric(), radObs, radObs, radObs*0.05, 20, 20);
      glTranslated(0,0,radObs*0.05);
      gluDisk(gluNewQuadric(), 0, radObs, 20, 1);
      glPopMatrix();
    }
}

                                    /**************************************************/
                                    /**************** Class BBoxObs2D *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
BBoxObs2D::BBoxObs2D()
{
  this->type_ = BOX_OBS2D;
  this->lastPosT() = PositionXYThT(0,0,0,0);
  this->lastSpd() = SpeedXYTh(0,0,0);
  this->bshape_ = new BBox(0,0);
  this->moving_ = false;
}
  
/**
 *  \brief init constructor
 *  \param post : the initial position-time of the obstacle
 *  \param spd : the initial speed of the obstacle
 *  \param lg : the length of the box
 *  \param wd : the width of the box
 */
BBoxObs2D::BBoxObs2D(const PositionXYThT& post,const SpeedXYTh& spd,double lg,double wd)
{
  this->type_ = BOX_OBS2D;
  this->lastPosT() = PositionXYThT(post.x(),post.y(),post.th(),post.t());
  this->lastSpd() = SpeedXYTh(spd.x(),spd.y(),spd.th());
  this->bshape_ = new BBox(lg,wd);

  if(spd.x() != 0 || spd.y() != 0 || spd.th() != 0) this->moving_ = true;
  else this->moving_ = false;
}

/**
 *  \brief init constructor (from a segment and a secure distance)
 *  \param p1 : the first end of the segment
 *  \param p2 : the second end of the segment
 *  \param dist : the secure distance from the segment to the boundaries of the box
 *  \param t : the initial time at which is defined the obstacle
 *  \param spd : the initial speed of the obstacle
 */
BBoxObs2D::BBoxObs2D(const PositionXY& p1,const PositionXY& p2,double dist,double t,const SpeedXYTh& spd)
{
  this->type_ = BOX_OBS2D;
  this->lastPosT() = PositionXYThT(0.5*(p1.x()+p2.x()),
				   0.5*(p1.y()+p2.y()),
				   atan2(p2.y()-p1.y(),p2.x()-p1.x()),
				   t);
  this->lastSpd() = SpeedXYTh(spd.x(),spd.y(),spd.th());

  double dx = p2.x()-p1.x();
  double dy = p2.y()-p1.y();
  this->bshape_ = new BBox(sqrt(dx*dx + dy*dy) + dist,dist);

  if(spd.x() != 0 || spd.y() != 0 || spd.th() != 0) this->moving_ = true;
  else this->moving_ = false;
}
  
/**
 *  \brief copy constructor
 *  \param o : the obstacle to copy
 */
BBoxObs2D::BBoxObs2D(const BBoxObs2D& o) : Obs2D(o)
{
  this->bshape_ = new BBox(o.bshape());
}

/**
 *  \brief destructor
 */
BBoxObs2D::~BBoxObs2D()
{
}

/**
 *  \brief return a pointer on a copy of the current obstacle
 *  \return a pointer on a copy of the current obstacle
 */
BBoxObs2D* BBoxObs2D::copy() const
{
  return new BBoxObs2D(*this);
}

/**
 *  \brief return a pointer on a copy of the current obstacle
 *  \param frame : the frame in which we want to express the coordinates of the obstacle
 *  \return a pointer on a copy of the current obstacle
 */
BBoxObs2D* BBoxObs2D::copyFromFrame(const PositionXYTh& frame) const
{
  // compute the position of the obstacle in the new frame
  PositionXYThT oldPosT = this->lastPosT(); 
  PositionXYTh posFromF = coordFromFrame(oldPosT,frame);
  PositionXYThT posTfromF(posFromF.x(),posFromF.y(),posFromF.th(),oldPosT.t());

  // compute the speed of the obstacle in the new frame
  SpeedXYTh spd = this->lastSpd();
  spd.rotate(frame.th());

  // copy the radius of the obstacle
  double lg = this->lg();
  double wd = this->wd();

  // construct the new obstacle
  BBoxObs2D* obs = new BBoxObs2D(posTfromF,spd,lg,wd);

  // return it
  return obs;
}

/**
 *  \brief return a pointer on a copy of the current obstacle
 *  \param frame : the frame in which we want to express the coordinates of the obstacle
 *  \return a pointer on a copy of the current obstacle
 */
BBoxObs2D* BBoxObs2D::copyFromFrame(const PositionXYThT& frame) const
{
  // compute the position of the obstacle in the new frame
  PositionXYThT oldPosT = this->lastPosT(); 
  PositionXYTh posFromF = coordFromFrame(oldPosT,frame);
  PositionXYThT posTfromF(posFromF.x(),posFromF.y(),posFromF.th(),frame.t());

  // compute the speed of the obstacle in the new frame
  SpeedXYTh spd = this->lastSpd();
  spd.rotate(frame.th());

  // copy the radius of the obstacle
  double lg = this->lg();
  double wd = this->wd();

  // construct the new obstacle
  BBoxObs2D* obs = new BBoxObs2D(posTfromF,spd,lg,wd);

  // return it
  return obs;
}

/**
 *  \brief return a pointer on a copy of the current obstacle
 *  \param frame : the frame in which we want to express the coordinates of the obstacle
 *  \return a pointer on a copy of the current obstacle
 */
BBoxObs2D* BBoxObs2D::copyInFrame(const PositionXYTh& frame) const
{
  // compute the position of the obstacle in the new frame
  PositionXYThT oldPosT = this->lastPosT(); 
  PositionXYTh posInF = coordInFrame(oldPosT,frame);
//   PositionXYThT posTinF(posInF.x(),posInF.y(),posInF.th(),oldPosT.t() + frame.t());
  PositionXYThT posTinF(posInF.x(),posInF.y(),posInF.th(),oldPosT.t());

  // compute the speed of the obstacle in the new frame
  SpeedXYTh spd = this->lastSpd();
  spd.rotate(-frame.th());

  // copy the radius of the obstacle
  double lg = this->lg();
  double wd = this->wd();

  // construct the new obstacle
  BBoxObs2D* obs = new BBoxObs2D(posTinF,spd,lg,wd);

  // return it
  return obs;
}

/**
 *  \brief return a pointer on a copy of the current obstacle
 *  \param frame : the frame in which we want to express the coordinates of the obstacle
 *  \return a pointer on a copy of the current obstacle
 */
BBoxObs2D* BBoxObs2D::copyInFrame(const PositionXYThT& frame) const
{
  // compute the position of the obstacle in the new frame
  PositionXYThT oldPosT = this->lastPosT(); 
  PositionXYTh posInF = coordInFrame(oldPosT,frame);
//   PositionXYThT posTinF(posInF.x(),posInF.y(),posInF.th(),oldPosT.t() + frame.t());
  PositionXYThT posTinF(posInF.x(),posInF.y(),posInF.th(),frame.t());

  // compute the speed of the obstacle in the new frame
  SpeedXYTh spd = this->lastSpd();
  spd.rotate(-frame.th());

  // copy the radius of the obstacle
  double lg = this->lg();
  double wd = this->wd();

  // construct the new obstacle
  BBoxObs2D* obs = new BBoxObs2D(posTinF,spd,lg,wd);

  // return it
  return obs;
}


/************* Obstruction functions **************/

/**
 *  \brief test if the environment obstruct the position pos
 *  \param pos : the position to test
 *  \param t : the input time at which to test collisions
 */
bool BBoxObs2D::obstruct(const PositionXY& pos,double t) const
{
  PositionXY near;
  PositionXY nearP2;
  double dist;
  bool res = areInCollision(this->forecastPosT(t), this->bshape(), pos, PointShape(), near, nearP2, dist);
  return res;
}

/**
 *  \brief test if the environment obstruct the position pos
 *  \param pos : the position to test
 *  \param t : the input time at which to test collisions
 *  \param near : the nearest position of the obstacle found
 *  \param dist : the minimal distance between pos and the nearest obstacle
 */
bool BBoxObs2D::obstruct(const PositionXY& pos,double t,PositionXY& near,double& dist) const
{
  PositionXY nearP2;
  bool res = areInCollision(this->forecastPosT(t), this->bshape(), PositionXYTh(pos.x(),pos.y(),0), PointShape(), near, nearP2, dist);
  return res;
}

/**
 *  \brief test if the environment obstruct the segment p1p2
 *  \param p1 : the first end position of the segment
 *  \param p2 : the second end position of the segment
 *  \param t : the input time at which to test collisions
 */
bool BBoxObs2D::obstruct(const PositionXY& p1,const PositionXY& p2,double t) const
{
  PositionXY itsc1;
  PositionXY itsc2; 
  int nbItsc = areInCollision(this->forecastPosT(t), this->bshape(), p1, p2, itsc1, itsc2);
  if(nbItsc == 0) return false;
  else return true;
}

/**
 *  \brief test if the environment obstruct the segment p1p2
 *  \param p1 : the first end position of the segment
 *  \param p2 : the second end position of the segment
 *  \param t : the input time at which to test collisions
 *  \param near : the nearest position of the obstacle found
 *  \param dist : the minimal distance between pos and the nearest obstacle
 */
bool BBoxObs2D::obstruct(const PositionXY& p1,const PositionXY& p2,double t,PositionXY& near,double& dist) const
{
  cerr << "BBoxObs2D::obstruct(segment,nearPos) : Not implemented yet!" << endl;
  exit(1);
}


/*************** Sampling functions ***************/

/**
 *  \brief sample the obstructed positions of the obstacle of the static environment
 *  \param map : the occupancy grid on which the obstacle must be sampled
 *  \param sampleX,sampleY : sampling scale of the resulting map
 *  \param minY,maxX,minY,maxY : the bounds of the window in which the obstacle must be sampled
 */
void BBoxObs2D::sample(Map* map,double minX,double maxX,double minY,double maxY,int sampleX,int sampleY) const
{
  // compute the homothety coefficients
  double hX = ((double)sampleX) / (maxX-minX);
  double hY = ((double)sampleY) / (maxY-minY);
  PositionXY hCoef(hX,hY);
  // compute the homothety center
  PositionXY hCenter(minX,minY);

  // get back the four corners of the box
  PositionXY ul = this->bshape().getULCorner(lastPosT());
  PositionXY ur = this->bshape().getURCorner(lastPosT());
  PositionXY ll = this->bshape().getLLCorner(lastPosT());
  PositionXY lr = this->bshape().getLRCorner(lastPosT());
  
  PositionXY ulH = mpc::homothety(ul,hCenter,hCoef);
  PositionXY urH = mpc::homothety(ur,hCenter,hCoef);
  PositionXY llH = mpc::homothety(ll,hCenter,hCoef);
  PositionXY lrH = mpc::homothety(lr,hCenter,hCoef);

  // draw each corner if it belongs to the output map
  map->setFreedom((int)(ulH.x()),(int)(ulH.y()),(int)(urH.x()),(int)(urH.y()),0.0);
  map->setFreedom((int)(urH.x()),(int)(urH.y()),(int)(lrH.x()),(int)(lrH.y()),0.0);
  map->setFreedom((int)(lrH.x()),(int)(lrH.y()),(int)(llH.x()),(int)(llH.y()),0.0);
  map->setFreedom((int)(llH.x()),(int)(llH.y()),(int)(ulH.x()),(int)(ulH.y()),0.0);

//   if(ulH.x() >= 0 && ulH.x() < sampleX && ulH.y() >= 0 && ulH.y() < sampleY) map->setFreedom((int)(ulH.x()),(int)(ulH.y()),0.0);
//   if(urH.x() >= 0 && urH.x() < sampleX && urH.y() >= 0 && urH.y() < sampleY) map->setFreedom((int)(urH.x()),(int)(urH.y()),0.0);
//   if(llH.x() >= 0 && llH.x() < sampleX && llH.y() >= 0 && llH.y() < sampleY) map->setFreedom((int)(llH.x()),(int)(llH.y()),0.0);
//   if(lrH.x() >= 0 && lrH.x() < sampleX && lrH.y() >= 0 && lrH.y() < sampleY) map->setFreedom((int)(lrH.x()),(int)(lrH.y()),0.0);
}


/*************** Display functions ****************/
  
/**
 *  \brief display the object's features
 */
void BBoxObs2D::display() const
{
  cout << "BBox Obstacle 2D -> lg : " << this->bshape().lg() 
       << "  wd : " << this->bshape().wd() << "  PosT : ";
  lastPosT().display(); 
  cout << "  Spd : "; 
  lastSpd().display(); 
  cout << endl;
}

/**
 *  \brief draw the object in an openGL viewer
 *  \param dispFM : true if the forecast model must be displayed
 */
void BBoxObs2D::toGL(bool dispFM) const
{
  double t = lastPosT().t();

  PositionXY ul = this->bshape().getULCorner(lastPosT());
  PositionXY ur = this->bshape().getURCorner(lastPosT());
  PositionXY ll = this->bshape().getLLCorner(lastPosT());
  PositionXY lr = this->bshape().getLRCorner(lastPosT());

  PositionXYT urul(ul.x()-ur.x(),ul.y()-ur.y(),0);
  PositionXYT ulll(ll.x()-ul.x(),ll.y()-ul.y(),0);
  PositionXYT lllr(lr.x()-ll.x(),lr.y()-ll.y(),0);
  PositionXYT lrur(ur.x()-lr.x(),ur.y()-lr.y(),0);
  PositionXYT vt(0.0,0.0,1.0);

  PositionXYT nTop = urul^vt;
  PositionXYT nLef = ulll^vt;
  PositionXYT nBot = lllr^vt;
  PositionXYT nRig = lrur^vt;

  if(dispFM)
    {
      cerr << "BBoxObs2D::toGL : Displaying of the forecast model not implemented yet!" << endl;
      exit(1);
    }
  else
    {
      glBegin(GL_TRIANGLE_STRIP);
      glNormal3f(nTop.x(),nTop.y(),nTop.t());
      glVertex3f(ul.x(), ul.y(), t + 0.5);
      glVertex3f(ul.x(), ul.y(), t - 0.5);
      glVertex3f(ur.x(), ur.y(), t + 0.5);
      glVertex3f(ur.x(), ur.y(), t - 0.5);
      glEnd();

      glBegin(GL_TRIANGLE_STRIP);
      glNormal3f(nLef.x(),nLef.y(),nLef.t());
      glVertex3f(ll.x(), ll.y(), t + 0.5);
      glVertex3f(ll.x(), ll.y(), t - 0.5);
      glVertex3f(ul.x(), ul.y(), t + 0.5);
      glVertex3f(ul.x(), ul.y(), t - 0.5);
      glEnd();

      glBegin(GL_TRIANGLE_STRIP);
      glNormal3f(nBot.x(),nBot.y(),nBot.t());
      glVertex3f(ul.x(), ul.y(), t + 0.5);
      glVertex3f(ul.x(), ul.y(), t - 0.5);
      glVertex3f(ur.x(), ur.y(), t + 0.5);
      glVertex3f(ur.x(), ur.y(), t - 0.5);
      glEnd();

      glBegin(GL_TRIANGLE_STRIP);
      glNormal3f(nTop.x(),nTop.y(),nTop.t());
      glVertex3f(ul.x(), ul.y(), t + 0.5);
      glVertex3f(ul.x(), ul.y(), t - 0.5);
      glVertex3f(ur.x(), ur.y(), t + 0.5);
      glVertex3f(ur.x(), ur.y(), t - 0.5);
      glEnd();
      
      glBegin(GL_TRIANGLE_STRIP);
      glNormal3f(0.0,0.0,-1.0);
      glVertex3f(ul.x(), ul.y(), t - 0.5);
      glVertex3f(ur.x(), ur.y(), t - 0.5);
      glVertex3f(ll.x(), ll.y(), t - 0.5);
      glVertex3f(lr.x(), lr.y(), t - 0.5);
      glEnd();

      glBegin(GL_TRIANGLE_STRIP);
      glNormal3f(0.0,0.0,1.0);
      glVertex3f(ul.x(), ul.y(), t + 0.5);
      glVertex3f(ur.x(), ur.y(), t + 0.5);
      glVertex3f(ll.x(), ll.y(), t + 0.5);
      glVertex3f(lr.x(), lr.y(), t + 0.5);
      glEnd();
    }
}

/**
 *  \brief draw the object in an openGL viewer
 *  \param frame : the frame of the perception system from which is construct the environment model
 *  \param dispFM : true if the forecast model must be displayed
 */
void BBoxObs2D::toGL(const PositionXYTh& frame,bool dispFM) const
{
  PositionXYThT frameT(frame.x(),frame.y(),frame.th(),0);
  this->BBoxObs2D::toGL(frameT,dispFM);
}

/**
 *  \brief draw the object in an openGL viewer
 *  \param frame : the frame of the perception system from which is construct the environment model
 *  \param dispFM : true if the forecast model must be displayed
 */
void BBoxObs2D::toGL(const PositionXYThT& frame,bool dispFM) const
{
  glPushMatrix();
  glTranslated(frame.x(),frame.y(),frame.t());
  glRotated(radToDeg(frame.th()),0,0,1);
  this->toGL(dispFM);
  glPopMatrix();
}
