#include "RoadMap.hpp"
using namespace mpc;

#include <GL/gl.h>
#include <GL/glu.h>

#include <fstream>
#include <iostream>
using namespace std;


/****************************************************************************************
 *********************************** class RoadNode *************************************
 ****************************************************************************************/

/********** Constructors and destructors **********/

// default constructor
RoadNode::RoadNode()
{
  this->pos_ = new PositionXY(0.0,0.0);
  this->lNeighb_ = new vector<RoadNode*>();
  this->visited_ = false;
  this->E_       = 0;
  this->L_       = 0;
  this->prev_    = NULL;
}

// copy constructor
RoadNode::RoadNode(const RoadNode& n)
{
  this->pos_ = new PositionXY(*(n.pos_));
  this->lNeighb_ = new vector<RoadNode*>();
//   if(n.lNeighb_ != NULL)
//     {
//       int N = n.lNeighb_->size();
//       for(int i=0; i<N; i++)
// 	{
// 	  this->lNeighb_->push_back(n.lNeighb_->at(i));
// 	}
//     }
  this->visited_ = false;
  this->E_       = 0;
  this->L_       = 0;
  this->prev_    = NULL;
}

// init constructor
RoadNode::RoadNode(PositionXY& pos)
{
  this->pos_ = new PositionXY(pos.x(),pos.y());
  this->lNeighb_ = new vector<RoadNode*>();
  this->visited_ = false;
  this->E_       = 0;
  this->L_       = 0;
  this->prev_    = NULL;
}

// init constructor
RoadNode::RoadNode(PositionXY& pos, RoadNode* ngh)
{
  this->pos_ = new PositionXY(pos.x(),pos.y());
  this->lNeighb_ = new vector<RoadNode*>();
  this->lNeighb_->push_back(ngh);
  this->visited_ = false;
  this->E_       = 0;
  this->L_       = 0;
  this->prev_    = NULL;
}

// destructor
RoadNode::~RoadNode()
{
  if(this->pos_ != NULL) delete this->pos_;
  if(this->lNeighb_ != NULL) delete this->lNeighb_;
}


/************** Getters and setters ***************/

// // return the position of the node
// ConfigR2D* RoadNode::getPos() const
// {
//   return this->pos_;
// }

// return the i-th neighbour of the node
RoadNode* RoadNode::getNeighbour(int i) const
{
  if(this->lNeighb_ != NULL && i < this->lNeighb_->size()) return this->lNeighb_->at(i);
  else return NULL;
}

// return the number of neighbours of n
int RoadNode::getNbNeighbour() const
{
  if(this->lNeighb_ != NULL) return this->lNeighb_->size();
  else return -1;
}

// add a new neighbour to the node
void RoadNode::addNeighbour(RoadNode* n)
{
  if(this->lNeighb_ != NULL) this->lNeighb_->push_back(n);
}

// return true if the roadnode has been already visited
bool RoadNode::isVisited() const
{
  return this->visited_;
}

// mark the node as visited
void RoadNode::visit()
{
  this->visited_ = true;
}


/**************** Travel functions ****************/

// reinitialize all parameters of the node for the travel of the roadmap
void RoadNode::reinit()
{
  this->visited_ = false;
  this->E_       = 0;
  this->L_       = 0;
  this->prev_    = NULL;
}

// compute and set the length of the path to the current node 
// from the previous node prev
void RoadNode::computeLfrom(RoadNode* prev)
{
  // compute the length between the current node and the previous one
  PositionXY curPos = this->getPos();
  PositionXY prevPos = prev->getPos();
  double dx = curPos.x() - prevPos.x();
  double dy = curPos.y() - prevPos.y();
  double dist = sqrt(dx*dx + dy*dy);

  // compute and set the length to the current node
  this->L_ = prev->getL() + dist;

  // add prev as the previous node of the current one
  this->prev_ = prev;
}

// compute and set the estimation from the current 
// node to the node goal
void RoadNode::computeEto(RoadNode* goal)
{
  PositionXY curPos = this->getPos();
  PositionXY goalPos = goal->getPos();
  double dx = curPos.x() - goalPos.x();
  double dy = curPos.y() - goalPos.y();
  double dist = sqrt(dx*dx + dy*dy);

  // set the estimation from the current node
  this->E_ = dist;
}

/*************** Display functions ****************/

// display the current node
void RoadNode::display()
{
  cout << "Pos : "; this->getPos().display(); cout << "  nbNgh : " << this->getNbNeighbour() << endl;
  cout << "  --> Prev : ";
  if(this->getPrev() == NULL) cout << "NULL";
  else { this->getPrev()->getPos().display(); }
//   for(int i=0; i<this->getNbNeighbour(); i++)
//     {
//       cout << "  --> Ngh : "; this->getNeighbour(i)->getPos().display(); cout << endl;
//     }
  cout << "  L : " << getL();
  cout << "  E : " << getE();
  cout << "  H : " << getH() << endl;;
}


/****************************************************************************************
 *********************************** class DRoadNode ************************************
 ****************************************************************************************/

double DRoadNode::maxDObs_ = 80.0;

/********** Constructors and destructors **********/

// default constructor
DRoadNode::DRoadNode() : RoadNode()
{
  this->DObs_ = 0;
  this->sumIDO_ = 0;
}

// copy constructor
DRoadNode::DRoadNode(const DRoadNode& n) : RoadNode(n)
{
  this->DObs_ = n.DObs_;
  this->sumIDO_ = 0;
}

// init constructor
DRoadNode::DRoadNode(PositionXY& pos,double dobs) : RoadNode(pos)
{
  this->DObs_ = dobs;
  this->sumIDO_ = 0;
}

// init constructor
DRoadNode::DRoadNode(PositionXY& pos, RoadNode* ngh,double dobs) : RoadNode(pos,ngh)
{
  this->DObs_ = dobs;
  this->sumIDO_ = 0;
}

// destructor
DRoadNode::~DRoadNode()
{
}


/************** Getters and setters ***************/

// set the maximal distance to which an additional weight is added
void DRoadNode::setMaxDistObs(double d)
{
  DRoadNode::maxDObs_ = d;
}

// get the maximal distance to which an additional weight is added
double DRoadNode::getMaxDistObs()
{
  return DRoadNode::maxDObs_;
}

// return the weight added to move away from obstacles
double DRoadNode::invD() const
{
  if(this->DObs_ > DRoadNode::maxDObs_) return 0;
  else return (DRoadNode::maxDObs_ - this->DObs_);
}


/**************** Travel functions ****************/

// reinitialize all parameters of the node for the travel of the roadmap
void DRoadNode::reinit()
{
  this->visited_ = false;
  this->E_       = 0;
  this->L_       = 0;
  this->prev_    = NULL;
  this->sumIDO_  = 0;
}

// compute and set the length of the path to the current node 
// from the previous node prev
void DRoadNode::computeLfrom(RoadNode* prev)
{
  // compute the length between the current node and the previous one
  PositionXY curPos = this->getPos();
  PositionXY prevPos = prev->getPos();
  double dx = curPos.x() - prevPos.x();
  double dy = curPos.y() - prevPos.y();
  double dist = sqrt(dx*dx + dy*dy);

  // compute and set the length to the current node
  this->L_ = prev->getL() + dist;

  // compute the sum of distance to obstacles
  this->computeSIDOfrom((DRoadNode*)prev);

  // add prev as the previous node of the current one
  this->prev_ = prev;

}

// compute the sum of distance to obstacles
void DRoadNode::computeSIDOfrom(DRoadNode* prev)
{
  // compute and set the length to the current node
  this->sumIDO_ = prev->getSIDO() + 3.0*this->invD();
}


/*************** Display functions ****************/

// display the current node
void DRoadNode::display()
{
  cout << "Pos : "; this->getPos().display(); cout << "  nbNgh : " << this->getNbNeighbour();
  cout << "  Prev : ";
  if(this->getPrev() == NULL) cout << "NULL";
  else { this->getPrev()->getPos().display(); }
  cout << "  NbNeighbour : " << getNbNeighbour() << endl;
//   for(int i=0; i<this->getNbNeighbour(); i++)
//     {
//       cout << "  --> Ngh : "; this->getNeighbour(i)->getPos().display(); cout << endl;
//     }
  cout << "  L : " << getL();
  cout << "  E : " << getE();
  cout << "  invD : " << invD();
  cout << "  SIDO : " << getSIDO();
  cout << "  H : " << getH() << endl;
}


/****************************************************************************************
 ************************************ class RoadMap *************************************
 ****************************************************************************************/

/********** Constructors and destructors **********/

// default constructor
RoadMap::RoadMap()
{
  this->lNode_         = new vector<RoadNode*>();
  this->env_           = NULL;
  this->maxDistNgh_    = 15.0;
  this->nbNodesNeeded_ = 2000;
}

// copy constructor
RoadMap::RoadMap(const RoadMap& rm)
{
  this->lNode_ = new vector<RoadNode*>();
  // copy all nodes of the roadmap but mark them as non visited
  int N = rm.getNbNodes();
  for(int i=0; i<N; i++)
    {
      RoadNode* oldNode = rm.getNode(i);
      RoadNode* newNode = new RoadNode(*oldNode);
      
      // copy all link coming from the old node
      int M = oldNode->getNbNeighbour();
      for(int j=0; j<M; j++)
	{
	  RoadNode* curNgh = oldNode->getNeighbour(j);
	  PositionXY nghPos = curNgh->getPos();

	  // search for position nghPos in the list of nodes we are constructing
	  RoadNode* nghInNewRM = this->getNode(nghPos);

	  // if the corresponding node already exists in the new list, add it
	  // as neighbour of the current node
	  if(nghInNewRM != NULL)
	    {
	      newNode->addNeighbour(nghInNewRM);
	    }
	}

      // add the node to the roadmap
      this->addNode(newNode);
    }
}

// init constructor
RoadMap::RoadMap(Environment* env)
{
  this->lNode_         = new vector<RoadNode*>();
  this->env_           = env;
  this->maxDistNgh_    = 15.0;
  this->nbNodesNeeded_ = 2000;
}


/************** Getters and setters ***************/

// return the i-th node of teh roadmap
RoadNode* RoadMap::getNode(int i) const
{
  if(this->lNode_ != NULL && i < this->lNode_->size())
    return this->lNode_->at(i);
  else
    return NULL;
}

// return the node of the roadmap which can be found in position cfg
// if it exists, return NULL else
RoadNode* RoadMap::getNode(PositionXY& cfg) const
{
  if(this->lNode_ != NULL)
    {
      int i=0;
      int N = this->getNbNodes();
      while(i<N)
	{
	  RoadNode* curNode = this->getNode(i);
	  PositionXY curPos = curNode->getPos();

	  // check if the current position is the same than the input position
	  if(curPos.x() == cfg.x() && curPos.y() == cfg.y())
	    {
	      return curNode;
	    }
	  else i++;
	}
      // if the prog arrive here, no corresponding node has been found
      return NULL;
    }
  else
    return NULL;
}

// return the index of the node in the roadmap which can be found in position cfg
// return -1 if it doesn't exists
int RoadMap::getIndNode(const PositionXY& cfg) const
{
  if(this->lNode_ != NULL)
    {
      int i=0;
      int N = this->getNbNodes();
      while(i<N)
	{
	  RoadNode* curNode = this->getNode(i);
	  PositionXY curPos = curNode->getPos();

	  // check if the current position is the same than the input position
	  if(curPos.x() == cfg.x() && curPos.y() == cfg.y())
	    {
	      return i;
	    }
	  else i++;
	}
      // if the prog arrive here, no corresponding node has been found
      return -1;
    }
  else
    return -1;
}

// return the number of nodes of the roadmap
int RoadMap::getNbNodes() const
{
  if(this->lNode_ != NULL)
    return this->lNode_->size();
  else
    return -1;
}

// return the maximal distance between two nodes
double RoadMap::getMaxDistNgh() const
{
  return this->maxDistNgh_;
}

// set the maximal distance between two nodes
void RoadMap::setMaxDistNgh(double d)
{
  this->maxDistNgh_ = d;
}

// return the number of nodes needed on the map
int RoadMap::getNbNodesNeeded() const
{
  return this->nbNodesNeeded_;
}

// set the number of nodes needed on the map
void RoadMap::setNbNodesNeeded(int n)
{
  this->nbNodesNeeded_ = n;
}

// return the environment of the roadmap
Environment* RoadMap::getEnv() const
{
  return this->env_;
}


/************** RoadMap Construction **************/

// construct the roadmap
void RoadMap::construct()
{
  // get back the dimensions of the map
  double Xmin = env_->minX();
  double Ymin = env_->minY();
  double Xmax = env_->maxX();
  double Ymax = env_->maxY();

  double sqMaxDistNgh = this->maxDistNgh_*this->maxDistNgh_;

  cout << "nbNodesNeeded_ : " << nbNodesNeeded_ << endl;
  
  while(this->lNode_->size() < nbNodesNeeded_)
    {
      cout << "this->lNode_->size() : " << this->lNode_->size() << endl;

      // generate a new position
      double newX = getDoubleRandom(Xmin, Xmax);
      double newY = getDoubleRandom(Ymin, Ymax);
      PositionXY newCfg(newX,newY);

      // check if the new position is obstructed
      bool obstructed = this->env_->staticObstruct(newCfg);

      if(!obstructed)
	{
	  // check if there is no point to close from the new one
	  int N = this->lNode_->size();
	  int j=0;
	  bool trouve = false;
	  while(j<N && !trouve)
	    {
	      RoadNode* curNode = this->lNode_->at(j);

	      // compute the distance between the new node and the current node
	      double dx = newCfg.x() - curNode->getPos().x();
	      double dy = newCfg.y() - curNode->getPos().y();
	      double sqDist = dx*dx + dy*dy;

	      if(sqDist < 0.09*sqMaxDistNgh)
		trouve = true;
	      else j++;
	    }

	  if(!trouve)
	    {
	      // create a new node corresponding to the new position
	      RoadNode* node = new RoadNode(newCfg);
	      
	      // check if there exists any neighbour of the node in the roadmap
	      for(int i=0; i<N; i++)
		{
		  RoadNode* curNode = this->lNode_->at(i);
		  
		  // compute the distance between the new node and the current node
		  double dx = node->getPos().x() - curNode->getPos().x();
		  double dy = node->getPos().y() - curNode->getPos().y();
		  double sqDist = dx*dx + dy*dy;
		  
		  // check if the new node is enoughly near from the current node
		  if(sqDist >= 0.09*sqMaxDistNgh && sqDist <= sqMaxDistNgh)
		    {
		      // check if there is no obstacle on the line between the new node and the current node
		      bool lineObstructed = this->env_->staticObstructSgmt(newCfg,(this->lNode_->at(i)->getPos()));
		      
		      if(!lineObstructed)
			{
			  // add a new link between the current node and the new node
			  node->addNeighbour(curNode);
			  curNode->addNeighbour(node);
			}
		    }
		}
	      
	      // add the new node to the roadmap
	      this->addNode(node);
	    }
	}
    }
}

// add a new node to the roadmap
void RoadMap::addNode(RoadNode* n)
{
  if(this->lNode_ != NULL)
    this->lNode_->push_back(n);
}

// find the nearest node of the roadmap from the position pos
RoadNode* RoadMap::findNearNode(const PositionXY& pos) const
{
  if(this->lNode_ != NULL && this->lNode_->size() != 0)
    {
      double minSqDist = -1;
      RoadNode* near = NULL;
      int N = this->getNbNodes();
      for(int i=0; i<N; i++)
	{
	  double dx = pos.x() - this->lNode_->at(i)->getPos().x();
	  double dy = pos.y() - this->lNode_->at(i)->getPos().y();
	  double sqDist = (dx*dx + dy*dy);

	  if(sqDist < minSqDist || i == 0)
	    {
	      minSqDist = sqDist;
	      near = this->lNode_->at(i);
	    }
	}
      return near;
    }
  else
    return NULL;
}

// reinitialize all parameters of the nodes of the roadmap for its travel
void RoadMap::reinit()
{
  int N = this->getNbNodes();
  for(int i=0; i<N; i++)
    {
      // reinit each node of the roadmap
      this->getNode(i)->reinit();
    }
}


/*************** Display functions ****************/

// display the main features of the roadmap
void RoadMap::display() const
{
  cout << "nbNodes : " << this->getNbNodes() << endl;
}

// display the roadmap in an openGL viewer
void RoadMap::toGL(double time) const
{
  if(this->lNode_ != NULL)
    {
      int N = this->lNode_->size();
      for(int i=0; i<N; i++)
	{
	  // display the current node
	  RoadNode* node = this->lNode_->at(i);

	  glPushMatrix();
	  glTranslated(node->getPos().x(), node->getPos().y(), time - 0.01);
	  gluDisk(gluNewQuadric(), 0, 0.5, 20, 1);
	  gluCylinder(gluNewQuadric(), 0.5, 0.5, 0.02, 20, 5);
	  glTranslated(0,0,0.02);
	  gluDisk(gluNewQuadric(), 0, 0.5, 20, 1);
	  glPopMatrix();

	  // display each link coming from the current node
	  int M = node->getNbNeighbour();
	  glLineWidth(((GLfloat)(1.0)));
	  glBegin(GL_LINES);
	  for(int j=0; j<M ; j++)
	    {
	      RoadNode* ngh = node->getNeighbour(j);

	      glVertex3f(node->getPos().x(), node->getPos().y(), time);
	      glVertex3f(ngh->getPos().x(), ngh->getPos().y(), time);
	    }
	  glEnd();
	}
    }    
}


/****************************************************************************************
 ************************************ class DRoadMap ************************************
 ****************************************************************************************/

/********** Constructors and destructors **********/

// default constructor
DRoadMap::DRoadMap() : RoadMap()
{
}

// copy constructor
DRoadMap::DRoadMap(const DRoadMap& rm) : RoadMap(rm)
{
}

// init constructor
DRoadMap::DRoadMap(Environment* env) : RoadMap(env)
{
}

/************** RoadMap Construction **************/

// construct the roadmap
void DRoadMap::construct()
{
  // get back the dimensions of the map
  double Xmin = env_->minX();
  double Ymin = env_->minY();
  double Xmax = env_->maxX();
  double Ymax = env_->maxY();

  double sqMaxDistNgh = this->maxDistNgh_*this->maxDistNgh_;

  cout << "nbNodesNeeded_ : " << nbNodesNeeded_ << endl;
  
  while(this->lNode_->size() < nbNodesNeeded_)
//   while(this->lNode_->size() < 500)
    {
//       cout << "this->lNode_->size() : " << this->lNode_->size() << endl;

      // generate a new position
      double newX = getDoubleRandom(Xmin, Xmax);
      double newY = getDoubleRandom(Ymin, Ymax);
      PositionXY newPos(newX,newY);
      PositionXY near;
      double mindist;

      // check if the new position is obstructed
      bool obstructed = this->env_->staticObstruct(newPos,near,mindist);

      if(!obstructed)
	{
	  // check if there is no point to close from the new one
	  int N = this->lNode_->size();
	  int j=0;
	  bool trouve = false;
	  while(j<N && !trouve)
	    {
	      RoadNode* curNode = this->lNode_->at(j);

	      // compute the distance between the new node and the current node
	      double dx = newPos.x() - curNode->getPos().x();
	      double dy = newPos.y() - curNode->getPos().y();
	      double sqDist = dx*dx + dy*dy;

	      if(sqDist < 0.09*sqMaxDistNgh)
		trouve = true;
	      else j++;
	    }

	  if(!trouve)
	    {
	      // create a new node corresponding to the new position
	      PositionXY newCfg(newPos.x(),newPos.y());
	      RoadNode* node = new DRoadNode(newCfg,mindist);
	      cout << "New node : (" << newPos.x() << "," << newPos.y() << ")  dist : " << mindist 
		   << "   (" << this->lNode_->size()+1 << "/" << nbNodesNeeded_ << ")" << endl;
	      
	      // check if there exists any neighbour of the node in the roadmap
	      for(int i=0; i<N; i++)
		{
		  RoadNode* curNode = this->lNode_->at(i);
		  
		  // compute the distance between the new node and the current node
		  double dx = node->getPos().x() - curNode->getPos().x();
		  double dy = node->getPos().y() - curNode->getPos().y();
		  double sqDist = dx*dx + dy*dy;
		  
		  // check if the new node is enoughly near from the current node
		  if(sqDist >= 0.09*sqMaxDistNgh && sqDist <= sqMaxDistNgh)
		    {
		      // check if there is no obstacle on the line between the new node and the current node
		      bool lineObstructed = this->env_->staticObstructSgmt(PositionXY(newPos.x(),newPos.y()),(this->lNode_->at(i)->getPos()));
		      
		      if(!lineObstructed)
			{
			  // add a new link between the current node and the new node
			  node->addNeighbour(curNode);
			  curNode->addNeighbour(node);
			}
		    }
		}
	      
	      // add the new node to the roadmap
	      this->addNode(node);
	    }
	}
    }
}


/***************** I/O functions ******************/

// load the roadmap from an input file
// return 0 if the process succeed
//       -1 if the input file cannot be opened
//       -2 if the input file is empty
//       -3 if NbNodes or DMaxObs or DMaxNgh commands are missing
//       -4 if an unavailable node index has been found
int DRoadMap::load(const char* nomfic)
{
  ifstream fic(nomfic, ios::in);

  if(!fic)
    {
      cerr << "DRoadMap::load : Unable to open the input file " << nomfic << endl;
      return -1;
    }
  if(fic.eof())
    {
      cerr << "DRoadMap::load : The input file is empty..." << endl;
      fic.close();
      return -2;
    }

  int NbNodes = -1;
  double DMaxObs = -1;
  double DMaxNgh = -1;

  char buf[255];
  while(fic.good())
    {
      fic.getline(buf, 255);

      // try to get back the number of nodes of the map
      int res1 = 0;
      if(NbNodes == -1) res1 = sscanf(buf, "NbNodes %d ~", &NbNodes);

      // try to get back the maximal distance to obstacles
      int res2 = 0;
      if(res1 != 1)
	{
	  if(DMaxObs == -1) res1 = sscanf(buf, "DMaxObs %lf ~", &DMaxObs);
	}

      // try to get back the maximal distance between neighbour nodes
      int res3 = 0;
      if(res1 != 1 && res2 != 1)
	{
	  if(DMaxNgh == -1) res1 = sscanf(buf, "DMaxNgh %lf ~", &DMaxNgh);
	}

      if(res1 != 1 && res2 != 1 && res3 != 1 && NbNodes != -1 && DMaxObs != -1 && DMaxNgh != -1)
	{
	  // try to get back a new node
	  int indNode = -1;
	  double posX = -1;
	  double posY = -1;
	  double minNearObs = -1;
	  int res4 = sscanf(buf, "Node %d %lf %lf %lf ~", &NbNodes, &posX, &posY, &minNearObs);
	  if(res4 == 4)
	    {
	      // create the new node and add it to the roadmap
	      PositionXY newPos(posX,posY);
	      RoadNode* node = new DRoadNode(newPos,minNearObs);
	      this->addNode(node);
	    }

	  // try to get back a new link
	  int indNode1 = -1;
	  int indNode2 = -1;
	  int res5 = 0;
	  if(res4 != 4)
	    {
	      res5 = sscanf(buf, "Link %d %d ~", &indNode1, &indNode2);
	      if(res5 == 2)
		{
		  // check if both index are valid
		  if(indNode1 < 0 || indNode1 >= this->getNbNodes() || indNode2 < 0 || indNode2 >= this->getNbNodes())
		    {
		      cerr << "DRoadMap::load : Unavailable node index" << endl;
		      fic.close();
		      return -4;
		    }

		  // create the new link between two nodes of the roadmap
		  RoadNode* node1 = this->lNode_->at(indNode1);
		  RoadNode* node2 = this->lNode_->at(indNode2);
		  if(node1 == NULL || node2 == NULL)
		    {
		      cerr << "DRoadMap::load : Undefined node index" << endl;
		      fic.close();
		      return -4;
		    }
		  node1->addNeighbour(node2);
		}
	    }

	  if(res4 != 4 && res5 != 2) cerr << "DRoadMap::load -> Skiped line: " << buf << endl;
	}
      else if(res1 != 1 && res2 != 1 && res3 != 1) cerr << "DRoadMap::load -> Skiped line: " << buf << endl;
    }

  if(NbNodes == -1 || DMaxObs == -1 || DMaxNgh == -1)
    {
      cerr << "DRoadMap::load : NbNodes or DMaxObs or DMaxNgh commands are missing" << endl;
      return -3;
    }
  
  // close the output file
  fic.close();
  return 0;
}

// save the roadmap into an output file
// return 0 if the process succeed 
//       -1 if the roadmap hasn't been built
//       -2 if the output file cannot be opened
int DRoadMap::save(const char* nomfic)
{
  // check if the roadmap has already been built
  if(lNode_ == NULL || lNode_->size() != nbNodesNeeded_) 
    {
      cerr << "DRoadMap::save : The roadmap hasn't been computed" << endl;
      return -1;
    }

  // open or create the output file
  ofstream fic(nomfic, ios::out | ios::trunc);
 
  if(fic)
    {
      // save the main features of the roadmap
      fic << "NbNodes " << lNode_->size() << " ~ " << endl;
      fic << "DMaxObs " << DRoadNode::getMaxDistObs() << " ~ " << endl;
      fic << "DMaxNgh " << maxDistNgh_ << " ~ " << endl;
      fic << endl;

      // save each node of the roadmap
      int N = lNode_->size();
      for(int i=0; i<N; i++)
	{
	  // get back the current node of the roadmap
	  DRoadNode* node = ((DRoadNode*) (this->getNode(i)));
	  fic << "Node " << i << " " << node->getPos().x() << " " << node->getPos().y() << " " << node->D() << " ~ " << endl;
	}
      fic << endl;

      // save each link of the roadmap
      for(int i=0; i<N; i++)
	{
	  // get back the current node of the roadmap
	  DRoadNode* node = ((DRoadNode*) (this->getNode(i)));
	  int M = node->getNbNeighbour();
	  for(int j=0; j<M; j++)
	    {
	      int ind2 = this->getIndNode(node->getNeighbour(j)->getPos());
	      fic << "Link " << i << " " << ind2 << " ~ " << endl;
	    }
	}
      
      // close the output file
      fic.close();

      return 0;
    }
  else 
    {
      cerr << "DRoadMap::save : The output file " << nomfic << " cannot be opened" << endl;
      return -2;
    }
}


/*************** Display functions ****************/

// display the roadmap in an openGL viewer
void DRoadMap::toGL(double time) const
{
  if(this->lNode_ != NULL)
    {
      int N = this->lNode_->size();
      for(int i=0; i<N; i++)
	{
	  // display the current node
	  RoadNode* node = this->lNode_->at(i);

	  glPushMatrix();
	  glTranslated(node->getPos().x(), node->getPos().y(), time - 0.01);
	  gluDisk(gluNewQuadric(), 0, 0.05, 20, 1);
	  gluCylinder(gluNewQuadric(), 0.05, 0.05, 0.02, 20, 5);
	  glTranslated(0,0,0.02);
	  gluDisk(gluNewQuadric(), 0, 0.05, 20, 1);
	  glPopMatrix();

	  // display each link coming from the current node
	  int M = node->getNbNeighbour();
	  glLineWidth(((GLfloat)(1.0)));
	  glBegin(GL_LINES);
	  for(int j=0; j<M ; j++)
	    { 
	      RoadNode* ngh = node->getNeighbour(j);

	      double coefG1 = ((DRoadNode*)node)->D() / DRoadNode::getMaxDistObs();
	      double coefG2 = ((DRoadNode*)ngh)->D() / DRoadNode::getMaxDistObs();

	      double coefG = 0.5*(coefG1+coefG2);
	      double coefR = 1.0 - coefG;
	      const GLfloat material[]={coefR,coefG,0.,1.};
	      glMaterialfv(GL_FRONT, GL_AMBIENT, material);
	      glMaterialfv(GL_FRONT, GL_DIFFUSE, material);

	      glVertex3f(node->getPos().x(), node->getPos().y(), time);
	      glVertex3f(ngh->getPos().x(), ngh->getPos().y(), time);
	    }
	  glEnd();
	}
    }    
}
