#include "gnuplot/inc/TrajPlot.hpp"
using namespace mpc;

#include "common/inc/Utils.hpp"

#include <iostream>
#include <string>
using namespace std;

                                    /**************************************************/
                                    /**************** Class TrajPlot ******************/
                                    /**************************************************/

/***************** Plot functions *****************/

/**
 *  \brief plot a math function
 *  \param plot : 
 *  \param fct : the function to plot
 *  \param title : the tite of the function
 *  \param sampling : the number of sampling points of the trajectory
 *  \param minBound : the minimal boundary to display
 *  \param maxBound : the maximal boundary to display
 */
void TrajPlot::plotFct(Gnuplot& plot, const Function& fct, const char* title, int sampling, double minBound, double maxBound)
{
  // sample the input function
  vector<double> tabT;
  vector<double> tabV;
  vector<double> tabMinB;
  vector<double> tabMaxB;

  // get back the boundaries of the function
  double binf = fct.getBInf();
  double bsup = fct.getBSup();

  double dt = (bsup - binf) / ((double) (sampling-1));

//   cout << "DT : " << dt << endl;

  // fill the tables
  for(int i=0; i<sampling; i++)
    {
      double t = binf + i*dt;
      double v = fct.getVal(t);
//       cout << "T : " << t << "  V : " << v << endl;

      tabT.push_back(t);
      tabV.push_back(v);
      if(minBound != INFINITY) tabMinB.push_back(minBound);
      if(maxBound != -INFINITY) tabMaxB.push_back(maxBound);
    }

  plot.set_style("lines").plot_xy(tabT,tabV,title);
  if(minBound != INFINITY) plot.set_style("points").plot_xy(tabT,tabMinB,"Min Bound");
  if(maxBound != -INFINITY) plot.set_style("points").plot_xy(tabT,tabMaxB,"Max Bound");
}

/**
 *  \brief plot a path from a Mana continuous state trajectory
 *  \param plot : the gnuplot window
 *  \param straj : the continuous state trajectory to display
 *  \param title : the tite of the function
 *  \param sampling : the number of sampling points of the trajectory
 */
void TrajPlot::plotDD2Path(Gnuplot& plot, const DD2PSTraj& straj, const char* title, int sampling)
{
  // sample the state trajectory
  DD2STraj* sampledTraj = straj.sample(sampling);

  if(sampledTraj != NULL)
    {
      // sample the input function
      vector<double> tabX;
      vector<double> tabY;
      
      int N = sampledTraj->getSize();
      for(int i=0; i<N; i++)
	{
	  tabX.push_back(sampledTraj->at(i).x());
	  tabY.push_back(sampledTraj->at(i).y());
	}
      
      // delete the sampled trajectory
      delete sampledTraj;

      plot.set_style("lines").plot_xy(tabX,tabY,title);      
    }
  else cerr << "TrajPlot::plotDD2Path : Unable to sample the state trajectory." << endl;
}

/**
 *  \brief plot the control trajectory of a Mana robotic system
 *  \param ctraj : the discrete control trajectory to display
 *  \param sys : the system constraints
 *  \param dispBounds : true if the bounds defined over the control must be displayed
 */
void TrajPlot::plotDD2CTraj(const DD2CTraj& ctraj, const DiffDrive2Sys* sys, bool dispBounds)
{
  // initialize the control vectors
  vector<double> tim;

  vector<double> vel;
  vector<double> velBinf;
  vector<double> velBsup;

  vector<double> avel;
  vector<double> avelBinf;
  vector<double> avelBsup;

  // fill the vectors with the control trajectory values
  int N = ctraj.getSize();
  for(int i=0; i<N; i++)
    {
      DD2ControlT ctrlt = ctraj.at(i);

      vel.push_back(ctrlt.v());
      avel.push_back(ctrlt.w());
      tim.push_back(ctrlt.t());

      if(sys != NULL && dispBounds)
	{
	  if(sys->canMoveBack()) velBinf.push_back(-sys->getVMax());
	  else velBinf.push_back(0.0);
	  velBsup.push_back(sys->getVMax());
	  avelBinf.push_back(-sys->getAVMax());
	  avelBsup.push_back(sys->getAVMax());
	}
    }

  // display the gnuplot windows
  Gnuplot gVel;
  gVel.reset_plot();
  gVel.set_style("lines").plot_xy(tim,vel,"Linear Velocity V(t)");
  if(sys != NULL && dispBounds)
    {
      gVel.set_style("points").plot_xy(tim,velBinf,"Min Bound");
      gVel.set_style("points").plot_xy(tim,velBsup,"Max Bound");
    }

  Gnuplot gAVel;
  gAVel.reset_plot();
  gAVel.set_style("lines").plot_xy(tim,avel,"Angular Velocity W(t)");
  if(sys != NULL && dispBounds)
    {
      gAVel.set_style("points").plot_xy(tim,avelBinf,"Min Bound");
      gAVel.set_style("points").plot_xy(tim,avelBsup,"Max Bound");
    }

  wait_for_key();
}

/**
 *  \brief plot the control trajectory of a Mana robotic system
 *  \param ctraj : the discrete control trajectory to display
 *  \param sys : the system constraints
 *  \param dispBounds : true if the bounds defined over the control must be displayed
 */
void TrajPlot::plotDD2AccCTraj(const DD2AccCTraj& ctraj, const DiffDrive2Sys* sys, bool dispBounds)
{
  // initialize the control vectors
  vector<double> tim;

  vector<double> acc;
  vector<double> accBinf;
  vector<double> accBsup;

  vector<double> aacc;
  vector<double> aaccBinf;
  vector<double> aaccBsup;

  // fill the vectors with the control trajectory values
  int N = ctraj.getSize();
  for(int i=0; i<N; i++)
    {
      DD2AccControlT ctrlt = ctraj.at(i);

      acc.push_back(ctrlt.acc());
      aacc.push_back(ctrlt.aacc());
      tim.push_back(ctrlt.t());

      if(sys != NULL && dispBounds)
	{
	  accBinf.push_back(-sys->getAMax());
	  accBsup.push_back(sys->getAMax());
	  aaccBinf.push_back(-sys->getAAMax());
	  aaccBsup.push_back(sys->getAAMax());
	}
    }

  // display the gnuplot windows
  Gnuplot gAcc;
  gAcc.reset_plot();
  gAcc.set_style("lines").plot_xy(tim,acc,"Linear Acceleration A(t)");
  if(sys != NULL && dispBounds)
    {
      gAcc.set_style("points").plot_xy(tim,accBinf,"Min Bound");
      gAcc.set_style("points").plot_xy(tim,accBsup,"Max Bound");
    }

  Gnuplot gAAcc;
  gAAcc.reset_plot();
  gAAcc.set_style("lines").plot_xy(tim,aacc,"Angular Acceleration MU(t)");
  if(sys != NULL && dispBounds)
    {
      gAAcc.set_style("points").plot_xy(tim,aaccBinf,"Min Bound");
      gAAcc.set_style("points").plot_xy(tim,aaccBsup,"Max Bound");
    }

  wait_for_key();
}

/**
 *  \brief plot the control trajectory of a Mana robotic system
 *  \param ctraj : the discrete control trajectory to display
 *  \param sys : the system constraints
 *  \param dispBounds : true if the bounds defined over the control must be displayed
 */
void TrajPlot::plotDD2STraj(const DD2STraj& straj, const DiffDrive2Sys* sys, bool dispBounds)
{
  // initialize the control vectors
  vector<double> tim;

  vector<double> posX;
  vector<double> posY;
  vector<double> headTH;

  vector<double> posItgX;
  vector<double> posItgY;

  vector<double> vel;
  vector<double> velBinf;
  vector<double> velBsup;

  vector<double> avel;
  vector<double> avelBinf;
  vector<double> avelBsup;

  // fill the vectors with the state trajectory values
  int N = straj.getSize();
  for(int i=0; i<N; i++)
    {
      DD2StateT statet = straj.at(i);

      posX.push_back(statet.x());
      posY.push_back(statet.y());
      posItgX.push_back(statet.v() * cos(statet.th()));
      posItgY.push_back(statet.v() * sin(statet.th()));
      headTH.push_back(statet.th());
      vel.push_back(statet.v());
      avel.push_back(statet.w());
      tim.push_back(statet.t());

      if(sys != NULL && dispBounds)
	{
	  if(sys->canMoveBack()) velBinf.push_back(-sys->getVMax());
	  else velBinf.push_back(0.0);
	  velBsup.push_back(sys->getVMax());
	  avelBinf.push_back(-sys->getAVMax());
	  avelBsup.push_back(sys->getAVMax());
	}
    }

  // display the gnuplot windows
  Gnuplot gXY;
  gXY.reset_plot();
  gXY.set_style("lines").plot_xy(posX,posY,"Trajectory XxY");

  Gnuplot gTimXY;
  gTimXY.reset_plot();
  gTimXY.set_style("lines").plot_xy(tim,posX,"X(t)");
  gTimXY.set_style("lines").plot_xy(tim,posY,"Y(t)");

  Gnuplot gItgXY;
  gItgXY.reset_plot();
  gItgXY.set_style("lines").plot_xy(tim,posItgX,"Itg X");
  gItgXY.set_style("lines").plot_xy(tim,posItgY,"Itg Y");

  Gnuplot gTH;
  gTH.reset_plot();
  gTH.set_style("lines").plot_xy(tim,headTH,"Heading Th(t)");

  Gnuplot gVel;
  gVel.reset_plot();
  gVel.set_style("lines").plot_xy(tim,vel,"Linear Velocity V(t)");
  if(sys != NULL && dispBounds)
    {
      gVel.set_style("points").plot_xy(tim,velBinf,"Min Bound");
      gVel.set_style("points").plot_xy(tim,velBsup,"Max Bound");
    }

  Gnuplot gAVel;
  gAVel.reset_plot();
  gAVel.set_style("lines").plot_xy(tim,avel,"Angular Velocity W(t)");
  if(sys != NULL && dispBounds)
    {
      gAVel.set_style("points").plot_xy(tim,avelBinf,"Min Bound");
      gAVel.set_style("points").plot_xy(tim,avelBsup,"Max Bound");
    }

  wait_for_key();
}

/**
 *  \brief plot the state trajectory of a Mana robotic system
 *  \param straj1 : the first discrete state trajectory to display
 *  \param straj2 : the second discrete state trajectory to display
 *  \param name1 : name of the first discrete state trajectory
 *  \param name2 : name of the second discrete state trajectory
 *  \param sys : the system constraints (robot constraints are displayed if the system is not null)
 */
void TrajPlot::plotDD2STraj(const DD2STraj& straj1, const DD2STraj& straj2, const char* name1, const char* name2, const DiffDrive2Sys* sys)
{
  // initialize the control vectors
  vector<double> tim1;

  vector<double> posX1;
  vector<double> posY1;
  vector<double> headTH1;

//   vector<double> posItgX1;
//   vector<double> posItgY1;

  vector<double> vel1;
  vector<double> velBinf;
  vector<double> velBsup;

  vector<double> avel1;
  vector<double> avelBinf;
  vector<double> avelBsup;

  // fill the vectors with the state trajectory values
  int N1 = straj1.getSize();
  for(int i=0; i<N1; i++)
    {
      DD2StateT statet = straj1.at(i);

      posX1.push_back(statet.x());
      posY1.push_back(statet.y());
//       posItgX1.push_back(statet.v() * cos(statet.th()));
//       posItgY1.push_back(statet.v() * sin(statet.th()));
      headTH1.push_back(statet.th());
      vel1.push_back(statet.v());
      avel1.push_back(statet.w());
      tim1.push_back(statet.t());

      if(sys != NULL)
	{
	  if(sys->canMoveBack()) velBinf.push_back(-sys->getVMax());
	  else velBinf.push_back(0.0);
	  velBsup.push_back(sys->getVMax());
	  avelBinf.push_back(-sys->getAVMax());
	  avelBsup.push_back(sys->getAVMax());
	}
    }

  // initialize the control vectors
  vector<double> tim2;
  vector<double> posX2;
  vector<double> posY2;
  vector<double> headTH2;
  vector<double> vel2;
  vector<double> avel2;

  vector<double> latShift;
  vector<double> angShift;

  // fill the vectors with the state trajectory values
  int N2 = straj2.getSize();
  for(int i=0; i<N2; i++)
    {
      DD2StateT statet = straj2.at(i);

      posX2.push_back(statet.x());
      posY2.push_back(statet.y());
      headTH2.push_back(statet.th());
      vel2.push_back(statet.v());
      avel2.push_back(statet.w());
      tim2.push_back(statet.t());

      DD2StateT s1;
      straj1.getStateAtTime(statet.t(), s1);
      double dx = statet.x() - s1.x();
      double dy = statet.y() - s1.y();
      double latS = sqrt(dx*dx + dy*dy);

      double dth = computeMinDiffAngle(statet.th(), s1.th());

      latShift.push_back(latS);
      angShift.push_back(dth);
    }

  // init titles
  string titleXY1(name1); titleXY1 += " XxY";
  string titleTh1(name1); titleTh1 += " Th(t) (rad)";
  string titleV1(name1); titleV1 += " V(t) (m/s)";
  string titleW1(name1); titleW1 += " W(t) (rad/s)";

  string titleXY2(name2); titleXY2 += " XxY";
  string titleTh2(name2); titleTh2 += " Th(t) (rad)";
  string titleV2(name2); titleV2 += " V(t) (m/s)";
  string titleW2(name2); titleW2 += " W(t) (rad/s)";

  // display the gnuplot windows
  Gnuplot gXY;
  gXY.reset_plot();
  gXY.set_style("lines").plot_xy(posX1,posY1, titleXY1);
  gXY.set_style("lines").plot_xy(posX2,posY2, titleXY2);

//   Gnuplot gTimXY;
//   gTimXY.reset_plot();
//   gTimXY.set_style("lines").plot_xy(tim1,posX1,"X(t)");
//   gTimXY.set_style("lines").plot_xy(tim1,posY1,"Y(t)");

//   Gnuplot gItgXY;
//   gItgXY.reset_plot();
//   gItgXY.set_style("lines").plot_xy(tim1,posItgX1,"Itg X");
//   gItgXY.set_style("lines").plot_xy(tim1,posItgY1,"Itg Y");

  Gnuplot gTH;
  gTH.reset_plot();
  gTH.set_style("lines").plot_xy(tim1,headTH1, titleTh1);
  gTH.set_style("lines").plot_xy(tim2,headTH2, titleTh2);

  Gnuplot gVel;
  gVel.reset_plot();
  gVel.set_style("lines").plot_xy(tim1,vel1, titleV1);
  gVel.set_style("lines").plot_xy(tim2,vel2, titleV2);
  if(sys != NULL)
    {
      gVel.set_style("points").plot_xy(tim1,velBinf,"Min Bound");
      gVel.set_style("points").plot_xy(tim1,velBsup,"Max Bound");
    }

  Gnuplot gAVel;
  gAVel.reset_plot();
  gAVel.set_style("lines").plot_xy(tim1,avel1, titleW1);
  gAVel.set_style("lines").plot_xy(tim2,avel2, titleW2);
  if(sys != NULL)
    {
      gAVel.set_style("points").plot_xy(tim1,avelBinf,"Min Bound");
      gAVel.set_style("points").plot_xy(tim1,avelBsup,"Max Bound");
    }

  Gnuplot gLatS;
  gLatS.reset_plot();
  gLatS.set_style("lines").plot_xy(tim2,latShift, "Lateral shift (m)");

  Gnuplot gAngS;
  gAngS.reset_plot();
  gAngS.set_style("lines").plot_xy(tim2,angShift, "Ang shift (rad)");

  wait_for_key();
}

/**
 *  \brief plot the control trajectory of a Mana robotic system
 *  \param ctraj : the discrete control trajectory to display
 *  \param sys : the system constraints
 *  \param dispBounds : true if the bounds defined over the control must be displayed
 *  \param sampling : the number of sampling poitns of the trajectory
 */
void TrajPlot::plotDD2PCTraj(const DD2PCTraj& ctraj, const DiffDrive2Sys* sys, bool dispBounds, int sampling)
{
  double minV = INFINITY;
  double maxV = -INFINITY;
  double minAV = INFINITY;
  double maxAV = -INFINITY;
  double minA = INFINITY;
  double maxA = -INFINITY;
  double minAA = INFINITY;
  double maxAA = -INFINITY;

  if(sys != NULL && dispBounds)
    {
      if(sys->canMoveBack()) minV = -sys->getVMax();
      else minV = 0.0;
      maxV = sys->getVMax();
      minAV = -sys->getAVMax();
      maxAV = sys->getAVMax();
      minA = -sys->getAMax();
      maxA = sys->getAMax();
      minAA = -sys->getAAMax();
      maxAA = sys->getAAMax();
    }

  Gnuplot gVel;
  gVel.reset_plot();
  TrajPlot::plotFct(gVel, *(ctraj.fctVel()), "Linear Velocity V(t)", sampling, minV, maxV);
    
  Gnuplot gAVel;
  gAVel.reset_plot();
  TrajPlot::plotFct(gAVel, *(ctraj.fctAVel()), "Angular Velocity W(t)", sampling, minAV, maxAV);

  Gnuplot gAcc;
  gAcc.reset_plot();
  TrajPlot::plotFct(gAcc, *(ctraj.fctAcc()), "Linear Acceleration A(t)", sampling, minA, maxA);

  Gnuplot gAAcc;
  gAAcc.reset_plot();
  TrajPlot::plotFct(gAAcc, *(ctraj.fctAAcc()), "Angular Acceleration MU(t)", sampling, minAA, maxAA);

  wait_for_key();
}

/**
 *  \brief plot the control trajectory of a Mana robotic system
 *  \param ctraj : the discrete control trajectory to display
 *  \param sys : the system constraints
 *  \param dispBounds : true if the bounds defined over the control must be displayed
 *  \param sampling : the number of sampling poitns of the trajectory
 */
void TrajPlot::plotDD2PSTraj(const DD2PSTraj& straj, const DiffDrive2Sys* sys, bool dispBounds, int sampling)
{
  // sample the state trajectory
  DD2STraj* sampledTraj = straj.sample(sampling);

  if(sampledTraj != NULL)
    {
      TrajPlot::plotDD2STraj(*sampledTraj, sys, dispBounds);
      
      // delete the sampled trajectory
      delete sampledTraj;
    }
  else cerr << "TrajPlot::plotDD2PSTraj : Unable to sample the state trajectory." << endl;
}

/**
 *  \brief plot the control trajectory of a Mana robotic system
 *  \param ctraj : the continuous control trajectory to display
 *  \param sys : the system constraints
 *  \param dispBounds : true if the bounds defined over the control must be displayed
 *  \param sampling : the number of sampling points of the trajectory
 */
void TrajPlot::plotDD2PCTraj(Gnuplot& plotVel, Gnuplot& plotAVel, Gnuplot& plotAcc, Gnuplot& plotAAcc, const DD2PCTraj& ctraj, const DiffDrive2Sys* sys, int sampling)
{
  double minV = INFINITY;
  double maxV = -INFINITY;
  double minAV = INFINITY;
  double maxAV = -INFINITY;
  double minA = INFINITY;
  double maxA = -INFINITY;
  double minAA = INFINITY;
  double maxAA = -INFINITY;

  if(sys != NULL)
    {
      if(sys->canMoveBack()) minV = -sys->getVMax();
      else minV = 0.0;
      maxV = sys->getVMax();
      minAV = -sys->getAVMax();
      maxAV = sys->getAVMax();
      minA = -sys->getAMax();
      maxA = sys->getAMax();
      minAA = -sys->getAAMax();
      maxAA = sys->getAAMax();

      TrajPlot::plotFct(plotVel, *(ctraj.fctVel()), "Linear Velocity V(t)", sampling, minV, maxV);
      TrajPlot::plotFct(plotAVel, *(ctraj.fctAVel()), "Angular Velocity W(t)", sampling, minAV, maxAV);
      TrajPlot::plotFct(plotAcc, *(ctraj.fctAcc()), "Linear Acceleration A(t)", sampling, minA, maxA);
      TrajPlot::plotFct(plotAAcc, *(ctraj.fctAAcc()), "Angular Acceleration MU(t)", sampling, minAA, maxAA);
    }  
  else
    {
      TrajPlot::plotFct(plotVel, *(ctraj.fctVel()), "", sampling);
      TrajPlot::plotFct(plotAVel, *(ctraj.fctAVel()), "", sampling);
      TrajPlot::plotFct(plotAcc, *(ctraj.fctAcc()), "", sampling);
      TrajPlot::plotFct(plotAAcc, *(ctraj.fctAAcc()), "", sampling);
    }
}

// /**
//  *  \brief plot the control trajectory of a Mana robotic system
//  *  \param straj : the continuous state trajectory to display
//  *  \param sys : the system constraints
//  *  \param dispBounds : true if the bounds defined over the control must be displayed
//  *  \param sampling : the number of sampling points of the trajectory
//  */
// void TrajPlot::plotDD2PSTraj(Gnuplot& plot, const DD2PSTraj& straj, const DiffDrive2Sys* sys=NULL, int sampling=50)
// {
// }
