#ifndef MPC_RGBIMAGE_HPP
#define MPC_RGBIMAGE_HPP

#include "graphics/inc/RGBColor.hpp"
#include <vector>

/*******************************************
 *              RGBImage frame :           *
 *                                         *
 *     (0,0)    -- j -->                   *
 *       +---------------------------      *
 *       |                          |      *
 *       |                          |      *
 *       |                          |      *
 *     | |                          |      *
 *     | |                          |      *
 *     i |                          |      *
 *     | |                          |      *
 *     | |                          |      *
 *     V |                          |      *
 *       |                          |      *
 *       |                          |      *
 *       |                          |      *
 *       ----------------------------      *
 *                                         *
 *                                         *
 *******************************************/


namespace mpc
{

  typedef struct IMAGE_INDEX
  {
    int line_i;
    int col_j;
  } Image_index;


                                    /**************************************************/
                                    /***************** Class RGBImage *****************/
                                    /**************************************************/

  /** \class RGBImage
   *  \brief RGB image
   */
  class RGBImage
  {
  protected : 
    RGBColor** tab_;                                           /**< table containing the RGB image */
    int nbCols_;                                               /**< number of columns */
    int nbLines_;                                              /**< number of lines */
    int depth_;                                                /**< color depth of the image */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */  
    RGBImage();

  public : 
    /**
     *  \brief init constructor
     */  
    RGBImage(int nbLines, int nbCols, int depth=255);

    /**
     *  \brief init constructor
     */  
    RGBImage(int nbLines, int nbCols, int depth, const RGBColor& color);

    /**
     *  \brief destructor
     */  
    ~RGBImage();


    /********************** Init **********************/

  private : 
    /*
     *  \brief init the image table
     */
    void initTab(int nbCols, int nbLines);

    /*
     *  \brief init the image table
     */
    void initTab(int nbCols, int nbLines, const RGBColor& color);

    /*
     *  \brief clear the image table
     */
    void clearTab();


    /******************** Accessors *******************/

  public : 
    /**
     *  \brief return the number of columns of the image
     *  \return the number of columns of the image
     */
    inline int nbCols() const { return this->nbCols_; }

    /**
     *  \brief return the number of columns of the image
     *  \return the number of columns of the image
     */
    inline int nbLines() const { return this->nbLines_; }

    /**
     *  \brief return the color defined at coord (col, line) = (i, j)
     *  \param (i, j) : (col, line) coord
     */
    const RGBColor& getCol(int i, int j) const;

    /**
     *  \brief set the color at coord (col, line) = (i, j)
     *  \param (i, j) : (col, line) coord
     *  \param col : the rgb color
     */
    void setCol(int i, int j, const RGBColor& col);

    /**
     *  \brief set the color at coord (col, line) = (i, j)
     *  \param (i, j) : (col, line) coord
     *  \param col : the rgb color
     */
    void setCol(int i, int j, int r, int g, int b);


    /****************** I/O functions *****************/

  public : 
    /**
     *  \brief save the current image into a ppm file
     *  \param filename : the input filename
     *  \return 0 if the process succeeded, -1 else
     */
    int saveIntoPPM(const char* filename) const;

    /**
     *  \brief load the current image from a ppm file
     *  \param filename : the input filename
     *  \return 0 if the process succeeded, -1 else
     */
    int loadFromPPM(const char* filename);


    /**************** Drawing functions ***************/

  public : 
    /**
     *  \brief draw a cross on the image
     *  \param (i,j) : the center of the cross
     *  \param rad : the radius of the cross (in pixels)
     *  \param col : the color of the cross
     */
    void drawCross(int i, int j, int rad, const RGBColor& col);

    /**
     *  \brief draw a line on the image from (i1,j1) to (i2,j2)
     *  \param (i1,j1) : the coord of the first end point of the line
     *  \param (i2,j2) : the coord of the second end point of the line
     *  \param col : the color of the cross
     */
    void drawLine(int i1, int j1, int i2, int j2, const RGBColor& col);


    /*************** Utility functions ***************/

  private : 
    /*
     *  \brief get back the list of indices of the grid corresponding to the line defined between the two input points
     *  \param i1 : first point line number
     *  \param j1 : first point column number
     *  \param i2 : second point line number
     *  \param j2 : second point column number
     *  \param res : the resulting list of indices
     */
    void getIndicesLine(int i1, int j1, int i2 , int j2, std::vector<Image_index>& res) const;
    
  public : 
    /**
     *  \brief display the header of the image
     */
    void displayHeader() const;
  };
};

#endif // MPC_RGBIMAGE_HPP
