#include "graphics/inc/RGBImage.hpp"
using namespace mpc;

#include "common/inc/Utils.hpp"

#include <iostream>
#include <stdlib.h>
#include <cstring>
#include <fstream>
using namespace std;

                                    /**************************************************/
                                    /***************** Class RGBImage *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */  
RGBImage::RGBImage()
{
  tab_ = NULL;
  nbLines_ = 0;
  nbCols_ = 0;
  depth_ = 0;
}

/**
 *  \brief init constructor
 */  
RGBImage::RGBImage(int nbLines, int nbCols, int depth)
{
  tab_ = NULL;
  nbLines_ = 0;
  nbCols_ = 0;
  depth_ = depth;

  this->initTab(nbCols, nbLines);
}

/**
 *  \brief init constructor
 */  
RGBImage::RGBImage(int nbLines, int nbCols, int depth, const RGBColor& color)
{
  tab_ = NULL;
  int nbLines_ = 0;
  int nbCols_ = 0;
  int depth_ = depth;

  this->initTab(nbCols, nbLines, color);
}

/**
 *  \brief destructor
 */  
RGBImage::~RGBImage()
{
  this->clearTab();
}


/********************** Init **********************/

/*
 *  \brief init the image table
 */
void RGBImage::initTab(int nbCols, int nbLines)
{
  this->clearTab();

  nbLines_ = nbLines;
  nbCols_ = nbCols;

  tab_ = new RGBColor* [nbLines];

  for(int i=0; i<nbLines; i++)
    {
      tab_[i] = new RGBColor [nbCols];
    }
}

/*
 *  \brief init the image table
 */
void RGBImage::initTab(int nbCols, int nbLines, const RGBColor& color)
{
  this->clearTab();

  tab_ = new RGBColor* [nbLines];

  for(int i=0; i<nbLines; i++)
    {
      tab_[i] = new RGBColor [nbCols];
    }

  for(int i=0; i<nbLines; i++)
    for(int j=0; j<nbCols; j++)
      {
	tab_[i][j] = RGBColor(color);
      }
}

/*
 *  \brief clear the image table
 */
void RGBImage::clearTab()
{
  if(tab_ != NULL)
    {
      for(int i=0; i<nbLines_; i++)
	{
	  if(tab_[i] != NULL) delete [] tab_[i];
	}
      delete [] tab_;
      tab_ = NULL;
    }
}


/******************** Accessors *******************/

/**
 *  \brief return the color defined at coord (col, line) = (i, j)
 *  \param (i, j) : (col, line) coord
 */
const RGBColor& RGBImage::getCol(int i, int j) const 
{
  if(i<0 || i>=nbLines_ || j<0 || j>=nbCols_) throw string("RGBImage::getCol : index out of bounds");

  return tab_[i][j];
}

/**
 *  \brief set the color at coord (col, line) = (i, j)
 *  \param (i, j) : (col, line) coord
 *  \param col : the rgb color
 */
void RGBImage::setCol(int i, int j, const RGBColor& col)
{
  if(i<0 || i>=nbLines_ || j<0 || j>=nbCols_) 
    {
      cerr << "i : " << i << "  j : " << j << "  nbLines : " << nbLines_ << "  nbCols : " << nbCols_ << endl;
      cout << "i : " << i << "  j : " << j << "  nbLines : " << nbLines_ << "  nbCols : " << nbCols_ << endl;
      throw string("RGBImage::setCol : index out of bounds");
    }

  tab_[i][j] = col;
}

/**
 *  \brief set the color at coord (col, line) = (i, j)
 *  \param (i, j) : (col, line) coord
 *  \param col : the rgb color
 */
void RGBImage::setCol(int i, int j, int r, int g, int b)
{
  if(i<0 || i>=nbLines_ || j<0 || j>=nbCols_) 
    {
      cerr << "i : " << i << "  j : " << j << "  nbLines : " << nbLines_ << "  nbCols : " << nbCols_ << endl;
      cout << "i : " << i << "  j : " << j << "  nbLines : " << nbLines_ << "  nbCols : " << nbCols_ << endl;
      throw string("RGBImage::setCol : index out of bounds");
    }

  tab_[i][j] = RGBColor(r,g,b,depth_);
}


/****************** I/O functions *****************/

/**
 *  \brief save the current image into a ppm file
 *  \param filename : the input filename
 *  \return 0 if the process succeeded, -1 else
 */
int RGBImage::saveIntoPPM(const char* filename) const 
{
  // open the output file
  ofstream fic(filename, ios::out | ios::trunc | ios::binary);

  // check if the opening succeeded
  if(fic)
    {
      // write the PGM magic number
      fic << "P6" << endl;
      
      // write a dump comment
      fic << "# MPC RGBImage saved into a PPM file" << endl;

      // write the dimensions of the map
      char colsString[20], linesString[20];
      int2String(nbLines_,colsString);
      int2String(nbCols_,linesString);
      fic << colsString << " " << linesString << endl;

      // write the depth of the map
//       char depthString[20];
//       int2String(depth_, depthString);
//       fic << depthString << endl;
      fic << "255" << endl;

      for(int i=0; i<nbLines_; i++)
	{
	  for(int j=0; j<nbCols_; j++)
	    {
	      const RGBColor col = this->getCol(i, j);

	      fic << ((char)col.r()) << ((char)col.g()) << ((char)col.b());
	    }
	}
      fic << endl;
      
      fic.close();

      return 0;
    }
  else
    {
      cerr << "RGBImage::saveIntoPPM : cannot open the output file : " << filename << endl;
      return -1;
    }
}

/**
 *  \brief load the current image from a ppm file
 *  \param filename : the input filename
 *  \return 0 if the process succeeded, -1 else
 */
int RGBImage::loadFromPPM(const char* filename)
{
  cerr << "RGBImage::loadFromPPM : Not implemented yet!" << endl;
  exit(1);
}


/**************** Drawing functions ***************/

/**
 *  \brief draw a cross on the image
 *  \param (i,j) : the center of the cross
 *  \param rad : the radius of the cross (in pixels)
 *  \param col : the color of the cross
 */
void RGBImage::drawCross(int i, int j, int rad, const RGBColor& col)
{
  // check if the center of the cross is in the frame
  if(i>=0 && i<nbLines_ && j>=0 && j<nbCols_)
    {
      for(int k=i-rad; k<=i+rad; k++)
	{
	  if(k>=0 && k<nbLines_ && j>=0 && j<nbCols_)
	    this->setCol(k, j, col);
	}
      for(int l=j-rad; l<=j+rad; l++)
	{
	  if(i>=0 && i<nbLines_ && l>=0 && l<nbCols_)
	    this->setCol(i, l, col);
	}
    }
}

/**
 *  \brief draw a line on the image from (i1,j1) to (i2,j2)
 *  \param (i1,j1) : the coord of the first end point of the line
 *  \param (i2,j2) : the coord of the second end point of the line
 *  \param col : the color of the cross
 */
void RGBImage::drawLine(int i1, int j1, int i2, int j2, const RGBColor& col)
{
  std::vector<Image_index> vec_res;
  this->getIndicesLine(i1, j1, i2 ,j2, vec_res);

  for(int k=0; k<vec_res.size(); k++) this->setCol(vec_res.at(k).line_i, vec_res.at(k).col_j, col);
}


/*************** Utility functions ***************/

/*
 *  \brief get back the list of indices of the grid corresponding to the line defined between the two input points
 *  \param i1 : first point line number
 *  \param j1 : first point column number
 *  \param i2 : second point line number
 *  \param j2 : second point column number
 *  \param res : the resulting list of indices
 */
void RGBImage::getIndicesLine(int i1, int j1, int i2 ,int j2, std::vector<Image_index>& res) const
{
  // check on which axis positions have to be sampled
  int di = i2 - i1;
  int dj = j2 - j1;
  int max_dij;
  if(abs(di) >= abs(dj)) max_dij = abs(di);
  else max_dij = abs(dj);

  // search for each index in the grid covered by the input line
  for(int k=0; k<=max_dij; k++)
    {
      Image_index curIndex;
      curIndex.line_i = (int) (i1 + ((float)k) / ((float)max_dij) * di );
      curIndex.col_j = (int) (j1 + ((float)k) / ((float)max_dij) * dj );

      if (curIndex.line_i >= 0 && curIndex.line_i < nbLines_ && curIndex.col_j >= 0 && curIndex.col_j < nbCols_) res.push_back(curIndex);
    }
}
    
/**
 *  \brief display the header of the image
 */
void RGBImage::displayHeader() const
{
  cout << "RGBImage header -> nbLines=" << nbLines_ << "  nbCols=" << nbCols_ << "  depth=" << depth_ << endl;
}
