#ifndef MPC_MATRIX2D_HPP
#define MPC_MATRIX2D_HPP

#include <stdlib.h>
#include <iostream>
using namespace std;

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class Matrix2D ****************/
                                    /**************************************************/

  /** \class Matrix2D
   *  \brief template 2-dimension matrix
   */
  template <typename T>
  class Matrix2D
  {
  protected : 
    int dim1_;                                      /**< size of the first dimension of the matrix */
    int dim2_;                                      /**< size of the second dimension of the matrix */
    T** tab_;                                       /**< table containing all the values of teh matrix */

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     *
     *  May only be used by subclasses
     */
    Matrix2D()
    {
      this->dim1_ = 0;
      this->dim2_ = 0;
      this->tab_ = NULL;
    }

  public : 
    /**
     *  \brief init constructor
     *  \param dim1 : size of the first dimension
     *  \param dim2 : size of the second dimension
     */
    Matrix2D(int dim1,int dim2)
    {
      this->tab_ = NULL;
    
      this->init(dim1,dim2);
    }

    /**
     *  \brief copy constructor
     *  \param m2 : the input matrix to copy
     */
    Matrix2D(const Matrix2D& m2)
    {
      this->tab_ = NULL;
      this->init(m2.dim1(),m2.dim2());

      for(int i=0; i<dim1_; i++)
	for(int j=0; j<dim2_; j++)      
	  tab_[i][j] = m2[i][j];
    }

    ~Matrix2D()
    {
      this->clear();
    }

    /***************** Initialization *****************/

  protected : 
    /**
     *  \brief clear the matrix
     */
    void clear()
    {
      if(tab_ != NULL)
	{
	  for(int i=0; i<dim1_; i++) delete tab_[i];
	  delete tab_;
	  tab_ = NULL;
	}
    }
  
  public : 
    /**
     *  \brief initialize the matrix table from its new dimensions
     *  \param dim1 : size of the first dimension
     *  \param dim2 : size of the second dimension
     */
    void init(int dim1,int dim2)
    {
      this->clear();

      dim1_ = dim1;
      dim2_ = dim2;

      tab_ = new T*[dim1];
      for(int i=0; i<dim1; i++)
	{
	  tab_[i] = new T[dim2];
	}
    }

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the size of the first dimension
     *  \return the size of the first dimension
     */
    inline int dim1() const { return this->dim1_; }

    /**
     *  \brief return the size of the first dimension
     *  \return the size of the first dimension
     */
    inline int dim2() const { return this->dim2_; }

    /******************** Operators *******************/

  public : 
    /**
     *  \brief return a value contained in the matrix
     *  \param i : index of the first dimension
     *  \param j : index of the second dimension
     *  \return the value of the matrix at index [i,j]
     */
    const T* operator[](int i) const
    {
      return tab_[i];
    }

    /**
     *  \brief return a reference to a value contained in the matrix
     *  \param i : index of the first dimension
     *  \param j : index of the second dimension
     *  \return a reference to the value of the matrix at index [i,j]
     */
    T* operator[](int i)
    {
      return tab_[i];
    }

    /**************** Display functions ***************/

  public : 
    /**
     *  \brief display the content of the matrix
     */
    void display() const
    {
      for(int i=0; i<dim1_; i++)
	{
	  for(int j=0; j<dim2_; j++)
	    cout << this->tab_[i][j] << " ";
	  cout << endl;
	}
    }
  };

};
#endif // MPC_MATRIX2D_HPP


#ifndef MPC_MATRIX3D_HPP
#define MPC_MATRIX3D_HPP

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class Matrix3D ****************/
                                    /**************************************************/

  /** \class Matrix3D
   *  \brief template 3-dimension matrix
   */
  template <typename T>
  class Matrix3D
  {
  protected : 
    int dim1_;                                      /**< size of the first dimension of the matrix */
    int dim2_;                                      /**< size of the second dimension of the matrix */
    int dim3_;                                      /**< size of the third dimension of the matrix */
    T*** tab_;                                      /**< table containing all the values of teh matrix */

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     *
     *  May only be used by subclasses
     */
    Matrix3D()
    {
      this->dim1_ = 0;
      this->dim2_ = 0;
      this->dim3_ = 0;
      this->tab_ = NULL;
    }

  public : 
    /**
     *  \brief init constructor
     *  \param dim1 : size of the first dimension
     *  \param dim2 : size of the second dimension
     */
    Matrix3D(int dim1,int dim2,int dim3)
    {
      this->tab_ = NULL;
    
      this->init(dim1,dim2,dim3);
    }

    /**
     *  \brief copy constructor
     *  \param m2 : the input matrix to copy
     */
    Matrix3D(const Matrix3D& m2)
    {
      this->tab_ = NULL;
      this->init(m2.dim1(),m2.dim2(),m2.dim3());

      for(int i=0; i<dim1_; i++)
	for(int j=0; j<dim2_; j++)      
	  for(int k=0; k<dim3_; k++)      
	    tab_[i][j][k] = m2[i][j][k];
    }

    /**
     *  \brief destructor
     */
    ~Matrix3D()
    {
      this->clear();
    }

    /***************** Initialization *****************/

  protected : 
    /**
     *  \brief clear the matrix
     */
    void clear()
    {
      if(tab_ != NULL)
	{
	  for(int i=0; i<dim1_; i++)
	    {
	      for(int j=0; j<dim2_; j++)
		{
		  delete tab_[i][j];
		}
	      delete tab_[i];
	    }
	  delete tab_;
	  tab_ = NULL;
	}
    }
  
  public : 
    /**
     *  \brief initialize the matrix table from its new dimensions
     *  \param dim1 : size of the first dimension
     *  \param dim2 : size of the second dimension
     *  \param dim3 : size of the third dimension
     */
    void init(int dim1,int dim2,int dim3)
    {
      this->clear();
    
      dim1_ = dim1;
      dim2_ = dim2;
      dim3_ = dim3;

      tab_ = new T**[dim1];
      for(int i=0; i<dim1; i++)
	{
	  tab_[i] = new T*[dim2];
	  for(int j=0; j<dim2; j++)
	    {
	      tab_[i][j] = new T[dim3];
	    }
	}
    }

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the size of the first dimension
     *  \return the size of the first dimension
     */
    inline int dim1() const { return this->dim1_; }

    /**
     *  \brief return the size of the first dimension
     *  \return the size of the first dimension
     */
    inline int dim2() const { return this->dim2_; }

    /**
     *  \brief return the size of the first dimension
     *  \return the size of the first dimension
     */
    inline int dim3() const { return this->dim3_; }

    /******************** Operators *******************/

  public : 
    /**
     *  \brief return a value contained in the matrix
     *  \param i : index of the first dimension
     *  \param j : index of the second dimension
     *  \return the value of the matrix at index [i,j]
     */
    T** operator[](int i) const
    {
      return tab_[i];
    }

    /**
     *  \brief return a reference to a value contained in the matrix
     *  \param i : index of the first dimension
     *  \param j : index of the second dimension
     *  \return a reference to the value of the matrix at index [i,j]
     */
    T** operator[](int i)
    {
      return tab_[i];
    }

    /**************** Display functions ***************/

  public : 
    /**
     *  \brief display the content of the matrix
     */
    void display() const
    {
      for(int i=0; i<dim1_; i++)
	{
	  for(int j=0; j<dim2_; j++)
	    {
	      for(int k=0; k<dim3_; k++)
		cout << this->tab_[i][j][k] << " ";
	      cout << endl;
	    }
	  cout << endl;
	}
    }
  };

};
#endif // MPC_MATRIX3D_HPP
