#ifndef OPERATION_HPP
#define OPERATION_HPP

#include "math/inc/Function.hpp"
#include <stdlib.h>

/**
 * \enum math_operator_t
 * \brief Enumeration of the different mathematical operator types
 *
 * ND_OP        : Undefined operator
 * ADD_OP       : Addition
 * MULT_OP      : Multiplication
 * COS_OP       : Cosinus
 * SIN_OP       : Sinus
 * ITG_OP       : Integration
 */
typedef enum {
  ND_OP     = 0,
  ADD_OP    = 1, 
  MULT_OP   = 2, 
  COS_OP    = 3,
  SIN_OP    = 4,
  ITG_OP    = 5
} math_operator_t;


                                    /**************************************************/
                                    /***************** Class Operation ****************/
                                    /**************************************************/

namespace mpc
{
  class Operation : public Function
  {
  protected : 
    int nbOperand_;                                    /**< Number of operands of the operation */
    math_operator_t op_;                               /**< Operator type */
    Function** tabOperand_;                            /**< Table containing the different operand of the operation */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    Operation();

    /**
     *  \brief init constructor (from the mumber of operands)
     *  \param nbOp : the number of operands
     */
    Operation(int nbOp);

  public : 
    /**
     *  \brief default constructor
     */
    Operation(const Operation& op);

    /**
     *  \brief destructor
     */
    virtual ~Operation();

    /**
     *  \brief return a copy of the current function
     *  \return a copy of the current function
     *
     *  A new instance of function is allocated and returned here
     */
    virtual Operation* copy() const = 0;

    /**************** Setters & Getters ***************/

    /**
     *  \brief return the value of the polynomial at a given parameter t
     *  \param t : the input parameter value
     *  \return the value of the polynomial at a given parameter t
     */
    virtual double getVal(double t) const = 0;
  
    /******************** Operators *******************/

    /**
     *  \brief multiply the current function by a constant value
     *  \param v : the input multiplication coefficient
     */
    virtual void operator*=(double v) = 0;

    /************ Integration & Derivation ************/

    /**
     *  \brief return the derivation of the current function
     *  \return the derivation of the current function
     */
    virtual Function* derivation() const = 0;

    //   virtual Function* derivation(const char* nameVar) = 0; // return the derivation of the current polynome wrt the variable named nameVar

    /**
     *  \brief return the integration of the current function
     *  \return the integration of the current function
     */
    virtual Function* integration() const = 0;

    /**
     *  \brief return the integration of the current function
     *  \param cste : the integration constant at t=0
     *  \return the integration of the current function
     */
    virtual Function* integration(double cste) const = 0;

    /**
     *  \brief return the integration of the current function
     *  \param cste,binf : cste is the integration constant value defined at t=binf
     *  \return the integration of the current function
     */
    virtual Function* integration(double binf,double cste) const = 0;

    /****************** Roots search ******************/

    /**
     *  \brief return the real roots of this function
     *  \param r : the list of roots of the function
     *  \return the number of roots of the function
     */
    virtual int realRoots(std::vector<double>& r) const = 0;

    /**
     *  \brief return the roots of the equation eq - v = 0
     *  \param r : the list of roots of the function
     *  \param v : the input constant value
     *  \return the number of roots of the function
     */
    virtual int realRoots(std::vector<double>& r,double v) const = 0;

    /**************** Utility functions ***************/

    /**
     *  \brief display the polynomial
     */
    virtual void display() const = 0;

    /**
     *  \brief shift the function on Y axis from value v
     *  \param v : the shifting value
     */
    virtual void shiftY(double v) = 0;
  };

};
#endif // OPERATION_HPP
