#include "gsl/gsl_matrix.h"
#include "gsl/gsl_vector.h"
#include "gsl/gsl_linalg.h"

#include <cmath>
#include <iostream>
using namespace std;

gsl_matrix* prodGSLMat(gsl_matrix* A, gsl_matrix* B)
{
  int M = A->size1;
  int N = B->size2;

  int M2 = A->size2;
  int N1 = B->size1;

  if(M2 != N1)
    {
      std::cerr << "The second dimension of A and the first dimension of B do not match. Cannot compute A*B." << std::endl;
      return NULL;
    }

  gsl_matrix* res = gsl_matrix_alloc(M,N);

  for(int i=0; i<M; i++)
    {
      for(int j=0; j<N; j++)
	{
	  double sum = 0;
	  for(int k=0; k<M2; k++)
	    {
	      sum += (gsl_matrix_get(A,i,k) * gsl_matrix_get(B,k,j));
	    }
	  if(fabs(sum) < 1e-10)
	    gsl_matrix_set(res,i,j,0.0);
	  else
	    gsl_matrix_set(res,i,j,sum);
	}
    }

  return res;
}

gsl_matrix* transpGSLMat(gsl_matrix* A)
{
  int M = A->size1;
  int N = A->size2;

  gsl_matrix* At = gsl_matrix_alloc(N,M);

  for(int i=0; i<M; i++)
    {
      for(int j=0; j<N; j++)
	{
	  gsl_matrix_set(At,j,i,gsl_matrix_get(A,i,j));
	}
    }

  return At;
}

void displayGSLMat(gsl_matrix* A)
{
  int M = A->size1;
  int N = A->size2;

  for(int i=0; i<M; i++)
    {
      for(int j=0; j<N; j++)
	{
	  cout << gsl_matrix_get(A, i, j) << " ";
	}
      cout << endl;
    }
}

int main()
{ 
  const int M=4;
  const int N=3;

  gsl_matrix* U = gsl_matrix_alloc(M,N);      // matrice initiale et matrice unitaire apres decomp
  gsl_matrix* A = gsl_matrix_alloc(M,N);      // copie de la matrice initiale
  gsl_vector* E = gsl_vector_alloc(N);        // vecteur de valeurs singulieres
  gsl_vector* work = gsl_vector_alloc(N);     // vecteur de travail pour la SVD
  gsl_matrix* V = gsl_matrix_alloc(N,N);      // matrice adjointe

  double d;
  int i,j,k;

  cout << "init" << endl;

  // initialise la matrice
  gsl_matrix_set(U, 0, 0, 1); gsl_matrix_set(U, 0, 1, 2); gsl_matrix_set(U, 0, 2, 3);
  gsl_matrix_set(U, 1, 0, 4); gsl_matrix_set(U, 1, 1, 2); gsl_matrix_set(U, 1, 2, 3);
  gsl_matrix_set(U, 2, 0, 1); gsl_matrix_set(U, 2, 1, 4); gsl_matrix_set(U, 2, 2, 5);
  gsl_matrix_set(U, 3, 0, 2); gsl_matrix_set(U, 3, 1, 0); gsl_matrix_set(U, 3, 2, 6);
  
  cout << "bouh" << endl;

  // effectue une copie de la matrice initiale
  for(i=0; i<M; i++)
    {
      for(j=0; j<N; j++)
	{
	  gsl_matrix_set(A, i, j, gsl_matrix_get(U, i, j));
	}
    }

  // affiche la matrice initiale A
  cout << "Matrice A : " << endl; 
  displayGSLMat(A);

  // Singular Values Decomposition
  gsl_linalg_SV_decomp (U, V, E, work);

  // affiche la matrice orthogonale U
  cout << "Matrice U : " << endl; 
  displayGSLMat(U);
  
  // calcule et la matrice des valeurs singulieres U
  gsl_matrix* matE = gsl_matrix_alloc(N,N); 
  for(i=0; i<N; i++)
    {
      for(j=0; j<N; j++)
	{
	  if(i == j) gsl_matrix_set(matE, i, j, gsl_vector_get(E, i));
	  else gsl_matrix_set(matE, i, j, 0);
	}
    }
  cout << "Matrice E : " << endl;
  displayGSLMat(matE);

  // affiche la matrice orthogonale V
  cout << "Matrice V : " << endl; 
  displayGSLMat(V);
  
  // calcule les transposees de U et V
  gsl_matrix* Ut = transpGSLMat(U);
  gsl_matrix* Vt = transpGSLMat(V);
  
  // calcul et affiche le produit U Ut
  gsl_matrix* prodUUt = prodGSLMat(Ut,U);
  cout << "Matrice Ut*U : " << endl;
  displayGSLMat(prodUUt);

  // calcul et affiche le produit V Vt
  gsl_matrix* prodVVt = prodGSLMat(V,Vt);
  cout << "Matrice V*Vt : " << endl;
  displayGSLMat(prodVVt); 
  
  // calcul et affiche le produit U E Vt
  gsl_matrix* prodUE = prodGSLMat(U,matE);
  gsl_matrix* prodUEVt = prodGSLMat(prodUE,Vt);

  cout << "Matrice U*E*Vt : " << endl;
  displayGSLMat(prodUEVt); 

  // cree la matrice contenant l'oppose de la matrice des valeurs singulieres
  gsl_matrix* Einv = gsl_matrix_alloc(N,N);
  for(i=0; i<N; i++)
    {
      for(j=0; j<N; j++)
	{
	  if(i == j) gsl_matrix_set(Einv,i,j,1.0/gsl_vector_get(E,i));
	  else gsl_matrix_set(Einv,i,j,0.0);
	}
    }

  // affiche la matrice Einv
  cout << "Matrice Einv : " << endl; 
  displayGSLMat(Einv);

  // calcule et affiche la matrice inverse de A
  gsl_matrix* prodVEinv = prodGSLMat(V,Einv);
  gsl_matrix* Ainv = prodGSLMat(prodVEinv,Ut);
  displayGSLMat(Ainv);

  // affiche le produit de A*Ainv
  cout << "Matrice Prod : " << endl;
  gsl_matrix* prod = prodGSLMat(A,Ainv);
  displayGSLMat(prod);

  return 0;
}
