#include "Function.hpp"
using namespace mpc;

#include <iostream>
#include <cmath>
using namespace std;


/********** Constructors and destructors **********/

// default constructor
Function::Function()
{
  this->binf_ = -INFINITY;
  this->bsup_ = INFINITY;
  
  this->type_ = ND_FUNC;
}

/**
 *  \brief copy constructor
 *  \param f :  the function to copy 
 */
Function::Function(const Function& f)
{
  this->binf_ = f.binf_;
  this->bsup_ = f.bsup_;

  this->type_ = f.type_;
}

// destructor
Function::~Function()
{
}


/**************** Setters & Getters ***************/

// set the bounds on which the function is defined
void Function::setBounds(double binf,double bsup)
{
  this->binf_ = binf;
  this->bsup_ = bsup;
}


/**************** Utility functions ***************/

// sample the function on interval [binf;bsup]
void Function::sample(double binf,double bsup,int step,vector<double>& sampling) const
{
  // compute the length of the interval
  double length = bsup - binf;
  
  // compute the input step
  double dt = length / ((double)step);

  // compute all values between binf and bsup
  for(int i=0; i<=step; i++)
    {
      double t = binf + i*dt;
      sampling.push_back(this->getVal(t));
    }
}

// return true if t \in [binf_;bsup_]
bool Function::isDefinedAt(double t) const
{
  if(t >= this->binf_ && t <= bsup_)
    return true;
  else 
    return false;
}


/**
 *  \brief merges and sorts two vectors of discontinuity points
 *  \param discPoints1 : the first discontinuity points vector
 *  \param discPoints2 : the second discontinuity points vector
 *  \param discPointsRes : the resulting discontinuity points vector
 */
void Function::mergeDiscPoints(const std::vector<Disc_Point>& discPoints1, 
			       const std::vector<Disc_Point>& discPoints2, 
			       std::vector<Disc_Point>& discPointsRes)
{
//   cout << "+++++++++++++++++++++++ MERGE DISC POINTS +++++++++++++++++++++++" << endl;

//   cout << "discPoints1 : "; Function::displayDiscPoints(discPoints1);
//   cout << "discPoints2 : "; Function::displayDiscPoints(discPoints2);

  std::vector<Disc_Point>::const_iterator it1 = discPoints1.begin();
  std::vector<Disc_Point>::const_iterator it2 = discPoints2.begin();

  while(it1 != discPoints1.end() || it2 != discPoints2.end())
    {
      if(it1 == discPoints1.end())
	{
	  // add the last elements of discPoints2
	  while(it2 != discPoints2.end())
	    {
	      discPointsRes.push_back(*it2);
	      it2++;
	    }
	  return;
	}

      else if(it2 == discPoints2.end())
	{
	  // add the last elements of discPoints1
	  while(it1 != discPoints1.end())
	    {
	      discPointsRes.push_back(*it1);
	      it1++;
	    }
	  return;
	}

      else 
	{
	  // check for the list with the smallest value
	  if((*it1).t < (*it2).t)
	    {
	      discPointsRes.push_back(*it1);
	      it1++;
	    }
	  else if((*it1).t > (*it2).t)
	    {
	      discPointsRes.push_back(*it2);
	      it2++;
	    }
	  else // (*it1.t == *it2.t)
	    {
	      discPointsRes.push_back(*it1);
	      it1++;
	      it2++;
	    }
	}
    }
}

/**
 *  \brief display a vector of discontinuity points
 */
void Function::displayDiscPoints(const std::vector<Disc_Point>& discPoints)
{
  cout << "Disc points : {"; 
  for(int i=0; i<discPoints.size(); i++)
    {
      cout << "(" << discPoints.at(i).t << "," << discPoints.at(i).v << ") ";
    }
  cout << "}" << endl;
}
