#include "Interval.hpp"

#include <cstdlib>
#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /***************** Class Interval *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
Interval::Interval()
{
  this->bLow_ = 0.0;
  this->bUp_ = 0.0;
  this->lowBT_ = OPEN_B;
  this->upBT_ = OPEN_B;
}

/**
 *  \brief copy constructor
 *  \param i : the interval to copy
 */
Interval::Interval(const Interval& i)
{
  this->bLow_  = i.bLow_;
  this->bUp_   = i.bUp_;
  this->lowBT_ = i.lowBT_;
  this->upBT_  = i.upBT_;
}

/**
 *  \brief init constructor
 *  \param bLow : the lower bound of the interval
 *  \param bLowT : the type of the lower bound of the interval
 *  \param bUp : the upper bound of the interval
 *  \param bUpT : the type of the upper bound of the interval
 */
Interval::Interval(double bLow,bound_t bLowT,double bUp,bound_t bUpT)
{
  this->bLow_  = bLow;
  this->bUp_   = bUp;
  this->lowBT_ = bLowT;
  this->upBT_  = bUpT;
}

/**
 *  \brief destructor
 */
Interval::~Interval()
{
  // nothing to do
}


/**************** Setters & Getters ***************/

/**
 *  \brief return true if the current interval is empty
 *  \return true if the current interval is empty
 */
bool Interval::isEmpty() const
{
  if((this->bUp() < this->bLow()) || ((this->bUp() == this->bLow()) && (this->lowBT() == OPEN_B || this->upBT() == OPEN_B))) return true;
  else return false;
}

/**
 *  \brief empty the current interval
 */
void Interval::empty()
{
  this->bLow_ = 0.0;
  this->bUp_ = 0.0;
  this->lowBT_ = OPEN_B;
  this->upBT_ = OPEN_B;
}


/*************** Inclusion functions **************/

/**
 *  \brief return true if the input value is include in the current interval
 *  \param t : the value to test
 *  \return true if the input value is include in the current interval
 */
bool Interval::include(double t) const
{
  if(((t > this->bLow()) || (t >= this->bLow() && this->lowBT() == CLOSE_B)) 
     && ((t < this->bUp()) || (t <= this->bUp() && this->upBT() == CLOSE_B))) return true;
  else return false;
}

/**
 *  \brief return the ordering of the input value wrt. the current interval
 *  \param t : the input value to order
 *  \return ordering of the input value wrt. the current interval
 *
 *  ordering : the value t is before, after, or include in the interval
 */
order_t Interval::order(double t) const
{
  if((t < this->bLow()) || (t == this->bLow() && this->lowBT() == OPEN_B))
    return BEFORE;
  else if((t > this->bUp()) || (t == this->bUp() && this->upBT() == OPEN_B))
    return AFTER;
  else return INCLUDE;
}


/*************** Shifting functions ***************/

/**
 *  \brief shift the interval of a value t
 *  \param t : the value to shift the interval
 */
void Interval::shift(double t)
{
  this->bLow() += t;
  this->bUp() += t;
}


/**************** Display functions ***************/

/**
 *  \brief display the current interval
 */
void Interval::display() const
{
  if(this->lowBT() == CLOSE_B) cout << "[";
  else cout << "]";
  cout << this->bLow();
  cout << ";";
  cout << this->bUp();
  if(this->upBT() == CLOSE_B) cout << "]";
  else cout << "[";
}


                                    /**************************************************/
                                    /*************** Class PiecewiseItv ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
PiecewiseItv::PiecewiseItv()
{
  // nothing to do
}

/**
 *  \brief copy constructor
 *  \param i : the interval to copy
 */
PiecewiseItv::PiecewiseItv(const PiecewiseItv& itv)
{
  for(int i=0; i<itv.vecItv_.size(); i++)
    {
      Interval curItv = itv.vecItv_.at(i);
      this->vecItv_.push_back(curItv);
    }
}

/**
 *  \brief init constructor (from a simple interval)
 *  \param i : the simple interval
 */
PiecewiseItv::PiecewiseItv(const Interval& itv)
{
  this->vecItv_.push_back(itv);
}

/**
 *  \brief int constructor (simple interval)
 *  \param bLow : the lower bound of the interval
 *  \param bLowT : the type of the lower bound of the interval
 *  \param bUp : the upper bound of the interval
 *  \param bUpT : the type of the upper bound of the interval
 */
PiecewiseItv::PiecewiseItv(double bLow,bound_t bLowT,double bUp,bound_t bUpT)
{
  Interval itv(bLow,bLowT,bUp,bUpT);
  this->vecItv_.push_back(itv);  
}

/**
 *  \brief destructor
 */
PiecewiseItv::~PiecewiseItv()
{
  // nothing to do
}


/**************** Setters & Getters ***************/

/**
 *  \brief add an interval to the current vector of intervals
 *  \param itv : the simple interval to add 
 *
 *  Before to insert it, must check if there is no intersection with the current intervals
 */
void PiecewiseItv::addSItv(const Interval& itv)
{
  // check if the input interval is not null
  if(itv.isEmpty()) return;

  if(this->getNbSItv() == 0) this->vecItv_.push_back(itv);
  else
    {
      int i=0;
      bool found = false;
      while(!found && i<this->getNbSItv())
	{
	  // check if the new interval must be inserted before the current one
	  if(itv.bUp() < this->at(i).bLow())
	    {
	      // the place to insert the new simple interval has been bound
	      found = true;
	      this->insertSItv(i,itv);
	    }
	  else i++;
	}

      // if the insertion place has not been found, we must append the new interval at the end
      if(!found) this->appendSItv(itv);
    }
}

/**
 *  \brief append an interval at the end of the current vector of intervals
 *  \param itv : the simple interval to append
 *
 *  Before to insert it, must check if there is no intersection with the current intervals
 */
void PiecewiseItv::appendSItv(const Interval& itv)
{
  this->vecItv_.push_back(itv);
}

/**
 *  \brief insert an interval in the current vector of intervals
 *  \param i : the index in the vector to insert the new simple interval
 *  \param itv : the simple interval to insert
 *
 *  Before to insert it, must check if there is no intersection with the current intervals
 */
void PiecewiseItv::insertSItv(int i,const Interval& itv)
{
  vector<Interval>::iterator it = this->vecItv_.begin();
  for(int j=0; j<i; j++) it++;
  this->vecItv_.insert(it,itv);
}

/**
 *  \brief access to the i-th simple interval of the piecewise current one
 *  \param i : index of the simple interval we wants to reach
 *  \return a const reference to the i-th simple interval of the piecewise current one
 */
const Interval& PiecewiseItv::at(int i) const
{
  return this->vecItv_.at(i);
}

/**
 *  \brief return true if the current interval is empty
 *  \return true if the current interval is empty
 */
bool PiecewiseItv::isEmpty() const
{
  return (this->vecItv_.size() == 0);
}

/**
 *  \brief empty the current interval
 */
void PiecewiseItv::empty()
{
  this->vecItv_.clear();
}


/******************** Operators *******************/

/**
 *  \brief access to the i-th simple interval of the piecewise current one
 *  \param i : index of the simple interval we wants to reach
 *  \return a reference to the i-th simple interval of the piecewise current one
 */
Interval& PiecewiseItv::operator[](int i)
{
  return (this->vecItv_)[i];
}


/**************** Ordering functions **************/

/**
 *  \brief return true if the input value is include in the current interval
 *  \param t : the value to test
 *  \return true if the input value is include in the current interval
 */
bool PiecewiseItv::include(double t) const
{
  bool trouve = false;
  int i=0;
  while(!trouve && i<this->vecItv_.size())
    {
      if(this->vecItv_.at(i).include(t)) trouve = true;
      else i++;
    }
  return trouve;
}


/*************** Shifting functions ***************/

/**
 *  \brief shift the interval of a value t
 *  \param t : the value to shift the interval
 */
void PiecewiseItv::shift(double t)
{
  int N = this->getNbSItv();
  for(int i=0; i<N; i++)
    {
      (*this)[i].shift(t);
    }
}


/**************** Display functions ***************/

/**
 *  \brief display the current interval
 */
void PiecewiseItv::display() const
{
  if(this->isEmpty())
    {
      cout << "[0]";
    }
  else
    {
      for(int i=0; i<this->getNbSItv()-1; i++)
	{
	  this->at(i).display();
	  cout << " U ";
	}
      this->at(this->getNbSItv()-1).display();
    }
}


                                    /**************************************************/
                                    /************* Intervals computations *************/
                                    /**************************************************/

/**
 *  \brief compute the intersection between two simple intervals
 *  \param itv1 : the first interval
 *  \param itv2 : the second interval
 *  \param res : the resulting interval
 */
void intersecItv(const Interval& itv1,const Interval& itv2,Interval& res)
{
  order_t ordT2low = itv1.order(itv2.bLow());
  order_t ordT2up = itv1.order(itv2.bUp());

  res.empty();
  
  switch(ordT2low)
    {
    case BEFORE : 
      switch(ordT2up)
	{
	case BEFORE : 
	  // no intersection exists bewteen both intervals : 
	  // return an empty interval
	  res = Interval();
	  break;
	case INCLUDE : 
	  // the intersection is [itv1.bLow;itv2.bUp]
	  res = Interval(itv1.bLow(),itv1.lowBT(),itv2.bUp(),itv2.upBT());
	  break;
	case AFTER : 
	  // the second interval includes completely the first one
	  // the intersection is then the first one
	  res = itv1;
	  break;
	}
      break;
    case INCLUDE : 
      switch(ordT2up)
	{
	case BEFORE : 
	  // the second intervel is not consistent, should not appear
	  // display an error and return an empty interval
	  cout << "intersecItv : Warning. The second interval is unconsistent" << endl;
	  res = Interval();
	  break;
	case INCLUDE : 
	  // the first interval completely includes the second one
	  // the intersection is then the second one
	  res = itv2;
	  break;
	case AFTER : 
	  // the intersection is [itv2.bLow;itv1.bUp]
	  res = Interval(itv2.bLow(),itv2.lowBT(),itv1.bUp(),itv1.upBT());
	  break;
	}
      break;
    case AFTER : 
      // no intersection exists bewteen both intervals : 
      // return an empty interval
      res = Interval();
      break;
    }
}

/**
 *  \brief compute the intersection between two piecewise intervals
 *  \param itv1 : the first interval
 *  \param itv2 : the second interval
 *  \param res : the resulting interval
 *
 *  Barbar method. Should be improved later.
 */
void intersecItv(const PiecewiseItv& itv1,const PiecewiseItv& itv2,PiecewiseItv& res)
{
  // reinit the result interval
  res.empty();

  int N1 = itv1.getNbSItv();
  int N2 = itv2.getNbSItv();

  for(int i=0; i<N1; i++)
    {
      // get back the current interval of the first piecewise itv
      Interval curItv1 = itv1.at(i);

      for(int j=0; j<N2; j++)
	{
	  // get back the current interval of the first piecewise itv
	  Interval curItv2 = itv2.at(j);

	  // compute the intersection between curItv1 and curItv2
	  Interval curItsc;
	  intersecItv(curItv1,curItv2,curItsc);
	  
	  // add the current intersection to the piecewise intersection interval
	  res.addSItv(curItsc);
	}
    }
}

/**
 *  \brief compute the intersection between two simple intervals
 *  \param itv1 : the first interval
 *  \param itv2 : the second interval
 *  \param res : the resulting interval
 */
void unionItv(const Interval& itv1,const Interval& itv2,PiecewiseItv& res)
{
  order_t ordT2low = itv1.order(itv2.bLow());
  order_t ordT2up = itv1.order(itv2.bUp());
  
  res.empty();

  switch(ordT2low)
    {
    case BEFORE : 
      switch(ordT2up)
	{
	case BEFORE : 
	  // the union is the sum of both intervals
	  res.appendSItv(itv2);
	  res.appendSItv(itv1);
	  break;
	case INCLUDE : 
	  // intersection between the end of the second interval and the beggining of the first one : 
	  // include a new interval [itv2.bLow;itv1.bUp]
	  res.appendSItv(Interval(itv2.bLow(),itv2.lowBT(),itv1.bUp(),itv1.upBT()));
	  break;
	case AFTER : 
	  // the second interval includes completely the first one : 
	  // append only the second interval
	  res.appendSItv(itv2);
	  break;
	}
      break;
    case INCLUDE : 
      switch(ordT2up)
	{
	case BEFORE : 
	  // the second interval is not consistent, should not appear
	  // display an error and return an empty interval
	  cout << "unionItv : Warning (1). The second interval is unconsistent" << endl;
	  break;
	case INCLUDE : 
	  // the first interval includes completely the second one : 
	  // append only the first interval
	  res.appendSItv(itv1);
	  break;
	case AFTER : 
	  // intersection between the end of the first interval and the beggining of the second one : 
	  // include a new interval [itv1.bLow;itv2.bUp]
	  res.appendSItv(Interval(itv1.bLow(),itv1.lowBT(),itv2.bUp(),itv2.upBT()));
	  break;
	}
      break;
    case AFTER : 
      switch(ordT2up)
	{
	case BEFORE : 
	  // the second interval is not consistent, should not appear
	  // display an error and return an empty interval
	  cout << "unionItv : Warning (2). The second interval is unconsistent" << endl;
	  break;
	case INCLUDE : 
	  // the second interval is not consistent, should not appear
	  // display an error and return an empty interval
	  cout << "unionItv : Warning (3). The second interval is unconsistent" << endl;
	  break;
	case AFTER : 
	  // the union is the sum of both intervals
	  res.appendSItv(itv1);
	  res.appendSItv(itv2);
	  break;
	}
      break;
    }
}

/**
 *  \brief compute the intersection between two piecewise intervals
 *  \param itv1 : the first interval
 *  \param itv2 : the second interval
 *  \param res : the resulting interval
 */
void unionItv(const PiecewiseItv& itv1,const PiecewiseItv& itv2,PiecewiseItv& res)
{
//   cout << "///////////////////////////////////////////" << endl;
//   cout << "itv1 : "; itv1.display(); cout << endl;
//   cout << "itv2 : "; itv2.display(); cout << endl;

  // reinit the result interval
  res.empty();

  int i1 = 0;
  int i2 = 0;

  while(i1 < itv1.getNbSItv() || i2 < itv2.getNbSItv())
    {
      double newBLow,newBUp;
      bound_t newLowBT,newUpBT;
      
      int followed;

//       cout << "i1 : " << i1 << "itv1.getNbSItv() : " << itv1.getNbSItv() << endl;
//       cout << "i2 : " << i2 << "itv2.getNbSItv() : " << itv2.getNbSItv() << endl;

      if(i1 >= itv1.getNbSItv())
	{
// 	  cout << "deb1" << endl;
	  followed = 2;
	  newBLow = itv2.at(i2).bLow();
	  newLowBT = itv2.at(i2).lowBT();
	}
      else if(i2 >= itv2.getNbSItv())
	{
// 	  cout << "deb2" << endl;
	  followed = 1;
	  newBLow = itv1.at(i1).bLow();
	  newLowBT = itv1.at(i1).lowBT();
	}
      else
	{
// 	  cout << "deb3" << endl;
	  // select the minimal bound from both piecewise intervals
	  double blow1 = itv1.at(i1).bLow();
	  double blow2 = itv2.at(i2).bLow();
	  
	  if(blow1 < blow2) 
	    {
// 	      cout << "deb31" << endl;
	      followed = 1;
	      newBLow = blow1;
	      newLowBT = itv1.at(i1).lowBT();
	    }
	  else 
	    {
// 	      cout << "deb32" << endl;
	      followed = 2;
	      newBLow = blow2;
	      newLowBT = itv2.at(i2).lowBT();
	    }
	}

      bool endItv = false;
      while(!endItv)
	{
	  if(i1 >= itv1.getNbSItv())
	    {
// 	      cout << "1" << endl;
	      newBUp = itv2.at(i2).bUp();
	      newUpBT = itv2.at(i2).upBT();
	      endItv = true;
	      i2++;
	    }
	  else if(i2 >= itv2.getNbSItv())
	    {
// 	      cout << "2" << endl;
	      newBUp = itv1.at(i1).bUp();
	      newUpBT = itv1.at(i1).upBT();
	      endItv = true;
	      i1++;
	    }
	  else if(followed == 1)
	    {
// 	      cout << "3" << endl;
	      order_t lowOrd = itv1.at(i1).order(itv2.at(i2).bLow());
	      order_t upOrd = itv1.at(i1).order(itv2.at(i2).bUp());
	      
	      if(lowOrd == INCLUDE && upOrd == INCLUDE)
		{
// 		  cout << "31" << endl;
		  // the second interval is completely included in the first one
		  i2++;
		}
	      else if(lowOrd == INCLUDE && upOrd == AFTER)
		{
// 		  cout << "32" << endl;
		  // the second interval finish after the first one : switch the interval followed
		  followed = 2;
		  i1++;
		}
	      else if(lowOrd == AFTER && upOrd == AFTER)
		{
// 		  cout << "33" << endl;
		  // the second interval is not linked : we found the end of the new interval to add
		  newBUp = itv1.at(i1).bUp();
		  newUpBT = itv1.at(i1).upBT();
		  endItv = true;
		  i1++;
// 		  i2++;
		}
	      else // should not appear
		{
		  cout << "unionItv : Warning. An error occured during the union of piecewise intervals." << endl;
		  exit(1);
		}
	    }
	  else // (followed == 2)
	    {
// 	      cout << "4" << endl;
	      order_t lowOrd = itv2.at(i2).order(itv1.at(i1).bLow());
	      order_t upOrd = itv2.at(i2).order(itv1.at(i1).bUp());
	      
	      if(lowOrd == INCLUDE && upOrd == INCLUDE)
		{
// 		  cout << "41" << endl;
		  // the first interval is completely included in the second one
		  i1++;
		}
	      else if(lowOrd == INCLUDE && upOrd == AFTER)
		{
// 		  cout << "42" << endl;
		  // the first interval finish after the second one : switch the interval followed
		  followed = 1;
		  i2++;
		}
	      else if(lowOrd == AFTER && upOrd == AFTER)
		{
// 		  cout << "43" << endl;
		  // the first interval is not linked : we found the end of the new interval to add
		  newBUp = itv2.at(i2).bUp();
		  newUpBT = itv2.at(i2).upBT();
		  endItv = true;
// 		  i1++;
		  i2++;
		}
	      else // should not appear
		{
		  cout << "unionItv : Warning. An error occured during the union of piecewise intervals." << endl;
		  exit(1);
		}
	    }
	}

      // create the new interval from the saved parameters
      Interval newItv(newBLow,newLowBT,newBUp,newUpBT);
      
//       cout << "itv append : "; newItv.display(); cout << endl;

      // append it to the end of the new piecewise interval
      res.appendSItv(newItv);
    }
}

/**
 *  \brief compute the complement of interval itv1 in interval itv2
 *  \param itv1 : the first interval
 *  \param itv2 : the second interval
 *  \param res : the resulting interval
 */
void complementItv(const Interval& itv1,const Interval& itv2,PiecewiseItv& res)
{
}

/**
 *  \brief compute the complement of interval itv1 in interval itv2
 *  \param itv1 : the first interval
 *  \param itv2 : the second interval
 *  \param res : the resulting interval
 */
void complementItv(const PiecewiseItv& itv1,const Interval& itv2,PiecewiseItv& res)
{
//   cout << "///////////////////////////////////////////" << endl;
//   cout << "itv1 : "; itv1.display(); cout << endl;
//   cout << "itv2 : "; itv2.display(); cout << endl;

  // reinit the result interval
  res.empty();

  // check if itv2 is empty
  if(itv2.isEmpty())
    {
      // the result is empty
      return;
    }

  if(itv1.isEmpty())
    {
      // the result is itv2
      res.appendSItv(itv2);
      return;
    }

  // check the first interval of itv1 included in itv2
  int i = 0;
  while(itv1.at(i).bUp() < itv2.bLow() && i < itv1.getNbSItv()) i++;

  if(i < itv1.getNbSItv() && itv2.bLow() < itv1.at(i).bLow())
    {
      // add the interval that is before the first interval of itv1
      bound_t bt;
      if(itv1.at(i).lowBT() == CLOSE_B) bt = OPEN_B;
      else bt = CLOSE_B;
      Interval firstItv(itv2.bLow(),itv2.lowBT(),itv1.at(i).bLow(),bt);
      res.appendSItv(firstItv);
    }

  bool fin = false;
  while(itv1.at(i).bUp() < itv2.bUp() && !fin)
    {
//       cout << "i : " << i << endl;

      if(i == itv1.getNbSItv()-1)
	{
// 	  cout << "1" << endl;

	  // case of the last interval of itv1
	  bound_t bt;
	  if(itv1.at(i).upBT() == CLOSE_B) bt = OPEN_B;
	  else bt = CLOSE_B;
	  Interval lastItv(itv1.at(i).bUp(),bt,itv2.bUp(),itv2.upBT());
	  res.appendSItv(lastItv);
	  fin = true;
	}
      else
	{
// 	  cout << "2" << endl;

	  // get back the last value of the current interval of itv1
	  double b1 = itv1.at(i).bUp();
	  bound_t bt1;
	  if(itv1.at(i).upBT() == CLOSE_B) bt1 = OPEN_B;
	  else bt1 = CLOSE_B;

	  // get back the first value of the next interval of itv1
	  double b2 = itv1.at(i+1).bLow();
	  bound_t bt2;
	  if(itv1.at(i+1).lowBT() == CLOSE_B) bt2 = OPEN_B;
	  else bt2 = CLOSE_B;

// 	  cout << "b1 : " << b1 << "  b2 : " << b2 << endl;

	  if(b2 < itv2.bUp())
	    {
// 	      cout << "21" << endl;
	      Interval newItv(b1,bt1,b2,bt2);
	      res.appendSItv(newItv);
	      i++;
	    }
	  else
	    {
// 	      cout << "22" << endl;
	      Interval newItv(b1,bt1,itv2.bUp(),itv2.upBT());
	      res.appendSItv(newItv);
	      fin = true;
	    }
	}
    }
}
