#include "math/inc/MultFunc.hpp"
using namespace mpc;

#include <iostream>
#include <stdlib.h>
using namespace std;

                                    /**************************************************/
                                    /***************** Class MultFunc *****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
MultFunc::MultFunc() : Operation(2)
{
  this->op_ = MULT_OP;

  this->setBounds(0.0,-1.0);

  this->type_ = MULT_FUNC;
}

/**
 *  \brief init constructor
 *  \param f1 : the first operand
 *  \param f2 : the second operand
 *
 *  Input pointers are directly assigned to the structure
 */
MultFunc::MultFunc(Function* f1, Function* f2) : Operation(2)
{
  this->op_ = MULT_OP;
  this->tabOperand_[0] = f1;
  this->tabOperand_[1] = f2;

  this->setBounds(max(f1->getBInf(),f2->getBInf()),min(f1->getBSup(),f2->getBSup()));

  this->type_ = MULT_FUNC;
}

// /**
//  *  \brief init constructor
//  *  \param f1 : the first operand
//  *  \param f2 : the second operand
//  *
//  *  A copy of the input functions is performed
//  */
// MultFunc::MultFunc(const Function& f1, const Function& f2) : Operation(2)
// {
//   this->op_ = MULT_OP;
//   this->tabOperand_[0] = f1.copy();
//   this->tabOperand_[1] = f2.copy();

//   this->setBounds(max(f1.getBInf(),f2.getBInf()),min(f1.getBSup(),f2.getBSup()));
// }

/**
 *  \brief copy constructor
 */
MultFunc::MultFunc(const MultFunc& op) : Operation(op)
{
}

/**
 *  \brief destructor
 */
MultFunc::~MultFunc()
{
}

/**
 *  \brief return a copy of the current function
 *  \return a copy of the current function
 *
 *  A new instance of function is allocated and returned here
 */
MultFunc* MultFunc::copy() const
{
  return new MultFunc(*this);
}


/**************** Setters & Getters ***************/

/**
 *  \brief return the value of the polynomial at a given parameter t
 *  \param t : the input parameter value
 *  \return the value of the polynomial at a given parameter t
 */
double MultFunc::getVal(double t) const
{
  if(t >= binf_ && t <= bsup_)
    {
      if(tabOperand_[0] != NULL && tabOperand_[1] != NULL)
	{
	  return tabOperand_[0]->getVal(t) * tabOperand_[1]->getVal(t);
	}
      else 
	{
	  cerr << "MultFunc::getVal : Operand functions are not defined." << endl;
	  exit(1);
	}
   }
  else return 0;
}

  
/******************** Operators *******************/

/**
 *  \brief multiply the current function by a constant value
 *  \param v : the input multiplication coefficient
 */
void MultFunc::operator*=(double v)
{
  cerr << "MultFunc::operator*= : Not defined yet!" << endl;
  exit(1);
}


/************ Integration & Derivation ************/

/**
 *  \brief return the derivation of the current function
 *  \return the derivation of the current function
 */
Function* MultFunc::derivation() const
{
  cerr << "MultFunc::derivation : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief return the integration of the current function
 *  \return the integration of the current function
 */
Function* MultFunc::integration() const
{
  cerr << "MultFunc::integration : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief return the integration of the current function
 *  \param cste : the integration constant at t=0
 *  \return the integration of the current function
 */
Function* MultFunc::integration(double cste) const
{
  cerr << "MultFunc::integration : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief return the integration of the current function
 *  \param cste,binf : cste is the integration constant value defined at t=binf
 *  \return the integration of the current function
 */
Function* MultFunc::integration(double binf,double cste) const
{
  cerr << "MultFunc::integration : Not defined yet!" << endl;
  exit(1);
}


/****************** Roots search ******************/

/**
 *  \brief return the real roots of this function
 *  \param r : the list of roots of the function
 *  \return the number of roots of the function
 */
int MultFunc::realRoots(std::vector<double>& r) const
{
  cerr << "MultFunc::realRoots : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief return the roots of the equation eq - v = 0
 *  \param r : the list of roots of the function
 *  \param v : the input constant value
 *  \return the number of roots of the function
 */
int MultFunc::realRoots(std::vector<double>& r,double v) const
{
  cerr << "MultFunc::realRoots : Not defined yet!" << endl;
  exit(1);
}


/**************** Utility functions ***************/

/**
 *  \brief display the polynomial
 */
void MultFunc::display() const
{
  cout << "[" << binf_ << ";" << bsup_ << "] -> Mult (*) : " << endl;
  if(tabOperand_ == NULL) cout << "tabOperand nul" << endl;
  else
    {
      cout << "First operand : " << endl;
      if(tabOperand_[0] == NULL) cout << "tabOperand[0] nul" << endl;
      else tabOperand_[0]->display();
      cout << "Second operand : " << endl;
      if(tabOperand_[1] == NULL) cout << "tabOperand[1] nul" << endl;
      else tabOperand_[1]->display();
    }
}

/**
 *  \brief shift the function on Y axis from value v
 *  \param v : the shifting value
 */
void MultFunc::shiftY(double v)
{
  cerr << "MultFunc::shiftY : Not defined yet!" << endl;
  exit(1);
}

/**
 *  \brief shift the time parameter of the function
 *  \param t : the shifting value
 */
void MultFunc::shiftTime(double t)
{
  if(tabOperand_ != NULL && tabOperand_[0] != NULL && tabOperand_[1] != NULL) 
    {
      tabOperand_[0]->shiftTime(t);
      tabOperand_[1]->shiftTime(t);
    }
}


/***************** Discontinuities ****************/

/**
 *  \brief search for the discontinuities of the function
 *  \param discPoints : resulting discontinuity points
 */
void MultFunc::getVecDiscPoints(std::vector<Disc_Point>& discPoints) const
{
  if(tabOperand_ != NULL)
    {
      if(tabOperand_[0] != NULL && tabOperand_[1] != NULL)
	{
	  vector<Disc_Point> discPoints1;
	  vector<Disc_Point> discPoints2;

	  tabOperand_[0]->getVecDiscPoints(discPoints1);
	  tabOperand_[1]->getVecDiscPoints(discPoints2);

	  Function::mergeDiscPoints(discPoints1,discPoints2,discPoints);
	}
    }
}
