#ifndef MPC_GPRIMPATH_HPP
#define MPC_GPRIMPATH_HPP

#include "motion/inc/Path.hpp"

/**
 * \enum path_t
 * \brief Enumeration of the different geometrical primitive path types
 *
 * STR_LINE  : Straight line
 * CIRC_BOW  : Circle bow
 * CLOTH_BOW : Clothoid bow
 */
typedef enum {
  ND_GPT    = -1,
  STR_LINE  =  0,
  CIRC_BOW  =  1,
  CLOTH_BOW =  2
} geom_path_t;

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class GPrimPath ****************/
                                    /**************************************************/

/** \class GPrimPath
 *  \brief Path defined by a geometric primitive
 */
class GPrimPath : public Path
{
protected : 
  DPMConfig initCfg_;                             /**< initial configuration of the path */
  double length_;                                 /**< distance travelled along the path */
  geom_path_t type_;                              /**< the type of the geometrical primitive path */

  /********** Constructors and destructors **********/

protected : 
  /**
   *  \brief default constructor
   *
   *  Only used by subclasses
   */
  GPrimPath();

  /**
   *  \brief copy constructor
   *
   *  Only used by subclasses
   */
  GPrimPath(const GPrimPath& p);

public : 
  /**
   *  \brief destructor
   */
  virtual ~GPrimPath();

  /************** Getters and setters ***************/

public : 
  /**
   *  \brief return the type of the geometrical primitive path
   *  \return the type of the geometrical primitive path
   */
  inline geom_path_t type() const { return this->type_; }

  /**
   *  \brief return the initial configuration of the path
   *  \return the initial configuration of the path
   */
  inline virtual const DPMConfig& initCfg() const { return this->initCfg_; }

  /**
   *  \brief return the goal configuration of the path
   *  \return the goal configuration of the path
   */
  virtual DPMConfig goalCfg() const = 0;

  /**
   *  \brief return the length of the path
   *  \return the length of the path
   */
  inline double getLength() const { return this->length_; }

  /**
   *  \brief return a pointer to a copy of the current primitive path
   *  \return a pointer to a copy of the current primitive path
   */
  GPrimPath* copy() const;
  
  /**************** Path modification ***************/

public : 
  /**
   *  \brief reverse the current path
   */
  virtual void reverse() = 0;
    
  /**************** Utility functions ***************/

public : 
  /**
   *  \brief diplay the path
   */
  virtual void display() const = 0;

  /**
   *  \brief display in an openGL viewer
   */
  virtual void toGL() const = 0;
  
  /**
   *  \brief return the barycenter of the path in the workspace
   */
  virtual PositionXY center() const = 0;

  /**
   *  \brief compute an approximative radius of the set of all configurations of the path
   *  \return an approximative radius of the set of all configurations of the path
   */
  virtual double radius() const = 0;

  /************* Input/Output functions *************/

public : 
  /**
   *  \brief load the path from an input file
   *  \param nomfic : name of the input file
   */
  virtual bool load(char* nomfic) = 0;
  
  /**
   *  \brief save the path in an output file
   *  \param nomfic : name of the output file
   */
  virtual bool save(char* nomfic) const = 0;
};

};
#endif // MPC_GPRIMPATH_HPP


#ifndef MPC_GEOMPATH_HPP
#define MPC_GEOMPATH_HPP

#include <list>

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class GeomPath *****************/
                                    /**************************************************/

/** \class GeomPath
 *  \brief Path composed of geometric primitive pathes
 */
class GeomPath : public Path
{
protected : 
  std::list<GPrimPath*> *listPrim_;                 /**< list of the primitives that compose the path */

  /********** Constructors and destructors **********/

protected : 
  /**
   *  \brief default constructor
   *
   *  Only used by subclasses
   */
  GeomPath();

  /**
   *  \brief copy constructor
   *
   *  Only used by subclasses
   */
  GeomPath(const GeomPath& p);

public : 
  /**
   *  \brief destructor
   */
  virtual ~GeomPath();

  /************** Getters and setters ***************/

public : 
  /**
   *  \brief return the initial configuration of the path
   *  \return the initial configuration of the path
   */
  virtual DPMConfig initCfg() const;

  /**
   *  \brief return the goal configuration of the path
   *  \return the goal configuration of the path
   */
  virtual DPMConfig goalCfg() const;

  /**
   *  \brief return the length of the path
   *  \return the length of the path
   */
  double getLength() const;
  
protected : 
  /**
   *  \brief getter : size of the list of primitives
   *  \return the size of the list of primitives
   */
  inline int size() const { return this->listPrim_->size(); }

  /**
   *  \brief return the i-th primitive of the path
   *  \param i : index of the primitive to return
   *  \return the i-th primitive of the path
   */
  virtual const GPrimPath& at(int i) const;

  /**************** Path modification ***************/

public : 
  /**
   *  \brief append a new primitive to the path
   *  \param c : the new primitive to append
   */
  void append(const GPrimPath& p);

  /**
   *  \brief prepend a new primitive to the path
   *  \param c : the the primitive to prepend
   */
  void prepend(const GPrimPath& c);

  /**
   *  \brief concatenate a path to the current one
   *  \param p : the input path to concatenate
   */
  void concat(const GeomPath& p);

  /**
   *  \brief reverse the current path
   */
  virtual void reverse();

  /**
   *  \brief clear the list of primitives
   */
  void clear();
    
  /**************** Utility functions ***************/

public : 
  /**
   *  \brief diplay the path
   */
  virtual void display() const;

  /**
   *  \brief display in an openGL viewer
   */
  virtual void toGL() const;
  
  /**
   *  \brief return the barycenter of the path in the workspace
   */
  virtual PositionXY center() const;

  /**
   *  \brief compute an approximative radius of the set of all configurations of the path
   *  \return an approximative radius of the set of all configurations of the path
   */
  virtual double radius() const;

  /************* Input/Output functions *************/

public : 
  /**
   *  \brief load the path from an input file
   *  \param nomfic : name of the input file
   */
  virtual bool load(char* nomfic);
  
  /**
   *  \brief save the path in an output file
   *  \param nomfic : name of the output file
   */
  virtual bool save(char* nomfic) const;
};

};
#endif // MPC_GEOMPATH_HPP


#ifndef MPC_STRLINEPATH_HPP
#define MPC_STRLINEPATH_HPP

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class StrLinePath ***************/
                                    /**************************************************/

/** \class StrLinePath
 *  \brief Straight line path followed by a system
 */
class StrLinePath : public GPrimPath
{
  /********** Constructors and destructors **********/

protected : 
  /**
   *  \brief default constructor
   */
  StrLinePath();

public : 
  /**
   *  \brief copy constructor
   */
  StrLinePath(const StrLinePath& p);
  
  /**
   *  \brief copy constructor
   */
  StrLinePath(const DPMConfig& initCfg,double dist);  

  /**
   *  \brief copy constructor
   */
  StrLinePath(const PMConfig& initCfg,const PMConfig& goalCfg);  

  /**
   *  \brief destructor
   */
  virtual ~StrLinePath();

  /************** Getters and setters ***************/

public : 
  /**
   *  \brief return the goal configuration of the path
   *  \return the goal configuration of the path
   */
  virtual DPMConfig goalCfg() const;

  /**************** Path modification ***************/

public : 
  /**
   *  \brief reverse the current path
   */
  virtual void reverse();

  /**************** Utility functions ***************/

public : 
  /**
   *  \brief diplay the path
   */
  virtual void display() const;

  /**
   *  \brief display in an openGL viewer
   */
  virtual void toGL() const;
  
  /**
   *  \brief return the barycenter of the path in the workspace
   */
  virtual PositionXY center() const;

  /**
   *  \brief compute an approximative radius of the set of all configurations of the path
   *  \return an approximative radius of the set of all configurations of the path
   */
  virtual double radius() const;

  /************* Input/Output functions *************/

public : 
  /**
   *  \brief load the path from an input file
   *  \param nomfic : name of the input file
   */
  virtual bool load(char* nomfic);
  
  /**
   *  \brief save the path in an output file
   *  \param nomfic : name of the output file
   */
  virtual bool save(char* nomfic) const;
};

};
#endif // MPC_STRLINEPATH_HPP
