#ifndef MPC_MANASIMPLESTRAJPLANNER_HPP
#define MPC_MANASIMPLESTRAJPLANNER_HPP

#include "planner/inc/SimpleSTrajPlanner.hpp"
#include "system/inc/ManaSys.hpp"
#include "trajGen/inc/ManaTrajGen.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /********** Class ManaSimpleSTrajPlanner **********/
                                    /**************************************************/

  /** \class ManaSimpleSTrajPlanner
   *  \brief Simple state trajectory planner for a Mana robotic system (no collision checking)
   */
  class ManaSimpleSTrajPlanner : public SimpleSTrajPlanner
  {
  protected : 

    /********** Constructors and destructors **********/
  
  protected : 
    /**
     *  \brief default constructor
     *  Only used by subclasses
     */
    ManaSimpleSTrajPlanner();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the robotic system handled
     *  \param coefVMax : maximal velocity coefficient
     *  \param coefAMax : maximal acceleration coefficient
     *  \param debug : the debugging level
     *  
     *  coefVMax \in [0;1]
     *  coefAMax \in [0;1]
     */
    ManaSimpleSTrajPlanner(ManaSys* sys, double coefVMax=0.8, double coefAMax=0.8, int debug=0);

    /**
     *  \brief destructor
     */
    virtual ~ManaSimpleSTrajPlanner();
  

    /*************** Getters and Setters **************/

  protected : 
    /**
     *  \brief return the system constraints
     *  \return the system constraints
     */
    virtual const ManaSys* sys() const;

    /**
     *  \brief return a pointer to the Mana trajectory generator
     *  \return a pointer to the Mana trajectory generator
     */
    ManaTrajGen* trajgen() const;

    /**
     *  \brief return a pointer to the Mana state trajectory
     *  \return a pointer to the Mana state trajectory
     */
    ManaPSTraj* straj() const;


    /*************** Trajectory planning **************/

  protected : 
    /**
     *  \brief Initialize the parameters of the Mana trajectory generator
     */
    void initManaTrajGen();

  public : 
    /**
     *  \brief compute the trajectory from s1 to s2
     *  \param s1 : the initial state
     *  \param s2 : the final state
     *  \param time : fixed duration of the trajectory
     *  \return 0 if the process succeed, -1 otherwise
     */
    virtual int computeTraj(const ManaState& s1,const ManaState& s2,double time=0);


    /***************** Debug / Display ****************/

  public : 
    /**
     *  \brief return a copy of the current control trajectory
     *  \return a copy of the current control trajectory
     * 
     *  A new instance of ManaPSTraj is returned
     */
    virtual const ManaPSTraj* getPSTraj();

    /**
     *  \brief return a sampled planned state trajectory
     *  \param sampling : the sampling of the sate trajectory
     *  \return a copy of a sampled planned state trajectory
     * 
     *  A new instance of ManaPSTraj is returned
     */
    virtual const ManaSTraj* getSTraj(int sampling=50);
  };

};

#endif // MPC_MANASIMPLESTRAJPLANNER_HPP
