#ifndef MPC_ROADMAPPP_HPP
#define MPC_ROADMAPPP_HPP

#include "planner/inc/PathPlanner.hpp"
#include "environment/inc/RoadMap.hpp"
#include "motion/inc/Path.hpp"
#include "common/inc/Config.hpp"

#include <list>

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class RoadMapPP ****************/
                                    /**************************************************/

  /** \class RoadMapPP
   *  \brief Path planner using a roadmap
   */
  class RoadMapPP : public PathPlanner
  {
  protected : 
    RoadMap* roadmap_;

    std::vector<RoadNode*>* vecNgh_;
    std::list<double>* vecMinDist_;

    /************* Constructors and destructors ************/

  protected : 
    /**
     *  \brief default constructor
     */
    RoadMapPP();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : robotic system handled
     *  \param e : environment in which the system evolve
     */
    RoadMapPP(System* sys,Environment* e);

    //   /**
    //    *  \brief init constructor
    //    *  \param sys : robotic system handled
    //    *  \param rm : roadmap defined over the environment
    //    */
    //   RoadMapPP(System* sys,RoadMap* rm);

    /**
     *  \brief init constructor
     *  \param sys : robotic system handled
     *  \param e : environment in which the system evolve
     *  \param rm : roadmap defined over the environment
     */
    RoadMapPP(System* sys,Environment* e,RoadMap* rm);

    /**
     *  \brief destructor
     */
    virtual ~RoadMapPP();

    /********************** Accessors **********************/

  public : 
    /**
     *  \brief return the current path
     *  \return the current path
     */
    inline virtual const PMPath* getPath() const { return ((PMPath*)path_); }

    /**
     *  \brief return the initial configuration
     *  \return the initial configuration
     */
    inline virtual const PMConfig& initCfg() const { return *((DPMConfig*)initCfg_); }

    /**
     *  \brief return the goal configuration
     *  \return the goal configuration
     */
    inline virtual const PMConfig& goalCfg() const { return *((DPMConfig*)goalCfg_); }

    /**
     *  \brief return the list of minimal distances to obstacles for all configurations of the path computed
     *  \return the list of minimal distances
     */
    inline std::list<double>* vecMinDist() const { return this->vecMinDist_; }
  
  protected : 
    /**
     *  \brief return the current path
     *  \return the current path
     */
    inline virtual PMPath* path() { return ((DPMPath*)path_); }  

    /**
     *  \brief return the initial configuration
     *  \return the initial configuration
     */
    inline virtual PMConfig& initCfg() { return *((DPMConfig*)initCfg_); }

    /**
     *  \brief return the goal configuration
     *  \return the goal configuration
     */
    inline virtual PMConfig& goalCfg() { return *((DPMConfig*)goalCfg_); }

    /******************** Initialization *******************/

  public : 
    /**
     *  \brief reinit the planner and the roadmap if needed
     */
    void reinit();

    /******************* Path Computing ********************/

  public : 
    /**
     *  \brief compute the path from cfg1 to cfg2 on map
     *  \param cfg1 : the initial configuration
     *  \param cfg2 : the goal configuration
     *  \return the path planning status
     *       ERR_STD_MAP    : the roadmap has not been defined
     *       ERR_GOAL_POS   : the goal configuration is not available or free 
     *       ERR_INIT_POS   : the init configuration is not available or free
     *       NO_PATH        : no path has been found
     *       PATH_CPTD      : the path has been computed
     */
    virtual int computePath(const PMConfig& cfg1, const PMConfig& cfg2);

  protected : 
    /**
     *  \brief compute the path between nodes init and goal on the current map
     *  \param init : the initial node of the roadmap
     *  \param goal : the goal node of the roadmap
     *  \return the path planning status
     *       ERR_STD_MAP    : the roadmap has not been defined
     *       ERR_GOAL_POS   : the goal configuration is not available or free 
     *       ERR_INIT_POS   : the init configuration is not available or free
     *       NO_PATH        : no path has been found
     *       PATH_CPTD      : the path has been computed
     */
    int computePath(RoadNode* init, RoadNode* goal);
  
    /**
     *  \brief construct the path to goalNode from the list of visited nodes
     *  \param goalNode : the goal node
     */
    void constructPath(RoadNode* goalNode);

    /**
     *  \brief add a new node to the list of visitable neighbours
     *  \param n : the new node to add
     */
    void addNgh(RoadNode* n);

    /**
     *  \brief return the node of the list of visitable neighbours with the best heuristic
     *  \return the node of the list of visitable neighbours with the best heuristic
     *
     *  The node returned is deleted from the list of neighbours
     */
    RoadNode* getBestNgh();

    /**
     *  \brief smooth the resulting path
     *  \return 0 if the process succeed, -1 if a collison has been found on one of the new nodes added
     */
    int smoothPath();
  
    /**
     *  delete the second and the second to last configurations of the path
     */
    void del2ndCfg();
  };

};
#endif // MPC_ROADMAPPP_HPP
