/********************************************************************************
*                                                                               *
*                              Simple FOX Viewer                                *
*                                                                               *
********************************************************************************/
#include "fx.h"

#include <time.h>
#include <sys/time.h>

#include <iostream>
using namespace std;

#define DEBUG 1

const FXuint TIMER_INTERVAL = 1000;

// Main Window
class SimpleViewer : public FXMainWindow {

  // Macro for class hierarchy declarations
  FXDECLARE(SimpleViewer)

private:

  // viewer components
  FXHorizontalFrame           *contents;                // Content frame
  FXVerticalFrame             *canvasFrame;             // Canvas frame
  FXVerticalFrame             *buttonFrame;             // Button frame
  FXCanvas                    *canvas;                  // Canvas to draw into
  FXColor                     drawColor;                // Color for drawing
  FXColor                     bgColor;                  // Color of the background

  // application components
  struct timeval              t1;                       // init time of the timer
  struct timeval              t2;                       // final time of the timer
  bool                        running;                  // true if the timer is running

  double                      time;                     // the current time
  double                      alreadyElapsedTime;       // the time which has been already elapsed

  int                         step;                     // the current step

protected:
  SimpleViewer(){}

public:

  // Message handlers
  long onStart(FXObject*,FXSelector,void*);
  long onStop(FXObject*,FXSelector,void*);
  long onReinit(FXObject*,FXSelector,void*);
  long onTimeout(FXObject*,FXSelector,void*);

  // Others methods
  void runCmd();                                        // run the main command
  void display();                                       // display the result of the main command

  void drawPoint(int x, int y);                         // draw a simple point on the canvas
  void drawLine(int x1, int y1, int x2, int y2);        // draw a line between (x1,y1) and (x2,y2) on the canvas
  void drawText(int x, int y, char* txt);               // draw the text txt on the canvas from (x,y)
  void setDrawColor(int r, int g, int b);               // set the drawing color
  void refresh();                                       // refresh the canvas

public:

  // Messages for our class
  enum{
    ID_CANVAS=FXMainWindow::ID_LAST,
    ID_START,
    ID_STOP,
    ID_REINIT,
    ID_TIMEOUT,
    ID_LAST
    };

public:

  // ScribbleWindow's constructor
  SimpleViewer(FXApp* a);

  // Initialize
  virtual void create();

  virtual ~SimpleViewer();
  };



// Message Map for the Scribble Window class
FXDEFMAP(SimpleViewer) SimpleViewerMap[]=
{
  //________Message_Type_____________________ID____________________Message_Handler____________
  FXMAPFUNC(SEL_COMMAND,        SimpleViewer::ID_START,              SimpleViewer::onStart),
  FXMAPFUNC(SEL_COMMAND,        SimpleViewer::ID_STOP,               SimpleViewer::onStop),
  FXMAPFUNC(SEL_COMMAND,        SimpleViewer::ID_REINIT,             SimpleViewer::onReinit),
  FXMAPFUNC(SEL_TIMEOUT,        SimpleViewer::ID_TIMEOUT,            SimpleViewer::onTimeout),
};


// Macro for the ScribbleApp class hierarchy implementation
FXIMPLEMENT(SimpleViewer,FXMainWindow,SimpleViewerMap,ARRAYNUMBER(SimpleViewerMap))




////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////// Constructors & Init functions //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Construct a ScribbleWindow
SimpleViewer::SimpleViewer(FXApp *a):FXMainWindow(a,"Simple Viewer",NULL,NULL,DECOR_ALL,0,0,800,600){

  /***************************************************************************************************************
   ********************************************** initialisation *************************************************
   ***************************************************************************************************************/

  drawColor          = FXRGB(255,0,0);
  bgColor            = FXRGB(255,255,255);

  running            = false;

  time               = 0;
  alreadyElapsedTime = 0;
  step               = 0;


  /***************************************************************************************************************
   ************************************************ main frame ***************************************************
   ***************************************************************************************************************/

  contents=new FXHorizontalFrame(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  /***************************************************************************************************************
   *********************************************** canvas frame **************************************************
   ***************************************************************************************************************/

  // LEFT pane to contain the canvas
  canvasFrame=new FXVerticalFrame(contents,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  // Label above the canvas
  new FXLabel(canvasFrame,"Canvas Frame",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  
  // Horizontal divider line
  new FXHorizontalSeparator(canvasFrame,SEPARATOR_GROOVE|LAYOUT_FILL_X);

  // Drawing canvas
  canvas=new FXCanvas(canvasFrame,this,ID_CANVAS,FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_FILL_ROW|LAYOUT_FILL_COLUMN);
  canvas->repaint();

  /***************************************************************************************************************
   *********************************************** button frame **************************************************
   ***************************************************************************************************************/

  // RIGHT pane for the buttons
  buttonFrame=new FXVerticalFrame(contents,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  // Label above the buttons
  new FXLabel(buttonFrame,"Button Frame",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  
  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);
  
  // Start button
  new FXButton(buttonFrame,"&Start",NULL,this,SimpleViewer::ID_START,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  
  // Stop button
  new FXButton(buttonFrame,"St&op",NULL,this,SimpleViewer::ID_STOP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  
  // Reinit button
  new FXButton(buttonFrame,"&Reinit",NULL,this,SimpleViewer::ID_REINIT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  
  
  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);
  
  // Exit button
  new FXButton(buttonFrame,"&Exit",NULL,getApp(),FXApp::ID_QUIT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  
}


SimpleViewer::~SimpleViewer()
{
}


// Create and initialize
void SimpleViewer::create()
{
  // Create the windows
  FXMainWindow::create();

  // Make the main window appear
  show(PLACEMENT_SCREEN);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////// Signal functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

long SimpleViewer::onStart(FXObject*,FXSelector,void*)
{
#if DEBUG >= 1
  cout << "starting timeout" << endl;
#endif

  // init the timer
  gettimeofday(&t1,NULL);
  running = true;
  getApp()->addTimeout(this,ID_TIMEOUT,TIMER_INTERVAL);

  return 1;
}

long SimpleViewer::onStop(FXObject*,FXSelector,void*)
{
#if DEBUG >= 1
  cout << "stoping timeout" << endl;
#endif

  running = false;
  getApp()->removeTimeout(this,ID_TIMEOUT);
  alreadyElapsedTime = time;
  
  return 1;
}

long SimpleViewer::onReinit(FXObject*,FXSelector,void*)
{
  this->refresh();

  // reinit the time already past
  time = 0;
  alreadyElapsedTime = 0;
  step = 0;

  return 1;
}

long SimpleViewer::onTimeout(FXObject*,FXSelector,void*)
{
  // compute the time past
  gettimeofday(&t2,NULL);
  time_t elapsedTs = t2.tv_sec - t1.tv_sec;
  suseconds_t elapsedTus = t2.tv_usec - t1.tv_usec;
  
  double elapsedT = alreadyElapsedTime + ((double)elapsedTs) + 0.000001 * ((double)elapsedTus);
  
  // set the current time
  time = elapsedT;
  step++;

#if DEBUG >= 1
  cout << "time : " << time << "\t step : " << step << endl;
#endif

  // run the main command
  this->runCmd();

  // display the result of the command
  this->display();

  getApp()->addTimeout(this,ID_TIMEOUT,TIMER_INTERVAL);

  return 1;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////// Other functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void SimpleViewer::runCmd()
{
}

void SimpleViewer::display()
{
}

// display a simple point (x,y) on the canvas
void SimpleViewer::drawPoint(int x, int y)
{
  FXDCWindow dc(canvas);
  dc.setForeground(drawColor);
  dc.drawPoint(x,y);
}

// display a line between (x1,y1) and (x2,y2) on the canvas
void SimpleViewer::drawLine(int x1, int y1, int x2, int y2)
{
  FXDCWindow dc(canvas);
  dc.setForeground(drawColor);
  dc.drawLine(x1, y1, x2, y2);
}

// set the drawing color
void SimpleViewer::setDrawColor(int r, int g, int b)
{
  drawColor=FXRGB(r,g,b);
}

// draw the text txt on the canvas from (x,y)
void SimpleViewer::drawText(int x, int y, char* txt)
{
  FXDCWindow dc(canvas);
  dc.setForeground(drawColor);
  dc.drawText(x,y,FXString(txt));
}

// refresh the canvas
void SimpleViewer::refresh()
{
  FXDCWindow dc(canvas);
  dc.setForeground(bgColor);
  dc.fillRectangle(0,0,canvas->getWidth(),canvas->getHeight());
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////// Main ///////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Here we begin
int main(int argc,char *argv[]){

  // Make application
  FXApp application("SimpleViewer","SimpleViewer");

  // Start app
  application.init(argc,argv);

  // create the simple viewer
  new SimpleViewer(&application);

  // Create the application's windows
  application.create();

  // Run the application
  return application.run();
  }

