#include "CTrajPlanner.hpp"
using namespace mpc;
                                    /**************************************************/
                                    /*************** Class CTrajPlanner ***************/
                                    /**************************************************/

/************* Constructors and destructors ************/
  
/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
CTrajPlanner::CTrajPlanner() : TrajPlanner()
{
  this->trajgen_ = NULL;
}

/**
 *  \brief init constructor
 *  \param sys : robotic system handled
 *  \param e   : environment in which the system evolve
 */
CTrajPlanner::CTrajPlanner(System* sys,Environment* e) : TrajPlanner(sys,e)
{
  this->trajgen_ = NULL;
}

/**
 *  \brief destructor
 */
CTrajPlanner::~CTrajPlanner()
{
  if(trajgen_ != NULL) delete trajgen_;
}


                                    /**************************************************/
                                    /************** Class CCCTrajPlanner **************/
                                    /**************************************************/

/************* Constructors and destructors ************/
  
/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
CCCTrajPlanner::CCCTrajPlanner() : CTrajPlanner()
{
}

/**
 *  \brief init constructor
 *  \param sys : robotic system handled
 *  \param e   : environment in which the system evolve
 */
CCCTrajPlanner::CCCTrajPlanner(System* sys,Environment* e) : CTrajPlanner(sys,e)
{
}

/**
 *  \brief destructor
 */
CCCTrajPlanner::~CCCTrajPlanner()
{
}

/********************** Accessors **********************/

/**
 *  \brief return the current trajectory
 *  \return the current trajectory
 */
const CLSTraj* CCCTrajPlanner::getTraj() const
{
  if(this->status_ == TRAJ_CPTD) return ((CLSTraj*)(this->traj_));
  else return NULL;
}

/**************** Trajectory Computation ***************/

/**
 *  \brief compute the state reach from an initial state and an input control
 *  \param s1 : the initial state
 *  \param ctrl : the input control time
 */
CLStateT CCCTrajPlanner::computeStateReachedFrom(const CLStateT& s1,const CCControlT& ctrl)
{
  // reinit the trajectory generator
  if(trajgen_ != NULL) delete trajgen_;
  trajgen_ = new CLTrajGen(sys());
  
  // copy the input state to modify its velocity and steering angle parameters
  CLStateT s1copy(s1);
  s1copy.s() = ctrl.v();
  s1copy.ph() = ctrl.ph();

//   cout << "CCCTrajPlanner::computeStateReachedFrom -> ctrl :";
//   ctrl.display(); cout << endl;

//   cout << "CCCTrajPlanner::computeStateReachedFrom -> scopy :";
//   s1copy.display(); cout << endl;

  // initialize the vector of parameters
  double param[7] = {0,0,0,0,0,0,ctrl.t()};

  // compute the state reached
  trajgen()->setDBG(1);
  trajgen()->ComputeFromIG(s1copy,param);
  CLStateT sreached = trajgen()->getFinalCLStateTReached();
  sreached.t() += s1.t();
  return sreached;
}

/**
 *  \brief compute the trajectory from a predefined input sequence of controls
 *  \param initS : the initial state of the trajectory
 */
void CCCTrajPlanner::computeDefinedTraj(const CLState& initS)
{
  // reinit the result trajectory
  if(traj_ != NULL) delete traj_;
  traj_ = new CLSTraj();

  // set the current state with the input state
  CLStateT curS;
  curS.x()  = initS.x();
  curS.y()  = initS.y();
  curS.th() = initS.th();
  curS.ph() = initS.ph();
  curS.s()  = initS.s();
  curS.t()  = 0;

  for(int i=0; i<30; i++)
    {
      // set the control to apply
      CCControlT ctrl;
      ctrl.t() = 0.5;

      switch(i)
	{
	case  0 : ctrl.v() = 0.0; ctrl.ph() = 0.0; break;
	case  1 : ctrl.v() = 0.2; ctrl.ph() = 0.0; break;
	case  2 : ctrl.v() = 0.5; ctrl.ph() = 0.0; break;
	case  3 : ctrl.v() = 0.8; ctrl.ph() = 0.0; break;
	case  4 : ctrl.v() = 1.2; ctrl.ph() = 0.1; break;
	case  5 : ctrl.v() = 1.6; ctrl.ph() = 0.2; break;
	case  6 : ctrl.v() = 2.0; ctrl.ph() = 0.3; break;
	case  7 : ctrl.v() = 2.0; ctrl.ph() = 0.4; break;
	case  8 : ctrl.v() = 2.0; ctrl.ph() = 0.4; break;
	case  9 : ctrl.v() = 2.0; ctrl.ph() = 0.3; break;
	case 10 : ctrl.v() = 2.0; ctrl.ph() = 0.2; break;
	case 11 : ctrl.v() = 2.0; ctrl.ph() = 0.1; break;
	case 12 : ctrl.v() = 2.0; ctrl.ph() = 0.0; break;
	case 13 : ctrl.v() = 2.0; ctrl.ph() = 0.0; break;
	case 14 : ctrl.v() = 2.0; ctrl.ph() = 0.0; break;
	case 15 : ctrl.v() = 2.0; ctrl.ph() = 0.0; break;
	case 16 : ctrl.v() = 2.0; ctrl.ph() = 0.0; break;
	case 17 : ctrl.v() = 2.0; ctrl.ph() = 0.0; break;
	case 18 : ctrl.v() = 2.0; ctrl.ph() = -0.1; break;
	case 19 : ctrl.v() = 2.0; ctrl.ph() = -0.2; break;
	case 20 : ctrl.v() = 2.0; ctrl.ph() = -0.3; break;
	case 21 : ctrl.v() = 2.0; ctrl.ph() = -0.4; break;
	case 22 : ctrl.v() = 2.0; ctrl.ph() = -0.4; break;
	case 23 : ctrl.v() = 2.0; ctrl.ph() = -0.3; break;
	case 24 : ctrl.v() = 1.6; ctrl.ph() = -0.2; break;
	case 25 : ctrl.v() = 1.2; ctrl.ph() = -0.1; break;
	case 26 : ctrl.v() = 0.8; ctrl.ph() = 0.0; break;
	case 27 : ctrl.v() = 0.5; ctrl.ph() = 0.0; break;
	case 28 : ctrl.v() = 0.2; ctrl.ph() = 0.0; break;
	case 29 : ctrl.v() = 0.0; ctrl.ph() = 0.0; break;
	}

      // compute the state reached from the current state and the input control
      CLStateT sreached = this->computeStateReachedFrom(curS,ctrl);

      // append the new state to the trajectory
      this->traj()->append(sreached);

      // set the current state as the new state reached
      curS = sreached;
    }

  this->status_ = TRAJ_CPTD;
  return;
}

/**
 *  \brief compute the trajectory from a predefined input sequence of controls
 *  \param initS : the initial state of the trajectory
 */
void CCCTrajPlanner::computeDefinedTraj2(const CLState& initS)
{
  // reinit the result trajectory
  if(traj_ != NULL) delete traj_;
  traj_ = new CLSTraj();

  // set the current state with the input state
  CLStateT curS;
  curS.x()  = initS.x();
  curS.y()  = initS.y();
  curS.th() = initS.th();
  curS.ph() = initS.ph();
  curS.s()  = initS.s();
  curS.t()  = 0;

  for(int i=0; i<30; i++)
    {
      // set the control to apply
      CCControlT ctrl;
      ctrl.t() = 0.5;

      switch(i)
	{
	case  0 : ctrl.v() = 0.0; ctrl.ph() = 0.0; break;
	case  1 : ctrl.v() = 2.0; ctrl.ph() = 0.0; break;
	case  2 : ctrl.v() = 5.0; ctrl.ph() = 0.0; break;
	case  3 : ctrl.v() = 8.0; ctrl.ph() = 0.0; break;
	case  4 : ctrl.v() = 12.0; ctrl.ph() = 0.1; break;
	case  5 : ctrl.v() = 16.0; ctrl.ph() = 0.2; break;
	case  6 : ctrl.v() = 20.0; ctrl.ph() = 0.3; break;
	case  7 : ctrl.v() = 20.0; ctrl.ph() = 0.4; break;
	case  8 : ctrl.v() = 20.0; ctrl.ph() = 0.4; break;
	case  9 : ctrl.v() = 20.0; ctrl.ph() = 0.3; break;
	case 10 : ctrl.v() = 20.0; ctrl.ph() = 0.2; break;
	case 11 : ctrl.v() = 20.0; ctrl.ph() = 0.1; break;
	case 12 : ctrl.v() = 20.0; ctrl.ph() = 0.0; break;
	case 13 : ctrl.v() = 20.0; ctrl.ph() = 0.0; break;
	case 14 : ctrl.v() = 20.0; ctrl.ph() = 0.0; break;
	case 15 : ctrl.v() = 20.0; ctrl.ph() = 0.0; break;
	case 16 : ctrl.v() = 20.0; ctrl.ph() = 0.0; break;
	case 17 : ctrl.v() = 20.0; ctrl.ph() = 0.0; break;
	case 18 : ctrl.v() = 20.0; ctrl.ph() = -0.1; break;
	case 19 : ctrl.v() = 20.0; ctrl.ph() = -0.2; break;
	case 20 : ctrl.v() = 20.0; ctrl.ph() = -0.3; break;
	case 21 : ctrl.v() = 20.0; ctrl.ph() = -0.4; break;
	case 22 : ctrl.v() = 20.0; ctrl.ph() = -0.4; break;
	case 23 : ctrl.v() = 20.0; ctrl.ph() = -0.3; break;
	case 24 : ctrl.v() = 16.0; ctrl.ph() = -0.2; break;
	case 25 : ctrl.v() = 12.0; ctrl.ph() = -0.1; break;
	case 26 : ctrl.v() = 8.0; ctrl.ph() = 0.0; break;
	case 27 : ctrl.v() = 5.0; ctrl.ph() = 0.0; break;
	case 28 : ctrl.v() = 2.0; ctrl.ph() = 0.0; break;
	case 29 : ctrl.v() = 0.0; ctrl.ph() = 0.0; break;
	}

      // compute the state reached from the current state and the input control
      CLStateT sreached = this->computeStateReachedFrom(curS,ctrl);

      // append the new state to the trajectory
      this->traj()->append(sreached);

      // set the current state as the new state reached
      curS = sreached;
    }

  this->status_ = TRAJ_CPTD;
  return;
}

/**
 *  \brief compute the trajectory from a predefined input sequence of controls
 *  \param initS : the initial state of the trajectory
 */
void CCCTrajPlanner::computeDefinedTraj3(const CLState& initS)
{
  // reinit the result trajectory
  if(traj_ != NULL) delete traj_;
  traj_ = new CLSTraj();

  // set the current state with the input state
  CLStateT curS;
  curS.x()  = initS.x();
  curS.y()  = initS.y();
  curS.th() = initS.th();
  curS.ph() = initS.ph();
  curS.s()  = initS.s();
  curS.t()  = 0;

  for(int i=0; i<30; i++)
    {
      // set the control to apply
      CCControlT ctrl;
      ctrl.t() = 0.5;

      switch(i)
	{
	case  0 : ctrl.v() = 0.0; ctrl.ph() = 0.8; break;
	case  1 : ctrl.v() = 2.0; ctrl.ph() = 0.8; break;
	case  2 : ctrl.v() = 5.0; ctrl.ph() = 0.8; break;
	case  3 : ctrl.v() = 8.0; ctrl.ph() = 0.8; break;
	case  4 : ctrl.v() = 12.0; ctrl.ph() = 0.8; break;
	case  5 : ctrl.v() = 16.0; ctrl.ph() = 0.8; break;
	case  6 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case  7 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case  8 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case  9 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 10 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 11 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 12 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 13 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 14 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 15 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 16 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 17 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 18 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 19 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 20 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 21 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 22 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 23 : ctrl.v() = 20.0; ctrl.ph() = 0.8; break;
	case 24 : ctrl.v() = 16.0; ctrl.ph() = 0.8; break;
	case 25 : ctrl.v() = 12.0; ctrl.ph() = 0.8; break;
	case 26 : ctrl.v() = 8.0; ctrl.ph() = 0.8; break;
	case 27 : ctrl.v() = 5.0; ctrl.ph() = 0.8; break;
	case 28 : ctrl.v() = 2.0; ctrl.ph() = 0.8; break;
	case 29 : ctrl.v() = 0.0; ctrl.ph() = 0.8; break;
	}

      // compute the state reached from the current state and the input control
      CLStateT sreached = this->computeStateReachedFrom(curS,ctrl);

      // append the new state to the trajectory
      this->traj()->append(sreached);

      // set the current state as the new state reached
      curS = sreached;
    }

  this->status_ = TRAJ_CPTD;
  return;
}
