#include "planner/inc/ManaSimpleSTrajPlanner.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /********** Class ManaSimpleSTrajPlanner **********/
                                    /**************************************************/

/********** Constructors and destructors **********/
  
/**
 *  \brief default constructor
 *  Only used by subclasses
 */
ManaSimpleSTrajPlanner::ManaSimpleSTrajPlanner() : SimpleSTrajPlanner()
{
}

/**
 *  \brief init constructor
 *  \param sys : the robotic system handled
 *  \param coefVMax : maximal velocity coefficient
 *  \param coefAMax : maximal acceleration coefficient
 *  \param debug : the debugging level
 *  
 *  coefVMax \in [0;1]
 *  coefAMax \in [0;1]
 */
ManaSimpleSTrajPlanner::ManaSimpleSTrajPlanner(ManaSys* sys, double coefVMax, double coefAMax, int debug) : SimpleSTrajPlanner(sys, coefVMax, coefAMax, debug)
{
  this->trajgen_ = new ManaTrajGen(sys);
  
  this->initTrajGen();
  this->initManaTrajGen();
}

/**
 *  \brief destructor
 */
ManaSimpleSTrajPlanner::~ManaSimpleSTrajPlanner()
{
}


/*************** Getters and Setters **************/

/**
 *  \brief return the system constraints
 *  \return the system constraints
 */
const ManaSys* ManaSimpleSTrajPlanner::sys() const 
{
  return ((ManaSys*)sys_); 
}

/**
 *  \brief return a pointer to the Mana trajectory generator
 *  \return a pointer to the Mana trajectory generator
 */
ManaTrajGen* ManaSimpleSTrajPlanner::trajgen() const
{
  return ((ManaTrajGen*) trajgen_);
}

/**
 *  \brief return a pointer to the Mana state trajectory
 *  \return a pointer to the Mana state trajectory
 */
ManaPSTraj* ManaSimpleSTrajPlanner::straj() const
{
  return ((ManaPSTraj*) straj_);
}


/*************** Trajectory planning **************/

/**
 *  \brief Initialize the parameters of the Mana trajectory generator
 */
void ManaSimpleSTrajPlanner::initManaTrajGen()
{
  /// Mana trajgen options

  // Mana constraints
  trajgen()->setUseVelCst(false);               // default = true
  trajgen()->setUseAccCst(false);               // default = true
  trajgen()->setUseAVelCst(false);              // default = true
  trajgen()->setUseAAccCst(false);              // default = true

  // final state constraints
  trajgen()->fixFinalXY(true);                 // default = true
  trajgen()->fixFinalTH(false);                 // default = true
  trajgen()->fixFinalV(false);                  // default = true
  trajgen()->fixFinalOM(false);                 // default = true
}

/**
 *  \brief compute the trajectory from s1 to s2
 *  \param s1 : the initial state
 *  \param s2 : the final state
 *  \param time : fixed duration of the trajectory
 *  \return 0 if the process succeed, -1 otherwise
 */
int ManaSimpleSTrajPlanner::computeTraj(const ManaState& s1,const ManaState& s2,double time)
{
  // compute the time needed to execute the trajectory
  double trajTime;
  if(time <= 0)
    {
      trajTime = this->computeTrajTime(s1.posXYTh(), s1.v(), s2.posXYTh(), s2.v());
    }
  else trajTime = time;

  cout << "time : " << time << endl;
  cout << "trajTime : " << trajTime << endl;
  
  // compute the corresponding state-times
  ManaStateT st1(s1.x(),s1.y(),s1.th(),s1.v(),s1.w(),0.0);
  ManaStateT st2(s2.x(),s2.y(),s2.th(),s2.v(),s2.w(),trajTime);

  // generation of a trajectory between s0 and sf
  // return 0 if the process succeed, -1 else
  int res = trajgen()->genTraj(st1,st2);

  /// computation of the control to apply

  // get back the resulting state trajectory
  if(res == 0)
    {
      boost::mutex::scoped_lock lock(mutexTraj_);
      if(straj_ != NULL) delete straj_;
      straj_ = trajgen()->getPSTraj();
    }
  
  return res;
}


/***************** Debug / Display ****************/

/**
 *  \brief return a copy of the current control trajectory
 *  \return a copy of the current control trajectory
 * 
 *  A new instance of ManaPSTraj is returned
 */
const ManaPSTraj* ManaSimpleSTrajPlanner::getPSTraj()
{
  if(straj_ != NULL)
    {
      boost::mutex::scoped_lock lock(mutexTraj_);
      ManaPSTraj* copyTraj = new ManaPSTraj(*straj());
      return copyTraj;
    }
  else return NULL;
}

/**
 *  \brief return a sampled planned state trajectory
 *  \param sampling : the sampling of the sate trajectory
 *  \return a copy of a sampled planned state trajectory
 * 
 *  A new instance of ManaPSTraj is returned
 */
const ManaSTraj* ManaSimpleSTrajPlanner::getSTraj(int sampling)
{
  if(straj_ != NULL)
    {
      ManaSTraj* sampledSTraj = straj()->sample(sampling);
      return sampledSTraj;
    }
  else return NULL;
}
