#include "PVDTrajPlanner.hpp"
using namespace mpc;

#include "gnuplot/inc/TrajPlot.hpp"

#include <stdio.h>

                                    /**************************************************/
                                    /************** Class PVDTrajPlanner **************/
                                    /**************************************************/

double PVDTrajPlanner::maxSpeedCoef_ = 0.6f;
double PVDTrajPlanner::maxAccCoef_ = 0.6f;

/************* Constructors and destructors ************/

/**
 *  \brief default constructor
 */
PVDTrajPlanner::PVDTrajPlanner() : TrajPlanner()
{
  this->path_     = NULL;
  this->timeStep_ = 0.5;
  this->type_     = PVD_TP;
}

/**
 *  \brief init constructor
 */
PVDTrajPlanner::PVDTrajPlanner(System* sys, Environment* e) : TrajPlanner(sys,e)
{
  this->path_     = NULL;
  this->timeStep_ = 0.5;
  this->type_     = PVD_TP;
}

/**
 *  \brief destructor
 */
PVDTrajPlanner::~PVDTrajPlanner()
{
}

// set the maximum speed coefficient
void PVDTrajPlanner::setMSC(double msc)
{
  if(msc > 0 && msc <= 1)
    {
      PVDTrajPlanner::maxSpeedCoef_ = msc;
    }
}

// // return the maximum acceleration coefficient
// double PVDTrajPlanner::getMAC()
// {
//   return PVDTrajPlanner::maxAccCoef_;
// }

// set the maximum acceleration coefficient
void PVDTrajPlanner::setMAC(double mac)
{
  if(mac > 0 && mac <= 1)
    {
      PVDTrajPlanner::maxAccCoef_ = mac;
    }
}


                                    /**************************************************/
                                    /************* Class PM_PVDTrajPlanner ************/
                                    /**************************************************/

/************* Constructors and destructors ************/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
PM_PVDTrajPlanner::PM_PVDTrajPlanner() : PVDTrajPlanner()
{
}

/**
 *  \brief init constructor
 */
PM_PVDTrajPlanner::PM_PVDTrajPlanner(PointMassSys* sys) : PVDTrajPlanner(sys,NULL)
{
}

/**
 *  \brief destructor
 */
PM_PVDTrajPlanner::~PM_PVDTrajPlanner()
{
}


/********************** Accessors **********************/

/**
 *  \brief return the current trajectory
 *  \return a copy of the planned trajectory
 *
 *  A new instance of PMSTraj is returned
 */
PMSTraj* PM_PVDTrajPlanner::getTraj() const
{
  if(this->status_ == TRAJ_CPTD) return new PMSTraj(*((PMSTraj*)(this->traj_)));
  else return NULL;
}


/**************** Trajectory Computing *****************/

/**
 *  \brief compute the trajectory from s1 to s2
 *  \param s1 : the initial state
 *  \param s2 : the final state
 *  \param p  : the precomputed path bewteen s1 and s2
 *  \return the trajectory planner status at the end of the planning process
 *            NO_PATH_LDD   : no path has been loaded
 *            ERR_ROBOT     : wrong robot features
 *            ERR_GOAL_ST   : the goal configuration is not available or free 
 *            ERR_INIT_ST   : the init configuration is not available or free
 *            NO_TRAJ       : no path can be computed
 *            TRAJ_CPTD     : the trajectory has been computed
 */
int PM_PVDTrajPlanner::computeTraj(const PMState& s1,const PMState& s2,const PMPath* p)
{
  // set the initial and final states
  if(initS_ != NULL) delete initS_;
  if(goalS_ != NULL) delete goalS_;
  initS_ = new PMState(s1);
  goalS_ = new PMState(s2);

  // init the path
  this->path_ = new PMPath(*p);

  // verify that the robot features have been loaded
  if(this->sys_ == NULL)
    {
      this->status_ = ERR_ROBOT;
      return ERR_ROBOT;
    }

  // verify that a path has been given
  if(this->path_ == NULL)
    {
      this->status_ = NO_PATH_LDD;
      return NO_PATH_LDD;
    }

  // allocation of the trajectory
  if(traj_ != NULL) delete traj_;
  this->traj_ = new PMSTraj();

  // init max speed and acceleration for the traj planning
  double maxSpeed = this->getMSC() * this->sys_->getVMax();
  double maxAcc = this->getMAC() * this->sys_->getAMax();

  // compute the length of the path
  double pathLength = this->path_->getLength();

  // computing of the initial speed
  double initSpd = sqrt(initS().sx() * initS().sx() + initS().sy() * initS().sy());
  // computing of the final speed
  double goalSpd = sqrt(goalS().sx() * goalS().sx() + goalS().sy() * goalS().sy());
  
  // computing of the time to reach the maximal speed from the initial speed
  double t1 = (maxSpeed - initSpd)/maxAcc;
  // computing of the time to reach the final speed from the maximal speed
  double t2 = (maxSpeed - goalSpd)/maxAcc;

  // computing of the length to reach the maximal speed
  double l1 = 0.5f * maxAcc * t1*t1 + initSpd * t1;
  // computing of the length to reach the goal speed
  double l2 = maxSpeed * t2 - 0.5 * maxAcc * t2*t2;

  if(pathLength < l1 + l2)
    {
      // case where the path is not enoughly long the reach the maximal speed
      
      // compute the middle length
      double lmid = 0.5 * pathLength;

      // compute the time needed to reach it
      double delta = initSpd*initSpd + 2*maxAcc*lmid;
      double tmid = (-initSpd + sqrt(delta)) / (maxAcc);

      // compute the total time needed
      double ttot = 2.0 * tmid;

      // compute the number of state-time of the trajectory
      int nbStep = ((int)(ttot / this->timeStep_)) + 1;
      // compute the effective time step
      double tStep = ttot / (nbStep - 1);

      // compute the maximal speed reached
      double maxSpeedRchd = initSpd + tmid * maxAcc;

      // init the trajectory computing
      double curTime = 0;
      double curLength = 0;
      double curSpeed = initSpd;
      int part = 1;
      int curCfg = 0;
      double lengthCurCfg = 0;

      // add the first state-time of the trajectory
      PMStateT first;
      first.x() = (this->path()->at(0).x());
      first.y() = (this->path()->at(0).y());
      first.t() = (0);
      first.sx() = (initS().sx());
      first.sy() = (initS().sy());
      this->traj()->append(first);

      // compute each state-time of the trajectory
      for(int i=1; i<nbStep-1; i++)
	{
	  // create a new state-time
	  PMStateT curS;
	  // set the previous time
	  double oldTime = curTime;
	  // set the current time
	  curTime += tStep;
	  // set the previous part
	  int oldPart = part;

	  // search on which part of the trajectory we are
	  if(curTime <= tmid)
	    part = 1;
	  else
	    part = 2;

	  // compute the current length and all parameters of the current state
	  if(part == 1)
	    {
	      // compute the new current length
	      curLength = 0.5 * maxAcc * curTime*curTime;
	      // compute the new current speed
	      curSpeed = maxAcc * curTime;
	    }
	  else if(part == 2)
	    {
	      // compute the new current length
	      curLength = lmid + (curTime - tmid) * maxSpeedRchd - 0.5 * (curTime - tmid)*(curTime - tmid) * maxAcc;
	      // compute the new current speed
	      curSpeed = maxSpeedRchd - (curTime - tmid) * maxAcc;
	    }

	  // search for the further configuration reached
	  bool trouve = false;
	  double fac = 0;
	  while(!trouve)
	    {
	      // get back the current configuration
	      PMConfig cfg = this->path()->at(curCfg);
	      // get back the next configuration
	      PMConfig nextCfg = this->path()->at(curCfg+1);
	      
	      // compute the length of the path to the next configuration
	      double dxCfg = nextCfg.x() - cfg.x();
	      double dyCfg = nextCfg.y() - cfg.y();
	      double distNextCfg = sqrt(dxCfg*dxCfg + dyCfg*dyCfg);
	      
	      // test if the current length has been reached
	      if(lengthCurCfg + distNextCfg > curLength)
		{
		  trouve = true;
		  fac = (curLength - lengthCurCfg) / distNextCfg;
		}
	      else
		{
		  lengthCurCfg = lengthCurCfg + distNextCfg;
		  curCfg++;
		}
	    }

	  // get back the current configuration
	  PMConfig cfg = this->path()->at(curCfg);
	  // get back the next configuration
	  PMConfig nextCfg = this->path()->at(curCfg+1);
	  
	  // compute the length of the path to the next configuration
	  double dxCfg = nextCfg.x() - cfg.x();
	  double dyCfg = nextCfg.y() - cfg.y();
	  double distNextCfg = sqrt(dxCfg*dxCfg + dyCfg*dyCfg);
	  
	  double dx = cfg.x() + fac * dxCfg - this->traj()->at(this->traj()->size()-1).x();
	  double dy = cfg.y() + fac * dyCfg - this->traj()->at(this->traj()->size()-1).y();

	  // compute the current state
	  curS.x() = (cfg.x() + fac * dxCfg);
	  curS.y() = (cfg.y() + fac * dyCfg);
	  curS.t() = (curTime);
	  curS.sx() = (dx / tStep);
	  curS.sy() = (dy / tStep);
	  // append it to the trajectory
	  this->traj()->append(curS);
	}

      // add the last state-time of the trajectory
      PMStateT last;
      last.x() = (this->path()->at(this->path()->size()-1).x());
      last.y() = (this->path()->at(this->path()->size()-1).y());
      last.t() = (ttot);
      last.sx() = (goalS().sx());
      last.sy() = (goalS().sy());
      this->traj()->append(last);
    }
  else
    {
      // case where the path is enoughly long the reach the maximal speed
            
      // compute the length of the intermediate part of the path (where the speed is constant and equals to maxSpeed)
      double lint = pathLength - l1 - l2;
      // compute the time duration of the intermediate part
      double tint = lint / maxSpeed;
      // compute the total time duration of the trajectory
      double ttot = t1 + tint + t2;

      // compute the number of state-time of the trajectory
      int nbStep = ((int)(ttot / this->timeStep_)) + 1;
      // compute the effective time step
      double tStep = ttot / (nbStep - 1);

      // init the trajectory computing
      double curTime = 0;
      double curLength = 0;
      double curSpeed = initSpd;
      int part = 1;
      int curCfg = 0;
      double lengthCurCfg = 0;

      // add the first state-time of the trajectory
      PMStateT first;
      first.x() = (this->path()->at(0).x());
      first.y() = (this->path()->at(0).y());
      first.t() = (0);
      first.sx() = (initS().sx());
      first.sy() = (initS().sy());
      this->traj()->append(first);

      // compute each state-time of the trajectory
      for(int i=1; i<nbStep-1; i++)
	{
	  // create a new state-time
	  PMStateT curS;
	  // set the previous time
	  double oldTime = curTime;
	  // set the current time
	  curTime += tStep;
	  // set the previous part
	  int oldPart = part;

	  // search on which part of the trajectory we are
	  if(curTime <= t1)
	    part = 1;
	  else if(curTime <= t1+tint)
	    part = 2;
	  else // curTime <= t1+tint+t2
	    part = 3;

	  // compute the current length and all parameters of the current state
	  if(part == 1)
	    {
	      // compute the new current length
	      curLength = 0.5 * maxAcc * curTime*curTime;
	      // compute the new current speed
	      curSpeed = maxAcc * curTime;
	    }
	  else if(part == 2)
	    {
	      // compute the new current length
	      curLength = l1 + (curTime - t1) * maxSpeed;
	      // compute the new current speed
	      curSpeed = maxSpeed;
	    }
	  else // part == 3
	    {
	      // compute the new current length
	      curLength = l1 + lint + (curTime - t1 - tint) * maxSpeed - 0.5 * (curTime - t1 -tint)*(curTime - t1 - tint) * maxAcc;
	      // compute the new current speed
	      curSpeed = maxSpeed - (curTime - t2) * maxAcc;
	    }

	  // search for the further configuration reached
	  bool trouve = false;
	  double fac = 0;
	  while(!trouve)
	    {
	      // get back the current configuration
	      PMConfig cfg = this->path()->at(curCfg);
	      // get back the next configuration
	      PMConfig nextCfg = this->path()->at(curCfg+1);
	      
	      // compute the length of the path to the next configuration
	      double dxCfg = nextCfg.x() - cfg.x();
	      double dyCfg = nextCfg.y() - cfg.y();
	      double distNextCfg = sqrt(dxCfg*dxCfg + dyCfg*dyCfg);
	      
	      // test if the current length has been reached
	      if(lengthCurCfg + distNextCfg > curLength)
		{
		  trouve = true;
		  fac = (curLength - lengthCurCfg) / distNextCfg;
		}
	      else
		{
		  lengthCurCfg = lengthCurCfg + distNextCfg;
		  curCfg++;
		}
	    }

	  // get back the current configuration
	  PMConfig cfg = this->path()->at(curCfg);
	  // get back the next configuration
	  PMConfig nextCfg = this->path()->at(curCfg+1);
	  
	  // compute the length of the path to the next configuration
	  double dxCfg = nextCfg.x() - cfg.x();
	  double dyCfg = nextCfg.y() - cfg.y();
	  double distNextCfg = sqrt(dxCfg*dxCfg + dyCfg*dyCfg);
	  
	  double dx = cfg.x() + fac * dxCfg - this->traj()->at(this->traj()->getSize()-1).x();
	  double dy = cfg.y() + fac * dyCfg - this->traj()->at(this->traj()->getSize()-1).y();

	  // compute the current state
	  curS.x() = (cfg.x() + fac * dxCfg);
	  curS.y() = (cfg.y() + fac * dyCfg);
	  curS.t() = (curTime);
	  curS.sx() = (dx / tStep);
	  curS.sy() = (dy / tStep);
	  // append it to the trajectory
	  this->traj()->append(curS);
	}
      
      // add the last state-time of the trajectory
      PMStateT last;
      last.x() = (this->path()->at(this->path()->size()-1).x());
      last.y() = (this->path()->at(this->path()->size()-1).y());
      last.t() = (ttot);
      last.sx() = (goalS().sx());
      last.sy() = (goalS().sy());
      this->traj()->append(last);
    }

  this->status_ = TRAJ_CPTD;
  return TRAJ_CPTD;
}


                                    /**************************************************/
                                    /************* Class DD2_PVDTrajPlanner ************/
                                    /**************************************************/

pvd_sampling_t DD2_PVDTrajPlanner::spg_type_ = PVD_SPACE_SPG;

/************* Constructors and destructors ************/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
DD2_PVDTrajPlanner::DD2_PVDTrajPlanner() : PVDTrajPlanner()
{
  this->maxVel_ = 1.0;
  this->maxAcc_ = 1.0;
}

/**
 *  \brief init constructor
 *  \param sys : the robotic system constraints
 */
DD2_PVDTrajPlanner::DD2_PVDTrajPlanner(DiffDrive2Sys* sys) : PVDTrajPlanner(sys, NULL)
{
}

/**
 *  \brief destructor
 */
DD2_PVDTrajPlanner::~DD2_PVDTrajPlanner()
{
}


/********************** Accessors **********************/

/**
 *  \brief return the current trajectory
 *  \return a copy of the planned trajectory
 *
 *  A new instance of DD2STraj is returned
 */
DD2STraj* DD2_PVDTrajPlanner::getTraj() const
{
  if(this->status_ == TRAJ_CPTD) return new DD2STraj(*((DD2STraj*)(this->traj_)));
  else return NULL;
}


/**************** Trajectory Computing *****************/

/*
 *  \brief 
 */
void DD2_PVDTrajPlanner::computeItdTime(double initSpd, double a1, double goalSpd, double pathLength, double& tmid, double& ttot)
{
  double t_diff = (goalSpd - initSpd) / a1;
  double delta = 4.0*initSpd*initSpd - 4.0*a1* (-initSpd*t_diff - 0.5*a1*t_diff*t_diff - pathLength);

  tmid = (-2.0*initSpd + sqrt(delta)) / (2.0*a1);

  double itdSpeedRchd = initSpd + tmid*a1;
  ttot = 2.0*tmid + ((initSpd - goalSpd) / a1);
}

/**
 *  \brief compute the trajectory from s1 to s2
 *  \param s0 : the initial state
 *  \param sf : the goal state
 *  \param p  : the precomputed path bewteen s1 and s2
 *  \param cruiseVel : the cruise velocity (that we try to keep all along the trajectory)
 *  \return the trajectory planner status at the end of the planning process
 *            NO_PATH_LDD   : no path has been loaded
 *            ERR_ROBOT     : wrong robot features
 *            ERR_GOAL_ST   : the goal configuration is not available or free 
 *            ERR_INIT_ST   : the init configuration is not available or free
 *            NO_TRAJ       : no path can be computed
 *            TRAJ_CPTD     : the trajectory has been computed
 */
int DD2_PVDTrajPlanner::computeTraj(const DD2State& s0, const DD2State& sf, const DPMPath* p, float cruiseVel)
{
//   cout << "DD2_PVDTrajPlanner::computeTraj 1" << endl;
  // set the initial and final states
  if(initS_ != NULL) delete initS_;
  if(goalS_ != NULL) delete goalS_;
  initS_ = new DD2State(s0);
  goalS_ = new DD2State(sf);

//   cout << "DD2_PVDTrajPlanner::computeTraj 2" << endl;
  // init the path
  this->path_ = new DPMPath(*p);
//   cout << "DD2_PVDTrajPlanner::computeTraj 3" << endl;

  // verify that a path has been given
  if(this->path_ == NULL) return NO_PATH_LDD;
//   cout << "DD2_PVDTrajPlanner::computeTraj 4" << endl;

  // allocation of the trajectory
  if(traj_ != NULL) delete traj_;
  this->traj_ = new WCSTraj();
//   cout << "DD2_PVDTrajPlanner::computeTraj 5" << endl;

  // init max speed and acceleration for the traj planning
//   double maxSpeed = this->getMSC() * this->sys_->getVMax();
  double maxSpeed = cruiseVel;
  double maxAcc = this->getMAC() * this->sys_->getAMax();

  // compute the length of the path
  double pathLength = this->path_->getLength();

  // computing of the initial speed
  double initSpd = initS().s();
  // computing of the final speed
  double goalSpd = goalS().s();
//   cout << "DD2_PVDTrajPlanner::computeTraj 6" << endl;
  
  // computing of the time to reach the maximal speed from the initial speed
  double t1 = fabs(maxSpeed - initSpd)/maxAcc;
  // computing of the time to reach the final speed from the maximal speed
  double t2 = fabs(maxSpeed - goalSpd)/maxAcc;

  // computing of the length to reach the maximal speed
  double l1;
  if(maxSpeed >= initSpd) l1 = 0.5f * maxAcc * t1*t1 + initSpd * t1;
  else l1 = 0.5f * -maxAcc * t1*t1 + initSpd * t1;
  // computing of the length to reach the goal speed
  double l2;
  if(maxSpeed >= goalSpd) l2 = maxSpeed * t2 - 0.5 * maxAcc * t2*t2;
  else l2 = maxSpeed * t2 + 0.5 * maxAcc * t2*t2;
//   cout << "DD2_PVDTrajPlanner::computeTraj 7" << endl;

//   cout << "t1 : " << t1 << "  t2 : " << t2 << "  l1 : " << l1 << "  l2 : " << l2 << endl;

  if(pathLength < l1 + l2)
    {
//       cout << "DD2_PVDTrajPlanner::computeTraj 81" << endl;
      // case where the path is not long enough to reach the maximal speed
      
      double a1;
      if(maxSpeed > initSpd) a1 = maxAcc;
      else a1 = -maxAcc;

      double tmid, ttot;
      this->computeItdTime(initSpd, a1, goalSpd, pathLength, tmid, ttot);

      double lmid = initSpd*tmid + 0.5*a1*tmid*tmid;

//       cout << "===================================================================================================" << endl;
//       double tmid2 = ttot - tmid;
//       cout << "tmid : " << tmid << "  tmid2 : " << tmid2 << "  ttot : " << ttot << "  lmid : " << lmid << "  pathLength : " << pathLength << endl;
//       cout << "initSpd*tmid + 0.5*a1*tmid*tmid + (initSpd + a1*tmid)*tmid2 + 0.5*(-a1)*tmid2*tmid2 - pathLength = " << initSpd*tmid + 0.5*a1*tmid*tmid + (initSpd + a1*tmid)*tmid2 + 0.5*(-a1)*tmid2*tmid2 - pathLength << endl;
//       cout << "===================================================================================================" << endl;

      // compute the number of state-time of the trajectory
      int nbStep = ((int)(ttot / this->timeStep_)) + 1;
      // compute the effective time step
      double tStep = ttot / (nbStep - 1);

      // compute the maximal speed reached
      double maxSpeedRchd;
      if(maxSpeed >= initSpd) maxSpeedRchd = initSpd + tmid * maxAcc;
      else maxSpeedRchd = initSpd - tmid * maxAcc;

      // init the trajectory computing
      double curTime = 0;
      double curLength = 0;
      double curSpeed = initSpd;
      int part = 1;
      int curCfg = 0;
      double lengthCurCfg = 0;

      // add the first state-time of the trajectory
      DD2StateT first;
      first.x() = (this->path()->at(0).x());
      first.y() = (this->path()->at(0).y());
      first.th() = (this->path()->at(0).th());
      first.v() = (initS().v());
      first.w() = (initS().w());
      first.t() = (0);
      this->traj()->append(first);
//       cout << "DD2_PVDTrajPlanner::computeTraj 82" << endl;

//       cout << "lmid : " << lmid << "  tmid : " << tmid << "  ttot : " << ttot <<  "  tStep : " << tStep << "  nbStep : " << nbStep << "  maxSpeedRchd : " << maxSpeedRchd << endl;

      // compute each state-time of the trajectory
      for(int i=1; i<nbStep-1; i++)
	{
// 	  cout << "DD2_PVDTrajPlanner::computeTraj 821" << endl;
	  // create a new state-time
	  DD2StateT curS;
	  // set the previous time
	  double oldTime = curTime;
	  // set the current time
	  curTime += tStep;
	  // set the previous part
	  int oldPart = part;

	  // search on which part of the trajectory we are
	  if(curTime <= tmid)
	    part = 1;
	  else // if(curTime > tmid)
	    part = 2;

// 	  cout << "DD2_PVDTrajPlanner::computeTraj 822" << endl;
	  // compute the current length and all parameters of the current state
	  if(part == 1)
	    {
	      if(maxSpeed >= initSpd)
		{
		  // compute the new current length
		  curLength = initSpd * curTime + 0.5 * maxAcc * curTime*curTime;
		  // compute the new current speed
		  curSpeed = initSpd + maxAcc * curTime;
		}
	      else
		{
		  // compute the new current length
		  curLength = initSpd * curTime + 0.5 * -maxAcc * curTime*curTime;
		  // compute the new current speed
		  curSpeed = initSpd - maxAcc * curTime;
		}
//   	      cout << "part1... initSpd : " << initSpd << "  curTime : " << curTime << "  curSpeed : " << curSpeed << "  curLength : " << curLength << "  pathLength : " << pathLength << endl;
	    }
	  else // if(part == 2)
	    {
	      if(maxSpeed >= goalSpd)
		{
		  // compute the new current length
		  curLength = lmid + (curTime - tmid) * maxSpeedRchd - 0.5 * (curTime - tmid)*(curTime - tmid) * maxAcc;
		  // compute the new current speed
		  curSpeed = maxSpeedRchd - (curTime - tmid) * maxAcc;
		}
	      else
		{
		  // compute the new current length
		  curLength = lmid + (curTime - tmid) * maxSpeedRchd + 0.5 * (curTime - tmid)*(curTime - tmid) * maxAcc;
		  // compute the new current speed
		  curSpeed = maxSpeedRchd + (curTime - tmid) * maxAcc;
		}
//   	      cout << "part2... maxSpeedRchd : " << maxSpeedRchd << "  tmid : " << tmid << "  curTime : " << curTime << "  curSpeed : " << curSpeed << "  curLength : " << curLength << "  pathLength : " << pathLength << endl;
	    }
// 	  cout << "DD2_PVDTrajPlanner::computeTraj 823" << endl;

	  // search for the further configuration reached
	  bool trouve = false;
	  double fac = 0;
	  while(!trouve)
	    {
// 	      cout << "curCfg : " << curCfg << "  curCfg+1 : " << curCfg+1 <<  "  this->path()->size() : " << this->path()->size() << "  i : " << i << "  nbStep : " << nbStep << "  curTime : " << curTime << "  ttot : " << ttot << endl;
	      // get back the current configuration
	      DPMConfig cfg = this->path()->at(curCfg);
	      // get back the next configuration
	      DPMConfig nextCfg = this->path()->at(curCfg+1);

// 	      cout << "CL_PVDTrajPlanner::computeTraj : Attention!!! Partie louche..." << endl;
	      
	      // compute the length of the path to the next configuration
	      double dxCfg = nextCfg.x() - cfg.x();
	      double dyCfg = nextCfg.y() - cfg.y();
	      double distNextCfg = sqrt(dxCfg*dxCfg + dyCfg*dyCfg);
	      
	      // test if the current length has been reached
	      if(lengthCurCfg + distNextCfg > curLength)
		{
		  trouve = true;
		  fac = (curLength - lengthCurCfg) / distNextCfg;
		}
	      else
		{
		  lengthCurCfg = lengthCurCfg + distNextCfg;
		  curCfg++;
		}
	    }

	  // get back the current configuration
	  DPMConfig cfg = this->path()->at(curCfg);
	  // get back the next configuration
	  DPMConfig nextCfg = this->path()->at(curCfg+1);
	  
	  // compute the length of the path to the next configuration
	  double dxCfg = nextCfg.x() - cfg.x();
	  double dyCfg = nextCfg.y() - cfg.y();
// 	  double dthetaCfg = nextCfg.th() - cfg.th();
	  double dthetaCfg = computeMinDiffAngle(nextCfg.th(), cfg.th());
	  
	  double distNextCfg = sqrt(dxCfg*dxCfg + dyCfg*dyCfg);
	  
	  double dx = cfg.x() + fac * dxCfg - this->traj()->at(this->traj()->getSize()-1).x();
	  double dy = cfg.y() + fac * dyCfg - this->traj()->at(this->traj()->getSize()-1).y();
	  double theta = cfg.th() + fac * dthetaCfg - this->traj()->at(this->traj()->getSize()-1).th();

	  // compute the current state
	  curS.x() = (cfg.x() + fac * dxCfg);
	  curS.y() = (cfg.y() + fac * dyCfg);
	  curS.th() = normalizedAngle(cfg.th() + fac * dthetaCfg);
//  	  curS.v() = (sqrt(dx*dx + dy*dy) / tStep);
 	  curS.v() = curSpeed;
	  curS.w() = ((curS.th() - this->path()->at(this->path()->size()-1).th())/tStep);
	  curS.t() = (curTime);
	  // append it to the trajectory
	  this->traj()->append(curS);
	}

      // add the last state-time of the trajectory
      DD2StateT last;
      last.x() = (this->path()->at(this->path()->size()-1).x());
      last.y() = (this->path()->at(this->path()->size()-1).y());
      last.th() = (this->path()->at(this->path()->size()-1).th());
      last.v() = (goalS().v());
      last.w() = (goalS().w());
      last.t() = (ttot);
      this->traj()->append(last);
    }
  else
    {
      // case where the path is enoughly long the reach the maximal speed

      // compute the length of the intermediate part of the path (where the speed is constant and equals to maxSpeed)
      double lint = pathLength - l1 - l2;
      // compute the time duration of the intermediate part
      double tint = lint / maxSpeed;
      // compute the total time duration of the trajectory
      double ttot = t1 + tint + t2;

      // compute the number of state-time of the trajectory
      int nbStep = ((int)(ttot / this->timeStep_)) + 1;
 //      cout << "nbStep : " << nbStep << endl;
      // compute the effective time step
      double tStep = ttot / (nbStep - 1);

      // init the trajectory computing
      double curTime = 0;
      double curLength = 0;
      double curSpeed = initSpd;
      int part = 1;
      int curCfg = 0;
      double lengthCurCfg = 0;

      // add the first state-time of the trajectory
      DD2StateT first;
      first.x() = (this->path()->at(0).x());
      first.y() = (this->path()->at(0).y());
      first.th() = (this->path()->at(0).th());
      first.v() = (initS().v());
      first.w() = (initS().w());
      first.t() = (0);
      this->traj()->append(first);
//       cout << "DD2_PVDTrajPlanner::computeTraj 92" << endl;

      // compute each state-time of the trajectory
      for(int i=1; i<nbStep-1; i++)
	{
	  // create a new state-time
	  DD2StateT curS;
	  // set the previous time
	  double oldTime = curTime;
	  // set the current time
	  curTime += tStep;
	  // set the previous part
	  int oldPart = part;

	  // search on which part of the trajectory we are
	  if(curTime <= t1)
	    part = 1;
	  else if(curTime <= t1+tint)
	    part = 2;
	  else // curTime <= t1+tint+t2
	    part = 3;

	  // compute the current length and all parameters of the current state
	  if(part == 1)
	    {
	      if(maxSpeed >= initSpd)
		{
		  // compute the new current length
		  curLength = initSpd * curTime + 0.5 * maxAcc * curTime*curTime;
		  // compute the new current speed
		  curSpeed = initSpd + maxAcc * curTime;
		}
	      else
		{
		  // compute the new current length
		  curLength = initSpd * curTime + 0.5 * -maxAcc * curTime*curTime;
		  // compute the new current speed
		  curSpeed = initSpd - maxAcc * curTime;
		}
// 	      cout << "part1... initSpd : " << initSpd << "  curTime : " << curTime << "  curSpeed : " << curSpeed << endl;
	    }
	  else if(part == 2)
	    {
	      // compute the new current length
	      curLength = l1 + (curTime - t1) * maxSpeed;
	      // compute the new current speed
	      curSpeed = maxSpeed;
	    }
	  else // part == 3
	    {
	      if(maxSpeed >= goalSpd)
		{
		  // compute the new current length
		  curLength = l1 + lint + (curTime - t1 - tint) * maxSpeed - 0.5 * (curTime - t1 -tint)*(curTime - t1 - tint) * maxAcc;
		  // compute the new current speed
		  curSpeed = maxSpeed - (curTime - t1 - tint) * maxAcc;
		}
	      else
		{
		  // compute the new current length
		  curLength = l1 + lint + (curTime - t1 - tint) * maxSpeed + 0.5 * (curTime - t1 -tint)*(curTime - t1 - tint) * maxAcc;
		  // compute the new current speed
		  curSpeed = maxSpeed + (curTime - t1 - tint) * maxAcc;
		}
// 	      cout << "part3... maxSpeed : " << maxSpeed << "  t2 : " << t2 << "  curTime : " << curTime << "  curSpeed : " << curSpeed << endl;
	    }

	  // search for the further configuration reached
	  bool trouve = false;
	  double fac = 0;
	  while(!trouve)
	    {
	      // get back the current configuration
	      DPMConfig cfg = this->path()->at(curCfg);
	      // get back the next configuration
	      DPMConfig nextCfg = this->path()->at(curCfg+1);

// 	      cout << "CL_PVDTrajPlanner::computeTraj : Attention!!! Partie louche..." << endl;
	      
	      // compute the length of the path to the next configuration
	      double dxCfg = nextCfg.x() - cfg.x();
	      double dyCfg = nextCfg.y() - cfg.y();
	      double distNextCfg = sqrt(dxCfg*dxCfg + dyCfg*dyCfg);
	      
	      // test if the current length has been reached
	      if(lengthCurCfg + distNextCfg > curLength)
		{
		  trouve = true;
		  fac = (curLength - lengthCurCfg) / distNextCfg;
		}
	      else
		{
		  lengthCurCfg = lengthCurCfg + distNextCfg;
		  curCfg++;
		}
	    }

	  // get back the current configuration
	  DPMConfig cfg = this->path()->at(curCfg);
	  // get back the next configuration
	  DPMConfig nextCfg = this->path()->at(curCfg+1);
	  
	  // compute the length of the path to the next configuration
	  double dxCfg = nextCfg.x() - cfg.x();
	  double dyCfg = nextCfg.y() - cfg.y();
// 	  double dthetaCfg = nextCfg.th() - cfg.th();
	  double dthetaCfg = computeMinDiffAngle(nextCfg.th(), cfg.th());
	  
	  double distNextCfg = sqrt(dxCfg*dxCfg + dyCfg*dyCfg);
	  
	  double dx = cfg.x() + fac * dxCfg - this->traj()->at(this->traj()->getSize()-1).x();
	  double dy = cfg.y() + fac * dyCfg - this->traj()->at(this->traj()->getSize()-1).y();
	  double theta = cfg.th() + fac * dthetaCfg - this->traj()->at(this->traj()->getSize()-1).th();

	  // compute the current state
	  curS.x() = (cfg.x() + fac * dxCfg);
	  curS.y() = (cfg.y() + fac * dyCfg);
	  curS.th() = normalizedAngle(cfg.th() + fac * dthetaCfg);
//  	  curS.v() = (sqrt(dx*dx + dy*dy) / tStep);
 	  curS.v() = curSpeed;
	  curS.w() = ((curS.th() - this->path()->at(this->path()->size()-1).th())/tStep);
	  curS.t() = (curTime);
	  // append it to the trajectory
	  this->traj()->append(curS);
	}

      // add the last state-time of the trajectory
      DD2StateT last;
      last.x() = (this->path()->at(this->path()->size()-1).x());
      last.y() = (this->path()->at(this->path()->size()-1).y());
      last.th() = (this->path()->at(this->path()->size()-1).th());
      last.v() = (goalS().v());
      last.w() = (goalS().w());
      last.t() = (ttot);
      this->traj()->append(last);
    }

  for(int i=1; i<this->traj_->getSize()-1; i++)
    {
      double dt = this->traj()->at(i+1).t() - this->traj()->at(i-1).t();
      double dTH = computeMinDiffAngle(this->traj()->at(i+1).th(), this->traj()->at(i-1).th());

      (*traj())[i].w() = (dTH / dt);

//       if(fabs((*traj())[i].w()) > 0.5*M_PI) cout << "t : " << this->traj()->at(i).t() << "  a1 : " << this->traj()->at(i+1).th() << "  a2 : " << this->traj()->at(i-1).th() 
// 						 << "  MinDiffAngle (a1-a2): " << dTH << "  dt : " << dt << " w : " << (*traj())[i].w() << endl;
    }

  this->status_ = TRAJ_CPTD;
  return TRAJ_CPTD;
}

/**
 *  \brief compute the trajectory from an input path
 *  \param p  : the precomputed path
 *  \param cruiseVel : the cruise velocity (that we try to keep all along the trajectory)
 *  \param startVel : the starting velocity
 *  \param endVel : the ending velocity
 *  \return the trajectory planner status at the end of the planning process
 *            NO_PATH_LDD   : no path has been loaded
 *            ERR_ROBOT     : wrong robot features
 *            ERR_GOAL_ST   : the goal configuration is not available or free 
 *            ERR_INIT_ST   : the init configuration is not available or free
 *            NO_TRAJ       : no path can be computed
 *            TRAJ_CPTD     : the trajectory has been computed
 */
int DD2_PVDTrajPlanner::computeTraj(const DPMPath* p, float cruiseVel, float startVel, float endVel)
{
  // check if the robot features have been loaded
  if(this->sys_ == NULL) return ERR_ROBOT;

  // check if the input path is not null
  if(p == NULL) return NO_PATH_LDD;

  // get back the initial and final configurations of the path
  const DPMConfig cfg0 = p->firstCfg();
  const DPMConfig cfgf = p->lastCfg();

  // build the initial and final states with null speeds
  DD2State s0(cfg0.x(), cfg0.y(), cfg0.th(), startVel, 0); 
  DD2State sf(cfgf.x(), cfgf.y(), cfgf.th(), endVel, 0);

  // check if the cruise velocity has been defined
  if(cruiseVel == 0) cruiseVel = this->getMSC() * this->sys_->getVMax();

  // compute the trajectory
  return this->computeTraj(s0, sf, p, cruiseVel);
}

/*
 *  \brief replan a part of the input state trajectory with a new maximal velocity
 *  \param straj : the input state trajectory
 *  \param maxVel : the new maximal velocity
 *  \param startTime : the starting time from which the trajectory must be modified
 *  \param startVel : the velocity registered at the starting time
 *  \param endTime : the ending time until which the trajectory must be modified
 *  \return the trajectory planner status at the end of the planning process
 *            ERR_ROBOT            : wrong robot features
 *            NO_INPUT_TRAJ        : no input trajectory
 *            PARAMS_NOT_VALID     : not valid input parameters
 *            TRAJ_CPTD            : the trajectory has been computed
 */
int DD2_PVDTrajPlanner::replanTraj(DD2STraj* straj, float maxVel, float startTime, float startVel, float endTime)
{
  // check if the robot features have been loaded
  if(this->sys_ == NULL) return ERR_ROBOT;

  // check if the input parameters are consistent
  if(straj == NULL) return NO_INPUT_TRAJ;

  // check the input parameters consistency
  if(startTime >= endTime || 
     maxVel > this->sys_->getVMax() ||
     startVel > this->sys_->getVMax()) return PARAMS_NOT_VALID;

  // get back both starting and ending states of the part of the trajectory to replace
  DD2StateT s1;
  straj->getStateAtTime(startTime, s1);
  DD2StateT s2;
  straj->getStateAtTime(endTime, s2);

//   cout << "DD2_PVDTrajPlanner::replanTraj -> s1.v() : " << s1.v() << endl;
//   cout << "DD2_PVDTrajPlanner::replanTraj -> s2.v() : " << s2.v() << endl;

  // get back the part of the trajectory defined after s2
  DD2STraj trajEnd;
  straj->cutAfterTime(endTime, &trajEnd);

//   // display trajEnd
//   cout << "======================================== trajEnd ========================================" << endl;
//   TrajPlot::plotDD2STraj(trajEnd, sys(), true);
//   cout << "================================= straj without trajEnd =================================" << endl;
//   TrajPlot::plotDD2STraj(*straj, sys(), true);

  // get back the path defined between s1 and s2
  DD2STraj trajMid;
  straj->cutAfterTime(startTime, &trajMid);
  DD2Path* path = trajMid.getPath();

//   // display trajMid
//   cout << "======================================== trajMid ========================================" << endl;
//   TrajPlot::plotDD2STraj(trajMid, sys(), true);
//   cout << "================================= straj without trajMid =================================" << endl;
//   TrajPlot::plotDD2STraj(*straj, sys(), true);

  // plan a trajectory between both state times with PVD
  int resPVD = this->computeTraj(path, maxVel, s1.v(), s2.v());
//   int resPVD = this->computeTraj(path, maxVel, startVel, s2.v());

  DD2STraj* newTrajMid = NULL;
  if(resPVD == TRAJ_CPTD)
    {
      /// replace the input trajectory with the concatenation of these three trajectories

      // get back the new trajectory
      newTrajMid = this->getTraj();
      // shift the trajectory to begin at startTime
      newTrajMid->shift(startTime);
      // concatenate it to the initial part of the trajectory
      straj->concat(*newTrajMid);

      // get back the final time of the new mid traj
      float newEndTime = newTrajMid->lastST().t();
      // shift the last part of the trajectory to begin at newEndTime
      trajEnd.shift(newEndTime - endTime);
      // concatenate it to the initial part of the trajectory
      straj->concat(trajEnd);
    }
  else
    {
      // the replanning failed : get back the initial profile of the state trajectory
      straj->concat(trajMid);
      straj->concat(trajEnd);

      cerr << "DD2_PVDTrajPlanner::replanTraj : Warning, the replanning failed. resPVD : " << resPVD << endl;
    }
  
  // delete every old path and trajectory
  if(path != NULL) delete path;
  if(newTrajMid != NULL) delete newTrajMid;

  return resPVD;
}


                                    /**************************************************/
                                    /************* Class CL_PVDTrajPlanner ************/
                                    /**************************************************/

/************* Constructors and destructors ************/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
CL_PVDTrajPlanner::CL_PVDTrajPlanner() : PVDTrajPlanner()
{
}

/**
 *  \brief init constructor
 */
CL_PVDTrajPlanner::CL_PVDTrajPlanner(CarLikeSys* sys) : PVDTrajPlanner(sys)
{
}

/**
 *  \brief destructor
 */
CL_PVDTrajPlanner::~CL_PVDTrajPlanner()
{
}


/********************** Accessors **********************/

/**
 *  \brief return the current trajectory
 *  \return a copy of the planned trajectory
 *
 *  A new instance of CLSTraj is returned
 */
CLSTraj* CL_PVDTrajPlanner::getTraj() const
{
  if(this->status_ == TRAJ_CPTD) return new CLSTraj(*((CLSTraj*)(this->traj_)));
  else return NULL;
}


/**************** Trajectory Computing *****************/

/**
 *  \brief compute the trajectory from s1 to s2
 *  \param s1 : the initial state
 *  \param s2 : the final state
 *  \param p  : the precomputed path bewteen s1 and s2
 *  \return the trajectory planner status at the end of the planning process
 *            NO_PATH_LDD   : no path has been loaded
 *            ERR_ROBOT     : wrong robot features
 *            ERR_ENV       : wrong environment
 *            ERR_ENV_MAP   : wrong static obstacles map
 *            ERR_ENV_OBS   : wrong dynamic obstacles list
 *            ERR_GOAL_ST   : the goal configuration is not available or free 
 *            ERR_INIT_ST   : the init configuration is not available or free
 *            NO_TRAJ       : no path can be computed
 *            TRAJ_CPTD     : the trajectory has been computed
 */
int CL_PVDTrajPlanner::computeTraj(const CLState& s1,const CLState& s2,const CLPath* p)
{
  // set the initial and final states
  if(initS_ != NULL) delete initS_;
  if(goalS_ != NULL) delete goalS_;
  initS_ = new CLState(s1);
  goalS_ = new CLState(s2);

  // init the path
  this->path_ = new CLPath(*p);

  // verify that the robot features have been loaded
  if(this->sys_ == NULL)
    {
      return ERR_ROBOT;
    }

  // verify that a path has been given
  if(this->path_ == NULL)
    {
      return NO_PATH_LDD;
    }

  // allocation of the trajectory
  if(traj_ != NULL) delete traj_;
  this->traj_ = new CLSTraj();

  // init max speed and acceleration for the traj planning
  double maxSpeed = this->getMSC() * this->sys_->getVMax();
  double maxAcc = this->getMAC() * this->sys_->getAMax();

  // compute the length of the path
  double pathLength = this->path_->getLength();

  // computing of the initial speed
  double initSpd = initS().s();
  // computing of the final speed
  double goalSpd = goalS().s();
  
  // computing of the time to reach the maximal speed from the initial speed
  double t1 = (maxSpeed - initSpd)/maxAcc;
  // computing of the time to reach the final speed from the maximal speed
  double t2 = (maxSpeed - goalSpd)/maxAcc;

  // computing of the length to reach the maximal speed
  double l1 = 0.5f * maxAcc * t1*t1 + initSpd * t1;
  // computing of the length to reach the goal speed
  double l2 = maxSpeed * t2 - 0.5 * maxAcc * t2*t2;

  if(pathLength < l1 + l2)
    {
      // case where the path is not enoughly long the reach the maximal speed
      
    }
  else
    {
      // case where the path is enoughly long the reach the maximal speed

      // compute the length of the intermediate part of the path (where the speed is constant and equals to maxSpeed)
      double lint = pathLength - l1 - l2;
      // compute the time duration of the intermediate part
      double tint = lint / maxSpeed;
      // compute the total time duration of the trajectory
      double ttot = t1 + tint + t2;

      // compute the number of state-time of the trajectory
      int nbStep = ((int)(ttot / this->timeStep_)) + 1;
 //      cout << "nbStep : " << nbStep << endl;
      // compute the effective time step
      double tStep = ttot / (nbStep - 1);

      // init the trajectory computing
      double curTime = 0;
      double curLength = 0;
      double curSpeed = initSpd;
      int part = 1;
      int curCfg = 0;
      double lengthCurCfg = 0;

      // add the first state-time of the trajectory
      CLStateT first;
      first.x() = (this->path()->at(0).x());
      first.y() = (this->path()->at(0).y());
      first.th() = (this->path()->at(0).th());
      first.s() = (initS().s());
      first.ph() = (initS().ph());
      first.t() = (0);
      this->traj()->append(first);

      // compute each state-time of the trajectory
      for(int i=1; i<nbStep-1; i++)
	{
	  // create a new state-time
	  CLStateT curS;
	  // set the previous time
	  double oldTime = curTime;
	  // set the current time
	  curTime += tStep;
	  // set the previous part
	  int oldPart = part;

	  // search on which part of the trajectory we are
	  if(curTime <= t1)
	    part = 1;
	  else if(curTime <= t1+tint)
	    part = 2;
	  else // curTime <= t1+tint+t2
	    part = 3;

	  cout << "11" << endl;

	  // compute the current length and all parameters of the current state
	  if(part == 1)
	    {
	      // compute the new current length
	      curLength = 0.5 * maxAcc * curTime*curTime;
	      // compute the new current speed
	      curSpeed = maxAcc * curTime;
	    }
	  else if(part == 2)
	    {
	      // compute the new current length
	      curLength = l1 + (curTime - t1) * maxSpeed;
	      // compute the new current speed
	      curSpeed = maxSpeed;
	    }
	  else // part == 3
	    {
	      // compute the new current length
	      curLength = l1 + lint + (curTime - t1 - tint) * maxSpeed - 0.5 * (curTime - t1 -tint)*(curTime - t1 - tint) * maxAcc;
	      // compute the new current speed
	      curSpeed = maxSpeed - (curTime - t2) * maxAcc;
	    }

	  // search for the further configuration reached
	  bool trouve = false;
	  double fac = 0;
	  while(!trouve)
	    {
	      cout << "curCfg : " << curCfg << " ";

	      // get back the current configuration
	      CLConfig cfg = this->path()->at(curCfg);
	      // get back the next configuration
	      CLConfig nextCfg = this->path()->at(curCfg+1);

	      cout << "CL_PVDTrajPlanner::computeTraj : Attention!!! Partie louche..." << endl;

// 	      if(cfg == NULL) cout << "cfg null" << endl;
// 	      if(nextCfg == NULL){
// 		cout << "nextCfg null" << endl;
// 		cout << "curCfg : " << curCfg << endl;
// 		cout << "this->path_->getSize() : " << this->path_->getSize() << endl;
// 		cout << "curLength : " << curLength << endl;
// 		cout << "lengthCurCfg : " << lengthCurCfg << endl;
// 		cout << "this->path_->getLength() : " << this->path_->getLength() << endl;
		
// 		double realCurLength = 0;
// 		for(int var=0; var<curCfg; var++)
// 		  {
// 		    ConfigR* c1 = (*(this->path_))[var];
// 		    ConfigR* c2 = (*(this->path_))[var+1];
// 		    double dX = c2->x() - c1->x();
// 		    double dY = c2->y() - c1->y();
// 		    realCurLength += sqrt(dX*dX + dY*dY);
// 		  }
// 		cout << "realCurLength : " << realCurLength << endl;
// 	      }
	      
	      // compute the length of the path to the next configuration
	      double dxCfg = nextCfg.x() - cfg.x();
	      double dyCfg = nextCfg.y() - cfg.y();
	      double distNextCfg = sqrt(dxCfg*dxCfg + dyCfg*dyCfg);
	      
	      // test if the current length has been reached
	      if(lengthCurCfg + distNextCfg > curLength)
		{
		  trouve = true;
		  fac = (curLength - lengthCurCfg) / distNextCfg;
		}
	      else
		{
		  lengthCurCfg = lengthCurCfg + distNextCfg;
		  curCfg++;
		}
	    }

	  // get back the current configuration
	  CLConfig cfg = this->path()->at(curCfg);
	  // get back the next configuration
	  CLConfig nextCfg = this->path()->at(curCfg+1);
	  
	  // compute the length of the path to the next configuration
	  double dxCfg = nextCfg.x() - cfg.x();
	  double dyCfg = nextCfg.y() - cfg.y();
	  double dthetaCfg = nextCfg.th() - cfg.th();
	  
	  double distNextCfg = sqrt(dxCfg*dxCfg + dyCfg*dyCfg);
	  
	  double dx = cfg.x() + fac * dxCfg - this->traj()->at(this->traj()->getSize()-1).x();
	  double dy = cfg.y() + fac * dyCfg - this->traj()->at(this->traj()->getSize()-1).y();
	  double theta = cfg.th() + fac * dthetaCfg - this->traj()->at(this->traj()->getSize()-1).th();

	  // compute the current state
	  curS.x() = (cfg.x() + fac * dxCfg);
	  curS.y() = (cfg.y() + fac * dyCfg);
	  curS.th() = normalizedAngle(cfg.th() + fac * dthetaCfg);
	  curS.s() = (sqrt(dx*dx + dy*dy) / tStep);
	  curS.ph() = ((curS.th() - this->path()->at(this->path()->size()-1).th())/curS.s());
	  curS.t() = (curTime);
	  // append it to the trajectory
	  this->traj()->append(curS);
	}

      // add the last state-time of the trajectory
      CLStateT last;
      last.x() = (this->path()->at(this->path()->size()-1).x());
      last.y() = (this->path()->at(this->path()->size()-1).y());
      last.th() = (this->path()->at(this->path()->size()-1).th());
      last.s() = (goalS().s());
      last.ph() = (goalS().ph());
      last.t() = (ttot);
      this->traj()->append(last);
    }

  for(int i=1; i<this->traj_->getSize()-1; i++)
    {
      double dx = this->traj()->at(i+1).x() - this->traj()->at(i-1).x();
      double dy = this->traj()->at(i+1).y() - this->traj()->at(i-1).y();
      double dist = sqrt(dx*dx + dy*dy);
      double dTH = this->traj()->at(i+1).th() - this->traj()->at(i-1).th();
      (*traj())[i].ph() = (dTH / dist);
    }

  return TRAJ_CPTD;
}
