#include "RoadMapPP.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /***************** Class RoadMapPP ****************/
                                    /**************************************************/

/************* Constructors and destructors ************/

// default constructor
RoadMapPP::RoadMapPP() : PathPlanner()
{
  this->roadmap_    = NULL;
  this->vecNgh_     = NULL;
  this->vecMinDist_ = NULL;
}

// init constructor
RoadMapPP::RoadMapPP(System* sys,Environment* e) : PathPlanner(sys,e)
{
  this->roadmap_    = NULL;
  this->vecNgh_     = NULL;
  this->vecMinDist_ = NULL;
}

// init constructor
RoadMapPP::RoadMapPP(System* sys,Environment* e,RoadMap* rm)
{
  this->path_       = NULL;
  this->env_        = e;
  this->sys_        = sys;
  this->initCfg_    = NULL;
  this->goalCfg_    = NULL;
  this->roadmap_    = rm;
  this->vecNgh_     = NULL;
  this->vecMinDist_ = NULL;
}

// destructor
RoadMapPP::~RoadMapPP()
{
//   if(initCfg_ != NULL) delete initCfg_;
//   if(goalCfg_ != NULL) delete goalCfg_;
}


/******************** Initialization *******************/

/**
 *  \brief reinit the planner and the roadmap if needed
 */
void RoadMapPP::reinit()
{
  // clear the vector of neighbours
  if(vecNgh_ != NULL)
    {
      delete vecNgh_;
      vecNgh_ = NULL;
    }
  
  // reinit the roadmap
  if(roadmap_ != NULL) roadmap_->reinit();
}


/******************* Path Computing ********************/

// compute the path from initCfg to goalCfg on map
int RoadMapPP::computePath(const PMConfig& cfg1, const PMConfig& cfg2)
{
  // set the initial and final configurations
  if(initCfg_ != NULL) delete initCfg_;
  if(goalCfg_ != NULL) delete goalCfg_;
  this->initCfg_ = new PMConfig(cfg1);
  this->goalCfg_ = new PMConfig(cfg2);

  // check if the roadmap is not null
  if(this->roadmap_ == NULL)
    {
      cout << "RoadMapPP::computePath() : Warning. No input roadmap." << endl;
      return ERR_STD_MAP;
    }

  // find the nearest visible node of the roadmap from the initial configuration  
  RoadNode* initRMNode = roadmap_->findNearNode(initCfg().posXY());

  // find the nearest visible node of the roadmap from the goal configuration
  RoadNode* goalRMNode = roadmap_->findNearNode(goalCfg().posXY());

  // ckeck if both initial and final nodes of the map have been correctly computed
  if(initRMNode == NULL)
    {
      cout << "RoadMapPP::computePath() : Warning. A problem occured with the initial position." << endl;
      return ERR_INIT_POS;
    }
  if(goalRMNode == NULL)
    {
      cout << "RoadMapPP::computePath() : Warning. A problem occured with the goal position." << endl;
      return ERR_GOAL_POS;
    }

  int res = this->computePath(initRMNode,goalRMNode);
  if(res == PATH_CPTD)
    {
      // if a path has been correctly computed : 
      // we connect both initial and final node to the path

      PMPath* itdPath = this->path();
      this->path_ = new PMPath();

      // concatenate all parts of the path
      this->path()->append(PMConfig(initCfg().x(),initCfg().y()));
      this->path()->concat(*itdPath);
      this->path()->append(PMConfig(goalCfg().x(),goalCfg().y()));

      // append the minimal distances corresponding to the initial and final configurations
      if(this->vecMinDist_->size() != 0)
	{
	  std::list<double>::iterator it = vecMinDist_->begin();
	  double vDeb = *it;
	  it = vecMinDist_->end(); it--;
	  double vFin = *it;
	    
	  vecMinDist_->push_front(vDeb);
	  vecMinDist_->push_back(vFin);
	}
      
      // delete the second and the second to last configurations of the path
      this->del2ndCfg();

      // smooth the actual path
      int res = 0;
      int maxNbSmooth = 3;
      int nbSmooth = 0;
      while(res == 0 && nbSmooth < maxNbSmooth) 
	{
	  this->smoothPath();
	  nbSmooth++;
	}

//       if(res == 0) cout << "no collision found" << endl;
//       else cout << "a collision has been found" << endl;
      
      return PATH_CPTD;
    }
  else return res;
}

// compute the path from init to goal on map
int RoadMapPP::computePath(RoadNode* init, RoadNode* goal)
{
  bool trouve = false;
  int nbVisitedNode = 0;

  RoadNode* goalFound = NULL;
  
  // reinit the vector of neighbours
  if(vecNgh_ != NULL)
    {
      delete vecNgh_;
    }
  vecNgh_ = new vector<RoadNode*>;

  // initialize the initial node
  init->reinit();
  init->computeEto(goal);
  // mark the initial node as visited
  init->visit();

  // check if the initial node is the goal node
  if(init == goal)
    {
      trouve = true;
      goalFound = init;
    }
  else
    {
      // initialize the vector of neighbours
      int N = init->getNbNeighbour();
      
      for(int i=0; i<N; i++)
	{
	  RoadNode* curNgh = init->getNeighbour(i);
	  curNgh->computeLfrom(init);
	  curNgh->computeEto(goal);
	  curNgh->visit();
	  
	  // check if this node is the goal node
	  if(curNgh->getE() == 0)
	    {
	      goalFound = curNgh;
	      trouve = true;
	    }
	  else
	    {
	      // add the current neighbour to the vector of neighbours
	      this->addNgh(curNgh);
	    }
	}
    }

  while(!trouve && vecNgh_->size() != 0)
    {
      // get back the best node from the list of neighbours
      RoadNode* curNode = this->getBestNgh();
      
      // get back all neighbours of the current node
      int M = curNode->getNbNeighbour();
      for(int i=0; i<M; i++)
	{
	  RoadNode* curNgh = curNode->getNeighbour(i);
	  
	  // check if the current neighbour has been already visited
	  if(!curNgh->isVisited())
	    {
	      curNgh->computeLfrom(curNode);
	      curNgh->computeEto(goal);
	      curNgh->visit();

	      // check if this node is the goal node
	      if(curNgh->getE() == 0)
		{
		  goalFound = curNgh;
		  trouve = true;
		}
	      else
		this->addNgh(curNgh);
	    }
	}
    }
  
  if(trouve)
    {
//       cout << "trouve!!!" << endl;
      // construct the path to the goal node
      this->constructPath(goalFound);
      return PATH_CPTD;
    }

  cout << "no path ;-(" << endl;
  return NO_PATH;
}

// construct the path to goalNode from the list of visited nodes
void RoadMapPP::constructPath(RoadNode* goalNode)
{
  // reinit the path if needed
  if(this->path_ != NULL)
    delete this->path_;
  this->path_ = new PMPath();

  // reinit the vector of minimal distances if needed
  if(this->vecMinDist_ != NULL)
    delete this->vecMinDist_;
  this->vecMinDist_ = new std::list<double>;
  
//   std::vector<double> vecInvMinDist;

  // get back the last node on which we stopped
  RoadNode* curNode = goalNode;
  vector<PMConfig> vecInvPath;

  // construct the path from the last node to the first one
  while(curNode != NULL)
    {
//       cout << "Node : "; curNode->display(); cout << endl;

      // get back the position of the current node
      PMConfig curCfg(curNode->getPos().x(),curNode->getPos().y());

      // append the node to the inverted path
      vecInvPath.push_back(curCfg);

//       // append the minimal distance to this node
//       vecMinDist_->push_front(((DRoadNode*)curNode)->D());

      // Move to the previous node of the current one
      curNode = curNode->getPrev();
    }

  // invert the path and append each config of it to the current path
  int N = vecInvPath.size();
  for(int i=N-1; i>=0; i--)
    {
      this->path()->append(vecInvPath[i]);
//       this->vecMinDist_->push_back(vecInvMinD[i]);
    }
}

// add a new node to the list of visitable neighbours
void RoadMapPP::addNgh(RoadNode* n)
{
  // check if the list is empty
  if(vecNgh_->size() == 0)
    {
      // add directly the node at the end of the list
      vecNgh_->push_back(n);
      return;
    }

  // get back the heuristic of the current node
  double h = n->getH();

  // sort the list by increasing heuristic
  bool trouve = false;
  int i=0;
  vector <RoadNode*>::iterator it = vecNgh_->begin();

  int N = vecNgh_->size();
  while(i<N && !trouve)
    {
      if(h < vecNgh_->at(i)->getH())
	trouve = true;
      else 
	{
	  it++;
	  i++;
	}
    }

  if(trouve)
    vecNgh_->insert(it,n);
  else 
    vecNgh_->push_back(n);

  return;
}

// return the node of the list of visitable neighbours 
// with the best heuristic and delete it from the list
RoadNode* RoadMapPP::getBestNgh()
{
  // get back the first element of the list
  RoadNode* res = vecNgh_->at(0);
  // remove it from the list
  vector <RoadNode*>::iterator it = vecNgh_->begin();
  vecNgh_->erase(it);
  // return it
  return res;
}

/**
 *  \brief smooth the resulting path
 *  \return 0 if the process succeed, -1 if a collison has been found on one of the new nodes added
 */
int RoadMapPP::smoothPath()
{
  // check if the path ahs been correctly computed
  if(this->path() == NULL) return 0;

  // save the actual path and vector of minimal distances
  PMPath* itdPath = this->path();
  list<double>* itdVMD = this->vecMinDist_;
  // reinit the resulting path and vector of minimal distances
  this->path_ = new PMPath();
  this->vecMinDist_ = new list<double>;

  // append the first configuration of the old path to the new path
  path()->append(itdPath->firstCfg());
  vecMinDist_->push_back(itdVMD->front());

  int N = itdPath->size();
  int i=0;
  bool col = false;

  while(i<N-2 && !col)
    {
      // get back the current position of the path
      PMConfig cur = itdPath->at(i);
      PMConfig next = itdPath->at(i+1);
      PMConfig nnext = itdPath->at(i+2);

      // compute the new intermediate configurations
      PMConfig p1(0.25*cur.x() + 0.75*next.x(),
		  0.25*cur.y() + 0.75*next.y());
      PMConfig p2(0.75*next.x() + 0.25*nnext.x(),
		  0.75*next.y() + 0.25*nnext.y());

      // check if these configurations are in collision with the environment
      PositionXY near1,near2;
      double dist1,dist2;
      bool col1 = env_->staticObstruct(PositionXY(p1.x(),p1.y()),near1,dist1);
      bool col2 = env_->staticObstruct(PositionXY(p2.x(),p2.y()),near2,dist2);

      path()->append(p1);
      vecMinDist_->push_back(dist1);
      path()->append(p2);
      vecMinDist_->push_back(dist2);

      i++;
    }

  if(col)
    {
    }
  else
    {
      // append the last configuration of the old path to the new path
      path()->append(itdPath->lastCfg());
      vecMinDist_->push_back(itdVMD->front());
      
      // delete the old path
      if(itdPath != NULL) delete itdPath;
      if(itdVMD != NULL)
	{
	  itdVMD->clear();
	  delete itdVMD;
	}
      
      return 0;
    }
}

/**
 *  delete the second and the second to last configurations of the path
 */
void RoadMapPP::del2ndCfg()
{
  // check if the path ahs been correctly computed
  if(this->path() == NULL) return;

  // save the actual path and vector of minimal distances
  PMPath* itdPath = this->path();
  list<double>* itdVMD = this->vecMinDist_;
  // reinit the resulting path and vector of minimal distances
  this->path_ = new PMPath();
  this->vecMinDist_ = new list<double>;

  int N = itdPath->size();
  list<double>::iterator it = itdVMD->begin();

  for(int i=0; i<N; i++)
    {
      if(i == 1)
	{
	  PMConfig first = itdPath->at(0);
	  PMConfig third = itdPath->at(2);
	  path()->append(PMConfig(0.5*(first.x()+third.x()),0.5*(first.y()+third.y())));
	  vecMinDist_->push_back(*it);
	}
      else if(i == N-2)
	{
	  PMConfig last = itdPath->at(N-1);
	  PMConfig pplast = itdPath->at(N-3);
	  path()->append(PMConfig(0.5*(last.x()+pplast.x()),0.5*(last.y()+pplast.y())));
	  vecMinDist_->push_back(*it);
	}
      else
	{
	  path()->append(itdPath->at(i));
	  vecMinDist_->push_back(*it);
	}
      it++;
    }
  return;
}
