#include "planner/inc/SimpleSTrajPlanner.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /************ Class SimpleSTrajPlanner ************/
                                    /**************************************************/

/********** Constructors and destructors **********/
  
/**
 *  \brief default constructor
 *  Only used by subclasses
 */
SimpleSTrajPlanner::SimpleSTrajPlanner()
{
  this->sys_ = NULL;

  this->trajgen_ = NULL;

  coefVMax_ = 0.8;
  coefAMax_ = 0.8;
  
  this->debug_ = 0;
  this->straj_ = NULL;
}

/**
 *  \brief init constructor
 *  \param sys : the robotic system handled
 *  \param coefVMax : maximal velocity coefficient
 *  \param coefAMax : maximal acceleration coefficient
 *  \param debug : the debugging level
 */
SimpleSTrajPlanner::SimpleSTrajPlanner(System* sys, double coefVMax, double coefAMax, int debug)
{
  this->sys_ = sys;

  this->trajgen_ = NULL;

  coefVMax_ = 0.8;
  coefAMax_ = 0.8;
  
  this->debug_ = debug;
  this->straj_ = NULL;
}

/**
 *  \brief destructor
 */
SimpleSTrajPlanner::~SimpleSTrajPlanner()
{
  if(trajgen_ != NULL) delete trajgen_;
  if(straj_ != NULL) delete straj_;
}


/*************** Control Computation **************/

/**
 *  \brief Initialize the standard parameters of the trajectory generator
 *  \param dbg : the trajectory generator debugging mode
 */
void SimpleSTrajPlanner::initTrajGen(int dbg)
{
  if(trajgen_ != NULL)
    {
      /// common trajgen options
      trajgen_->setTcStatus(TIME_FREE,0,999999999.9);
      
      // init options of the trajectory generation
      trajgen_->setMaxNbStep(20);                 // default = 20
      trajgen_->setMaxResidual(0.05);             // default = 0.05
      
      // init the debug level mode
      trajgen_->setDBG(dbg);                      // default = 0
    }
}

/**
 *  \brief compute the time needed to execute the trajectory between s1 and s2
 *  \param s1 : the initial state
 *  \param s2 : the goal state
 */
double SimpleSTrajPlanner::computeTrajTime(const PositionXYTh& p1, double v1, const PositionXYTh& p2, double v2) const
{
//   cout << "p1 : "; p1.display(); 
//   cout << "  v1 : " << v1; 
//   cout << "  p2 : "; p2.display(); 
//   cout << "  v2 : " << v2 << endl; 

  // compute the distance between p1 and p2
  double dx = p2.x() - p1.x();
  double dy = p2.y() - p1.y();
  double dist = sqrt(dx*dx + dy*dy);

  // compute the approximate distance we would like to travel
  double dir = atan2(dy,dx);
  double difDir = computeMinDiffAngle(normalizedAngle(dir),normalizedAngle(p2.th()));
  double coefDist = 1.0 + (0.5*M_PI - 1)*(fabs(difDir)/M_PI);
  double distTraj = coefDist*dist;

//   cout << "dist : " << dist << "  distTraj : " << distTraj << endl;

  // compute the intermediate acceleration and speed used to travel on the trajectory
  double itdAcc = this->sys_->getAMax() * coefAMax_;
  double itdSpd = this->sys_->getVMax() * coefVMax_;

//   cout << "coefAMax_ : " << coefAMax_ << "  coefVMax_ : " << coefVMax_ << endl;
//   cout << "sys_->getAMax() : " << sys_->getAMax() << "  sys_->getVMax() : " << sys_->getVMax() << endl;
//   cout << "itdAcc : " << itdAcc << "  itdSpd : " << itdSpd << endl;

  // compute the times and distances needed to accelerate and decelerate to reach the speed itdSpd
  double tacc,dacc;
  double tdec,ddec;
  minTDtoReachSpd(v1,itdSpd,itdAcc,tacc,dacc);
  minTDtoReachSpd(itdSpd,v2,itdAcc,tdec,ddec);

//   cout << "tacc : " << tacc << "  tdec : " << tdec << "dacc : " << dacc << "ddec : " << ddec << endl;

  // compute the time needed to execute the trajectory
  if(distTraj <= dacc + ddec)
    {
//       cout << "(tacc + tdec) : " << (tacc + tdec) << endl;
      return (tacc + tdec);
    }
  else
    {
      double dmid = distTraj - (dacc + ddec);
      double tmid = dmid / itdSpd;

//       cout << "(tacc + tmid + tdec) : " << (tacc + tmid + tdec) << endl;
      return (tacc + tmid + tdec);
    }
}
