#ifndef WC_SSLATTICE_HPP
#define WC_SSLATTICE_HPP

#include "sslattice/inc/SSLattice.hpp"
#include "system/inc/WheelChairSys.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class WC_SSLattice **************/
                                    /**************************************************/

  /** \class WC_SSLattice
   *  \brief State Space Lattice for a Wheelchair system
   */
  class WC_SSLattice : public SSLattice
  {
    vector<double>* tabPosTrajX_;
    vector<double>* tabPosTrajY_;
    vector<double>* tabPosTrajT_;
    vector<double>* tabPosTrajV_;
    vector<double>* tabPosTrajAV_;

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    WC_SSLattice();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the wheelchair system for which the state space lattice is built
     */
    WC_SSLattice(WheelChairSys* sys);

    /**
     *  \brief destructor
     */
    virtual ~WC_SSLattice();

    /************** Getters and Setters ***************/

  protected : 
    /**
     *  \brief return the system constraints of the car-like
     *  \return the system constraints of the car-like
     */
    WheelChairSys* sys() const { return ((WheelChairSys*)sys_); }

    /************** Lattice construction **************/

  protected : 
    /**
     *  \brief allocate a new state-time and copy the input one
     *  \param s : input state-time
     *  \return a pointer to the copy of the input state-time
     */
    virtual WCStateT* copyStateT(const StateTime& s);

    /**
     *  \brief generate a trajectory to compute the state reached by s by ct
     *  \param st : the initial state-time of the trajectory
     *  \param ct : the control-time applied on the system
     *  \return the final state-time reached
     */
    virtual WCStateT* genTraj(const StateTime* st,const ControlTime* ct);

    /**
     *  \brief sample the available controls
     *  \param ctrlSp : number of samples used for each control
     *  \param dt : the time of application of each control
     *  \return the control sampling vector
     */
    virtual void sampleControl(int nbCtrlSp,double dt);

    /************** Feasibility functions *************/

  protected : 
    /**
     *  \brief compute the available error between two state-times
     *  \return the available error between two state-times
     */
    virtual StateTime* computeAvError(const StateTime* s1,const StateTime* s2) const;

    /**
     *  \brief compute the difference between two state-times
     *  \return the difference between two state-times
     *
     *  A new instance of a state-time is computed, it must be deleted
     */
    virtual StateTime* diffStateT(const StateTime* s1,const StateTime* s2) const;

    // public : 
    //   /**
    //    *  \brief check if the input state-time is reachable according to the sslattice
    //    *  \param s : the input state to evaluate
    //    */
    //   virtual bool isReachable(const StateTime& s) const;

    /**************** Utility functions ***************/

  protected : 
    /**
     *  \allocate the trajectory vectors
     *  \param nb : the number of trajectories to allocate
     */
    virtual void allocTrajVec(int nb);

    /**************** Display functions ***************/

  public : 
    /**
     *  \brief draw the lattice in a gnuplot window
     *  \return 0 if the process succeed, -1 else
     */
    virtual int drawGP() const;
  };

};
#endif // WC_SSLATTICE_HPP
