#include "SSLattice.hpp"
#include "Tree.hpp"
using namespace mpc;

#include <cmath>

using namespace mpc;

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */
SSLattice::SSLattice()
{
  this->height_         = 7;
  this->nbCtrlSampling_ = 3;
  this->sys_            = NULL;
  this->tree_           = NULL;
  this->ctrlSp_         = NULL;
  this->tabLeaves_      = NULL;
  this->avError_        = NULL;

  this->displayTraj_    = false;
  this->trajSp_         = 30;

  this->curDT_          = 0.0;
  this->nbLeaves_       = -1;
}

/**
 *  \brief init constructor
 *  \param sys : the system for which the state space lattice is built
 */
SSLattice::SSLattice(System* sys)
{
  this->height_         = 7;
  this->nbCtrlSampling_ = 3;
  this->sys_            = sys;
  this->tree_           = NULL;
  this->ctrlSp_         = NULL;
  this->tabLeaves_      = NULL;
  this->avError_        = NULL;

  this->displayTraj_    = false;
  this->trajSp_         = 30;

  this->curDT_          = 0.0;
  this->nbLeaves_       = -1;
}

/**
 *  \brief destructor
 */
SSLattice::~SSLattice()
{
  // delete the tree
  if(tree_ != NULL) delete tree_;

  // delete the control sampling
  this->deleteCtrlSp();

  // delete the leaves tab
  this->deleteTabLeaves();
}


/************** Getters and Setters ***************/

/**
 *  \brief return the i-th leaf of the lattice tree
 *  \return the i-th leaf of the lattice tree
 */
StateTime* SSLattice::leafAt(int i) const 
{
  if(tabLeaves_ != NULL && i >= 0 && i < nbLeaves_)
    return tabLeaves_[i];
  else return NULL;
}


/************** Lattice construction **************/

/**
 *  \brief recursive method that construct one node of the tree
 *  \param s : the initial state-time from which the current node is built
 *  \param dt : the time duration of a subtrajectory
 *  \param lvl : the remaining level to build up to the leaves
 *  \return the root of the branch built
 */
mpc::Node<StateTime*>* SSLattice::constructNode(StateTime* s,double dt,int lvl)
{
  cout << "SSLattice::constructNode(s="; s->display(); cout << ", dt=" << dt << ", lvl=" << lvl << " )" << endl;

  mpc::Node<StateTime*>* n = new mpc::Node<StateTime*>(s);

  // check if the current node is a leaf
  if(lvl == 0)
    {
      // add the current node to the tab of leaves of the tree
      tabLeaves_[J_] = s;
      J_++;
    }
  else
    {
      // generate all the subnodes of the current one
      int N = ctrlSp_->size();
      for(int i=0; i<N; i++)
	{
	  // get back the current control-time
	  ControlTime* ctrlt = ctrlSp_->at(i);

// 	  cout << "ctrlt : " << endl;
// 	  ctrlt->display(); cout << endl;

	  // compute the final state-time reached
	  StateTime* finalState = this->genTraj(s,ctrlt);
	  
	  // construct a node from the new state computed
	  mpc::Node<StateTime*>* newNode = this->constructNode(finalState,dt,lvl-1);

	  // add the new node as a child of the current node
	  n->addSon(newNode);
	}
    }

  return n;
}

/**
 *  \brief construct the state space lattice
 *  \param s : the initial state-time from which the state space lattice is built
 *  \param t : the time until which the lattice is built
 *  \return 0 if the state space lattice has been constructed successfully
 *  \return -1 if no input system has been defined
 */
int SSLattice::construct(const StateTime& s,double t)
{
  // check if the system constraints have been defined
  if(sys_ == NULL)
    {
      cerr << "SSLattice::construct : No input system has been defined" << endl;
      return -1;
    }

  cout << "System : " << endl;
  sys_->display();
  cout << endl;

  // inititalize the vectors to display trajectories if needed
  if(displayTraj_) this->allocTrajVec(this->computeNbTraj());

  // initialize the vector of leaves
  this->deleteTabLeaves();
  nbLeaves_ = this->computeNbLeaves();
  cout << "nbLeaves_ : " << nbLeaves_ << endl;
  tabLeaves_ = new StateTime* [nbLeaves_];
  J_ = 0;

  // copy the input state-time to define the root of the tree
  StateTime* sroot = this->copyStateT(s);

  double dt = t / ((double)height_-1);
  curDT_ = dt;

  // build the control sampling vector
  this->sampleControl(nbCtrlSampling_,dt);
  cout << "ctrl sampling : "; this->displayCtrlSp();

  // delete the old tree if needed
  if(tree_ != NULL) delete tree_;

  // construct the tree recursively from its root
  tree_ = this->constructNode(sroot,dt,height_-1);

  return 0;
}


/************** Feasibility functions *************/

/**
 *  \brief check if the input state-time is reachable according to the sslattice
 *  \param s : the input state to evaluate
 */
bool SSLattice::isReachable(const StateTime& s) const
{
  bool trouve = false;
  int i = 0;
  int N = this->computeNbLeaves();

  cout << "NbLeaves : " << N << endl;
//   cout << "Leaves : " << endl;
//   for(int i=0; i<N; i++)
//     {
//       cout << i << ": "; (tabLeaves_[i])->display(); cout << endl;
//     }

  i=0;

  while(!trouve && i<N)
    {
//       cout << "i : " << i << endl;
//       cout << "tabLeaves_[i]: "; (tabLeaves_[i])->display(); cout << endl;

      bool trouve2 = false;
      // compute the difference between both states
      StateTime* diffS = this->diffStateT(&s,tabLeaves_[i]);

      // compute the error allowed between both states
      StateTime* avError = this->computeAvError(&s,tabLeaves_[i]);

      int j=0;
      while(!trouve2 && j<diffS->size())
	{
	  if(diffS->at(j) > avError->at(j)) trouve2 = true;
	  else j++;
	}

      if(!trouve2) 
	{
	  cout << "State to reach : "; s.display(); cout << endl;
	  cout << "State found : "; (tabLeaves_[i])->display(); cout << endl;
	  cout << "diffS : "; diffS->display(); cout << endl;
	  cout << "avError : "; avError->display(); cout << endl;
	  trouve = true;
	}
      else i++;

      delete diffS;
      delete avError;
    }

  return trouve;
}


/**************** Utility functions ***************/

/**
 *  \brief delete the control sampling
 */
void SSLattice::deleteCtrlSp()
{
  if(ctrlSp_ != NULL)
    {
      for(int i=0; i<ctrlSp_->size(); i++)
	delete ctrlSp_->at(i);
      delete ctrlSp_;
    }
}

/**
 *  \brief delete the leaves tab
 */
void SSLattice::deleteTabLeaves()
{
  if(tabLeaves_ != NULL) delete tabLeaves_;
}

/**
 *  \brief compute the number of trajectories that will be computed
 *  \return the number of trajectories that will be computed
 */
int SSLattice::computeNbTraj() const
{
  int res = 0;
  for(int i=1; i<height_; i++)
    res += std::pow((double)nbCtrlSampling_, ctrlSize_*i);

  return res;
}

/**
 *  \brief compute the number of leaves of the tree
 *  \return the number of leaves of the tree
 */
int SSLattice::computeNbLeaves() const
{
  cout << "=========== SSLattice::computeNbLeaves ===========" << endl;
  cout << "nbCtrlSampling_ : " << nbCtrlSampling_ << endl;
  cout << "ctrlSize_ : " << ctrlSize_ << endl;
  cout << "height_ : " << height_ << endl;
  cout << "==================================================" << endl;
  return std::pow((double)nbCtrlSampling_, ctrlSize_*(height_-1));
}


/**************** Display functions ***************/

/**
 *  \brief display the list of controls computed
 */
void SSLattice::displayCtrlSp() const
{
    if(ctrlSp_ != NULL)
    {
      for(int i=0; i<ctrlSp_->size(); i++)
	{
	  ctrlSp_->at(i)->display();
	  cout << "  ";
	}
      cout << endl;
    }
}
