#include "TrajFollow.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /**************** Class TrajFollow ****************/
                                    /**************************************************/

float TrajFollow::minDistToObstacles_ = TF_DFLT_MIN_DIST_TO_OBS;
float TrajFollow::maxDistToObstacles_ = TF_DFLT_MAX_DIST_TO_OBS;
float TrajFollow::minCoefSpeed_ = TF_DFLT_MIN_SPEED_COEF;

float TrajFollow::maxDistTgtReached_ = TF_DFLT_MAX_DIST_TGT_REACHED;
float TrajFollow::maxDistWpReached_ = TF_DFLT_MAX_DIST_WP_REACHED;
float TrajFollow::maxAngleWpReached_ = TF_DFLT_MAX_ANGLE_WP_REACHED;


/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 *  Only used by subclasses
 */
TrajFollow::TrajFollow()
{
  this->traj_          = NULL;
  this->curRegST_      = NULL;
  this->trajgen_       = NULL;
  this->system_        = NULL;
  this->dtReach_       = 2.0;
  this->timeMeeting_   = 0.0;
  this->lastDuration_  = 0.0;
  this->endTraj_       = false;

  this->distmap_       = NULL;
  this->tooCloseToObs_ = false;
}

/*
 *  \brief init constructor
 *  \param sys : the robotic system handeled
 *  \param traj : the trajectory to follow
 */
TrajFollow::TrajFollow(System* sys,STraj* traj)
{
  this->traj_          = traj;
  this->curRegST_      = NULL;
  this->trajgen_       = NULL;
  this->system_        = sys;
  this->dtReach_       = 2.0;
  this->timeMeeting_   = 0.0;
  this->lastDuration_  = 0.0;

  this->endTraj_       = false;

  this->distmap_       = NULL;
  this->tooCloseToObs_ = false;
}

/*
 *  \brief destructor
 */
TrajFollow::~TrajFollow()
{
  if(this->curRegST_ != NULL) delete this->curRegST_;
  if(this->trajgen_ != NULL) delete this->trajgen_;

  if(this->distmap_ != NULL) delete this->distmap_;
}


/*************** Getters and Setters **************/

/*
 *  \brief return the time of the last internal update
 *  \return time of the last internal update
 */
double TrajFollow::lastTimeUp() const
{
  if(this->curRegST_ != NULL)
    return this->curRegST_->t();
  else return 0.0;
}


/*************** Environment update ***************/

/**
 *  \brief update an input distance to obstacles map
 *  \param distmap : the input distmap
 *  \return 0 if the process succeed, -1 else
 */
int TrajFollow::updateDistMap(const DistObsMap* distmap)
{
  bool distMapChanged = false;

  {
    boost::mutex::scoped_lock lock(mutexDOM_);
    if(distmap_ != NULL) delete distmap_;
    
    if(distmap == NULL) 
      {
	distmap_ = NULL;
	distMapChanged = false;
      }
    else 
      {
	distmap_ = distmap->copy();
	distMapChanged = true;
      }
  }

  if(distMapChanged)
    {
      // compute a coefficient to apply on the maximal speed wrt. the maximal distance to obstacles
      float coefMaxVel = this->computeCoefMaxSpeedWrtEnv();

      // update the trajectory wrt. the maximal allowed speed
      this->updateTrajVelProfile(coefMaxVel);

      return 0;
    }
  else return -1;
}

/*
 *  \brief search for the states (poses) of the trajectory that the system is to take until it stops after a breaking maneuver
 *  \return a vector containing the successive poses
 */
vector<PositionXYT> TrajFollow::posesBreakingManeuver()
{
  boost::mutex::scoped_lock lock(mutexTraj_);

  // compute the distance required to the system to break from its current state
  float minBD = this->minBreakDist();

  // get back the current position of the robot
  PositionXYT curPos = curRegST_->posXYT();

  vector<PositionXYT> vecRes = traj_->vecPosesFromPosToDist(curPos, minBD);
//   cout << "TrajFollow::posesBreakingManeuver vecRes.size() : " << vecRes.size() << endl;

  // push the current position in the front of the resulting vector
  vector<PositionXYT>::iterator it = vecRes.begin();
  vecRes.insert(it, curPos);

  return vecRes;
}

/*
 *  \brief compute a coefficient to apply on the maximal speed wrt. the maximal distance to obstacles
 *  \return compute a coefficient to apply on the maximal speed wrt. the environment
 */
float TrajFollow::computeCoefMaxSpeedWrtEnv()
{

  boost::mutex::scoped_lock lock(mutexDOM_);

  if(distmap_ == NULL) 
    {
      cout << "TrajFollow::computeCoefMaxSpeedWrtEnv : The distance to obstacles map has not been defined yet!" << endl;
      return 1.0;
    }
  else if(curRegST_ == NULL)
    {
      cout << "TrajFollow::computeCoefMaxSpeedWrtEnv : The location of the robot has not been defined yet!" << endl;
      return 0.0;
    }
  else
    {
//       cout << "current state : "; curRegST_->display(); cout << endl;

      // search for the states of the trajectory that the system is to take until it stops after a breaking maneuver
      vector<PositionXYT> vecPoses = this->posesBreakingManeuver();

      // search for the closest state to obstacles
      float minDist = 999999999;
      int N = vecPoses.size();

      for(int i=0; i<N; i++)
	{
	  PositionXYT pos = vecPoses.at(i);

	  // search for the distance to obstacles of the current position on the distmap
	  obs2D_cell_t cellSt;
	  float dist;
	  PositionXY near;
	  distmap_->getInfo(pos.x(), pos.y(), cellSt, dist, near);

	  if(i == 0 || dist < minDist)
	    {
	      minDist = dist;
	    }
	}

      // compute the coef to apply on the maximal speed
      if(minDist < TrajFollow::minDistToObstacles_)
	{
	  this->tooCloseToObs_ = true;
//  	  cout << "TrajFollow::computeCoefMaxSpeedWrtEnv -> coef : 0.0  minDist : " << minDist << endl;
	  return 0.0;
	}
      else if(minDist < TrajFollow::maxDistToObstacles_)
	{
	  this->tooCloseToObs_ = false;
	  float coef = (TrajFollow::minCoefSpeed_ + ((1.0 - TrajFollow::minCoefSpeed_) * ((minDist - TrajFollow::minDistToObstacles_) / (TrajFollow::maxDistToObstacles_ - TrajFollow::minDistToObstacles_))));
	  if(coef < 0.1) coef = 0.1;  // ... trick to avoid trajectory generation problems
	  
//  	  cout << "TrajFollow::computeCoefMaxSpeedWrtEnv -> coef : " << coef << "  minDist : " << minDist << endl;
	  return coef;
	}
      else
	{
	  this->tooCloseToObs_ = false;
//  	  cout << "TrajFollow::computeCoefMaxSpeedWrtEnv -> coef : 1.0  minDist : " << minDist  << endl;
	  return 1.0;
	}
    }
}


/**************** Utility functions ***************/

/*
 *  \brief get back the last time of the trajectory
 *  \return the last time of the trajectory
 */
double TrajFollow::lastTimeTraj()
{
  boost::mutex::scoped_lock lock(mutexTraj_);

  return traj_->lastST().t();
}


/****************** I/O functions *****************/

/**
 *  \brief return true if the trajectory following process is running
 *  \return true if the trajectory following process is running
 */
bool TrajFollow::isRunning() const
{
  return timer_.isRunning();
}

/**
 *  \brief start the trajectory following process
 */
void TrajFollow::start()
{
  // start the timer
//   timer_.reinit();
  timer_.start();

  this->tooCloseToObs_ = false;
}

/**
 *  \brief stop the trajectory following process
 */
void TrajFollow::stop()
{
  // stop the timer
  timer_.stop();
}

/**
 *  \brief start the trajectory following process
 *  \param traj : the trajectory to follow
 */
void TrajFollow::reinit(STraj* traj)
{
  // reinit the timer
  timer_.reinit();

  // replace the current trajectory with the input one if needed
  if (traj != NULL)
    {
      boost::mutex::scoped_lock lock(mutexTraj_);
      this->traj_ = traj;
    }
}
