#include "DDTrajGen.hpp"
using namespace mpc;

#include "Utils.hpp"

#include "gnuplot/inc/gnuplot_i.hpp"

#include <iostream>
using namespace std;

// Parameters
#define R      R_
#define UnSurR UnSurR_

#define A1     A1_
#define A2     A2_

#define B1     (*vecParameters_p)[0]
#define C1     (*vecParameters_p)[1]
#define D1     (*vecParameters_p)[2]
#define B2     (*vecParameters_p)[3]
#define C2     (*vecParameters_p)[4]
#define D2     (*vecParameters_p)[5]
#define T      (*vecParameters_p)[6]

// States 
#define X0     vecState_x0[0]
#define Y0     vecState_x0[1]
#define TH0    vecState_x0[2]
#define V0     vecState_x0[3]
#define OM0    vecState_x0[4]
#define T0     vecState_x0[5]

#define Xf     vecState_xf[0]
#define Yf     vecState_xf[1]
#define THf    vecState_xf[2]
#define Vf     vecState_xf[3]
#define OMf    vecState_xf[4]
#define Tf     vecState_xf[5]

#define Xsim   vecState_xsim[0]
#define Ysim   vecState_xsim[1]
#define THsim  vecState_xsim[2]
#define Vsim   vecState_xsim[3]
#define OMsim  vecState_xsim[4]
#define Tsim   vecState_xsim[5]

#define Xerr   vecConstraints_g[0]
#define Yerr   vecConstraints_g[1]
#define THerr  vecConstraints_g[2]
#define Verr   vecConstraints_g[3]
#define OMerr  vecConstraints_g[4]

#define TIME_FIXED true

#define COEF_COR 0.1
#define VARIABLE_COEF_COR true

#define DFLT_DERIV_VAL -999999999.99

#define FAST_PART_DERIV true

#define COEF_WHENBDD 0.2

#define MAXABS_POS_ERR    0.1
#define MAXREL_POS_ERR    0.01


/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 */  
DDTrajGen::DDTrajGen()
{
  xLength = 6;       // State length (x, y, th, v, om, t)
  pLength = 7;       // Parameters length (B1, B2, C1, C2, D1, D2, T)
  gLength = 5;       // Constraints length (x(t)=x_f, y(t)=y_f, th(t)=th_f, v(t)=v_f, om(t)=om_f)
  dLength = 3;       // Dynamic Constraints length (maxWAcc, maxWVel, R)

  vecState_x0 = Vec_DP(xLength);
  vecState_xf = Vec_DP(xLength);
  if(tcStatus_ == TIME_FIXED) vecState_xsim = Vec_DP(xLength-1);
  else vecState_xsim = Vec_DP(xLength);
  vecConstraints_g = Vec_DP(gLength);
  vecParameters_p = NULL;

  posX   = NULL;
  posY   = NULL;
  headTH = NULL;
  velV  = NULL;
  avelOM = NULL;
  timeT  = NULL;

//   if(TIME_FIXED)
//     {
//       vecCorrParameters_p = Vec_DP(pLength-1);
//       matJacobianG = Mat_DP(gLength,pLength-1);
//       matInvJacobianG = Mat_DP(pLength-1,gLength);
//     }
//   else
//     {
//       vecCorrParameters_p = Vec_DP(pLength);
//       matJacobianG = Mat_DP(gLength,pLength);
//       matInvJacobianG = Mat_DP(pLength,gLength);
//     }
  vecCorrParameters_p = Vec_DP(pLengthFree());
  matJacobianG = Mat_DP(gLength,pLengthFree());
  //matInvJacobianG = Mat_DP(pLengthFree(),gLength);

  finalResidual_ = -1;
  finalNbSteps_ = -1;

  A1 = 0;
  A2 = 0;
  R  = 0;

  fctWLAcc_ = NULL;
  fctWRAcc_ = NULL;
  fctWLVel_ = NULL;
  fctWRVel_ = NULL;
  fctVel_ = NULL;
  fctAVel_ = NULL;
  fctTH_ = NULL;

  fctWLAccWB_ = NULL;
  fctWRAccWB_ = NULL;
  fctWLVelWB_ = NULL;
  fctWRVelWB_ = NULL;
  fctVelWB_ = NULL;
  fctAVelWB_ = NULL;
  fctTHWB_ = NULL;

}

/**
 *  \brief destructor
 */  
DDTrajGen::~DDTrajGen()
{
}


/*************** Getters and Setters **************/


/********* Optimization related routines **********/

/**
 *  \brief set the initial guess with a vector of parameters
 *  \param parameters : the vector of parameters
 */
void DDTrajGen::ComputeIG(Vec_DP** parameters)
{
  // allocate the parameters vector
  if(parameters == NULL)
    {
      vecParameters_p = new Vec_DP(pLength);
    }
  else
    {
      *parameters = new Vec_DP(pLength);
      vecParameters_p = *parameters;
    }

  T = Tf - T0;
  
  // compute the distance between the initial and final positions
  double dist = sqrt(pow(Xf-X0,2) + pow(Yf-Y0,2));
  // compute the direction of the vector defined between the initial and final positions
  double THvec = atan2(Yf-Y0,Xf-X0);

  // compute the heading change between the initial and final positions
  double dTH = computeMinDiffAngle(normalizedAngle(THf),normalizedAngle(TH0));
  // compute the heading change between the initial heading and the global direction
  double dTH1 = computeMinDiffAngle(normalizedAngle(THvec),normalizedAngle(TH0));
  // compute the heading change between the global direction and the final heading
  double dTH2 = computeMinDiffAngle(normalizedAngle(THf),normalizedAngle(THvec));
  
  double Vmoy = 0.5 * (V0 + Vf);

//   if(fabs(dTH) > 0.5*M_PI)
//     {
//       // we search for a backward trajectory
//     }
//   else
//     {
//       // we search for a forward trajectory
      
//     }

//   //B1 = dTH / (T*T) * R;
  B1 = 0.5 * ((Vf - V0) + (Vmoy - V0)) / T;
  C1 = 0.0;
  D1 = 0.0;

//   //B2 = -dTH / (T*T) * R;
//   B2 = 2.0 * (Vmoy - V0) / T;
  B2 = 0.5 * ((Vf - V0) + (Vmoy - V0)) / T;
  C2 = 0.0;
  D2 = 0.0;  

//   if(dist < 5*R)
//     {


//       B1 = (dist + 1.0*dTH1*R) / (2.0*T*T);
//       C1 = (1.0*(dTH2-dTH1)*R) / (2.0*T*T*T);
//       D1 = 0.0;
      
//       B2 = (dist - 1.0*dTH1*R) / (2.0*T*T);
//       C2 = (1.0*(dTH1-dTH2)*R) / (2.0*T*T*T);
//       D2 = 0.0;



//     }
//   else
//     {
//       B1 = ((5 + dTH1)*dist) / (10.0*T*T);
//       C1 = ((dTH2-dTH1)*dist) / (2.0*T*T*T);
//       D1 = 0.0;
      
//       B2 = ((5 - dTH1)*dist) / (10.0*T*T);
//       C2 = ((dTH1-dTH2)*dist) / (2.0*T*T*T);
//       D2 = 0.0;
//     }
}

/**
 *  \brief compute the error between the state reached and the state we want to reach
 */
void DDTrajGen::ComputeG()
{
  Xerr  = Xsim - Xf;
  Yerr  = Ysim - Yf;
  THerr = computeMinDiffAngle(normalizedAngle(THsim),normalizedAngle(THf));
  Verr  = Vsim - Vf;
  OMerr = OMsim - OMf;
}

/**
 *  \brief compute the partial derivatives of the state wrt. the parameters by finite differences
 */
void DDTrajGen::ComputeJacobianG_FD()
{
  this->computeAllFct();
  double oldX  = this->funcX(T);
  double oldY  = this->funcY(T);
  double oldTH = this->funcTheta(T);
  double oldV  = this->funcVel(T);
  double oldOM = this->funcAVel(T);

  B1 += E;
  this->computeAllFct();
  double X_B1  = this->funcX(T);
  double Y_B1  = this->funcY(T);
  double TH_B1 = this->funcTheta(T);
  double V_B1  = this->funcVel(T);
  double OM_B1 = this->funcAVel(T);  
  B1 -= E;

  // compute dS/dB1
  matJacobianG[0][0] = (X_B1 - oldX)*UnSurE;            // dX/dB1
  matJacobianG[1][0] = (Y_B1 - oldY)*UnSurE;            // dY/dB1
  matJacobianG[2][0] = (TH_B1 - oldTH)*UnSurE;          // dTH/dB1
  matJacobianG[3][0] = (V_B1 - oldV)*UnSurE;            // dV/dB1
  matJacobianG[4][0] = (OM_B1 - oldOM)*UnSurE;          // dOM/dB1

  C1 += E;
  this->computeAllFct();
  double X_C1  = this->funcX(T);
  double Y_C1  = this->funcY(T);
  double TH_C1 = this->funcTheta(T);
  double V_C1  = this->funcVel(T);
  double OM_C1 = this->funcAVel(T);  
  C1 -= E;

  // compute dS/dC1
  matJacobianG[0][1] = (X_C1 - oldX)*UnSurE;            // dX/dC1
  matJacobianG[1][1] = (Y_C1 - oldY)*UnSurE;            // dY/dC1
  matJacobianG[2][1] = (TH_C1 - oldTH)*UnSurE;          // dTH/dC1
  matJacobianG[3][1] = (V_C1 - oldV)*UnSurE;            // dV/dC1
  matJacobianG[4][1] = (OM_C1 - oldOM)*UnSurE;          // dOM/dC1

  D1 += E;
  this->computeAllFct();
  double X_D1  = this->funcX(T);
  double Y_D1  = this->funcY(T);
  double TH_D1 = this->funcTheta(T);
  double V_D1  = this->funcVel(T);
  double OM_D1 = this->funcAVel(T);  
  D1 -= E;

  // compute dS/dD1
  matJacobianG[0][2] = (X_D1 - oldX)*UnSurE;            // dX/dD1
  matJacobianG[1][2] = (Y_D1 - oldY)*UnSurE;            // dY/dD1
  matJacobianG[2][2] = (TH_D1 - oldTH)*UnSurE;          // dTH/dD1
  matJacobianG[3][2] = (V_D1 - oldV)*UnSurE;            // dV/dD1
  matJacobianG[4][2] = (OM_D1 - oldOM)*UnSurE;          // dOM/dD1

  B2 += E;
  this->computeAllFct();
  double X_B2  = this->funcX(T);
  double Y_B2  = this->funcY(T);
  double TH_B2 = this->funcTheta(T);
  double V_B2  = this->funcVel(T);
  double OM_B2 = this->funcAVel(T);  
  B2 -= E;

  // compute dS/dB2
  matJacobianG[0][3] = (X_B2 - oldX)*UnSurE;            // dX/dB2
  matJacobianG[1][3] = (Y_B2 - oldY)*UnSurE;            // dY/dB2
  matJacobianG[2][3] = (TH_B2 - oldTH)*UnSurE;          // dTH/dB2
  matJacobianG[3][3] = (V_B2 - oldV)*UnSurE;            // dV/dB2
  matJacobianG[4][3] = (OM_B2 - oldOM)*UnSurE;          // dOM/dB2

  C2 += E;
  this->computeAllFct();
  double X_C2  = this->funcX(T);
  double Y_C2  = this->funcY(T);
  double TH_C2 = this->funcTheta(T);
  double V_C2  = this->funcVel(T);
  double OM_C2 = this->funcAVel(T);  
  C2 -= E;

  // compute dS/dC2
  matJacobianG[0][4] = (X_C2 - oldX)*UnSurE;            // dX/dC2
  matJacobianG[1][4] = (Y_C2 - oldY)*UnSurE;            // dY/dC2
  matJacobianG[2][4] = (TH_C2 - oldTH)*UnSurE;          // dTH/dC2
  matJacobianG[3][4] = (V_C2 - oldV)*UnSurE;            // dV/dC2
  matJacobianG[4][4] = (OM_C2 - oldOM)*UnSurE;          // dOM/dC2

  D2 += E;
  this->computeAllFct();
  double X_D2  = this->funcX(T);
  double Y_D2  = this->funcY(T);
  double TH_D2 = this->funcTheta(T);
  double V_D2  = this->funcVel(T);
  double OM_D2 = this->funcAVel(T);  
  D2 -= E;

  // compute dS/dD2
  matJacobianG[0][5] = (X_D2 - oldX)*UnSurE;            // dX/dD2
  matJacobianG[1][5] = (Y_D2 - oldY)*UnSurE;            // dY/dD2
  matJacobianG[2][5] = (TH_D2 - oldTH)*UnSurE;          // dTH/dD2
  matJacobianG[3][5] = (V_D2 - oldV)*UnSurE;            // dV/dD2
  matJacobianG[4][5] = (OM_D2 - oldOM)*UnSurE;          // dOM/dD2
}

/**
 *  \brief compute the partial derivatives of the state wrt. the parameters
 */
void DDTrajGen::ComputeJacobianG()
{
  // computation of dOM/dP
  matJacobianG[4][0] = UnSurR * T;      // dOM/dB1
  matJacobianG[4][1] = UnSurR * T*T;    // dOM/dC1
  matJacobianG[4][2] = UnSurR * T*T*T;  // dOM/dD1
  matJacobianG[4][3] = -UnSurR * T;     // dOM/dB2
  matJacobianG[4][4] = -UnSurR * T*T;   // dOM/dC2
  matJacobianG[4][5] = -UnSurR * T*T*T; // dOM/dD2
  if(tcStatus_ != TIME_FIXED) matJacobianG[4][6] = (0.5 * UnSurR) * (funcWRAcc(T) - funcWLAcc(T)); // dOM/dT

  // computation of dTH/dP
  matJacobianG[2][0] = 0.5*UnSurR*T*T;            // dTH/dB1
  matJacobianG[2][1] = 0.333333333*UnSurR*T*T*T;  // dTH/dC1
  matJacobianG[2][2] = 0.25*UnSurR*T*T*T*T;       // dTH/dD1
  matJacobianG[2][3] = -0.5*UnSurR*T*T;           // dTH/dB2
  matJacobianG[2][4] = -0.333333333*UnSurR*T*T*T; // dTH/dC2
  matJacobianG[2][5] = -0.25*UnSurR*T*T*T*T;      // dTH/dD2
  if(tcStatus_ != TIME_FIXED) matJacobianG[2][6] = funcAVel(T);               // dTH/dT

  // computation of dVX/dP
  matJacobianG[3][0] = T;      // dVX/dB1
  matJacobianG[3][1] = T*T;    // dVX/dC1
  matJacobianG[3][2] = T*T*T;  // dVX/dD1
  matJacobianG[3][3] = T;      // dVX/dB2
  matJacobianG[3][4] = T*T;    // dVX/dC2
  matJacobianG[3][5] = T*T*T;  // dVX/dD2
  if(tcStatus_ != TIME_FIXED) matJacobianG[3][6] = 0.5 * (funcWRAcc(T) + funcWLAcc(T));  // dVX/dT

  // computation of dX/dP
  matJacobianG[0][0] = qgaus(&DDTrajGen::itgdXdB1,0,T); // dX/dB
  matJacobianG[0][1] = qgaus(&DDTrajGen::itgdXdC1,0,T); // dX/dC
  matJacobianG[0][2] = qgaus(&DDTrajGen::itgdXdD1,0,T); // dX/dD
  matJacobianG[0][3] = qgaus(&DDTrajGen::itgdXdB2,0,T); // dX/dB2
  matJacobianG[0][4] = qgaus(&DDTrajGen::itgdXdC2,0,T); // dX/dC2
  matJacobianG[0][5] = qgaus(&DDTrajGen::itgdXdD2,0,T); // dX/dD2
  if(tcStatus_ != TIME_FIXED) matJacobianG[0][6] = funcItgX(T);                     // dX/dT

  // computation of dY/dP
  matJacobianG[1][0] = qgaus(&DDTrajGen::itgdYdB1,0,T); // dY/dB
  matJacobianG[1][1] = qgaus(&DDTrajGen::itgdYdC1,0,T); // dY/dC
  matJacobianG[1][2] = qgaus(&DDTrajGen::itgdYdD1,0,T); // dY/dD
  matJacobianG[1][3] = qgaus(&DDTrajGen::itgdYdB2,0,T); // dY/dB2
  matJacobianG[1][4] = qgaus(&DDTrajGen::itgdYdC2,0,T); // dY/dC2
  matJacobianG[1][5] = qgaus(&DDTrajGen::itgdYdD2,0,T); // dY/dD2
  if(tcStatus_ != TIME_FIXED) matJacobianG[1][6] = funcItgY(T);                     // dY/dT
}

double DDTrajGen::fct1(const double t) const
{
  return ( (A1+A2) + (B1+B2)*t + (C1+C2)*t*t + (D1+D2)*t*t*t );
}

double DDTrajGen::fct2(const double t) const
{
  return normalizedAngle( TH0 + UnSurR*(A1-A2)*t + 0.5*UnSurR*(B1-B2)*t*t + 0.33333333*UnSurR*(C1-C2)*t*t*t + 0.25* UnSurR*(D1-D2)*t*t*t*t );
}

double DDTrajGen::itgdXdB1(const double t) const
{
  return (t * cos(fct2(t))) + (fct1(t) * (-0.5 * UnSurR * t*t) * sin(fct2(t)));
}

double DDTrajGen::itgdXdC1(const double t) const
{
  return (t*t * cos(fct2(t))) + (fct1(t) * (-0.33333333 * UnSurR * t*t*t) * sin(fct2(t)));
}

double DDTrajGen::itgdXdD1(const double t) const
{
  return (t*t*t * cos(fct2(t))) + (fct1(t) * (-0.25 * UnSurR * t*t*t*t) * sin(fct2(t)));
}

double DDTrajGen::itgdXdB2(const double t) const
{
  return (t * cos(fct2(t))) + (fct1(t) * (0.5 * UnSurR * t*t) * sin(fct2(t)));
}

double DDTrajGen::itgdXdC2(const double t) const
{
  return (t*t * cos(fct2(t))) + (fct1(t) * (0.33333333 * UnSurR * t*t*t) * sin(fct2(t)));
}

double DDTrajGen::itgdXdD2(const double t) const
{
  return (t*t*t * cos(fct2(t))) + (fct1(t) * (0.25 * UnSurR * t*t*t*t) * sin(fct2(t)));
}

double DDTrajGen::itgdYdB1(const double t) const
{
  return (t * sin(fct2(t))) + (fct1(t) * (0.5 * UnSurR * t*t) * cos(fct2(t)));
}

double DDTrajGen::itgdYdC1(const double t) const
{
  return (t*t * sin(fct2(t))) + (fct1(t) * (0.33333333 * UnSurR * t*t*t) * cos(fct2(t)));
}

double DDTrajGen::itgdYdD1(const double t) const
{
  return (t*t*t * sin(fct2(t))) + (fct1(t) * (0.25 * UnSurR * t*t*t*t) * cos(fct2(t)));
}

double DDTrajGen::itgdYdB2(const double t) const
{
  return (t * sin(fct2(t))) + (fct1(t) * (-0.5 * UnSurR * t*t) * cos(fct2(t)));
}

double DDTrajGen::itgdYdC2(const double t) const
{
  return (t*t * sin(fct2(t))) + (fct1(t) * (-0.33333333 * UnSurR * t*t*t) * cos(fct2(t)));
}

double DDTrajGen::itgdYdD2(const double t) const
{
  return (t*t*t * sin(fct2(t))) + (fct1(t) * (-0.25 * UnSurR * t*t*t*t) * cos(fct2(t)));
}

/**
 *  \brief compute the residual between the goal state and the stae reached
 *  \return the residual between the goal state and the stae reached
 */
double DDTrajGen::ComputeResidual()
{
  double sqRes = 0;

//   cout << "B1 : " << B1;
//   cout << "  C1 : " << C1;
//   cout << "  D1 : " << D1;
//   cout << "  B2 : " << B2;
//   cout << "  C2 : " << C2;
//   cout << "  D2 : " << D2;
//   cout << "  T : " << T << endl;

//   cout << "Xf : " << Xf;
//   cout << "  Yf : " << Yf;
//   cout << "  THf : " << THf;
//   cout << "  OMf : " << OMf;
//   cout << "  Vf : " << Vf << endl;

//   cout << "Xsim : " << Xsim;
//   cout << "  Ysim : " << Ysim;
//   cout << "  THsim : " << THsim;
//   cout << "  OMsim : " << OMsim;
//   cout << "  Vsim : " << Vsim << endl;
  
//   cout << "Xerr : " << Xerr;
//   cout << "  Yerr : " << Yerr;
//   cout << "  THerr : " << THerr;
//   cout << "  OMerr : " << OMerr;
//   cout << "  Verr : " << Verr << endl;

  sqRes += (Xerr * Xerr);
  sqRes += (Yerr * Yerr);
  sqRes += (50.0 * (THerr * THerr));
  sqRes += (10.0 * (OMerr * OMerr));
  sqRes += (10.0 * (Verr * Verr));
  
  return sqrt(sqRes);
}


/**************** Quadrature tools ****************/

/**
 *  \brief Quadrature tools...
 */
double DDTrajGen::qgaus(double (DDTrajGen::*func)(const double) const, const double a, const double b) const
{
  static const int n = 20;
  double x[n];
  double w[n];
  int j;
  double s;
  
  gauleg(a, b, x, w, n);      
  s=0;
  for (j=0;j<n;j++) {        
    s += w[j]*((this->*func)(x[j]));
  }
  return s;
}


/*********** State profiles computations **********/

/**
 *  \brief compute the angular acceleration profile of the left wheel
 */
void DDTrajGen::computeFctWLAcc()
{
  // delete the previous acceleration function
  if(fctWLAcc_ != NULL) delete fctWLAcc_;

  // create the polynomial acceleration function
  fctWLAccWB_ = new Polynomial(6.0*D2,4.0*C2,2.0*B2);
  fctWLAcc_ = new MultiItvFunction(fctWLAccWB_,0,INFINITY);

//   cout << "fctWLAcc_ : " << endl;
//   fctWLAcc_->display();

  list<ItvFunction> lMinBItv;
  fctWLAcc_->addMinBound(-maxWAcc,&lMinBItv);
  list<ItvFunction> lMaxBItv;
  fctWLAcc_->addMaxBound(maxWAcc,&lMaxBItv);
}

/**
 *  \brief compute the angular acceleration profile of the right wheel
 */
void DDTrajGen::computeFctWRAcc()
{
  // delete the previous acceleration function
  if(fctWRAcc_ != NULL) delete fctWRAcc_;

  // create the polynomial acceleration function
  fctWRAccWB_ = new Polynomial(6.0*D1,4.0*C1,2.0*B1);
  fctWRAcc_ = new MultiItvFunction(fctWRAccWB_,0,INFINITY);

//   cout << "fctWRAcc_ : " << endl;
//   fctWRAcc_->display();

  list<ItvFunction> lMinBItv;
  fctWRAcc_->addMinBound(-maxWAcc,&lMinBItv);
  list<ItvFunction> lMaxBItv;
  fctWRAcc_->addMaxBound(maxWAcc,&lMaxBItv);
}

/**
 *  \brief compute the angular velocity profile of the left wheel
 */
void DDTrajGen::computeFctWLVel()
{
  fctWLVel_ = (MultiItvFunction*) fctWLAcc_->integration(0.0,2.0*A2);

//   cout << "fctWLVel_ : " << endl;
//   fctWLVel_->display();

  list<ItvFunction> lMinBItv;
  fctWLVel_->addMinBound(-maxWVel,&lMinBItv);
  //list<ItvFunction>* lMinBItv = fctWLVel_->addMinBound(0.0);
  list<ItvFunction> lMaxBItv;
  fctWLVel_->addMaxBound(maxWVel,&lMaxBItv);
}

/**
 *  \brief compute the angular velocity profile of the right wheel
 */
void DDTrajGen::computeFctWRVel()
{
  fctWRVel_ = (MultiItvFunction*) fctWRAcc_->integration(0.0,2.0*A1);

//   cout << "fctWRVel_ : " << endl;
//   fctWRVel_->display();

  list<ItvFunction> lMinBItv;
  fctWRVel_->addMinBound(-maxWVel);
  //list<ItvFunction>* lMinBItv = fctWRVel_->addMinBound(0.0);
  list<ItvFunction> lMaxBItv;
  fctWRVel_->addMaxBound(maxWVel);
}

/**
 *  \brief compute the linear velocity profile of the robot
 */
void DDTrajGen::computeFctVel()
{
  fctVel_ = new MultiItvFunction(*fctWLVel_);
  (*fctVel_) += (*fctWRVel_);
  (*fctVel_) *= 0.5;

//   cout << "fctVel_ : " << endl;
//   fctVel_->display();
}

/**
 *  \brief compute the angular velocity profile of the robot
 */
void DDTrajGen::computeFctAVel()
{
  fctAVel_ = new MultiItvFunction(*fctWRVel_);
  (*fctAVel_) -= (*fctWLVel_);
  (*fctAVel_) *= (0.5 / R);

//   cout << "fctAVel_ : " << endl;
//   fctAVel_->display();
}

/**
 *  \brief compute the heading profile of the robot
 */
void DDTrajGen::computeFctTH()
{
//   cout << "##################### TH ######################" << endl;

//   cout << "fctAVel_ : " << endl;
//   fctAVel_->display();
  
  fctTH_ = (MultiItvFunction*) fctAVel_->integration(0.0,TH0);

//   cout << "fctTH_ : " << endl;
//   fctTH_->display();

//   cout << "###############################################" << endl;
}

/**
 *  \brief compute all profiles
 */
void DDTrajGen::computeAllFct()
{
  this->computeFctWLAcc();
  this->computeFctWRAcc();
  this->computeFctWLVel();
  this->computeFctWRVel();
  this->computeFctVel();
  this->computeFctAVel();
  this->computeFctTH();
}


/********** State Parameters Computation **********/  

/**
 *  \brief get the value of the angular acceleration profile of the left wheel at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the angular acceleration profile of the left wheel at time t
 */
double DDTrajGen::funcWLAcc(const double t) const
{
  return this->fctWLAcc_->getVal(t);
}

/**
 *  \brief get the value of the angular acceleration profile of the right wheel at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the angular acceleration profile of the right wheel at time t
 */
double DDTrajGen::funcWRAcc(const double t) const
{
  return this->fctWRAcc_->getVal(t);
}

/**
 *  \brief get the value of the angular velocity profile of the left wheel at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the angular velocity profile of the left wheel at time t
 */
double DDTrajGen::funcWLVel(const double t) const
{
  return this->fctWLVel_->getVal(t);
}

/**
 *  \brief get the value of the angular velocity profile of the right wheel at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the angular velocity profile of the right wheel at time t
 */
double DDTrajGen::funcWRVel(const double t) const
{
  return this->fctWRVel_->getVal(t);
}

/**
 *  \brief get the value of the linear velocity profile of the robot at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the linear velocity profile of the robot at time t
 */
double DDTrajGen::funcVel(const double t) const
{
  return this->fctVel_->getVal(t);
}

/**
 *  \brief get the value of the angular velocity profile of the robot at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the angular velocity profile of the robot at time t
 */
double DDTrajGen::funcAVel(const double t) const
{
  return this->fctAVel_->getVal(t);
}

/**
 *  \brief get the value of the heading profile of the robot at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the heading profile of the robot at time t
 */
double DDTrajGen::funcTheta(const double t) const
{
  return this->fctTH_->getVal(t);
}

/**
 *  \brief integration function for the position X profile computation
 *  \param t : the time at which the function is evaluated
 *  \return the value at time t of the integration function
 */  
double DDTrajGen::funcItgX(const double t) const
{
  return funcVel(t) * cos(normalizedAngle(funcTheta(t)));
}

/**
 *  \brief integration function for the position Y profile computation
 *  \param t : the time at which the function is evaluated
 *  \return the value at time t of the integration function
 */  
double DDTrajGen::funcItgY(const double t) const
{
  return funcVel(t) * sin(normalizedAngle(funcTheta(t)));
}

/**
 *  \brief get the value of the position X profile at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the position X profile at time t
 */
double DDTrajGen::funcX(const double t) const
{
  return X0 + qgaus(&DDTrajGen::funcItgX,0,t);
}

/**
 *  \brief get the value of the position Y profile at time t
 *  \param t : the time at which the function is evaluated
 *  \return value of the position Y profile at time t
 */
double DDTrajGen::funcY(const double t) const
{
  return Y0 + qgaus(&DDTrajGen::funcItgY,0,t);
}


/*************** Forward Solution *****************/

/**
 *  \brief compute the state reached by the current set of parameters
 */
void DDTrajGen::ForwardSolution()
{
  // initialize the partial derivatives
  for(int i=0; i<gLength; i++)
//     for(int j=0; j<pLength-1; j++)
    for(int j=0; j<pLengthFree(); j++)
      matJacobianG[i][j] = DFLT_DERIV_VAL;

//   cout << "///////////////////////////////////////////////////////////////////////////" << endl;
//   cout << "X0:" << X0 << " Y0:" << Y0 << " TH0:" << TH0 << " V0:" << V0 << " OM0:" << OM0 << " T0:" << T0 << endl;
//   cout << "B1:" << B1 << " C1:" << C1 << " D1:" << D1 << " B2:" << B1 << " C2:" << C2 << " D2:" << D2 << " T:" << T << endl;
  
  // compute the different functions' profiles
  computeAllFct();
  
  // compute the simulated state
  Xsim  = funcX(T);
  Ysim  = funcY(T);
  THsim = normalizedAngle(funcTheta(T));
  Vsim  = funcVel(T);
  OMsim = funcAVel(T);
  if(tcStatus() != TIME_FIXED) Tsim = T;

//   cout << "Xsim:" << Xsim << " Ysim:" << Ysim << " THsim:" << THsim << " Vsim:" << Vsim << " OMsim:" << OMsim << endl;
//   cout << "OMLsim : " << funcWLVel(T) << "  OMRsim : " << funcWRVel(T) << endl;

//   cout << "fctWLAcc : " << endl;
//   fctWLAcc_->display();
//   cout << "fctWRAcc : " << endl;
//   fctWRAcc_->display();
//   cout << "fctWLVel : " << endl;
//   fctWLVel_->display();
//   cout << "fctWRVel : " << endl;
//   fctWRVel_->display();
//   exit(1);
//   cout << "///////////////////////////////////////////////////////////////////////////" << endl;
}


/***************** Initialization *****************/

/**
 *  \brief set the initial guess of parameters
 *  \param b,c,d : coefficients of the polynomial right wheel acceleration profile
 *  \param b2,c2,d2 : coefficients of the polynomial left wheel acceleration profile
 *  \param t : final time to compute the trajectory (time of the final state)
 */
void DDTrajGen::setInitialGuess(double b1,double c1,double d1,double b2,double c2,double d2,double t)
{
  vecParameters_p = new Vec_DP(pLength);

  B1 = b1;
  C1 = c1;
  D1 = d1;
  B2 = b2;
  C2 = c2;
  D2 = d2;
  T = t;
}

/**
 *  \brief set the initial state, final state and constraints vectors
 *  \param x0 : initial state
 *  \param xf : final state
 *  \param constraints : constraints defined over the system
 *
 *  constraints[0] : max angular acceleration of the wheels
 *  constraints[1] : max linear acceleration of the wheels
 *  constraints[2] : radius of the robot
 */
void DDTrajGen::setInitialParam(Vec_DP& x0,Vec_DP& xf,Vec_DP& constraints)
{
  //Set initial state, final state and the constraints defined over the system
  vecState_x0 = x0;
  vecState_xf = xf;

  maxWAcc = constraints[0];
  maxWVel = constraints[1];
  R       = constraints[2];
  UnSurR  = 1.0 / R;
  
  A1 = 0.5 * (V0 + R*OM0);
  A2 = 0.5 * (V0 - R*OM0);
}

/**
 *  \brief set the initial guess with a vector of parameters
 *  \param parameters : the vector of parameters
 */
void DDTrajGen::setInitialGuess(double* p)
{
  this->setInitialGuess(p[0],p[1],p[2],p[3],p[4],p[5],p[6]);
}

/**
 *  \brief set the initial guess with a vector of parameters
 *  \param parameters : the vector of parameters
 */
void DDTrajGen::setInitialConfig(double* x0,double* constraints)
{
  // Set initial and final states
  vecState_x0 = Vec_DP(xLength);
  vecState_x0[0] = x0[0];
  vecState_x0[1] = x0[1];
  vecState_x0[2] = x0[2];
  vecState_x0[3] = x0[3];
  vecState_x0[4] = x0[4];
  vecState_x0[5] = x0[5];
  
  vecState_xf = vecState_x0;

  // Set the constraints defined over the system
  maxWAcc = constraints[0];
  maxWVel = constraints[1];
  R       = constraints[2];
  UnSurR  = 1.0 / R;
  
  A1 = 0.5 * (V0 + R*OM0);
  A2 = 0.5 * (V0 - R*OM0);

//   cout << "maxWAcc : " << maxWAcc << endl;
//   cout << "maxWVel : " << maxWVel << endl;
//   cout << "R : " << R << endl;
}


/************* Trajectory Generation **************/

/**
 *  \brief init a vector of the constraints defined over the system
 *  \param sys : the robotic system features
 *  \param vcstt : the resulting vector of constraints
 */
void DDTrajGen::initCstt(System* sys,Vec_DP& vcstt)
{
  vcstt[0] = ((DiffDriveSys*)sys)->getWAMax();
  vcstt[1] = ((DiffDriveSys*)sys)->getWVMax();
  vcstt[2] = ((DiffDriveSys*)sys)->getLength();
}

// /**
//  *  \brief generate a trajectory by constraints optimiation
//  *  \param x0 : initial state
//  *  \param xf : final state
//  *  \param constraints : constraints defined over the system
//  *  \param parameters : initial guess of parameters
//  *
//  *  x0 = (X0,Y0,TH0,V0,OM0,T0)
//  *  xf = (Xf,Yf,THf,Vf,OMf,Tf)
//  *  constraints = (maxWAcc,maxWVel,R)
//  */
// int DDTrajGen::ComputeTrajectory(Vec_DP x0,Vec_DP xf,Vec_DP constraints,Vec_DP** parameters)
// {
//   // initialize the correction coefficient
//   double coef_cor = COEF_COR;
//   int nbDeacreaseStep = 0;
//   double oldResidual = 0;

//   // Init the initial best parameters and residual
//   double bestB1=0, bestC1=0, bestD1=0, bestB2=0, bestC2=0, bestD2=0;
//   double bestResidual = -1;
//   int bestCNT = -1;

//   //Set initial state, final state and system constraints
//   setInitialParam(x0,xf,constraints);

//   if(DBG_ >= 1) {
//     cout << "trajGen : (X0=" << X0 << ",Y0=" << Y0 << ",TH0=" << TH0 << 
//       ",V0=" << V0 << ",OM0=" << OM0 << ",T0=" << T0 << ") -> ";
//     cout << "(Xf=" << Xf << ",Yf=" << Yf << ",THf=" << THf << 
//       ",Vf=" << Vf << ",OMf=" << OMf << ",Tf=" << Tf << ")" << endl;
//   }

//   //Heuristic guess
//   if(parameters == NULL || *parameters == NULL)
//     InitialGuess(parameters);
//   else
//     {
//       vecParameters_p = *parameters;
//       T = Tf - T0;
//     }

//   if(DBG_ >= 1) {
//     cout << "Initial Guess :  ";
//     cout << "B1:" << B1 << "  C1:" << C1 << "  D1:" << D1 << "  B2:" << B2 << "  C2:" << C2 << "  D2:" << D2 << "  T:" << T << endl;
//   }

//   // Init the number of the algorithm's step and the residual
//   double residual = 1000000;  
//   int cnt=0;
//   CNT_ = 0;
  
//   vecCorrParameters_p = 0;  

//   while(residual > maxResidual_)
//     {
//       // save the best parameters
//       if(CNT_ == 0 || residual < bestResidual)
// 	{
// 	  bestB1 = B1;
// 	  bestC1 = C1;
// 	  bestD1 = D1;
// 	  bestB2 = B2;
// 	  bestC2 = C2;
// 	  bestD2 = D2;
// 	  bestResidual = residual;
// 	  bestCNT = CNT_;
// 	}

// #if VARIABLE_COEF_COR
//       // update the correction coefficient
//       if(residual <= oldResidual)
// 	{
// 	  nbDeacreaseStep++;
	  
// 	  if(nbDeacreaseStep >= 10) coef_cor = 0.75;
// 	  else if(nbDeacreaseStep >= 9) coef_cor = 0.60;
// 	  else if(nbDeacreaseStep >= 8) coef_cor = 0.45;
// 	  else if(nbDeacreaseStep >= 7) coef_cor = 0.35;
// 	  else if(nbDeacreaseStep >= 6) coef_cor = 0.25;
// 	  else if(nbDeacreaseStep >= 5) coef_cor = 0.20;
// 	  else if(nbDeacreaseStep >= 3) coef_cor = 0.15;	
// 	}
//       else
// 	{
// 	  coef_cor = COEF_COR;
// 	  nbDeacreaseStep = 0;
// 	}
//       oldResidual = residual;
// #endif

//       //Apply correction
// #if TIME_FIXED
//       for(int i=0; i<pLength-1;i++)
// #else
//       for(int i=0; i<pLength;i++)
// #endif
// 	(*vecParameters_p)[i] += coef_cor*vecCorrParameters_p[i];            

//       printf ("Parameters = (,B1=%f,C1=%f,D1=%f,B2=%f,C2=%f,D2=%f,T=%f)\n", 
// 	      B1,C1,D1,B2,C2,D2,T);

      
//       //MUST_DISPLAY = false;
//       ForwardSolution();
//       //MUST_DISPLAY = false;
//       CNT_++;

//       if(DBG_ >= 1) {
// 	printf ("Simulated state = (Xsim=%f,Ysim=%f,Vsim=%f,OMsim=%f)\n", 
// 		Xsim,Ysim,THsim,Vsim,OMsim);
//       }

//       ComputeG();

//       if(DBG_ >= 1) {
// 	printf ("Errors = (%f,%f,%f,%f,%f)\n", 
// 		Xerr,Yerr,THerr,Verr,OMerr);
// // 	printf ("Parameters K = (OM0=%f,B=%f,C=%f,D=%f,T=%f)\n", 
// // 		OM0,B,C,D,T);
// // 	printf ("Parameters A = (B2=%f,C2=%f,D2=%f,T=%f)\n", 
// // 		B2,C2,D2,T);
//       }

//       ComputeJacobianG_FD();
//       ComputeInvJacobianG();

//       //Compute the correction
// #if TIME_FIXED
//       for(int i=0; i<pLength-1;i++)
// #else
// 	for(int i=0; i<pLength;i++)
// #endif
// 	  {
// 	    vecCorrParameters_p[i] = 0;
	    
// // 	    if(!TIME_FIXED || i != 3)
// // 	      {
// 		for(int j=0;j<gLength;j++)
// 		  {
// 		    vecCorrParameters_p[i] += (-1.0)*vecConstraints_g[j]*matInvJacobianG[i][j]; 
// 		  }
// // 	      }	    
// 	  }
      
// //       if(DBG_ >= 1) {
// // 	printf ("Correction = (B=%f,C=%f,D=%f,B2=%f,C2=%f,D2=%f)\n", 
// // 		vecCorrParameters_p[0],vecCorrParameters_p[1],vecCorrParameters_p[2],
// // 		vecCorrParameters_p[3],vecCorrParameters_p[4],vecCorrParameters_p[5]);
// //       }

//       residual = sqrt( pow(vecConstraints_g[0],2) + 
// 		       pow(vecConstraints_g[1],2) +
// 		       pow(vecConstraints_g[2],2) +
// 		       pow(vecConstraints_g[3],2) +
// 		       pow(vecConstraints_g[4],2) + 
// 		       pow(vecConstraints_g[5],2));
      
//       if(DBG_ >= 1) {        
// 	printf ("Residual = %f\t CNT : %d\t COEF_COR : %f\n", 
// 		residual,CNT_,coef_cor); 
//       }      

//       if(DBG_ >= 2) {
// 	displayProfiles();
//       }

//       cnt++;
//       if(cnt >= maxNbStep_)
// 	{
// 	  if(DBG_ >= 1) {
// 	    cout << "  No traj found in " << maxNbStep_ << "iterations" << endl;
// 	  }
	  
// 	  // get back the best parameters found
// 	  B1 = bestB1;
// 	  C1 = bestC1;
// 	  D1 = bestD1;
// 	  B2 = bestB2;
// 	  C2 = bestC2;
// 	  D2 = bestD2;
	  
// 	  // compute the forward method with these parameters
// 	  ForwardSolution();
// 	  // compute the associate error
// 	  ComputeG();
	  
// 	  // get back errors on X, Y and TH
// 	  finalResidual_ = bestResidual;
// 	  finalNbSteps_ = CNT_;
// 	  finalErrX_ = vecConstraints_g[0];
// 	  finalErrY_ = vecConstraints_g[1];
// 	  finalErrTH_ = vecConstraints_g[2];
	  
// 	  return -1;
// 	}
//     }
//   finalResidual_ = residual;
//   finalNbSteps_ = CNT_;
//   finalErrX_ = vecConstraints_g[0];
//   finalErrY_ = vecConstraints_g[1];
//   finalErrTH_ = vecConstraints_g[2];
//   return 0;
// }

/**
 *  \brief generate a trajectory by constraints optimiation
 *  \param statet : initial state-time
 *  \param sys : constraints defined over the robotic system
 *  \param ctrlt : control-time applied on the system
 */
void DDTrajGen::ComputeFromCtrl(const StateTime& statet,const System* sys,const ControlTime& ctrlt)
{
  cerr << "DDTrajGen::ComputeFromCtrl : Not implemented yet!" << endl;
  exit(1);
}


/**************** Output Functions ****************/

/**
 *  \brief get back the state reached at time t
 *  \param rstatet : the resulting reached state
 *  \param t : time at which we evaluate the reached state
 *
 *  The resulting reached state must be already allocated
 *  Reached state-time : (w,y,th,omr,oml,t)
 */
void DDTrajGen::getStateTReachedAt(double* rstatet,double t) const
{
  rstatet[0] = funcX(t);
  rstatet[1] = funcY(t);
  rstatet[2] = normalizedAngle(funcTheta(t));
  rstatet[3] = funcWRVel(t);
  rstatet[4] = funcWLVel(t);
  rstatet[5] = t;
}

/**
 *  \brief get back the state reached at time t
 *  \param rstatet : the resulting reached state
 *  \param t : time at which we evaluate the reached state
 *
 *  The resulting reached state must be already allocated
 *  Reached state-time : (w,y,th,omr,oml,t)
 */
void DDTrajGen::getFinalStateTReached(double* rstatet) const
{
  rstatet[0] = funcX(T);
  rstatet[1] = funcY(T);
  rstatet[2] = normalizedAngle(funcTheta(T));
  rstatet[3] = funcWRVel(T);
  rstatet[4] = funcWLVel(T);
  rstatet[5] = T;
}

/**
 *  \brief save a sampling of the trajectory profile
 *  \param nbp : number of sampling points
 */
void DDTrajGen::SaveTrajectory(int nbp)
{
  if(posX != NULL) delete posX;
  if(posY != NULL) delete posY;
  if(headTH != NULL) delete headTH;
  if(timeT != NULL) delete timeT;
  if(velV != NULL) delete velV;
  if(avelOM != NULL) delete avelOM;

  posX = new vector<double>;
  posY = new vector<double>;
  headTH = new vector<double>;
  timeT = new vector<double>;
  velV = new vector<double>;
  avelOM = new vector<double>;

  posX->clear();
  posY->clear();
  headTH->clear();
  timeT->clear();
  velV->clear();
  avelOM->clear();

  //Computing trajectory
  double step = 1.0 / ((double)nbp);
  for(int j=0;j<=nbp;j++)
    {
      if(fctWLAcc_ != NULL && fctWRAcc_ != NULL && fctWLVel_ != NULL && fctWRVel_ != NULL)
	{
	  posX->push_back(funcX(j*(step*T)));
	  posY->push_back(funcY(j*(step*T)));
	  headTH->push_back(funcTheta(j*(step*T)));
	  timeT->push_back(j*step*T);
	  velV->push_back(funcVel(j*step*T));
	  avelOM->push_back(funcAVel(j*step*T));
	}
      else
	{
	  posX->push_back(0.0);
	  posY->push_back(0.0);
	  headTH->push_back(0.0);
	  timeT->push_back(0.0);
	  velV->push_back(0.0);
	  avelOM->push_back(0.0);
	}
    }
}

/**
 *  \brief trace the generated trajectory on a map
 *  \param map : the map on which to trace the path
 *  \param nbp : number of sampling points
 */
void DDTrajGen::traceOn(Map *map,int nbp)
{
  if(map == NULL)
    {
      cout << "DDTrajGen::traceOn : Warning. The input map is null. Can't trace the path." << endl;
      return;
    }
  else
    {
      int pathCol = (int) (map->getDepth() / 2);
      for(int i=0; i<=nbp; i++)
	{
	  double x = funcX(i*(1.0/((double)nbp)*T));
	  double y = funcY(i*(1.0/((double)nbp)*T));

	  if( (x >= 0) && (x < map->getWidth()) && (y >= 0) && (y < map->getHeight()))
	    map->setCol(((int)x),((int)y),pathCol);
	}
    }
}

/**
 *  \brief trace the generated trajectory on a map
 *  \param map : the map on which to trace the path
 *  \param nbp : number of sampling points
 *  \param initPos : initial position
 */  
void DDTrajGen::traceOn(Map *map,int nbp,const PositionXYTh& initPos)
{
  if(map == NULL)
    {
      cout << "DDTrajGen::traceOn : Warning. The input map is null. Can't trace the path." << endl;
      return;
    }
  else
    {
      int pathCol = (int) (map->getDepth() / 2);
      for(int i=0; i<=nbp; i++)
	{
	  double x1 = funcX(i*(1.0/((double)nbp)*T));
	  double y1 = funcY(i*(1.0/((double)nbp)*T));
	  double th1 = funcTheta(i*(1.0/((double)nbp)*T));
	  PositionXYTh pos1(x1,y1,th1);

	  PositionXYTh pos = coordInFrame(pos1,initPos);
	  double x = pos.x();
	  double y = pos.y();

	  if( (x >= 0) && (x < map->getWidth()) && (y >= 0) && (y < map->getHeight()))
	    map->setCol(((int)x),((int)y),pathCol);
	}
    }
}

/**
 *  \brief return a sample trajectory from the parameterized one
 *  \param ts : the sampling time step
 *  \return a sample trajectory from the parameterized one
 *
 *  A new instance of sample trajectory is returned here
 */
DDSTraj* DDTrajGen::sampledTraj(double ts) const
{
  cout << "DDTrajGen::sampledTraj : Not implemented yet!" << endl;
  exit(1);
}


/*************** Utility Functions ****************/

/**
 *  \brief display both wheels velocity, linear and angular velocity, heading and position profiles
 */
void DDTrajGen::displayProfiles()
{
  // compute the forward solution with the new parameters
  this->ForwardSolution();
  
  //Computing Position
  this->SaveTrajectory(100);
  vector<double> xIGpos, yIGpos;
  xIGpos.push_back(X0);
  yIGpos.push_back(Y0);
  xIGpos.push_back(Xf);
  yIGpos.push_back(Yf);

  Gnuplot gPosition;
  gPosition.reset_plot();
  gPosition.set_style("lines").plot_xy(*(this->posX),*(this->posY),"Position");
  gPosition.set_style("points").plot_xy(xIGpos,yIGpos,"Initial and Final Pos");

  //Computing Left and Right Wheels Acceleration
  vector<double> wlacc;
  vector<double> wracc;
  vector<double> wlaccBinf;
  vector<double> wlaccBsup;
  vector<double> distancewlacc;
  for(int j=0;j<=50;j++){
    wlacc.push_back((double)this->funcWLAcc(j*(T/50)));
    wracc.push_back((double)this->funcWRAcc(j*(T/50)));
    wlaccBinf.push_back(-maxWAcc);
    wlaccBsup.push_back(maxWAcc);
    distancewlacc.push_back((j*(T/50)));
  }    
  Gnuplot gWLAcc;
  gWLAcc.reset_plot();
  gWLAcc.set_style("lines").plot_xy(distancewlacc,wlacc,"Left Wheel Acceleration");
  gWLAcc.set_style("lines").plot_xy(distancewlacc,wracc,"Right Wheel Acceleration");
  gWLAcc.set_style("points").plot_xy(distancewlacc,wlaccBinf,"Min Bound");
  gWLAcc.set_style("points").plot_xy(distancewlacc,wlaccBsup,"Max Bound");

  //Computing Left Wheel Velocity
  vector<double> wlvel;
  vector<double> wrvel;
  vector<double> wlvelBinf;
  vector<double> wlvelBsup;
  vector<double> distancewl;
  for(int j=0;j<=50;j++){
    wlvel.push_back((double)this->funcWLVel(j*(T/50)));
    wrvel.push_back((double)this->funcWRVel(j*(T/50)));
    wlvelBinf.push_back(-maxWVel);
    wlvelBsup.push_back(maxWVel);
    distancewl.push_back((j*(T/50)));
  }    
  Gnuplot gWLVel;
  gWLVel.reset_plot();
  gWLVel.set_style("lines").plot_xy(distancewl,wlvel,"Left Wheel Velocity");
  gWLVel.set_style("lines").plot_xy(distancewl,wrvel,"Right Wheel Velocity");
  gWLVel.set_style("points").plot_xy(distancewl,wlvelBinf,"Min Bound");
  gWLVel.set_style("points").plot_xy(distancewl,wlvelBsup,"Max Bound");
  
  //Computing Velocity
  vector<double> vel;
  vector<double> distanceV;
  for(int j=0;j<=50;j++){
    vel.push_back((double)this->funcVel(j*(T/50)));
    distanceV.push_back((j*(T/50)));
  }    
  Gnuplot gVel;
  gVel.reset_plot();
  gVel.set_style("lines").plot_xy(distanceV,vel,"Velocity");
  
  //Computing Angular Velocity
  vector<double> avel;
  vector<double> distanceAV;
  for(int j=0;j<=50;j++){
    avel.push_back((double)this->funcAVel(j*(T/50)));
    distanceAV.push_back((j*(T/50)));
  }    
  Gnuplot gAVel;
  gAVel.reset_plot();
  gAVel.set_style("lines").plot_xy(distanceAV,avel,"Angular Velocity");
  
  //Computing Heading
  vector<double> dir2;
  vector<double> dir2n;
  vector<double> cosdir2n;
  vector<double> sindir2n;
  vector<double> distance2;
  for(int j=0;j<=50;j++){
    double thtmp = (double)this->funcTheta(j*(T/50.0));
    double thtmpn = normalizedAngle(thtmp);
    double disttmp = (j*(T/50.0));
    dir2.push_back(thtmp);
    dir2n.push_back(thtmpn);
    cosdir2n.push_back(cos(thtmpn));
    sindir2n.push_back(sin(thtmpn));
    distance2.push_back(disttmp);
  } 
  Gnuplot gTH;
  gTH.reset_plot();
  gTH.set_style("lines").plot_xy(distance2,dir2,"Heading");
  

  wait_for_key();
}

/**
 *  \brief display all functions
 */
void DDTrajGen::displayAllFct()
{
  cout << "Left Wheel Acceleration : " << endl;
  fctWLAcc_->display();
  cout << "Right Wheel Acceleration : " << endl;
  fctWRAcc_->display();
  cout << "Left Wheel Velocity : " << endl;
  fctWLVel_->display();
  cout << "Right Wheel Velocity : " << endl;
  fctWRVel_->display();
  cout << "Linear Velocity : " << endl;
  fctVel_->display();
  cout << "Angular Velocity : " << endl;
  fctAVel_->display();
  cout << "Heading : " << endl;
  fctTH_->display();
}
