#include "UtilsTG.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /************** Conversion functions **************/
                                    /**************************************************/

/**
 *  \brief Convert car-like states into vectors for the trajectory generation process
 *  \param initS : initial car-like state
 *  \param goalS : goal car-like state
 *  \param initTS : resulting initial state table
 *  \param goalTS : resulting goal state table
 *
 *  initTS = (X0,Y0,TH0,PH0,V0,T0)
 *  goalTS = (Xf,Yf,THf,PHf,Vf,Tf)
 */
void mpc::CLStateT_2_CL2tabD(const CLStateT& initS,const CLStateT& goalS,double** initTS,double** goalTS)
{
  (*initTS) = new double[6]; 
  (*initTS)[0] = initS.x(); 
  (*initTS)[1] = initS.y(); 
  (*initTS)[2] = initS.th(); 
  (*initTS)[3] = initS.ph(); 
  (*initTS)[4] = initS.s(); 
  (*initTS)[5] = 0.0;

  (*goalTS) = new double[6];
  (*goalTS)[0] = goalS.x(); 
  (*goalTS)[1] = goalS.y(); 
  (*goalTS)[2] = goalS.th(); 
  (*goalTS)[3] = goalS.ph(); 
  (*goalTS)[4] = goalS.s(); 
  (*goalTS)[5] = goalS.t() - initS.t();
}

/**
 *  \brief Convert a car-like states into a vectors for the trajectory generation process
 *  \param initS : initial car-like state
 *  \param goalS : goal car-like state
 *  \param initVS : resulting initial state vector
 *  \param goalVS : resulting goal state vector
 *  \param L : length of the car-like vehicle
 *
 *  initVS = (X0,Y0,TH0,OM0,V0,T0)
 *  goalVS = (Xf,Yf,THf,OMf,Vf,Tf)
 */
void mpc::CLStateT_2_CLVecDP(const CLStateT& initS,const CLStateT& goalS,Vec_DP& initVS,Vec_DP& goalVS,double L)
{
  initVS = Vec_DP(6); 
  initVS[0] = initS.x(); 
  initVS[1] = initS.y(); 
  initVS[2] = initS.th(); 
  initVS[3] = initS.s()*tan(initS.ph())/L; 
  initVS[4] = initS.s(); 
  initVS[5] = 0.0;

  goalVS = Vec_DP(6); 
  goalVS[0] = goalS.x(); 
  goalVS[1] = goalS.y(); 
  goalVS[2] = goalS.th(); 
  goalVS[3] = goalS.s()*tan(goalS.ph())/L; 
  goalVS[4] = goalS.s(); 
  goalVS[5] = goalS.t() - initS.t();
}

/**
 *  \brief Convert a car-like state into a vector for the trajectory generation process
 *  \param initS : initial car-like state
 *  \param initTS : resulting initial state table
 *
 *  initTS = (X0,Y0,TH0,PH0,V0,T0)
 */
void mpc::CLStateT_1_CL2tabD(const CLStateT& initS,double** initTS,double L)
{
  (*initTS) = new double[6]; 
  (*initTS)[0] = initS.x(); 
  (*initTS)[1] = initS.y(); 
  (*initTS)[2] = initS.th(); 
  (*initTS)[3] = initS.ph(); 
  (*initTS)[4] = initS.s(); 
  (*initTS)[5] = 0.0;
}

/**
 *  \brief Convert a car-like state into a vector for the trajectory generation process
 *  \param initS : initial car-like state
 *  \param initTS : resulting initial state table
 *
 *  initTS = (X0,Y0,TH0,OM0,V0,T0)
 */
void mpc::CLStateT_1_CLtabD(const CLStateT& initS,double** initTS,double L)
{
  (*initTS) = new double[6]; 
  (*initTS)[0] = initS.x(); 
  (*initTS)[1] = initS.y(); 
  (*initTS)[2] = initS.th(); 
//   cout << "initS.th() : " << initS.th() << "  tan(initS.th()) : " << tan(initS.th()) << "  L : " << L << "  tan(initS.th())/L : " << tan(initS.th())/L << endl;
  (*initTS)[3] = initS.s()*tan(initS.ph())/L;
  (*initTS)[4] = initS.s(); 
  (*initTS)[5] = 0.0;
}

/**
 *  \brief Convert wheelchair states into vectors for the trajectory generation process
 *  \param initS : initial wheelchair state
 *  \param goalS : goal wheelchair state
 *  \param initVS : resulting initial state vector
 *  \param goalVS : resulting goal state vector
 *
 *  initVS = (X0,Y0,TH0,V0,OM0,T0)
 *  goalVS = (Xf,Yf,THf,Vf,OM0,Tf)
 */
void mpc::WCStateT_2_WCVecDP(const WCStateT& initS,const WCStateT& goalS,Vec_DP& initVS,Vec_DP& goalVS)
{
  initVS = Vec_DP(6); 
  initVS[0] = initS.x(); 
  initVS[1] = initS.y(); 
  initVS[2] = initS.th(); 
  initVS[3] = initS.s(); 
  initVS[4] = initS.om(); 
  initVS[5] = 0.0;

  goalVS = Vec_DP(6); 
  goalVS[0] = goalS.x(); 
  goalVS[1] = goalS.y(); 
  goalVS[2] = goalS.th(); 
  goalVS[3] = goalS.s(); 
  goalVS[4] = goalS.om(); 
  goalVS[5] = goalS.t() - initS.t();
}

/**
 *  \brief Convert a car-like system into a vector of constraints for the trajectory generation process
 *  \param sys : input car-like system
 *  \param cstt : output vector of constraints
 *
 *  constraints = (maxCurv,maxAcc,maxVel,robotLength,maxSteer,maxSteerSpd)
 */
void mpc::CLSys_2_CLVecCstt(const CarLikeSys& sys,Vec_DP& cstt)
{
  cstt = Vec_DP(6); 
  cstt[0] = 0.0;
  cstt[1] = sys.getAMax();
  cstt[2] = sys.getVMax();
  cstt[3] = sys.getLength();
  cstt[4] = sys.getMaxSteering();
  cstt[5] = sys.getMaxSteerSpd();
}

/**
 *  \brief Convert a car-like system into a vector of constraints for the trajectory generation process
 *  \param sys : input car-like system
 *  \param cstt : output vector of constraints
 *
 *  constraints = (maxAcc,maxVel,robotLength,maxSteer,maxSteerSpd)
 */
void mpc::CLSys_2_CL3VecCstt(const CarLikeSys& sys,Vec_DP& cstt)
{
  cstt = Vec_DP(5); 
  cstt[0] = sys.getAMax();
  cstt[1] = sys.getVMax();
  cstt[2] = sys.getLength();
  cstt[3] = sys.getMaxSteering();
  cstt[4] = sys.getMaxSteerSpd();
}

/**
 *  \brief Convert a wheelchair system into a vector of constraints for the trajectory generation process
 *  \param sys : input wheelchair system
 *  \param cstt : output vector of constraints
 *
 *  constraints = (maxAcc,maxAAcc,maxVel,maxAVel)
 */
void mpc::WCSys_2_WCVecCstt(const WheelChairSys& sys,Vec_DP& cstt)
{
  cstt = Vec_DP(4); 
  cstt[0] = sys.getAMax();
  cstt[1] = sys.getAAMax();
  cstt[2] = sys.getVMax();
  cstt[3] = sys.getAVMax();
}

/**
 *  \brief Convert a car-like system into a vector of constraints for the trajectory generation process
 *  \param sys : input car-like system
 *  \param cstt : output vector of constraints
 *
 *  constraints = (maxAcc,maxVel,maxSteerSpd,maxSteer,robotLength)
 */
void mpc::CLSys_2_CL2TabCstt(const CarLikeSys& sys,double** cstt)
{
  (*cstt) = new double[5]; 
  (*cstt)[0] = sys.getAMax();
  (*cstt)[1] = sys.getVMax();
  (*cstt)[2] = sys.getMaxSteerSpd();
  (*cstt)[3] = sys.getMaxSteering();
  (*cstt)[4] = sys.getLength();
}

/**
 *  \brief Convert a car-like system into a vector of constraints for the trajectory generation process
 *  \param sys : input car-like system
 *  \param cstt : output vector of constraints
 *
 *  constraints = (maxAcc,maxSpd,L,maxSteer,maxSteerSpd)
 */
void mpc::CLSys_2_CL3TabCstt(const CarLikeSys& sys,double** cstt)
{
  (*cstt) = new double[5]; 
  (*cstt)[0] = sys.getAMax();
  (*cstt)[1] = sys.getVMax();
  (*cstt)[2] = sys.getLength();
  (*cstt)[3] = sys.getMaxSteering();
  (*cstt)[4] = sys.getMaxSteerSpd();
}

// /**
//  *  \brief Convert a car-like control into a vector of parameters IG for the trajectory generation process
//  *  \param ctrl : input car-like control
//  *  \param t : input time
//  *  \param param : output parameters IG for the trajectory generation process
//  *  \param s : the state on which the control is applied
//  *  \param L : input length of the car-like
//  */
// void mpc::CLctrlt_2_CLIG(const CLControl& ctrl,double t,double** param,const StateS6D& s,double L)
// {
//   // compute the angular velocity corresponding to the input steering speed
//   double acc    = ctrl.acc();
//   double sts    = ctrl.ss();
//   double vel    = s.s();
//   double sta    = s.ph();
//   double tanSta = tan(sta);

//   double aacc = (1.0 / L) * (acc*tanSta + vel*sts*(1.0 + tanSta*tanSta));
    
//   (*param) = new double[7];
//   (*param)[0] = aacc;
//   (*param)[1] = 0.0;
//   (*param)[2] = 0.0;
//   (*param)[3] = ctrl.acc();
//   (*param)[4] = 0.0;
//   (*param)[5] = 0.0;
//   (*param)[6] = t;
// }

/**
 *  \brief Convert a car-like control into a vector of parameters IG for the trajectory generation process
 *  \param ctrl : input car-like control
 *  \param t : input time
 *  \param param : output parameters IG for the trajectory generation process
 */
void mpc::CLctrlt_2_CL2IG(const CLControl& ctrl,double t,double** param)
{    
  (*param) = new double[7];
  (*param)[0] = ctrl.ss();
  (*param)[1] = 0.0;
  (*param)[2] = 0.0;
  (*param)[3] = ctrl.acc();
  (*param)[4] = 0.0;
  (*param)[5] = 0.0;
  (*param)[6] = t;
}


// /**
//  *  \brief Convert a car-like control into a vector of parameters IG for the trajectory generation process
//  *  \param param : input parameters IG for the trajectory generation process
//  *  \param ctrl : output car-like control
//  *  \param t : output time
//  *  \param L : input length of the car-like
//  */
// void mpc::IG_2_CLctrlt(double* param,const CLControl& ctrl,double& t,const StateS6D& s,double L)
// {
// }

/**
 *  \brief Convert a car-like control into a vector of parameters IG for the trajectory generation process
 *  \param ctrl : input cycab control
 *  \param t : input time
 *  \param param : output parameters IG for the trajectory generation process
 *  \param s : the state on which the control is applied
 *  \param L : input length of the car-like
 */
void mpc::CCctrlt_2_CL3IG(const CCControl& ctrl,double t,double** param,const CLStateT& s,double L)
{
  (*param) = new double[7];
  (*param)[0] = 0.0;
  (*param)[1] = 0.0;
  (*param)[2] = 0.0;
  (*param)[3] = 0.0;
  (*param)[4] = 0.0;
  (*param)[5] = 0.0;
  (*param)[6] = t;
}

/**
 *  \brief Convert a wheelchair control into a vector of parameters IG for the trajectory generation process
 *  \param ctrl : input wheelchair control
 *  \param t : input time
 *  \param param : output parameters IG for the trajectory generation process
 */
void mpc::WCctrlt_2_WCIG(const WCControl& ctrl,double t,double** param)
{
  (*param) = new double[7];
  (*param)[0] = 0.0;
  (*param)[1] = 0.0;
  (*param)[2] = 0.0;
  (*param)[3] = 0.0;
  (*param)[4] = 0.0;
  (*param)[5] = 0.0;
  (*param)[6] = t;
}


                                    /**************************************************/
                                    /***************** Bends functions ****************/
                                    /**************************************************/

/**
 *  \brief compute the center of curvature of a left bend follow by a car-like
 *  \param p : initial position of the car-like
 *  \param maxSt : maximal steering angle of the car-like
 *  \param L : length of the car-like
 *  \return the center of curvature of a left bend follow by a car-like
 */
PositionXY mpc::leftCurvCenter(PositionXYTh p,double maxSt,double L)
{
  // compute the minimal curvature radius the system can follow
  double curvRad = L / tan(maxSt);

  double angle = normalizedAngle(p.th() + 0.5*M_PI);

  PositionXY res(p.x() + curvRad*cos(angle),
		 p.y() + curvRad*sin(angle));
  return res;
}

/**
 *  \brief compute the center of curvature of a right bend follow by a car-like
 *  \param p : initial position of the car-like
 *  \param maxSt : maximal steering angle of the car-like
 *  \param L : length of the car-like
 *  \return the center of curvature of a right bend follow by a car-like
 */
PositionXY mpc::rightCurvCenter(PositionXYTh p,double maxSt,double L)
{
  // compute the minimal curvature radius the system can follow
  double curvRad = L / tan(maxSt);

  double angle = normalizedAngle(p.th() - 0.5*M_PI);

  PositionXY res(p.x() + curvRad*cos(angle),
		 p.y() + curvRad*sin(angle));
  return res;
}
