#ifndef MANATRAJSEL_HPP
#define MANATRAJSEL_HPP

#include "trajSel/inc/TrajSel.hpp"

namespace mpc
{

                                    /**************************************************/
                                    /**************** Class ManaTrajSel ***************/
                                    /**************************************************/

  /** \class ManaTrajSel
   * \brief reactive navigation method selecting the best trajectory to follow from an input DEM (for a Mana robotic system)
   */
  class ManaTrajSel : public TrajSel
  {
  protected : 
    int spgAcc_;                                       /**< Sampling over the linear acceleration control */
    int spgAAcc_;                                      /**< Sampling over the angular acceleration control */
    int spgJerk_;                                      /**< Sampling over the linear jerk control */
    int spgAJerk_;                                     /**< Sampling over the angular jerk control */

    int modeCtrlSpg_;                                  /**< control sampling mode */

  protected : 
    double coefAcc_;                                   /**< coefficient of the linear acceleration control used */
    double coefAAcc_;                                  /**< coefficient of the angular acceleration control used */
    double coefJerk_;                                  /**< coefficient of the linear jerk control used */
    double coefAJerk_;                                 /**< coefficient of the angular jerk control used */

  public : 
    static double TS_WG;                               /**< Goal position heuritic weight */
    static double TS_WO;                               /**< Obstacles heuritic weight */
    static double TS_WH;                               /**< Goal heading heuritic weight */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    ManaTrajSel();

  public : 
    /**
     *  \brief default constructor
     *  \param sys : the robotic system constraints
     *  \param spgAcc : sampling over the linear acceleration control
     *  \param spgAAcc : sampling over the angular acceleration control
     */
    ManaTrajSel(ManaSys* sys, int spgAcc=5, int spgAAcc=5);

    /**
     *  \brief default constructor
     *  \param sys : the robotic system constraints
     *  \param spgAcc : sampling over the linear acceleration control
     *  \param spgAAcc : sampling over the angular acceleration control
     *  \param spgJerk : sampling over the linear jerk control
     *  \param spgAJerk : sampling over the angular jerk control
     *  \param coefAcc : coefficient of the linear acceleration control used
     *  \param coefAAcc : coefficient of the angular acceleration control used
     *  \param coefJerk : coefficient of the linear jerk control used
     *  \param coefAJerk : coefficient of the angular jerk control used
     */
    ManaTrajSel(ManaSys* sys, int spgAcc, int spgAAcc, int spgJerk, int spgAJerk, double coefAcc=0.3, double coefAAcc=0.5, double coefJerk=2.8, double coefAJerk=4.2);
    
    /**
     *  \brief destructor
     */
    virtual ~ManaTrajSel();


    /******************** Accessors *******************/

  protected : 
    /*
     *  \brief return a pointer to the robotic system constraints
     *  \return a pointer to the robotic system constraints
     */
    virtual ManaSys* sys() const { return ((ManaSys*)sys_); }

    /*
     *  \brief return a pointer to the localization of the robot
     *  \return a pointer to the localization of the robot
     */
    virtual ManaState* loc() const { return ((ManaState*)loc_); }

    /*
     *  \brief return a pointer to the best found control trajectory
     *  \return a pointer to the best found control trajectory
     */
    virtual ManaPCTraj* bestCTraj() const { return ((ManaPCTraj*)bestPCTraj_); }

    /*
     *  \brief return a pointer to the best found control trajectory
     *  \return a pointer to the best found control trajectory
     */
    virtual ManaPSTraj* bestSTraj() const { return ((ManaPSTraj*)bestPSTraj_); }


    /**************** Vector management ***************/

    /*
     *  \brief return a pointer the the i-th candidate control trajectory
     *  \return a pointer the the i-th candidate control trajectory
     */
    virtual const ManaPCTraj* getCandCTraj(int i) const;

    /*
     *  \brief return a pointer the the i-th candidate state trajectory
     *  \return a pointer the the i-th candidate state trajectory
     */
    virtual const ManaPSTraj* getCandSTraj(int i) const;


    /************* Trajectory computation *************/

  protected : 
    /*
     *  \brief compute the set of control trajectories
     *  \param distToGoal : distance from the current position of the robot to the goal location
     *  \param distToObs : distance from the current position of the robot to the closest obstacle
     *  \return the number of computed control trajectories
     */
    virtual int computeCTrajSet(double distToGoal, double distToObs);

    /*
     *  \brief compute the set of control trajectories (with acceleration control)
     *  \return the number of computed control trajectories
     */
    virtual int computeCTrajSetAcc();

    /*
     *  \brief compute the set of control trajectories (with jerk control)
     *  \param distToGoal : distance from the current position of the robot to the goal location
     *  \param distToObs : distance from the current position of the robot to the closest obstacle
     *  \return the number of computed control trajectories
     */
    virtual int computeCTrajSetJerk(double distToGoal, double distToObs);

    /*
     *  \brief compute the set of state trajectories
     *  \return the number of computed state trajectories
     */
    virtual int computeSTrajSet();


    /************* Heurisctics computation ************/

  private : 
    /*
     *  \brief compute the heuristic of an input trajectory
     *  \param curS : the current state of the robot
     *  \param targetPos : the target location
     *  \param ctraj : the input control trajectory
     *  \param straj : the input state trajectory
     */
    float computeH(const ManaState& curS, const PositionXY& targetPos, 
		   const ManaPCTraj* ctraj, const ManaPSTraj* straj) const;

    /*
     *  \brief compute the heuristic of an input trajectory : distance to the goal
     *  \param curS : the current state of the robot
     *  \param finalS : the final state of a trajectory
     *  \param finalS_Obs : state of a trajectory up to which the collision with obstacles is checked
     *  \param targetPos : the target location
     *  \param straj : the input state trajectory
     */
    float computeHG(const ManaState& curS, const ManaStateT& finalS, const ManaStateT& finalS_Obs, const PositionXY& targetPos, const ManaPSTraj* straj) const;

    /*
     *  \brief compute the heuristic of an input trajectory : distance to obstacles
     *  \param wp : waypoint (state-time) of a trajectory
     *  \param coefVel : velocity coeficient used to prevent high velocities when the robot is close to obstacles
     */
    float computeHO_pos(const ManaStateT& wp, double coefVel) const;

    /*
     *  \brief compute the heuristic of an input trajectory : distance to obstacles
     *  \param curS : the current state of the robot
     *  \param finalS : the final state of a trajectory
     *  \param straj : the input state trajectory
     */
    float computeHO(const ManaState& curS, const ManaStateT& finalS_Obs, const ManaPSTraj* straj) const;

    /*
     *  \brief compute the heuristic of an input trajectory : final heading
     *  \param finalS : the final state of a trajectory
     *  \param targetPos : the target location
     */
    float computeHH(const ManaStateT& finalS, const PositionXY& targetPos) const;

    /*
     *  \brief compute an heuristic that penalize trajectory with no motion
     *  \param curS : the current state of the robot
     *  \param finalS : the final state of a trajectory
     */
    float computeH_static(const ManaState& curS, const ManaStateT& finalS) const;


    /************** Trajectory selection **************/

  protected : 
    /*
     *  \brief compute heuristic of the trajectory corresponding to index i
     *  \param i : the input index
     *  \return heuristic corresponding to the i-th trajectory
     */
    virtual float computeH(int i) const;

  public : 
    /**
     *  \brief compute the best trajectory to follow
     *  \return 0 if the process succeedm -1 else
     *
     *  If the input dem is not null, update the dem
     *  Do not use this option if you want asynchronous trajectory selesction and DEM update
     */
    int computeBestTraj(const ManaState& curS, const PositionXY& tgt, const DEM* dem=NULL);

    /****************** I/O functions *****************/

  public : 
    /**
     *  \brief return a copy of the best control trajectory found
     *  \return a copy of the best control trajectory found, or NULL if the process failed
     */
    ManaPCTraj* getBestCTraj() const;

    /**
     *  \brief return a copy of the best state trajectory found
     *  \return a copy of the best state trajectory found, or NULL if the process failed
     */
    ManaPSTraj* getBestSTraj() const;

    /**
     *  \brief return a dated copy of the best control trajectory found
     *  \param pctraj : the resulting copy of the best control trajectory found, or NULL if the process failed
     *  \param date : the date at which the trajectory has been registered
     */
    void getBestCTraj(ManaPCTraj*& pctraj, struct timeval& date) const;

    /**
     *  \brief return a dated copy of the best state trajectory found
     *  \param pstraj : the resulting copy of the best state trajectory found, or NULL if the process failed
     *  \param date : the date at which the trajectory has been registered
     */
    void getBestSTraj(ManaPSTraj*& pstraj, struct timeval& date) const;

    /**
     *  \brief return the best control to apply
     *  \return the best control to apply
     */
    ManaControl getCtrl();


    /**************** Display functions ***************/

    /**
     *  \brief display all the candidate trajectories in a gnuplot viewer
     */
    void displayGP_CandTraj() const;
  };

};

#endif // MANATRAJSEL_HPP
