#ifndef TRAJSEL_HPP
#define TRAJSEL_HPP

#include "environment/inc/DEM.hpp"
#include "environment/inc/SlopeMap.hpp"
#include "environment/inc/DistObsMap.hpp"

#include "motion/inc/PCTraj.hpp"
#include "motion/inc/PSTraj.hpp"

#include "common/inc/Timer.hpp"

#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>

#include <vector>

// #define TS_SLOPEMAP_DIST 0.3           // Maximal distance considered between two cells of the slope map to compute the slope
// #define TS_MAX_SLOPE 0.3*M_PI          // Maximal slope that the robot can climb
// #define TS_SAFE_DIST 0.1               // Distance that the robot must keep between its center of rotation and the obstacle to be safe

// #define TS_TIME_TRAJ 9.0
// #define TS_TIME_TRAJ_OBS_FREE 5.0

// #define SAMPLING_CAND_TRAJ 15
// #define DIST_OBS_MAX 2.0

namespace mpc
{

                                    /**************************************************/
                                    /****************** Class TrajSel *****************/
                                    /**************************************************/

  /** \class TrajSel
   * \brief reactive navigation method selecting the best trajectory to follow from an input DEM
   */
  class TrajSel
  {
  protected : 
    System* sys_;                             /**< Robotic system constraints */

    DEM* dem_;                                /**< Data Elevation Map on which the robot is navigating */
    VarSlopeMap* slopeMap_;                   /**< Slope map corresponding to the DEM */
    DistObsMap* distMap_;                     /**< Distance to obstacles map corresponding to the DEM */

    State* loc_;                              /**< Current location */
    PCTraj* bestPCTraj_;                      /**< Best control trajectory found */
    PSTraj* bestPSTraj_;                      /**< Best state trajectory found */
    PositionXY targetPos_;                    /**< The target position to reach */
    struct timeval dateLoc_;                  /**< Date of the last registered localization */
    struct timeval dateTraj_;                 /**< Date of the last registered trajectory */

    std::vector<PCTraj*> vecPCTraj_;          /**< vector containing the candidate control trajectories */
    std::vector<PSTraj*> vecPSTraj_;          /**< vector containing the candidate state trajectories */
    std::vector<float> vecH_;                 /**< vector containing heuristics corresponding to the candidate trajectories */

    Timer* timerTrajUsed_;                    /**< Timer used to compute the control */
    Timer* timerTrajCpted_;                   /**< Timer of the trajectory currently computed */

    boost::mutex mutexDEM_;                   /**< mutex used to prevent DEM access conflicts */
    boost::mutex mutexDistMap_;               /**< mutex used to prevent dist map access conflicts */
    boost::mutex mutexBestTraj_;              /**< mutex used to prevent best trajectories access conflicts */
    boost::mutex mutexTimers_;                /**< mutex used to prevent timers access conflicts */

  public : 
    static double TS_SLOPEMAP_DIST;           /**< Maximal distance considered between two cells of the slope map to compute the slope */
    static double TS_MAX_SLOPE;               /**< Maximal slope that the robot can climb */
    static double TS_MAX_VAR_SLOPE;           /**< Maximal slope derivative that the robot can climb */
    static double TS_SAFE_DIST;               /**< Distance that the robot must keep between its center of rotation and the obstacle to be safe */

    static double TS_TIME_TRAJ;               /**< Duration of the candidate trajectories */
    static double TS_TIME_TRAJ_OBS_FREE;      /**< Time of the candidate trajectories until which collision check is performed */

    static int TS_SAMPLING_CAND_TRAJ;         /**< Sampling of the candidate trajectories */
    static double TS_DIST_OBS_MAX;            /**< Maximal distance from obstacles until which a repulsive potential is applied */
    static double TS_DIST_OBS_DECREASED_SPD;  /**< Maximal distance from obstacles until which the robot's speed is decreased */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     *  \param sys : the robotic system constraints
     */
    TrajSel(System* sys = NULL);
    
  public : 
    /**
     *  \brief destructor
     */
    virtual ~TrajSel();


    /******************** Accessors *******************/

  protected : 
    /*
     *  \brief return a pointer to the robotic system constraints
     *  \return a pointer to the robotic system constraints
     */
    virtual System* sys() const = 0;

    /*
     *  \brief return a pointer to the localization of the robot
     *  \return a pointer to the localization of the robot
     */
    virtual State* loc() const = 0;

    /*
     *  \brief return a pointer to the best found control trajectory
     *  \return a pointer to the best found control trajectory
     */
    virtual PCTraj* bestCTraj() const = 0;

    /*
     *  \brief return a pointer to the best found control trajectory
     *  \return a pointer to the best found control trajectory
     */
    virtual PSTraj* bestSTraj() const = 0;

    /*
     *  \brief return the number of candidate trajectories
     *  \return the number of candidate trajectories
     */
    int nbCandTraj() const;


    /**************** Vector management ***************/

  private : 
    /*
     *  \brief clear the control trajectory vector
     */
    void clearVecCTraj();

    /*
     *  \brief clear the state trajectory vector
     */
    void clearVecSTraj();

    /*
     *  \brief clear the heuristic vector
     */
    void clearVecH();

  protected : 
    /*
     *  \brief add a new candidate control trajectory
     *  \param ctraj : the new candidate control trajectory
     */
    void addCandCTraj(PCTraj* ctraj);

    /*
     *  \brief return a pointer the the i-th candidate control trajectory
     *  \return a pointer the the i-th candidate control trajectory
     */
    virtual const PCTraj* getCandCTraj(int i) const;

    /*
     *  \brief add a new candidate state trajectory
     *  \param ctraj : the new candidate state trajectory
     */
    void addCandSTraj(PSTraj* straj);

    /*
     *  \brief return a pointer the the i-th candidate state trajectory
     *  \return a pointer the the i-th candidate state trajectory
     */
    virtual const PSTraj* getCandSTraj(int i) const;

    /*
     *  \brief add a heuristic corresponding to a candidate trajectory
     *  \param the new heuristic
     */
    void addCandH(float H);


    /**************** Timers management ***************/

  protected : 
    /*
     *  \brief switch the timers
     */
    void switchTimers();

    /*
     *  \brief start the timer of the currently computed trajectory
     */
    void startTimerTrajCpted();

    /*
     *  \brief stop the timer of the currently computed trajectory (when the computation failed)
     */
    void stopTimerTrajCpted();


    /************* Trajectory computation *************/

  protected : 
    /*
     *  \brief compute the set of control trajectories
     *  \param distToGoal : distance from the current position of the robot to the goal location
     *  \param distToObs : distance from the current position of the robot to the closest obstacle
     *  \return the number of computed control trajectories
     */
    virtual int computeCTrajSet(double distToGoal, double distToObs) = 0;

    /*
     *  \brief compute the set of state trajectories
     *  \return the number of computed state trajectories
     */
    virtual int computeSTrajSet() = 0;


    /************** Trajectory selection **************/

  protected : 
    /*
     *  \brief reinit the trajectory selection process
     */
    void reinitTrajSel();

    /*
     *  \brief compute heuristics for every candidate trajectory
     */
    void computeH_trajSet();

    /*
     *  \brief compute heuristic of the trajectory corresponding to index i
     *  \param i : the input index
     *  \return heuristic corresponding to the i-th trajectory
     */
    virtual float computeH(int i) const = 0;

    /*
     *  \brief return index corresponding to the best computed trajectory
     *  \return index corresponding to the best computed trajectory
     */
    int selBestHindex() const;

    /*
     *  \brief stop the trajectory selection process
     */
    void stopTrajSel();

    /*
     *  \brief compute the best trajectory to follow
     *  \return 0 if the process succeed 
     *  \return -1 if an error occured
     *  \return -2 if the target position is unreachable
     *
     *  The DEM, the robot state, and the target location must have been initialized
     */
    virtual int computeBestTrajectory();

    /****************** I/O functions *****************/

  protected : 
    /*
     *  \brief update the localization of the robot
     *  \param loc : the current localization
     */
    void updateLoc(const State& loc);

    /*
     *  \brief update the target position of the robot
     *  \param tgt : the input target position
     */
    void updateTgt(const PositionXY& tgt);

//     /*
//      *  \brief reinit the DEM update process
//      */
//     void reinitDEMupdate();

  public : 
    /**
     *  \brief update the Data Elevation Map
     *  \param dem : the input DEM
     *  \return 0 if the process succeed, -1 else
     */
    int updateDEM(const DEM* dem);

    /**
     *  \brief update the DistObsMap
     *  \param dem : the input DistObsMap
     *  \return 0 if the process succeed, -1 else
     */
    int updateDistMap(const DistObsMap* dem);


    /************ Display /Debug functions ************/

    /*
     *  \brief return a pointer to the vector containing the candidate control trajectories
     *  \return a pointer to the vector containing the candidate control trajectories
     */
    const std::vector<PCTraj*>* getVecPCTraj() const { return &vecPCTraj_; }

    /*
     *  \brief return a pointer to the vector containing the candidate state trajectories
     *  \return a pointer to the vector containing the candidate state trajectories
     */
    const std::vector<PSTraj*>* getVecPSTraj() const { return &vecPSTraj_; }
  };

};

#endif // TRAJSEL_HPP
