#include "trajSel/inc/TrajSel.hpp"
using namespace mpc;

#include <cmath>
#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /****************** Class TrajSel *****************/
                                    /**************************************************/

double TrajSel::TS_SLOPEMAP_DIST = 0.3;           /**< Maximal distance considered between two cells of the slope map to compute the slope */
double TrajSel::TS_MAX_SLOPE = 0.35*M_PI;         /**< Maximal slope that the robot can climb */
double TrajSel::TS_MAX_VAR_SLOPE = 2.0*M_PI;      /**< Maximal slope derivative that the robot can climb */
double TrajSel::TS_SAFE_DIST = 0.1;               /**< Distance that the robot must keep between its center of rotation and the obstacle to be safe */

double TrajSel::TS_TIME_TRAJ = 7.0;               /**< Duration of the candidate trajectories */
double TrajSel::TS_TIME_TRAJ_OBS_FREE = 5.0;      /**< Time of the candidate trajectories until which collision check is performed */

int TrajSel::TS_SAMPLING_CAND_TRAJ = 15;          /**< Sampling of the candidate trajectories */
double TrajSel::TS_DIST_OBS_MAX = 2.0;            /**< Maximal distance from obstacles until which a repulsive potential is applied */
double TrajSel::TS_DIST_OBS_DECREASED_SPD = 1.0;  /**< Maximal distance from obstacles until which the robot's speed is decreased */

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 *  \param sys : the robotic system constraints
 */
TrajSel::TrajSel(System* sys)
{
  sys_ = sys;
  
  dem_ = NULL;
  slopeMap_ = NULL;
  distMap_ = NULL;
  
  loc_ = NULL;
  bestPCTraj_ = NULL;
  bestPSTraj_ = NULL;

  timerTrajUsed_ = new Timer;
  timerTrajCpted_ = new Timer;

  dateTraj_.tv_sec = -1;
  dateTraj_.tv_usec = -1;
  dateLoc_.tv_sec = -1;
  dateLoc_.tv_usec = -1;
}

/**
 *  \brief destructor
 */
TrajSel::~TrajSel()
{
  this->clearVecCTraj();
  this->clearVecSTraj();
  this->clearVecH();
}


/******************** Accessors *******************/

/*
 *  \brief return the number of candidate trajectories
 *  \return the number of candidate trajectories
 */
int TrajSel::nbCandTraj() const
{
  return vecPCTraj_.size();
}


/**************** Vector management ***************/

/*
 *  \brief clear the control trajectory vector
 */
void TrajSel::clearVecCTraj()
{
  vector<PCTraj*>::iterator it = vecPCTraj_.begin();
  while(it != vecPCTraj_.end())
    {
      // delete the current traj
      if(*it != NULL) delete *it;
      it++;
    }
  vecPCTraj_.clear();
}

/*
 *  \brief clear the state trajectory vector
 */
void TrajSel::clearVecSTraj()
{
  vector<PSTraj*>::iterator it = vecPSTraj_.begin();
  while(it != vecPSTraj_.end())
    {
      // delete the current traj
      if(*it != NULL) delete *it;
      it++;
    }
  vecPSTraj_.clear();
}

/*
 *  \brief clear the heuristic vector
 */
void TrajSel::clearVecH()
{
  vecH_.clear();
}

/*
 *  \brief add a new candidate control trajectory
 *  \param ctraj : the new candidate control trajectory
 */
void TrajSel::addCandCTraj(PCTraj* ctraj)
{
  vecPCTraj_.push_back(ctraj);  
}

/*
 *  \brief return a pointer the the i-th candidate control trajectory
 *  \return a pointer the the i-th candidate control trajectory
 */
const PCTraj* TrajSel::getCandCTraj(int i) const
{
  return vecPCTraj_.at(i);
}

/*
 *  \brief add a new candidate state trajectory
 *  \param ctraj : the new candidate state trajectory
 */
void TrajSel::addCandSTraj(PSTraj* straj)
{
  vecPSTraj_.push_back(straj);  
}

/*
 *  \brief return a pointer the the i-th candidate state trajectory
 *  \return a pointer the the i-th candidate state trajectory
 */
const PSTraj* TrajSel::getCandSTraj(int i) const
{
  return vecPSTraj_.at(i);
}

/*
 *  \brief add a heuristic corresponding to a candidate trajectory
 *  \param the new heuristic
 */
void TrajSel::addCandH(float H)
{
  vecH_.push_back(H);
}


/**************** Timers management ***************/

/*
 *  \brief switch the timers
 */
void TrajSel::switchTimers()
{
  boost::mutex::scoped_lock lock(mutexTimers_);

  timerTrajUsed_->stop();
  timerTrajUsed_->reinit();

  Timer* oldTimer = timerTrajUsed_;
  timerTrajUsed_ = timerTrajCpted_;
  timerTrajCpted_ = oldTimer;
}

/*
 *  \brief start the timer of the currently computed trajectory
 */
void TrajSel::startTimerTrajCpted()
{
  boost::mutex::scoped_lock lock(mutexTimers_);

  timerTrajCpted_->start();
}

/*
 *  \brief stop the timer of the currently computed trajectory (when the computation failed)
 */
void TrajSel::stopTimerTrajCpted()
{
  boost::mutex::scoped_lock lock(mutexTimers_);

  timerTrajCpted_->stop();
  timerTrajCpted_->reinit();
}


/************** Trajectory selection **************/

/*
 *  \brief reinit the trajectory selection process
 */
void TrajSel::reinitTrajSel()
{
  this->clearVecCTraj();
  this->clearVecSTraj();
  this->clearVecH();
}

/*
 *  \brief compute heuristics for every candidate trajectory
 */
void TrajSel::computeH_trajSet()
{
  // block the distmap to prevent changing it during the heuristic evaluation
  boost::mutex::scoped_lock lock(mutexDistMap_);

  int N = this->nbCandTraj();
  for(int i=0; i<N; i++)
    {
      float H = this->computeH(i);
      vecH_.push_back(H);
    }
}

/*
 *  \brief return index corresponding to the best computed trajectory
 *  \return index corresponding to the best computed trajectory
 */
int TrajSel::selBestHindex() const
{
  int bestInd = -1;
  float bestH = 0;

  int N = this->nbCandTraj();
  for(int i=0; i<N; i++)
    {
      float H = vecH_[i];
      if(bestInd == -1 || H < bestH)
	{
	  bestInd = i;
	  bestH = H;
	}
    }
  return bestInd;
}

/*
 *  \brief stop the trajectory selection process
 */
void TrajSel::stopTrajSel()
{
  boost::mutex::scoped_lock lock(mutexBestTraj_);
	  
  // set a null control trajectory
  if(bestPCTraj_ != NULL) delete bestPCTraj_;
  if(bestPSTraj_ != NULL) delete bestPSTraj_;
  bestPCTraj_ = NULL;
  bestPSTraj_ = NULL;
  dateTraj_.tv_sec = -1;
  dateTraj_.tv_usec = -1;
  
  // stop the timer
  this->stopTimerTrajCpted();
}

/*
 *  \brief compute the best trajectory to follow
 *  \return 0 if the process succeed
 *  \return -1 an error occured (the loc or the DEM is undefined)
 *  \return -2 if the target position is unreachable (when the robot is close to it)
 *  \return -3 if the robot is in collision with obstacles
 *
 *  The DEM, the robot state, and the target location must have been initialized
 */
int TrajSel::computeBestTrajectory()
{
  // check if the DEM and the distance map have been initialized
  if(distMap_ == NULL) return -1;

  // check if the state of the robot has been initialized
  if(loc_ == NULL) return -1;

  // check if the robot is in collision with obstacles
  PositionXYTh curPos = loc_->posXYTh();

  obs2D_cell_t cellSt_curPos; 
  float distObs_curPos; 
  PositionXY nearObs_curPos;
  distMap_->getInfo(curPos.x(), curPos.y(), cellSt_curPos, distObs_curPos, nearObs_curPos);

//   if(cellSt_curPos == CELL_OBS || distObs_curPos < TrajSel::TS_SAFE_DIST)
  if(cellSt_curPos == CELL_OBS)
    {
      cerr << "TrajSel::computeBestTrajectory : The robot is too close to obstacles. Stopping the TrajSel process (distance = " << distObs_curPos << " < TrajSel::TS_SAFE_DIST = " << TrajSel::TS_SAFE_DIST << ")" << endl;
      this->stopTrajSel();
      return -3;
    }

  // check if the target position is obstructed (when the robot is close to it)
  double dxCT = targetPos_.x() - curPos.x();
  double dyCT = targetPos_.y() - curPos.y();
  double distCT = dxCT*dxCT + dyCT*dyCT;
//   cout << "curPos : "; curPos.display();  cout << "  targetPos : "; targetPos_.display(); cout << "  distToGoal : " << sqrt(distCT) << "  distToObs : " << distObs_curPos << endl;

  if(distCT < TrajSel::TS_DIST_OBS_MAX)
    {
      // check if the target position is obstructed
      obs2D_cell_t cellSt; 
      float distObs; 
      PositionXY nearObs;
      distMap_->getInfo(targetPos_.x(), targetPos_.y(), cellSt, distObs, nearObs);

//       cout << "targetPos distObs : " << distObs << endl;

      if(cellSt == CELL_OBS || distObs < 0.25*TrajSel::TS_DIST_OBS_MAX) 
	{
	  cerr << "TrajSel::computeBestTrajectory : The goal position is obstructed. Stopping the TrajSel process" << endl;
	  this->stopTrajSel();
	  return -2;
	}
    }

  // reinit the trajectory selection process
  this->reinitTrajSel();

  // compute the set of control trajectories
  int nbCTraj = this->computeCTrajSet(sqrt(distCT), distObs_curPos);

  // compute the set of state trajectories
  int nbSTraj = this->computeSTrajSet();

  // compute an heurisctic for each candidate trajectory
  this->computeH_trajSet();
  
  // search for the best candidate trajectory
  int ind = this->selBestHindex();

  // save the best control and state trajectories
  if(ind == -1)
    {
      this->stopTrajSel();
      return -1;
    }
  else 
    {
      boost::mutex::scoped_lock lock(mutexBestTraj_);
      
      if(bestPCTraj_ != NULL) delete bestPCTraj_;
      if(bestPSTraj_ != NULL) delete bestPSTraj_;
      
      bestPCTraj_ = vecPCTraj_.at(ind)->copy();
      bestPSTraj_ = vecPSTraj_.at(ind)->copy();
      dateTraj_.tv_sec = dateLoc_.tv_sec;
      dateTraj_.tv_usec = dateLoc_.tv_usec;
      
      this->switchTimers();
      return 0;
    }
}


/****************** I/O functions *****************/

/*
 *  \brief update the localization of the robot
 *  \param loc : the current localization
 */
void TrajSel::updateLoc(const State& loc)
{
  if(loc_ != NULL) delete loc_;
  loc_ = loc.copy();

  Timer::absoluteTime(dateLoc_);
}

/*
 *  \brief update the target position of the robot
 *  \param tgt : the input target position
 */
void TrajSel::updateTgt(const PositionXY& tgt)
{
  this->targetPos_ = tgt;
}

/**
 *  \brief update the Data Elevation Map
 *  \param dem : the input DEM
 *  \return 0 if the process succeed, -1 else
 */
int TrajSel::updateDEM(const DEM* dem)
{
//   this->reinitDEMupdate();

  if(dem == NULL || dem->nbCols() == 0 || dem->nbLines() == 0 || dem->xScale() == 0 || dem->yScale() == 0)
    {
      {
	boost::mutex::scoped_lock lock(mutexDEM_);
	if(dem_ != NULL) delete dem_;
	this->dem_ = NULL;
      }

      if(slopeMap_ != NULL) delete slopeMap_;
      this->slopeMap_ = NULL;

      {
	boost::mutex::scoped_lock lock(mutexDistMap_);
	if(distMap_ != NULL) delete distMap_;
	this->distMap_ = NULL;      
      }
      return -1;
    }
  else
    {
      {
	// copy the input DEM
	boost::mutex::scoped_lock lock(mutexDEM_);
	if(dem_ != NULL) delete dem_;
	dem_ = dem->copy();
	
	// compute a slope map from the input DEM
	if(slopeMap_ != NULL) delete slopeMap_;
	slopeMap_ = new VarSlopeMap(dem_,TS_SLOPEMAP_DIST);
      }

      {
	boost::mutex::scoped_lock lock(mutexDistMap_);
	// compute a distance to obstacles map from the slopemap
	if(distMap_ != NULL) delete distMap_;
	distMap_ = new DistObsMap(slopeMap_,TS_MAX_SLOPE,TS_MAX_VAR_SLOPE,TS_SAFE_DIST);
      }
      return 0;
    }
}

/**
 *  \brief update the DistObsMap
 *  \param dem : the input DistObsMap
 *  \return 0 if the process succeed, -1 else
 */
int TrajSel::updateDistMap(const DistObsMap* distmap)
{
  boost::mutex::scoped_lock lock(mutexDistMap_);
  if(distMap_ != NULL) delete distMap_;

  if(distmap == NULL)
    {
      this->distMap_ = NULL;      
      return -1;
    }
  else
    {
//       distMap_ = new DistObsMap(slopeMap_,TS_MAX_SLOPE,TS_MAX_VAR_SLOPE,TS_SAFE_DIST);
      distMap_ = new DistObsMap(*distmap);

//       distMap_->saveIntoPPM("DistMap_updated.ppm", -15, 22);

      return 0;
    }
}
