#include "MPC_HugrIfc.hpp"
#include "common/inc/Position.hpp"

#include "common/inc/Utils.hpp"

#include <time.h>
#include <sys/time.h>

#include "math/inc/gnuplot_i.hpp"

#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>
namespace po = boost::program_options;
namespace fs = boost::filesystem;

#define THRESHOLD 200
#define INCREASING 20


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////// Constructors & Init functions //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// default constructor (only used by subclasses)
MPC_HugrIfc::MPC_HugrIfc(){}

// init constructor (creation of the window)
MPC_HugrIfc::MPC_HugrIfc(FXApp* a):FXMainWindow(a,"Hugr Wheelchair and Cycab Control",NULL,NULL,DECOR_ALL,0,0,950,680)
{
  /***************************************************************************************************************
   ********************************************** initialisation *************************************************
   ***************************************************************************************************************/

  TrajGen::setDBG(0);

  this->resize(1160,730);

  // Orthogonal / Perspective view
  orthoView = true;

  // the map used
  map = NULL;
  wtI = 0;
  htI = 0;

  // sampling of the static environment
  samplingStEnvFilename = FXString("../map/hall.pgm");
  sampleX = 400;
  sampleY = 400;
  stEnvLoaded = false;

  minFieldX = -30;
  maxFieldX =  44;
  minFieldY =   0;
  maxFieldY =  40;

  rb_       = NULL;
  typeSys_  = ND_SYS;

//   env_ = NULL;

  initColor = FXRGB(255,0,0);
  goalColor = FXRGB(0,0,255);

  // variables used to set the initial configuration of the robot
  initX = 2.0;
  initY = 20.0;
  initTH = 0.0;

  initX_target.connect(initX);
  initY_target.connect(initY);
  initTH_target.connect(initTH);

  // wheelchair state parameters
  WC_x = 0.0;
  WC_y = 0.0;
  WC_th = 0.0;
  WC_v = 0.0;
  WC_av = 0.0;

  WC_x_target.connect(WC_x);
  WC_y_target.connect(WC_y);
  WC_th_target.connect(WC_th);
  WC_v_target.connect(WC_v);
  WC_av_target.connect(WC_av);

  WC_namePos = FXString("wuweiPose");
  WC_nameCmd = FXString("wuweiCmd");
  WC_namePos_target.connect(WC_namePos);
  WC_nameCmd_target.connect(WC_nameCmd);

  // wheelchair system configuration
  WC_wheel_radius = 0.0;
  WC_wheel_base = 0.0;
  WC_max_speed = 0.0;
  WC_max_acc = 0.0;
  WC_max_rotation_speed = 0.0;
  WC_max_rotation_acc = 0.0;
  WC_braking_dist = 0.0;
  WC_length = 0.0;
  WC_width = 0.0;

  WC_wheel_radius_target.connect(WC_wheel_radius);
  WC_wheel_base_target.connect(WC_wheel_base);
  WC_max_speed_target.connect(WC_max_speed);
  WC_max_acc_target.connect(WC_max_acc);
  WC_max_rotation_speed_target.connect(WC_max_rotation_speed);
  WC_max_rotation_acc_target.connect(WC_max_rotation_acc);
  WC_braking_dist_target.connect(WC_braking_dist);
  WC_length_target.connect(WC_length);
  WC_width_target.connect(WC_width);

  // wheelchair/cycab simple controller parameters
  Ctrl_v = 0.0;
  Ctrl_avph = 0.0;
  Coef_Ctrl_v = 0.3;
  Coef_Ctrl_avph = 0.25;

  Ctrl_v_target.connect(Ctrl_v);
  Ctrl_avph_target.connect(Ctrl_avph);
  Coef_Ctrl_v_target.connect(Coef_Ctrl_v);
  Coef_Ctrl_avph_target.connect(Coef_Ctrl_avph);

  // car-like state parameters
  CC_x = 0.0;
  CC_y = 0.0;
  CC_th = 0.0;
  CC_v = 0.0;
  CC_ph = 0.0;

  CC_x_target.connect(CC_x);
  CC_y_target.connect(CC_y);
  CC_th_target.connect(CC_th);
  CC_v_target.connect(CC_v);
  CC_ph_target.connect(CC_ph);

  CC_namePos = FXString("cycabPose");
  CC_nameCmd = FXString("cycabCmd");
  CC_namePos_target.connect(CC_namePos);
  CC_nameCmd_target.connect(CC_nameCmd);

  // car-like system configuration
  CC_length = 0.0;
  CC_max_speed = 0.0;
  CC_max_acc = 0.0;
  CC_max_steer = 0.0;
  CC_max_stspd = 0.0;

  CC_length_target.connect(CC_length);
  CC_max_speed_target.connect(CC_max_speed);
  CC_max_acc_target.connect(CC_max_acc);
  CC_max_steer_target.connect(CC_max_steer);
  CC_max_stspd_target.connect(CC_max_stspd);

//   // car-like simple controller parameters
//   CC_Ctrl_v = 0.0;
//   CC_Ctrl_ph = 0.0;

//   CC_Ctrl_v_target.connect(CC_Ctrl_v);
//   CC_Ctrl_ph_target.connect(CC_Ctrl_ph);

  // goal configuration of the planner
  planner_goalX  = 0.0;
  planner_goalY  = 0.0;
  planner_goalTH = 0.0;

  planner_goalX_target.connect(planner_goalX);
  planner_goalY_target.connect(planner_goalY);
  planner_goalTH_target.connect(planner_goalTH);

  fixed_T = false;                  // true if the final time must be fixed
  planner_goalT = 10.0;             // value of the fixed final time
  planner_DT = 0.3;                 // value of the delta time between each state-time of the computed trajectory

  planner_goalT_target.connect(planner_goalT);
  planner_DT_target.connect(planner_DT);
  
  planner_CoefSpd = 0.35;
  planner_CoefAcc = 0.35;

  planner_CoefSpd_target.connect(planner_CoefSpd);
  planner_CoefAcc_target.connect(planner_CoefAcc);

  // controller type
  // 0 -> Simple controller
  // 1 -> Trajectory follower
  // 2 -> Teddy (Trajectory Deformation)
  ctrlerType = 2;
  ctrlerType_target.connect(ctrlerType);

  // controller debug level
  ctrlerDebugLevel = 0;
  ctrlerDebugLevel_target.connect(ctrlerDebugLevel);

  // deformer components
  k_int = 1.0;                    // internal forces coefficient
  k_ext = 0.5;                   // external forces coefficient (dynamic obstacles)
  k_stext = 0.2;                 // external forces coefficient (static obstacles)
  dyn_med = 3.0;                  // maximal effective distance (dynamic obstacles)
  st_med = 2.0;                   // maximal effective distance (static obstacles)
  tmp_med = 6.0;                  // maximal temporal effective distance

  k_int_target.connect(k_int);
  k_ext_target.connect(k_ext);
  k_stext_target.connect(k_stext);
  dyn_med_target.connect(dyn_med);
  st_med_target.connect(st_med);
  tmp_med_target.connect(tmp_med);

  // recording of the trajectory during the execution of the motion
  recordTraj = false;

  locActive = false;

  vecTrajSt = NULL;             // vector of state-times of the trajectory that a robot has to follow
  vecLoc = NULL;                // vector of recorded localization state-times
  vecDev = NULL;                // vector of state deviation recorded during a trajectoy following process
  vecCtrlT = NULL;              // vector of state control-times recorded during a trajectoy following process
  
  // planner type
  // SIMPLE_TP -> Simple trajectory planner
  // PVD_TP -> Path-velocity decomposition trajectory planner
  // ROADMAP_TP -> Trajectory planner on a roadmap
  plannerType = ROADMAP_TP;
  plannerType_target.connect(plannerType);

  // time intervals between localization and control requests
  dtLoc = 0.3;
  dtCtrl = 0.1;

  dtLoc_target.connect(dtLoc);
  dtCtrl_target.connect(dtCtrl);

  // hugr localization and mapping variables
  nameDynObs = FXString("DynamicObstacles");
  nameDynObs_target.connect(nameDynObs);

  // set the viewer parameters
  dynamicView = true;

  // name of the file containing a static environment to load
  stEnvFilename = FXString("../map/robotic_hall.map2");
  stEnvFilename_target.connect(stEnvFilename);

  // roadmap parameters
  autoBuildRoadmap = true;
  dispRoadmap = true;

  optRoadmap = 2;                               // roadmap option : 0 no roadmap used, 1 autobuild roadmap, 2 load roadmap from file
  RMfilename = FXString("../map/RM_robotic_hall.rdm");  // roadmap filename
  optRoadmap_target.connect(optRoadmap);
  RMfilename_target.connect(RMfilename);

  // perception process options
  distFilter = 0.5;
  distFilter_target.connect(distFilter);
  securDist = 0.5;
  securDist_target.connect(securDist);


  // trajectory deformation visualization
  dispFExt = false;                 // true if the external forces from dynamic obstacles must be displayed (during deformation process)
  dispStFExt = false;               // true if the external forces from static obstacles must be displayed (during deformation process)
  dispNearDynObs = false;           // true if the position of the nearest dynamic obstacles must be displayed (during deformation process)
  dispNearStObs = false;            // true if the position of the nearest static obstacles must be displayed (during deformation process)


  /***************************************************************************************************************
   ********************************** Initialization from configuration files ************************************
   ***************************************************************************************************************/

  // init the gui main parameters
  this->initHugrIfcFromFile("../cfg/MPC_HugrIfc.cfg");

  
  /***************************************************************************************************************
   ************************************************ main frame ***************************************************
   ***************************************************************************************************************/

  contents=new FXHorizontalFrame(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);
//   contents->setHeight(1000);
//   contents->setWidth();
//   contents->resize(1700,1050);
  
  /***************************************************************************************************************
   *********************************************** viewer frame **************************************************
   ***************************************************************************************************************/

  // LEFT pane to contain the canvas
  imageFrame=new FXVerticalFrame(contents,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  
  // Label above the canvas
  new FXLabel(imageFrame,"Viewer",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  
  // Horizontal divider line
  new FXHorizontalSeparator(imageFrame,SEPARATOR_GROOVE|LAYOUT_FILL_X);

//   // Tab book with switchable panels
//   FXTabBook* panelsViewer=new FXTabBook(imageFrame,NULL,0,LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  /*********************************************************
   ********************* 2D Scanner Map ********************
   *********************************************************/

//   new FXTabItem(panelsViewer,"2D Scanner Map");

//   FXVerticalFrame *tabScanMap=new FXVerticalFrame(panelsViewer,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  // Make image widget
  //imageview=new FXImageView(imageFrame,NULL,this,ID_IMAGE,LAYOUT_FILL_X|LAYOUT_FILL_Y);

  // Horizontal divider line
  new FXHorizontalSeparator(imageFrame,SEPARATOR_GROOVE|LAYOUT_FILL_X);

  /*********************************************************
   ******************** 3D OpenGL Viewer *******************
   *********************************************************/

//   new FXTabItem(panelsViewer,"3D OpenGL Viewer");

//   FXVerticalFrame *tabGLViewer=new FXVerticalFrame(panelsViewer,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  // A Visual to drag OpenGL
  glvisual=new FXGLVisual(getApp(),VISUAL_DOUBLEBUFFER|VISUAL_STEREO);
  //glvisual=new FXGLVisual(tabGLViewer,VISUAL_DOUBLEBUFFER|VISUAL_STEREO);

  // Drawing glcanvas
  glcanvas=new FXGLDynCanvas(imageFrame,glvisual,this,(FXSelector)ID_CANVAS,(FXuint)(LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT));
  glcanvas->setMainWin(this,ID_REPAINT);
  //glcanvas=new FXGLViewer(glpanel,glvisual,this,ID_GLVIEWER,LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT);

  // Horizontal divider line
  new FXHorizontalSeparator(imageFrame,SEPARATOR_GROOVE|LAYOUT_FILL_X);

  msgBox = new FXLabel(imageFrame,"No map loaded. Click on \"Load map\" to load a new map.",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);

  /***************************************************************************************************************
   ********************************************** commands frame *************************************************
   ***************************************************************************************************************/

  // RIGHT pane for the buttons
  buttonFrame=new FXVerticalFrame(contents,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  
  // Label above the buttons
  new FXLabel(buttonFrame,"Commands",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  
  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  // Tab book with switchable panels
  FXTabBook* panels=new FXTabBook(buttonFrame,NULL,0,LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  /***************************************************************************************************************
   ************************************************ status tab ***************************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"System status");

  FXVerticalFrame *tabSysStatus=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  currentSystemLabel = new FXLabel(tabSysStatus,"System: No connected system",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  new FXHorizontalSeparator(tabSysStatus,SEPARATOR_RIDGE|LAYOUT_FILL_X);
  
  // Tab book with switchable panels
  FXTabBook* panelsSys=new FXTabBook(tabSysStatus,NULL,0,LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);

  /*********************************************************
   *********************** wheelchair **********************
   *********************************************************/

  new FXTabItem(panelsSys,"Wheelchair");

  FXVerticalFrame *tabWCSysStatus=new FXVerticalFrame(panelsSys,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabWCSysStatus,"System configuration:",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  FXMatrix *matSysConf=new FXMatrix(tabWCSysStatus,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matSysConf,"Wheel radius:");
  new FXTextField(matSysConf,6,&WC_wheel_radius_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysConf,"Wheel base:");
  new FXTextField(matSysConf,6,&WC_wheel_base_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysConf,"Max speed:");
  new FXTextField(matSysConf,6,&WC_max_speed_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysConf,"Max acc:");
  new FXTextField(matSysConf,6,&WC_max_acc_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysConf,"Max rotation speed:");
  new FXTextField(matSysConf,6,&WC_max_rotation_speed_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysConf,"Max rotation acc:");
  new FXTextField(matSysConf,6,&WC_max_rotation_acc_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysConf,"Braking distance:");
  new FXTextField(matSysConf,6,&WC_braking_dist_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysConf,"Length:");
  new FXTextField(matSysConf,6,&WC_length_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysConf,"Width:");
  new FXTextField(matSysConf,6,&WC_width_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  //sysConfLabel = new FXLabel(matSysConf,"No system loaded",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);

//   new FXLabel(tabWCSysStatus,"System state:",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
//   FXMatrix *matSysState=new FXMatrix(tabWCSysStatus,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matSysState,"x:");
//   new FXTextField(matSysState,6,&WC_x_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matSysState,"y:");
//   new FXTextField(matSysState,6,&WC_y_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matSysState,"th:");
//   new FXTextField(matSysState,6,&WC_th_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matSysState,"v:");
//   new FXTextField(matSysState,6,&WC_v_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matSysState,"av:");
//   new FXTextField(matSysState,6,&WC_av_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(tabWCSysStatus,"Hugr varibles:",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  FXMatrix *matWCHugrVar=new FXMatrix(tabWCSysStatus,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matWCHugrVar,"Pose:");
  new FXTextField(matWCHugrVar,20,&WC_namePos_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|LAYOUT_FILL_X|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matWCHugrVar,"Command:");
  new FXTextField(matWCHugrVar,20,&WC_nameCmd_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|LAYOUT_FILL_X|FRAME_SUNKEN|FRAME_THICK);

  // Button to take control of a wheelchair
  FXMatrix *matWCHugrVarSel=new FXMatrix(tabWCSysStatus,2,MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);
  new FXButton(matWCHugrVarSel,"Simulation Var",NULL,this,ID_WC_SIMUVAR,JUSTIFY_CENTER_X|FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matWCHugrVarSel,"Robot Var",NULL,this,ID_WC_ROBOTVAR,JUSTIFY_CENTER_X|FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(tabWCSysStatus,"Connect HUGR Wheelchair",NULL,this,ID_CONNECT_HUGR_WC,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXLabel(tabWCSysStatus,"Initial Config:",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  FXMatrix *matWCInitCfg=new FXMatrix(tabWCSysStatus,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matWCInitCfg,"x:");
  new FXTextField(matWCInitCfg,6,&initX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matWCInitCfg,"y:");
  new FXTextField(matWCInitCfg,6,&initY_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matWCInitCfg,"th:");
  new FXTextField(matWCInitCfg,6,&initTH_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  // Button to set the initial configuration of the wheelchair
  new FXButton(tabWCSysStatus,"Set Wheelchair Init Config",NULL,this,ID_INIT_WCCFG,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);


  /*********************************************************
   ************************* cycab *************************
   *********************************************************/

  new FXTabItem(panelsSys,"Cycab");

  FXVerticalFrame *tabCLSysStatus=new FXVerticalFrame(panelsSys,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabCLSysStatus,"System configuration:",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  FXMatrix *matCLSysConf=new FXMatrix(tabCLSysStatus,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matCLSysConf,"Length:");
  new FXTextField(matCLSysConf,6,&CC_length_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCLSysConf,"Max speed:");
  new FXTextField(matCLSysConf,6,&CC_max_speed_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCLSysConf,"Max acc:");
  new FXTextField(matCLSysConf,6,&CC_max_acc_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCLSysConf,"Max steer:");
  new FXTextField(matCLSysConf,6,&CC_max_steer_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCLSysConf,"Max steer speed:");
  new FXTextField(matCLSysConf,6,&CC_max_stspd_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXLabel(tabCLSysStatus,"System state:",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  FXMatrix *matCLSysState=new FXMatrix(tabCLSysStatus,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matCLSysState,"x:");
  new FXTextField(matCLSysState,6,&CC_x_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCLSysState,"y:");
  new FXTextField(matCLSysState,6,&CC_y_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCLSysState,"th:");
  new FXTextField(matCLSysState,6,&CC_th_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCLSysState,"v:");
  new FXTextField(matCLSysState,6,&CC_v_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCLSysState,"ph:");
  new FXTextField(matCLSysState,6,&CC_ph_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  // Buttons to take control of a cycab
  new FXButton(tabCLSysStatus,"Connect HUGR Cycab",NULL,this,ID_CONNECT_HUGR_CC,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  //sysStateLabel = new FXLabel(matSysState,"No system loaded",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);

  new FXLabel(tabCLSysStatus,"Initial Config:",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  FXMatrix *matCLInitCfg=new FXMatrix(tabCLSysStatus,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matCLInitCfg,"x:");
  new FXTextField(matCLInitCfg,6,&initX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCLInitCfg,"y:");
  new FXTextField(matCLInitCfg,6,&initY_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matCLInitCfg,"th:");
  new FXTextField(matCLInitCfg,6,&initTH_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  // Button to set the initial configuration of the wheelchair
  new FXButton(tabCLSysStatus,"Set Cycab &Init Config",NULL,this,ID_INIT_CCCFG,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);


  /***************************************************************************************************************
   ************************************************ planning tab *************************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Motion Planning");

  FXVerticalFrame *tabMotionPlanning=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  currentPlannerLabel = new FXLabel(tabMotionPlanning,"Planner: No connected planner",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  new FXHorizontalSeparator(tabMotionPlanning,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  /*********************************************************
   ******************** Simple Planner *********************
   *********************************************************/

  new FXRadioButton(tabMotionPlanning,"Simple planner",&plannerType_target,FXDataTarget::ID_OPTION+SIMPLE_TP,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXMatrix *matOptSimplePlanner=new FXMatrix(tabMotionPlanning,4,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  FXCheckButton *fixedTimeCB = new FXCheckButton(matOptSimplePlanner,"Fixed final time T:",this,ID_FIXED_TIME_CB);
  fixedTimeCB->setCheck(fixed_T,false);
  new FXTextField(matOptSimplePlanner,6,&planner_goalT_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matOptSimplePlanner,"DT:");
  new FXTextField(matOptSimplePlanner,6,&planner_DT_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabMotionPlanning,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  /*********************************************************
   ******************** RoadMap Planner ********************
   *********************************************************/

  new FXRadioButton(tabMotionPlanning,"RoadMap planner",&plannerType_target,FXDataTarget::ID_OPTION+ROADMAP_TP,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXMatrix *matOptRMPlanner=new FXMatrix(tabMotionPlanning,6,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matOptRMPlanner,"Coef Speed:");
  new FXTextField(matOptRMPlanner,6,&planner_CoefSpd_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matOptRMPlanner,"Coef Acc:");
  new FXTextField(matOptRMPlanner,6,&planner_CoefAcc_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matOptRMPlanner," DT:");
  new FXTextField(matOptRMPlanner,6,&planner_DT_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(tabMotionPlanning,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  /*********************************************************
   ******************* Planning Commands *******************
   *********************************************************/

  // Image view to select easily the target configuration
  FXVerticalFrame* imageFrame = new FXVerticalFrame(tabMotionPlanning,FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);
  imageview=new FXImageView(imageFrame,NULL,this,ID_IMAGE,LAYOUT_FILL_X|LAYOUT_FILL_Y);
  imageview->setHeight(200);
  imageview->setWidth(200);

//   //new FXLabel(tabMotionPlanning,"Goal config:",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
//   FXMatrix *matGoalPlannerCfg=new FXMatrix(tabMotionPlanning,7,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matGoalPlannerCfg,"x:");
//   new FXTextField(matGoalPlannerCfg,6,&planner_goalX_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalPlannerCfg,"y:");
//   new FXTextField(matGoalPlannerCfg,6,&planner_goalY_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matGoalPlannerCfg,"th:");
//   new FXTextField(matGoalPlannerCfg,6,&planner_goalTH_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   new FXButton(matGoalPlannerCfg,"Init Goal Pos",NULL,this,ID_INIT_PL_GOALPOS,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXHorizontalSeparator(tabMotionPlanning,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matPlannerCmd=new FXMatrix(tabMotionPlanning,2,MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);
  new FXButton(matPlannerCmd,"           Init New Planner           ",NULL,this,ID_INIT_PLANNER,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matPlannerCmd,"             Start Planner            ",NULL,this,ID_START_PLANNER,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  /***************************************************************************************************************
   ************************************************ control tab **************************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Motion Control");

  FXVerticalFrame *tabMotionControl=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  currentControllerLabel = new FXLabel(tabMotionControl,"Controller: No connected controller",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  new FXHorizontalSeparator(tabMotionControl,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matCtrlerType=new FXMatrix(tabMotionControl,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);  

  new FXRadioButton(matCtrlerType,"Simple controller",&ctrlerType_target,FXDataTarget::ID_OPTION+SIMPLE_CTRLER,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXHorizontalFrame *tabHzMotionControl=new FXHorizontalFrame(matCtrlerType,LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);

  FXVerticalFrame *tabVtMotionControl1=new FXVerticalFrame(tabHzMotionControl,LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);

  FXMatrix *matSysCtrl=new FXMatrix(tabVtMotionControl1,2,FRAME_THICK|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matSysCtrl,"vel (m/s):");
//   new FXTextField(matSysCtrl,6,&Ctrl_v_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
//   new FXLabel(matSysCtrl,"avel (rad/s):");
//   new FXTextField(matSysCtrl,6,&Ctrl_avph_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysCtrl,"coef vel:");
  new FXTextField(matSysCtrl,6,&Coef_Ctrl_v_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matSysCtrl,"coef avel:");
  new FXTextField(matSysCtrl,6,&Coef_Ctrl_avph_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   new FXButton(matCtrlerType,"Update Control",NULL,this,ID_UPDATE_CTRL,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  FXVerticalFrame *tabVtMotionControl2=new FXVerticalFrame(tabHzMotionControl,LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);

  FXMatrix *matSysCtrl2=new FXMatrix(tabVtMotionControl2,3,FRAME_THICK|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);
  new FXButton(matSysCtrl2,"&A",NULL,this,ID_MOVE_FWD_LEFT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matSysCtrl2,"&Z",NULL,this,ID_MOVE_FWD,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matSysCtrl2,"&E",NULL,this,ID_MOVE_FWD_RIGHT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matSysCtrl2,"&Q",NULL,this,ID_TURN_LEFT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matSysCtrl2,"&S",NULL,this,ID_STOP_MOTION,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matSysCtrl2,"&D",NULL,this,ID_TURN_RIGHT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matSysCtrl2,"&W",NULL,this,ID_MOVE_BWD_LEFT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matSysCtrl2,"&X",NULL,this,ID_MOVE_BWD,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matSysCtrl2,"&C",NULL,this,ID_MOVE_BWD_RIGHT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Horizontal divider line
  new FXHorizontalSeparator(matCtrlerType,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXRadioButton(matCtrlerType,"Trajectory follower",&ctrlerType_target,FXDataTarget::ID_OPTION+TRAJFOLLOW_CTRLER,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

//   FXMatrix *matDtCtrl=new FXMatrix(tabMotionControl,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
//   new FXLabel(matDtCtrl,"dt control (s):");
//   new FXTextField(matDtCtrl,6,&dtCtrl_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

//   new FXButton(tabMotionControl,"Init &New Controller",NULL,this,ID_INIT_CTRLER,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
//   new FXButton(tabMotionControl,"&Start Controller",NULL,this,ID_START_CTRLER,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
//   new FXButton(tabMotionControl,"&Stop Controller",NULL,this,ID_STOP_CTRLER,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  // Horizontal divider line
  new FXHorizontalSeparator(matCtrlerType,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXRadioButton(matCtrlerType,"Trajectory deformer",&ctrlerType_target,FXDataTarget::ID_OPTION+TEDDY_CTRLER,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXMatrix *matTeddyParams=new FXMatrix(matCtrlerType,6,FRAME_THICK|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matTeddyParams,"K_INT:");
  new FXTextField(matTeddyParams,6,&k_int_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matTeddyParams,"K_EXT:");
  new FXTextField(matTeddyParams,6,&k_ext_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matTeddyParams,"K_STEXT:");
  new FXTextField(matTeddyParams,6,&k_stext_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matTeddyParams,"DYN_MED:");
  new FXTextField(matTeddyParams,6,&dyn_med_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matTeddyParams,"ST_MED:");
  new FXTextField(matTeddyParams,6,&st_med_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matTeddyParams,"TMP_MED:");
  new FXTextField(matTeddyParams,6,&tmp_med_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  new FXHorizontalSeparator(matCtrlerType,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXButton(matCtrlerType,"Init New Controller",NULL,this,ID_INIT_CTRLER,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);


  // Horizontal divider line
//   new FXHorizontalSeparator(tabMotionControl,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXHorizontalFrame *frameControlCmd=new FXHorizontalFrame(tabMotionControl,LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);

  FXMatrix *matCtrlCmd=new FXMatrix(frameControlCmd,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(matCtrlCmd,"Controller command:",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);
  new FXHorizontalSeparator(matCtrlCmd,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  FXMatrix *matCtrlOpt=new FXMatrix(matCtrlCmd,1,FRAME_THICK|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matDebugCtrl=new FXMatrix(matCtrlOpt,3,MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);
  new FXLabel(matDebugCtrl,"debug level:");
  //FXMatrix *matDebugLevelCtrl=new FXMatrix(matDebugCtrl,2,MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,5,5,5,5);
  new FXRadioButton(matDebugCtrl,"0",&ctrlerDebugLevel_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matDebugCtrl,"1",&ctrlerDebugLevel_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXMatrix *matDtCtrl=new FXMatrix(matCtrlOpt,2,MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);
  new FXLabel(matDtCtrl,"dt control (s):");
  new FXTextField(matDtCtrl,6,&dtCtrl_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  
  FXMatrix *matCtrlerCmd=new FXMatrix(matCtrlCmd,2,FRAME_THICK|MATRIX_BY_COLUMNS|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXButton(matCtrlerCmd,"Start Controller",NULL,this,ID_START_CTRLER,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matCtrlerCmd,"Stop Controller",NULL,this,ID_STOP_CTRLER,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXLabel(matCtrlerCmd,"status:");
  ctrlerStatusLabel = new FXLabel(matCtrlerCmd,"disconnected");


  FXMatrix *matCtrlDisp=new FXMatrix(frameControlCmd,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  nbDeformLabel = new FXLabel(matCtrlDisp,"nbDeform:");
  avgDeformDurLabel = new FXLabel(matCtrlDisp,"avg duration:");
  DPSLabel = new FXLabel(matCtrlDisp,"DPS:");

//   new FXHorizontalSeparator(tabMotionControl,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXButton(tabMotionControl,"Stop Robot Motion (&V)",NULL,this,ID_STOP_ROBOT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  

  /***************************************************************************************************************
   ********************************************** environment tab ************************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Slammot");

  FXVerticalFrame *tabSlammot=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  /*********************************************************
   *********************** Mapping *************************
   *********************************************************/

  new FXLabel(tabSlammot,"Mapping");

  FXMatrix *matStaticEnv=new FXMatrix(tabSlammot,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matLoadStaticEnv=new FXMatrix(matStaticEnv,2,MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);
  new FXTextField(matLoadStaticEnv,35,&stEnvFilename_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|LAYOUT_FILL_X|FRAME_SUNKEN|FRAME_THICK);
  new FXButton(matLoadStaticEnv,"...",NULL,this,ID_CHANGE_STENV,FRAME_THICK|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);  

  // Button to load the static environment
  new FXButton(matStaticEnv,"Load Static Environment",NULL,this,ID_LOAD_STATICENV,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

//   FXCheckButton *autoBuildRoadmapChBox = new FXCheckButton(matStaticEnv,"autobuild roadmap",this,ID_AUTOBUILD_ROADMAP);
//   autoBuildRoadmapChBox->setCheck(autoBuildRoadmap,false);

  new FXRadioButton(matStaticEnv,"No roadmap used",&optRoadmap_target,FXDataTarget::ID_OPTION+0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matStaticEnv,"Autobuild roadmap",&optRoadmap_target,FXDataTarget::ID_OPTION+1,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  new FXRadioButton(matStaticEnv,"Load roadmap from file",&optRoadmap_target,FXDataTarget::ID_OPTION+2,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXMatrix *matLoadRoadmap=new FXMatrix(matStaticEnv,2,MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);
  new FXTextField(matLoadRoadmap,35,&RMfilename_target,FXDataTarget::ID_VALUE,JUSTIFY_LEFT|LAYOUT_FILL_X|FRAME_SUNKEN|FRAME_THICK);
  new FXButton(matLoadRoadmap,"...",NULL,this,ID_CHANGE_ROADMAP,FRAME_THICK|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);  

  FXMatrix *matLoadSaveRM=new FXMatrix(matStaticEnv,2,MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);
  new FXButton(matLoadSaveRM,"Load Roadmap",NULL,this,ID_LOAD_ROADMAP,FRAME_THICK|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);  
  new FXButton(matLoadSaveRM,"Save Roadmap",NULL,this,ID_SAVE_ROADMAP,FRAME_THICK|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);  

  FXCheckButton *dispRoadmapChBox = new FXCheckButton(matStaticEnv,"display roadmap",this,ID_DISP_ROADMAP);
  dispRoadmapChBox->setCheck(dispRoadmap,false);

  new FXHorizontalSeparator(tabSlammot,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  /*********************************************************
   ********************* localization **********************
   *********************************************************/

  new FXLabel(tabSlammot,"Localization");

  FXMatrix *matDtLoc=new FXMatrix(tabSlammot,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matDtLoc,"dt localization (s):");
  new FXTextField(matDtLoc,6,&dtLoc_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);

  FXMatrix *matLoc=new FXMatrix(tabSlammot,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXMatrix *matLocCmd=new FXMatrix(matLoc,2,FRAME_THICK|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXButton(matLocCmd,"Start Localization",NULL,this,ID_START_LOC,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matLocCmd,"Stop Localization",NULL,this,ID_STOP_LOC,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXLabel(matLocCmd,"status:");
  locStatusLabel = new FXLabel(matLocCmd,"disconnected");

  locLabel = new FXLabel(matLoc,"Current state : ()");

  // Horizontal divider line
  new FXHorizontalSeparator(tabSlammot,SEPARATOR_RIDGE|LAYOUT_FILL_X);


  /*********************************************************
   **************** moving object tracking *****************
   *********************************************************/

  new FXLabel(tabSlammot,"Moving Objects Tracking");

  FXMatrix *matSlammot=new FXMatrix(tabSlammot,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXHorizontalFrame *frameDATMO=new FXHorizontalFrame(matSlammot,LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);

  FXMatrix *matSlammotCmd=new FXMatrix(frameDATMO,2,FRAME_THICK|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXButton(matSlammotCmd,"Start MOT",NULL,this,ID_START_SLAMMOT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matSlammotCmd,"Stop MOT",NULL,this,ID_STOP_SLAMMOT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXLabel(matSlammotCmd,"status:");
  slammotStatusLabel = new FXLabel(matSlammotCmd,"disconnected");

  FXMatrix *matDATMOoptions=new FXMatrix(frameDATMO,2,FRAME_THICK|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);
  new FXLabel(matDATMOoptions,"filtering distance:");
  new FXTextField(matDATMOoptions,6,&distFilter_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);
  new FXLabel(matDATMOoptions,"security distance:");
  new FXTextField(matDATMOoptions,6,&securDist_target,FXDataTarget::ID_VALUE,TEXTFIELD_REAL|JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK);


//   new FXButton(tabSlammot,"Connect HUGR &Environment",NULL,this,ID_CONNECT_ENV,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);


  /***************************************************************************************************************
   *********************************************** viewer options ************************************************
   ***************************************************************************************************************/

  new FXTabItem(panels,"Viewer Opt.");

  FXVerticalFrame *tabViewerOpt=new FXVerticalFrame(panels,FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  new FXLabel(tabViewerOpt,"General Viewer Options",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);

  FXMatrix *matGeneralView=new FXMatrix(tabViewerOpt,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXCheckButton *dynamicViewChBox = new FXCheckButton(matGeneralView,"dynamic view",this,ID_DYNVIEW_CB);
  dynamicViewChBox->setCheck(dynamicView,false);

  new FXButton(matGeneralView,"Center View",NULL,this,ID_CENTER_VIEW,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXHorizontalSeparator(tabViewerOpt,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabViewerOpt,"Trajectory Recording",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);

  FXMatrix *matRecordTraj=new FXMatrix(tabViewerOpt,1,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXCheckButton *recordTrajCBox = new FXCheckButton(matRecordTraj,"Record trajectory",this,ID_RECORD_TRAJ_BTN);
  recordTrajCBox->setCheck(recordTraj,false);

  new FXButton(matRecordTraj,"Display trajectory record",NULL,this,ID_DISP_TRAJREC,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matRecordTraj,"Display all traj records",NULL,this,ID_DISP_ALLTRAJREC,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matRecordTraj,"Empty current record",NULL,this,ID_EMPTY_DEVREC,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  new FXButton(matRecordTraj,"Empty all records",NULL,this,ID_EMPTY_ALLREC,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);

  new FXHorizontalSeparator(tabViewerOpt,SEPARATOR_RIDGE|LAYOUT_FILL_X);

  new FXLabel(tabViewerOpt,"Trajectory Deformation Visualization",NULL,JUSTIFY_CENTER_X|LAYOUT_FILL_X);

  FXMatrix *matDeformationView=new FXMatrix(tabViewerOpt,2,FRAME_THICK|FRAME_RAISED|MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,10,10);

  FXCheckButton *dispFExtCBox = new FXCheckButton(matDeformationView,"display FExt",this,ID_DISP_FEXT_UPD);
  dispFExtCBox->setCheck(dispFExt,false);
  FXCheckButton *dispNearDynObsCBox = new FXCheckButton(matDeformationView,"display near DynObs",this,ID_DISP_NEAR_DYNOBS_UPD);
  dispNearDynObsCBox->setCheck(dispNearDynObs,false);
  FXCheckButton *dispStFExtCBox = new FXCheckButton(matDeformationView,"display StFExt",this,ID_DISP_STFEXT_UPD);
  dispStFExtCBox->setCheck(dispStFExt,false);
  FXCheckButton *dispNearStObsCBox = new FXCheckButton(matDeformationView,"display near StObs",this,ID_DISP_NEAR_STOBS_UPD);
  dispNearStObsCBox->setCheck(dispNearStObs,false);


  /***************************************************************************************************************
   ********************************************** general commands ***********************************************
   ***************************************************************************************************************/

  // Horizontal divider line
  new FXHorizontalSeparator(buttonFrame,SEPARATOR_RIDGE|LAYOUT_FILL_X);

//   // Button to load a map from a file
//   new FXButton(buttonFrame,"Load Map from file",NULL,this,ID_LOAD_MAP,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);
  
  // Exit button
  new FXButton(buttonFrame,"Exit",NULL,getApp(),FXApp::ID_QUIT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,10,10,5,5);


  /***************************************************************************************************************
   ******************************************** post initializations *********************************************
   ***************************************************************************************************************/

  this->glcanvas->init(7.37287, -119.219, 113.566, 7.32997, 26, 0.14861, 0, 1, 0); 

  getApp()->addTimeout(this,ID_TIMEOUT_DRAW,1);
}

// destructor
MPC_HugrIfc::~MPC_HugrIfc()
{
}

// Create and initialize
void MPC_HugrIfc::create(){

  // Create the windows
  FXMainWindow::create();

  // Make the main window appear
  show(PLACEMENT_SCREEN);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////// HUGR store connection functions ///////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// // establish a connection to the hugr store 
// // and try to get back informations about the environment
// // return 0 if the connection has been established, -1 else
// int MPC_HugrIfc::connectSLAMMOT()
// {
//   store.connect();

//   if(this->env_ != NULL) delete env_;
//   env_ = new HEnvironment(&store);
//   int res = env_->connect();
//   if(res == -1) slammotStatusLabel->setText("disconnected");
//   else slammotStatusLabel->setText("stopped");
//   return res;
// }

// set the wheelchair simulation variables for the hugr store
long MPC_HugrIfc::setWCSimuVar(FXObject*,FXSelector,void*)
{
  WC_namePos = FXString("wheelchair_pose");
  WC_nameCmd = FXString("wheelchair_command");
  return 1;
}

// set the wheelchair robot variables for the hugr store
long MPC_HugrIfc::setWCRobotVar(FXObject*,FXSelector,void*)
{
  WC_namePos = FXString("wuweiPose");
  WC_nameCmd = FXString("wuweiCmd");
  return 1;
}

// establish a connection to the hugr store
// and try to take control of the wheelchair
long MPC_HugrIfc::connectHugrWheelChair(FXObject*,FXSelector,void*)
{
  store.connect();

  if(this->WheelChairExists())
    {
      typeSys_ = WHEELCHAIR_SYS;
      msgBox->setText("Connection to the wheelchair established.");
      
//       rb_ = new WheelChair(string("wheelchair"),&store);
      rb_ = new WheelChair(&store,string(WC_namePos.text()),string(WC_nameCmd.text()));

      FXString str(rb_->sys()->toString());
      //sysConfLabel->setText(str);

//       cout << "Wheelchair parameters : " << endl;
//       rb_->sys()->display();

      WC_wheel_radius       = ((WheelChairSys*)(rb_->sys()))->getWheelRad();
      WC_wheel_base         = ((WheelChairSys*)(rb_->sys()))->getWheelBase();
      WC_max_speed          = ((WheelChairSys*)(rb_->sys()))->getVMax();
      WC_max_acc            = ((WheelChairSys*)(rb_->sys()))->getAMax();
      WC_max_rotation_speed = ((WheelChairSys*)(rb_->sys()))->getAVMax();
      WC_max_rotation_acc   = ((WheelChairSys*)(rb_->sys()))->getAAMax();
      WC_braking_dist       = ((WheelChairSys*)(rb_->sys()))->getBrakingDist();
      WC_length             = ((WheelChairSys*)(rb_->sys()))->getLength();
      WC_width              = ((WheelChairSys*)(rb_->sys()))->getWidth();

      // launch the localization process
      this->startLoc(NULL,0,NULL);

      // launch the moving object tracking process
      this->startSLAMMOT(NULL,0,NULL);

      // init the environment perception process
      rb_->setAutobuildRM(autoBuildRoadmap);
      rb_->setDispRM(dispRoadmap);
      
      // start the controller status update
      new boost::thread(boost::bind(&MPC_HugrIfc::loopStatus,this));

      // update the GUI
      this->updateCurSystemLabel();
      this->updateCurPlannerLabel();
      this->updateCurControllerLabel();
    }
  else
    {
      msgBox->setText("Connection to the wheelchair failed.");
      FXMessageBox::error(this,MBOX_OK,"Connection to the wheelchair failed.","The wheelchair doesn't exist in the HUGR store");

      // update the GUI
      this->updateCurSystemLabel();
      this->updateCurPlannerLabel();
      this->updateCurControllerLabel();
      return -1;
    }

  return 1;
}

// establish a connection to the hugr store 
// and try to take control of the cycab
long MPC_HugrIfc::connectHugrCycab(FXObject*,FXSelector,void*)
{
  store.connect();

  if(this->CycabExists())
    {
      typeSys_ = CARLIKE_SYS;
      msgBox->setText("Connection to the cycab established.");

      rb_ = new Cycab(string("cycab"),&store);

//       cout << "Cycab parameters : " << endl;
//       rb_->sys()->display();

      CC_length = ((CarLikeSys*)(rb_->sys()))->getLength();
      CC_max_speed = ((CarLikeSys*)(rb_->sys()))->getVMax();
      CC_max_acc = ((CarLikeSys*)(rb_->sys()))->getAMax();
      CC_max_steer = ((CarLikeSys*)(rb_->sys()))->getMaxSteering();
      CC_max_stspd = ((CarLikeSys*)(rb_->sys()))->getMaxSteerSpd();

      // launch the localization process
      this->startLoc(NULL,0,NULL);

      // init the environment perception process
      rb_->setAutobuildRM(autoBuildRoadmap);
      rb_->setDispRM(dispRoadmap);

      // start the controller status update
      new boost::thread(boost::bind(&MPC_HugrIfc::loopStatus,this));
    }
  else
    {
      msgBox->setText("Connection to the cycab failed.");
      FXMessageBox::error(this,MBOX_OK,"Connection to the cycab failed.","The cycab doesn't exist in the HUGR store");
      return -1;
    }

  return 1;
}

// check if the wheelchair exists in the HUGR store
bool MPC_HugrIfc::WheelChairExists()
{
  try 
    { 
      WC_poseId = store.lookupVariable(WC_namePos.text());
      WC_commandId = store.lookupVariable(WC_nameCmd.text());
    }
  catch(runtime_error e) { return false; }
  return true;
}

// check if the cycab exists in the HUGR store
bool MPC_HugrIfc::CycabExists()
{
  try { CC_constantsId = store.lookupVariable("cycab_state"); }
  catch(runtime_error e) { return false; }
  return true;
}

// loop to update the localization and controller status of the robot
void MPC_HugrIfc::loopStatus()
{
  // check if the robot is defined
  while(rb_ != NULL)
    {
      // update the controller status
      switch(rb_->ctrlerStatus())
	{
	case CTRLER_NULL : 
	  ctrlerStatusLabel->setText("null");
	  break;
	case CTRLER_DISCONNECTED : 
	  ctrlerStatusLabel->setText("disconnected");
	  break;
	case CTRLER_STOPPED : 
	  ctrlerStatusLabel->setText("stopped");
	  break;
	case CTRLER_STARTED : 
	  ctrlerStatusLabel->setText("started");
	  break;
	}

      if((rb_->ctrlerType() == TEDDY_CTRLER) && rb_->ctrlerStatus() == CTRLER_STARTED)
	{
	  int nbDeform;
	  double avgDeformDur;
	  double DPS;
	  rb_->getDeformInfo(nbDeform,avgDeformDur,DPS);

	  FXString msgNdDeform("nbDeform: ");
	  msgNdDeform += FXStringVal(((double)nbDeform),0,FALSE);
	  nbDeformLabel->setText(msgNdDeform);
	  
	  FXString msgAvgDeformDur("avg duration: ");
	  msgAvgDeformDur += FXStringVal(avgDeformDur,4,FALSE);
	  avgDeformDurLabel->setText(msgAvgDeformDur);
	  
	  FXString msgDPS("DPS: ");
	  msgDPS += FXStringVal(DPS,2,FALSE);
	  DPSLabel->setText(msgDPS);	  
	}

      usleep(300000); // sleep 0.3 sec
    }
}

// loop to update the wheelchair localization display
void MPC_HugrIfc::loopLocWC()
{  
  while(locActive)
    {
      WCStateT st = *((WCStateT*)(rb_->statet()));
      WC_x = st.x();
      WC_y = st.y();
      WC_th = st.th();
      WC_v = st.s();
      WC_av = st.om();

//       cout << "st.x() : " << st.x() << "  st.y() : " << st.y() << "  st.th() : " << st.th() << endl;
      FXString msg("Current state (x: ");
      msg += FXStringVal(st.x(),2,FALSE);
      msg += "  y: ";
      msg += FXStringVal(st.y(),2,FALSE);
      msg += "  th: ";
      msg += FXStringVal(normalizedAngle(st.th()),2,FALSE);
      msg += ")";
      locLabel->setText(msg);

//       boost::mutex::scoped_lock lock(mutexDrawScene);
//       drawScene();

      if(recordTraj) 
	{
	  if((rb_->ctrlerType() == TRAJFOLLOW_CTRLER || rb_->ctrlerType() == TEDDY_CTRLER) && rb_->ctrlerStatus() == CTRLER_STARTED)
	    this->recordTrajDev();
	}

      usleep(1000000*dtLoc);
    }
  locStatusLabel->setText("stopped");
}

// loop to update the wheelchair localization display
void MPC_HugrIfc::loopLocCC()
{  
  while(locActive)
    {
      CLStateT st = *((CLStateT*)(rb_->statet()));
      CC_x = st.x();
      CC_y = st.y();
      CC_th = st.th();
      CC_v = st.s();
      CC_ph = st.ph();

//       cout << "st.x() : " << st.x() << "  st.y() : " << st.y() << "  st.th() : " << st.th() << endl;
      FXString msg("Current state (x: ");
      msg += FXStringVal(st.x(),2,FALSE);
      msg += "  y: ";
      msg += FXStringVal(st.y(),2,FALSE);
      msg += "  th: ";
      msg += FXStringVal(normalizedAngle(st.th()),2,FALSE);
      msg += ")";
      locLabel->setText(msg);

      if(recordTraj) 
	{
	  if((rb_->ctrlerType() == TRAJFOLLOW_CTRLER || rb_->ctrlerType() == TEDDY_CTRLER) && rb_->ctrlerStatus() == CTRLER_STARTED)
	    this->recordTrajDev();
	}

      usleep(1000000*dtLoc);

//       boost::mutex::scoped_lock lock(mutexDrawScene);
//       drawScene();
    }
  locStatusLabel->setText("stopped");
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////// Signal functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////
////////////// GL viewer signals ////////////////
/////////////////////////////////////////////////

long MPC_HugrIfc::onConfigure(FXObject*,FXSelector,void*)
{
  return 1;
}

long MPC_HugrIfc::onExpose(FXObject*,FXSelector,void*)
{
//   double EX = glcanvas->getEyeX();
//   double EY = glcanvas->getEyeY();
//   double EZ = glcanvas->getEyeZ();
//   double CX = glcanvas->getCenterX();
//   double CY = glcanvas->getCenterY();
//   double CZ = glcanvas->getCenterZ();
//   double UX = glcanvas->getUpX();
//   double UY = glcanvas->getUpY();
//   double UZ = glcanvas->getUpZ();

//   FXString str("EX:");
//   str += FXStringVal(EX,2,FALSE);
//   str += FXString("  EY:");
//   str += FXStringVal(EY,2,FALSE);
//   str += FXString("  EZ:");
//   str += FXStringVal(EZ,2,FALSE);
//   str += FXString("  CX:");
//   str += FXStringVal(CX,2,FALSE);
//   str += FXString("  CY:");
//   str += FXStringVal(CY,2,FALSE);
//   str += FXString("  CZ:");
//   str += FXStringVal(CZ,2,FALSE);
//   str += FXString("  UX:");
//   str += FXStringVal(UX,2,FALSE);
//   str += FXString("  UY:");
//   str += FXStringVal(UY,2,FALSE);
//   str += FXString("  UZ:");
//   str += FXStringVal(UZ,2,FALSE);
//   msgBox->setText(str);

  if(MUTEX_DEBUG == 1) cout << "MPC_HugrIfc::onExpose : waiting for mutex DrawScene" << endl;
  boost::mutex::scoped_lock lock(mutexDrawScene);
  if(MUTEX_DEBUG == 1) cout << "MPC_HugrIfc::onExpose : mutex DrawScene locked" << endl;

  drawScene();

  if(MUTEX_DEBUG == 1) cout << "MPC_HugrIfc::onExpose : mutex DrawScene released" << endl;
  return 1;
}

long MPC_HugrIfc::repaint(FXObject*,FXSelector,void*)
{
  if(MUTEX_DEBUG == 1) cout << "MPC_HugrIfc::repaint : waiting for mutex DrawScene" << endl;
  boost::mutex::scoped_lock lock(mutexDrawScene);
  if(MUTEX_DEBUG == 1) cout << "MPC_HugrIfc::repaint : mutex DrawScene locked" << endl;

  drawScene();

  if(MUTEX_DEBUG == 1) cout << "MPC_HugrIfc::repaint : mutex DrawScene released" << endl;
  return 1;
}

long MPC_HugrIfc::timeoutDraw(FXObject*,FXSelector,void*)
{
  drawScene();
  usleep(50000);
  getApp()->addTimeout(this,ID_TIMEOUT_DRAW,1);
  return 1;
}


/////////////////////////////////////////////////
///////////// Map loading signals ///////////////
/////////////////////////////////////////////////


// long MPC_HugrIfc::loadHugrMap(FXObject*,FXSelector,void*)
// {
//   cout << "Load Hugr Map" << endl;
  
//   FXImage *img=NULL;
//   FXImage *old;

//   // delete the old map
//   if(map != NULL) delete map;
//   map = new MapHugr();

//   img=new FXPPMImage(getApp(),NULL,IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
  
//   // Load the map
//   map->load("mapImage");

//   // Save the map in a local file named HugrMap.pgm
// //   map->thresholdPGM(THRESHOLD);
//   //map->save("$DEFORMATION_DIR/HugrMap.pgm");
//   map->save("/tmp/HugrMap.pgm");

//   // Load the image
//   FXFileStream stream;
//   //if(stream.open("$DEFORMATION_DIR/HugrMap.pgm",FXStreamLoad)){
//   if(stream.open("/tmp/HugrMap.pgm",FXStreamLoad)){
//     getApp()->beginWaitCursor();
//     img->loadPixels(stream);
//     stream.close();

//     img->create();

//     // rescaling the image
//     int wtF = imageview->getWidth();
//     int htF = imageview->getHeight();

//     wtI = img->getWidth();
//     htI = img->getHeight();

//     double ratioW = ((double) wtI) / ((double) wtF);
//     double ratioH = ((double) htI) / ((double) htF);

//     if(ratioW > ratioH)
//       {
// 	img->scale((FXint)(((double)img->getWidth())/ratioW), (FXint)(((double)img->getHeight())/ratioW));
//       }
//     else
//       {
// 	img->scale((FXint)(((double)img->getWidth())/ratioH), (FXint)((double)img->getHeight())/ratioH));
//       }

//     old=imageview->getImage();
//     imageview->setImage(img);
//     delete old;
//     getApp()->endWaitCursor();

//     // display both configuration if they are defined
// //     drawInitCfg();
// //     drawGoalCfg();
//     }
//   return 1;
// }

// long MPC_HugrIfc::loadMapFromFile(FXObject*,FXSelector,void*)
// {
//   cout << "Load Map from file" << endl;
//   FXFileDialog open(this,"Load map (PGM Image)");
//   open.setFilename(filename);
//   if(open.execute()){
//     filename=open.getFilename();
//     loadimage(filename);

//     msgBox->setText("A map has been Loaded. Click left to select init configuration and click right to select target configuration");
//   }
//   return 1;
// }

long MPC_HugrIfc::loadStaticEnvBtn(FXObject*,FXSelector,void*)
{
  int res = this->loadStaticEnv(stEnvFilename.text());
  cout << "loadStaticEnvBtn -> res : " << res << endl;

  switch(res)
    {
    case 0 : 
      {
	if(strcmp(stEnvFilename.text(),"/home/vdelsart/these/AM2/MPC/map/robotic_hall.map2") == 0 || 
	   strcmp(stEnvFilename.text(),"../map/robotic_hall.map2") == 0)
	  {
	    minFieldX = -3.0;
	    maxFieldX =  13.0;
	    minFieldY = -14.0;
	    maxFieldY =  3.5;
	  }
	else if(strcmp(stEnvFilename.text(),"../map/accueil.map2") == 0)
	  {
	    minFieldX = -8.6;
	    maxFieldX =  9.7;
	    minFieldY = -3.7;
	    maxFieldY =  13.9;
	  }
	else
	  {
	    minFieldX = -10.6;
	    maxFieldX =  12.7;
	    minFieldY = -9.9;
	    maxFieldY =  13.9;
	  }
	
	// center the view on the defined environment
	this->centerView();
	
	// sample the static environment to select easily the goal configuration
	map = this->sampleStaticEnv();
	if(map != NULL) 
	  {
	    map->save(samplingStEnvFilename.text());
	    this->loadMapInViewer();
	    stEnvLoaded = true;
	  }
	break;      
      }
    case -1 : 
      msgBox->setText("Unable the load the static environment. The input file has not been found.");
      FXMessageBox::error(this,MBOX_OK,"Unable the load the static environment","The input file has not been found.");
      break;
    case -2 : 
      msgBox->setText("Unable the load the static environment. The localization must be started first.");
      FXMessageBox::error(this,MBOX_OK,"Unable the load the static environment","The localization must be started first.");
      break;
    case -3 : 
      msgBox->setText("Unable the load the static environment. The environment perception has not been initialized.");
      FXMessageBox::error(this,MBOX_OK,"Unable the load the static environment","The environment perception has not been initialized.");
      break;
    case -4 : 
      msgBox->setText("Unable the load the static environment. An unknown error occured.");
      FXMessageBox::error(this,MBOX_OK,"Unable the load the static environment","An unknown error occured.");
      break;
    case -5 : 
      msgBox->setText("Unable the load the static environment. A system must be connected first.");
      FXMessageBox::error(this,MBOX_OK,"Unable the load the static environment","A system must be connected first.");
      break;
    }

  if(res == 0) return 1;
  else return -1;
}

long MPC_HugrIfc::loadRoadmapBtn(FXObject*,FXSelector,void*)
{
  int res = this->loadRoadmap(RMfilename.text());

  switch(res)
    {
    case 0 : 
      // refresh the viewer
      this->drawScene();

      msgBox->setText("The roadmap loading process succeed");
      break;
    case -4 : 
      msgBox->setText("The roadmap loading process failed");
      FXMessageBox::error(this,MBOX_OK,"Unable to load the roadmap","The roadmap loading process failed.");
      break;
    case -5 : 
      msgBox->setText("Unable to load the roadmap. A system must be connected first.");
      FXMessageBox::error(this,MBOX_OK,"Unable to load the roadmap","A system must be connected first.");
      break;
    }

  return 1;
}

long MPC_HugrIfc::saveRoadmapBtn(FXObject*,FXSelector,void*)
{
  int res = this->saveRoadmap(RMfilename.text());

  switch(res)
    {
    case 0 : 
      // refresh the viewer
      this->drawScene();

      msgBox->setText("The roadmap saving process succeed");
      break;
    case -4 : 
      msgBox->setText("The roadmap saving process failed");
      FXMessageBox::error(this,MBOX_OK,"Unable to save the roadmap","The roadmap saving process failed.");
      break;
    case -5 : 
      msgBox->setText("Unable to save the roadmap. A system must be connected first.");
      FXMessageBox::error(this,MBOX_OK,"Unable to save the roadmap","A system must be connected first.");
      break;
    }

  return 1;
}

long MPC_HugrIfc::changeStEnvFileName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Load static environment (*.map)");
  open.setFilename(stEnvFilename);
  if(open.execute()){
    stEnvFilename=open.getFilename();
  }
  return 1;
}

long MPC_HugrIfc::changeRoadmapFileName(FXObject*,FXSelector,void*)
{
  FXFileDialog open(this,"Load roadmap (*.rdm)");
  open.setFilename(RMfilename);
  if(open.execute()){
    RMfilename=open.getFilename();
  }
  return 1;
}


// long MPC_HugrIfc::onLeftMouseDown(FXObject*,FXSelector,void* ptr)
// {
//   // get back the event of the signal
//   FXEvent *ev=(FXEvent*)ptr;

//   // grab the image window
//   imageview->grab();

//   // the left button is pressed
//   mlbFlag = true;

//   // create the drawing window
//   FXDCWindow dc(imageview);

//   // draw the init position on the map
//   dc.setForeground(initColor);
//   dc.drawEllipse(ev->win_x - 2, ev->win_y - 2, 4, 4);

//   // set the coordinates of init position
//   int xM=0,yM=0;
//   this->coordFrameToCoordMap(ev->win_x, ev->win_y, xM, yM);

//   x1 = xM;
//   y1 = yM;
//   return 1;
// }

// long MPC_HugrIfc::onLeftMouseUp(FXObject*,FXSelector,void* ptr)
// {
//   // get back the event of the signal
//   FXEvent *ev=(FXEvent*)ptr;

//   // ungrab the image window
//   imageview->ungrab();

//   // the left button is unpressed
//   mlbFlag = false;
  
//   // compute the init direction
//   int xM=0,yM=0;
//   this->coordFrameToCoordMap(ev->win_x, ev->win_y, xM, yM);

//   // compute the vector between init position and the new position
//   int dx = (int)(xM - x1);
//   int dy = (int)(yM - y1);
  
//   PositionXY vect(dx,dy);
//   PositionXY vectN = vect.normalized();
//   PositionXY vectN20(vectN.x()*20.0f,vectN.y()*20.0f);

//   int xTgtMap = (int)(x1 + vectN20.x());
//   int yTgtMap = (int)(y1 + vectN20.y());

//   // compute the direction from the vector
//   double ang = atan2(dy,dx);
//   theta1 = ang / M_PI * 180.0f;

//   // draw the direction on the map
//   int x1Frame=0,y1Frame=0;
//   coordMapToCoordFrame((int)x1, (int)y1, x1Frame, y1Frame);

//   int xTgtFrame=0,yTgtFrame=0;
//   coordMapToCoordFrame(xTgtMap, yTgtMap, xTgtFrame, yTgtFrame);

//   // create the drawing window
//   FXDCWindow dc(imageview);

//   // draw the init direction on the map
//   dc.setForeground(initColor);
//   dc.drawLine(x1Frame, y1Frame, xTgtFrame, yTgtFrame);

//   this->loadimage(filename);

//   return 1;
// }

long MPC_HugrIfc::onRightMouseDown(FXObject*,FXSelector,void* ptr)
{
  if(stEnvLoaded)
    {
      // get back the event of the signal
      FXEvent *ev=(FXEvent*)ptr;

      // grab the image window
      imageview->grab();

      // the right button is pressed
      mrbFlag = true;

//       // clear the drawing window
//       this->loadMapInViewer();

      // create the drawing window
      FXDCWindow dc(imageview);

      // draw the init position on the map
      dc.setForeground(goalColor);
      dc.drawEllipse(ev->win_x - 2, ev->win_y - 2, 4, 4);

      // set the coordinates of init position
      double xM=0,yM=0;
      this->coordFrameToCoordMap(ev->win_x, ev->win_y, xM, yM);

      planner_goalX = xM;
      planner_goalY = yM;
    }
  return 1;
}

long MPC_HugrIfc::onRightMouseUp(FXObject*,FXSelector,void* ptr)
{
  if(stEnvLoaded)
    {
      // get back the event of the signal
      FXEvent *ev=(FXEvent*)ptr;

      // ungrab the image window
      imageview->ungrab();

      // the right button is unpressed
      mrbFlag = false;

      // compute the goal direction
      double xM=0,yM=0;
      this->coordFrameToCoordMap(ev->win_x, ev->win_y, xM, yM);

      // compute the vector between init position and the new position
      double dx = (xM - planner_goalX);
      double dy = (yM - planner_goalY);
  
      PositionXY vect(dx,dy);
      PositionXY vectN = vect.normalized();
      PositionXY vectN20(vectN.x()*20.0f,vectN.y()*20.0f);

      double xTgtMap = (planner_goalX + vectN20.x());
      double yTgtMap = (planner_goalY + vectN20.y());

      // compute the direction from the vector
      double ang = atan2(dy,dx);
      planner_goalTH = ang / M_PI * 180.0f;

      // draw the direction on the map
      int x2Frame=0,y2Frame=0;
      coordMapToCoordFrame(planner_goalX, planner_goalY, x2Frame, y2Frame);

      int xTgtFrame=0,yTgtFrame=0;
      coordMapToCoordFrame(xTgtMap, yTgtMap, xTgtFrame, yTgtFrame);

      // create the drawing window
      FXDCWindow dc(imageview);

      // draw the init direction on the map
      dc.setForeground(goalColor);
      dc.drawLine(x2Frame, y2Frame, xTgtFrame, yTgtFrame);
    }
  return 1;
}

long MPC_HugrIfc::onMouseMove(FXObject*,FXSelector,void*)
{
  return 1;
}


/////////////////////////////////////////////////
/////// Init of the robot config signals ////////
/////////////////////////////////////////////////

long MPC_HugrIfc::initWCCfg(FXObject*,FXSelector,void*)
{
  cout << "initWCCfg" << endl;

  // connect the hugr store if it doesn't have been done yet
  store.connect();

  hugr::VariableId WC_poseId;

//   std::string str("wheelchair");
  try
    {
//       str += "_pose";
//       WC_poseId = store.lookupVariable(str.c_str());
      WC_poseId = store.lookupVariable(WC_namePos.text());
    }
  catch(runtime_error e)
    {
      cerr << "WheelChair::updatePose : Warning. " << WC_namePos.text() << " cannot be found in the hugr store" << endl;
      return -1;
    }

  cout << "WC_poseId : " << WC_poseId << endl;
  cout << "initX : " << initX << "initY : " << initY << "initTH : " << initTH << endl;

  WheelchairPose WC_pose;
  WC_pose.position.x = initX;
  WC_pose.position.y = initY;
  WC_pose.position.theta = initTH;
  
  try
    {
      store.writeVariable(WC_poseId, WC_pose);
    }
  catch(runtime_error e)
    {
      cerr << "WheelChair::updatePose : Warning. Unable the write the wheelchair pose variable." << endl;
      return -1;
    }
  
  cout << "WheelChair::updatePose : Wheelchair pose written...?" << endl;
  return 1;
}

long MPC_HugrIfc::initCCCfg(FXObject*,FXSelector,void*)
{
  // connect the hugr store if it doesn't have been done yet
  store.connect();

  hugr::VariableId CC_poseId;

  // get back the cycab pose id
  std::string str("cycab");
  try
    {
      str += "_simuState";
      CC_poseId = store.lookupVariable(str.c_str());
    }
  catch(runtime_error e)
    {
      cerr << "Cycab::updatePose : Warning. " << str.c_str() << " cannot be found in the hugr store" << endl;
      return -1;
    }
  
  SimulatorState CC_pose;
  CC_pose.x = initX;
  CC_pose.y = initY;
  CC_pose.theta = initTH;
  store.writeVariable(CC_poseId, CC_pose);

  return 1;
}


/////////////////////////////////////////////////
///////////// Localization signals //////////////
/////////////////////////////////////////////////

long MPC_HugrIfc::startLoc(FXObject*,FXSelector,void*)
{
  locActive = true;
  if(rb_ != NULL) 
    {
      locStatusLabel->setText("started");
      rb_->setDtLoczer(dtLoc);
      int res = rb_->startLoc();
      if(res == 0)
	{
	  locStatusLabel->setText("started");
	  
	  switch(typeSys_)
	    {
	    case WHEELCHAIR_SYS : 
	      {
		//boost::thread* threadCtrl_ = 
		new boost::thread(boost::bind(&MPC_HugrIfc::loopLocWC,this));
		break;
	      }
	    case CARLIKE_SYS : 
	      {
		//boost::thread* threadCtrl_ = 
		new boost::thread(boost::bind(&MPC_HugrIfc::loopLocCC,this));
		break;
	      }
	    default : 
	      FXMessageBox::error(this,MBOX_OK,"Cannot start the localization","Robotic system type not handled");
	      msgBox->setText("Cannot start the localization. Robotic system type not handled.");
	      break;
	    }
	}
      else locStatusLabel->setText("disconnected");      
    }
  else
    {
      FXMessageBox::error(this,MBOX_OK,"Cannot start the localization","No robotic system has been defined");
      msgBox->setText("Cannot start the localization. No robotic system has been defined.");
      return -1;
    }

  msgBox->setText("The localization process has been started.");
  return 1;
}

long MPC_HugrIfc::stopLoc(FXObject*,FXSelector,void*)
{
  locActive = false;
  if(rb_ != NULL)
    {
      int res = rb_->stopLoc();
      if(res == 0) locStatusLabel->setText("stopped");
      else locStatusLabel->setText("disconnected");
    }

  msgBox->setText("The localization process has been stopped.");
  return 1;
}

long MPC_HugrIfc::displayTrajDevRecBtn(FXObject*,FXSelector,void*)
{
  this->displayTrajDevRec();
  return 1;
}

long MPC_HugrIfc::displayAllTrajRecBtn(FXObject*,FXSelector,void*)
{
  this->displayAllTrajRec();
  return 1;
}

long MPC_HugrIfc::deleteTrajDevRecBtn(FXObject*,FXSelector,void*)
{
  this->reinitTrajDevRec();
  return 1;
}

long MPC_HugrIfc::deleteAllTrajRecBtn(FXObject*,FXSelector,void*)
{
  this->deleteAllTrajRec();
  return 1;
}


/////////////////////////////////////////////////
////////////// Controller signals ///////////////
/////////////////////////////////////////////////

long MPC_HugrIfc::initCtrler(FXObject*,FXSelector,void*)
{
  if(rb_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Cannot initialize the controller","No robotic system has been defined");
      msgBox->setText("Cannot initialize the controller. No robotic system has been defined.");
      return -1;
    }

  // empty the vectors used to record the trajectory
  this->reinitTrajDevRec();

  // init the controller depending on its type
  int res;
  switch(ctrlerType)
    {
    case SIMPLE_CTRLER : 
      res = rb_->initCtrler(SIMPLE_CTRLER);
      break;
    case TRAJFOLLOW_CTRLER : 
      res = rb_->initCtrler(TRAJFOLLOW_CTRLER);
      break;
    case TEDDY_CTRLER : 
      res = rb_->initCtrler(TEDDY_CTRLER);
      if(res == 0) rb_->setTeddyParams(k_int,k_ext,k_stext,dyn_med,st_med,tmp_med);

      break;
    default : 
      FXMessageBox::error(this,MBOX_OK,"Cannot initialize the controller","Unknown controller type");
      msgBox->setText("Cannot initialize the controller. Unknown controller type.");

      // update the GUI
      this->updateCurControllerLabel();
      return -1;
    }

  switch(res)
    {
    case 0 : 
      // init the controller debug level
      //if(rb_ != NULL) rb_->setCtrlerDebugLevel(ctrlerDebugLevel);
      if(rb_ != NULL) rb_->setDebug(ctrlerDebugLevel);
      msgBox->setText("The controller has been initialized");

      // update the GUI
      this->updateCurControllerLabel();
      return 1;
    case -1 :
    case -2 : 
      return 1;
    case -3 : 
      FXMessageBox::error(this,MBOX_OK,"Cannot initialize the controller","Unknown controller type");

      // update the GUI
      msgBox->setText("Cannot initialize the controller. Unknown controller type.");
      this->updateCurControllerLabel();
      return -1;
    case -4 : 
      FXMessageBox::error(this,MBOX_OK,"Cannot initialize the controller","An unknown problem occured");

      // update the GUI
      msgBox->setText("Cannot initialize the controller. An unknown problem occured.");
      this->updateCurControllerLabel();
      return -1;
    case -5 : 
      FXMessageBox::error(this,MBOX_OK,"Cannot initialize the controller","The connection to the store cannot be established");

      // update the GUI
      msgBox->setText("Cannot initialize the controller. The connection to the store cannot be established.");
      this->updateCurControllerLabel();
      return -1;
    case -6 : 
      FXMessageBox::error(this,MBOX_OK,"Cannot initialize the controller","An input trajectory is required");

      // update the GUI
      msgBox->setText("Cannot initialize the controller. An input trajectory is required.");      
      this->updateCurControllerLabel();
      return -1;
    }
  return -1;
}

long MPC_HugrIfc::startCtrler(FXObject*,FXSelector,void*)
{
  if(rb_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Cannot start the controller","No robotic system has been defined");
      msgBox->setText("Cannot start the controller. No robotic system has been defined.");
      return -1;
    }

  rb_->setDtCtrler(dtCtrl);
  rb_->startCtrler();

  switch(rb_->ctrlerStatus())
    {
    case CTRLER_NULL : 
      ctrlerStatusLabel->setText("null");
      break;
    case CTRLER_DISCONNECTED : 
      ctrlerStatusLabel->setText("disconnected");
      break;
    case CTRLER_STOPPED : 
      ctrlerStatusLabel->setText("stopped");
      break;
    case CTRLER_STARTED : 
      ctrlerStatusLabel->setText("started");
      break;
    }

  msgBox->setText("The controller has been started.");
  return 1;
}

long MPC_HugrIfc::stopCtrler(FXObject*,FXSelector,void*)
{
  if(rb_ != NULL) rb_->stopCtrler();
  switch(rb_->ctrlerStatus())
    {
    case CTRLER_NULL : 
      ctrlerStatusLabel->setText("null");
      break;
    case CTRLER_DISCONNECTED : 
      ctrlerStatusLabel->setText("disconnected");
      break;
    case CTRLER_STOPPED : 
      ctrlerStatusLabel->setText("stopped");
      break;
    case CTRLER_STARTED : 
      ctrlerStatusLabel->setText("started");
      break;
    }

  msgBox->setText("The controller has been stopped.");
  return 1;
}

long MPC_HugrIfc::updateCtrl(FXObject*,FXSelector,void*)
{
  if(rb_ != NULL)
    {
      switch(typeSys_)
	{
	case WHEELCHAIR_SYS : 
	  {
	    rb_->updateCtrl(WCControl(Ctrl_v,Ctrl_avph));
	    break;
	  }
	case CARLIKE_SYS : 
	  {
	    rb_->updateCtrl(CCControl(Ctrl_v,Ctrl_avph));
	    break;
	  }
	default : 
	  FXMessageBox::error(this,MBOX_OK,"Cannot update the control","Robotic system type not handled");
	  msgBox->setText("Cannot update the control. Robotic system type not handled.");
	  break;
	}
    }

  return 1;
}

long MPC_HugrIfc::stopRobotMotion(FXObject*,FXSelector,void*)
{
  if(rb_ != NULL) rb_->stopMoving();
  return 1;
}

long MPC_HugrIfc::updateRecordTrajBtn(FXObject*,FXSelector,void*)
{
  // update the trajectory registration option
  if(recordTraj == false)
    recordTraj = true;
  else
    recordTraj = false;

  return 1;
}

long MPC_HugrIfc::moveForwardBtn(FXObject*,FXSelector,void*)
{
  if(rb_ != NULL)
    {
      switch(typeSys_)
	{
	case WHEELCHAIR_SYS : 
	  {
	    // 	  WCControlT* curCtrl = ((WheelChair*)rb_)->controlt();
	    rb_->updateCtrl(WCControl(Coef_Ctrl_v*WC_max_speed,0.0));
	    break;
	  }
	case CARLIKE_SYS : 
	  {
	    break;
	  }
	default : 
	  FXMessageBox::error(this,MBOX_OK,"Cannot move","Robotic system type not handled or not defined");
	  msgBox->setText("Cannot move. Robotic system type not handled or not defined.");
	  break;
	}
    }
  
  return 1;
}

long MPC_HugrIfc::moveBackwardBtn(FXObject*,FXSelector,void*)
{
  if(rb_ != NULL)
    {
      switch(typeSys_)
	{
	case WHEELCHAIR_SYS : 
	  {
	    rb_->updateCtrl(WCControl(-Coef_Ctrl_v*WC_max_speed,0.0));
	    break;
	  }
	case CARLIKE_SYS : 
	  {
	    break;
	  }
	default : 
	  FXMessageBox::error(this,MBOX_OK,"Cannot move","Robotic system type not handled or not defined");
	  msgBox->setText("Cannot move. Robotic system type not handled or not defined.");
	  break;
	}
    }
  
  return 1;
}

long MPC_HugrIfc::turnLeftBtn(FXObject*,FXSelector,void*)
{
  if(rb_ != NULL)
    {
      switch(typeSys_)
	{
	case WHEELCHAIR_SYS : 
	  {
	    rb_->updateCtrl(WCControl(0.0,Coef_Ctrl_avph*WC_max_rotation_speed));
	    break;
	  }
	case CARLIKE_SYS : 
	  {
	    break;
	  }
	default : 
	  FXMessageBox::error(this,MBOX_OK,"Cannot move","Robotic system type not handled or not defined");
	  msgBox->setText("Cannot move. Robotic system type not handled or not defined.");
	  break;
	}
    }
  
  return 1;
}

long MPC_HugrIfc::turnRightBtn(FXObject*,FXSelector,void*)
{
  if(rb_ != NULL)
    {
      switch(typeSys_)
	{
	case WHEELCHAIR_SYS : 
	  {
	    rb_->updateCtrl(WCControl(0.0,-Coef_Ctrl_avph*WC_max_rotation_speed));
	    break;
	  }
	case CARLIKE_SYS : 
	  {
	    break;
	  }
	default : 
	  FXMessageBox::error(this,MBOX_OK,"Cannot move","Robotic system type not handled or not defined");
	  msgBox->setText("Cannot move. Robotic system type not handled or not defined.");
	  break;
	}
    }
  
  return 1;
}

long MPC_HugrIfc::moveFwdLeft(FXObject*,FXSelector,void*)
{
  if(rb_ != NULL)
    {
      switch(typeSys_)
	{
	case WHEELCHAIR_SYS : 
	  {
	    rb_->updateCtrl(WCControl(Coef_Ctrl_v*WC_max_speed,Coef_Ctrl_avph*WC_max_rotation_speed));
	    break;
	  }
	case CARLIKE_SYS : 
	  {
	    break;
	  }
	default : 
	  FXMessageBox::error(this,MBOX_OK,"Cannot move","Robotic system type not handled or not defined");
	  msgBox->setText("Cannot move. Robotic system type not handled or not defined.");
	  break;
	}
    }
  
  return 1;
}

long MPC_HugrIfc::moveFwdRight(FXObject*,FXSelector,void*)
{
  if(rb_ != NULL)
    {
      switch(typeSys_)
	{
	case WHEELCHAIR_SYS : 
	  {
	    rb_->updateCtrl(WCControl(Coef_Ctrl_v*WC_max_speed,-Coef_Ctrl_avph*WC_max_rotation_speed));
	    break;
	  }
	case CARLIKE_SYS : 
	  {
	    break;
	  }
	default : 
	  FXMessageBox::error(this,MBOX_OK,"Cannot move","Robotic system type not handled or not defined");
	  msgBox->setText("Cannot move. Robotic system type not handled or not defined.");
	  break;
	}
    }
  
  return 1;
}

long MPC_HugrIfc::moveBwdLeft(FXObject*,FXSelector,void*)
{
  if(rb_ != NULL)
    {
      switch(typeSys_)
	{
	case WHEELCHAIR_SYS : 
	  {
	    rb_->updateCtrl(WCControl(-Coef_Ctrl_v*WC_max_speed,Coef_Ctrl_avph*WC_max_rotation_speed));
	    break;
	  }
	case CARLIKE_SYS : 
	  {
	    break;
	  }
	default : 
	  FXMessageBox::error(this,MBOX_OK,"Cannot move","Robotic system type not handled or not defined");
	  msgBox->setText("Cannot move. Robotic system type not handled or not defined.");
	  break;
	}
    }
  
  return 1;
}

long MPC_HugrIfc::moveBwdRight(FXObject*,FXSelector,void*)
{
  if(rb_ != NULL)
    {
      switch(typeSys_)
	{
	case WHEELCHAIR_SYS : 
	  {
	    rb_->updateCtrl(WCControl(-Coef_Ctrl_v*WC_max_speed,-Coef_Ctrl_avph*WC_max_rotation_speed));
	    break;
	  }
	case CARLIKE_SYS : 
	  {
	    break;
	  }
	default : 
	  FXMessageBox::error(this,MBOX_OK,"Cannot move","Robotic system type not handled or not defined");
	  msgBox->setText("Cannot move. Robotic system type not handled or not defined.");
	  break;
	}
    }
  
  return 1;
}

long MPC_HugrIfc::stopMotionBtn(FXObject*,FXSelector,void*)
{
  if(rb_ != NULL)
    {
      switch(typeSys_)
	{
	case WHEELCHAIR_SYS : 
	  {
	    rb_->updateCtrl(WCControl(0.0,0.0));
	    break;
	  }
	case CARLIKE_SYS : 
	  {
	    break;
	  }
	default : 
	  FXMessageBox::error(this,MBOX_OK,"Cannot move","Robotic system type not handled or not defined");
	  msgBox->setText("Cannot move. Robotic system type not handled or not defined.");
	  break;
	}
    }
  
  return 1;
}


/////////////////////////////////////////////////
//////////////// PLANNER signals ////////////////
/////////////////////////////////////////////////

long MPC_HugrIfc::initGoalPosPlanner(FXObject*,FXSelector,void*)
{
  // set the current pose as the goal position of the planner
  switch(typeSys_)
    {
    case WHEELCHAIR_SYS : 
      {
	planner_goalX = WC_x;
	planner_goalY = WC_y;
	planner_goalTH = radToDeg(WC_th);
	break;
      }
    case CARLIKE_SYS : 
      {
	planner_goalX = CC_x;
	planner_goalY = CC_y;
	planner_goalTH = radToDeg(CC_th);	
	break;
      }
    default : 
      {
	FXMessageBox::error(this,MBOX_OK,"Cannot init the planner goal position","Robotic system type not handled");
	msgBox->setText("Cannot init the planner goal position. Robotic system type not handled.");
	return -1;
      }
      break;
    }

  return 1;
}

long MPC_HugrIfc::initPlanner(FXObject*,FXSelector,void*)
{
  if(rb_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Cannot initialize the planner","No robotic system has been defined");
      msgBox->setText("Cannot initialize the planner. No robotic system has been defined.");

      // update the GUI
      msgBox->setText("Cannot initialize the planner. No robotic system has been defined.");
      this->updateCurPlannerLabel();
      return -1;
    }

  int res = rb_->initPlanner((trajPlanner_t)plannerType);

  switch(res)
    {
    case  0 : 
      // update the GUI
      msgBox->setText("The planner has been initialized.");
      this->updateCurPlannerLabel();
      return 1;
    case -2 : 
      // update the GUI
      this->updateCurPlannerLabel();
      return 1;
    case -3 : 
    case -4 : 
    default : 
      FXMessageBox::error(this,MBOX_OK,"Cannot initialize the planner","An error occured during the planning process");

      // update the GUI
      msgBox->setText("Cannot initialize the planner. An error occured during the planning process.");
      this->updateCurPlannerLabel();
      return -1;
    }
}

long MPC_HugrIfc::startPlanner(FXObject*,FXSelector,void*)
{
  if(rb_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Cannot start the planning process","No robotic system has been defined");
      msgBox->setText("Cannot start the planning process. No robotic system has been defined.");
      return -1;
    }

  int res;
  switch(typeSys_)
    {
    case WHEELCHAIR_SYS : 
      {
	WCState goalS(planner_goalX,planner_goalY,degToRad(planner_goalTH),0.0,0.0);
	double goalT;
	if(fixed_T) goalT = planner_goalT;
	else goalT = 0;

	res = ((WheelChair*)rb_)->planTraj(goalS,goalT,planner_DT,planner_CoefSpd,planner_CoefAcc);
	break;
      }
    case CARLIKE_SYS : 
      {	
	CLState goalS(planner_goalX,planner_goalY,degToRad(planner_goalTH),0.0,0.0);
	if(fixed_T) res = ((Cycab*)rb_)->planTraj(goalS,planner_goalT,planner_DT);
	else res = ((Cycab*)rb_)->planTraj(goalS,0.0,planner_DT);
	break;
      }
    default : 
      {
	FXMessageBox::error(this,MBOX_OK,"Cannot start the planner","Robotic system type not handled");
	msgBox->setText("Cannot start the planner. Robotic system type not handled.");
	return -1;
      }
    }

  switch(res)
    {
    case 0 : 
      msgBox->setText("A trajectory has been planned.");
      cout << "Trajectory planned : " << endl;
      rb_->traj()->display();
      return 1;
    case -1 : 
      FXMessageBox::error(this,MBOX_OK,"Trajectory planning error","No trajectory has been found");
      msgBox->setText("Trajectory planning error. No trajectory has been found.");
      return -1;
    case -2 : 
      FXMessageBox::error(this,MBOX_OK,"Trajectory planning error","The planner hasn't been initialized");
      msgBox->setText("Trajectory planning error. The planner hasn't been initialized.");
      return -1;
    case -3 : 
      FXMessageBox::error(this,MBOX_OK,"Trajectory planning error","Planner type not handled");
      msgBox->setText("Trajectory planning error. Planner type not handled.");
      return -1;
    case -4 : 
      FXMessageBox::error(this,MBOX_OK,"Trajectory planning error","A roadmap is required. No roadmap found.");
      msgBox->setText("Trajectory planning error. A roadmap is required. No roadmap found.");
      return -1;
    case -5 : 
      FXMessageBox::error(this,MBOX_OK,"Trajectory planning error","The environment must be initialized first.");
      msgBox->setText("Trajectory planning error. The environment must be initialized first.");
      return -1;
    case -6 : 
      FXMessageBox::error(this,MBOX_OK,"Trajectory planning error","The robot is still moving.");
      msgBox->setText("Trajectory planning error. The robot is still moving.");
      return -1;
    default : 
      FXMessageBox::error(this,MBOX_OK,"Trajectory planning error","Unknown error");
      msgBox->setText("Trajectory planning error. Unknown error.");
      return -1;
    }
  return 0;
}


/////////////////////////////////////////////////
//////////////// SLAMMOT signals ////////////////
/////////////////////////////////////////////////

long MPC_HugrIfc::startSLAMMOT(FXObject*,FXSelector,void*)
{
  if(rb_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Cannot start the environment perception process","No robotic system has been defined");
      msgBox->setText("Cannot start the environment perception process. No robotic system has been defined.");
      return -1;
    }

  rb_->initDATMO(distFilter,securDist);
  rb_->startEnvPercept();

  switch(rb_->envpStatus())
    {
    case ENVUP_NULL : 
      slammotStatusLabel->setText("null");
      break;
    case ENVUP_DISCONNECTED : 
      slammotStatusLabel->setText("disconnected");
      break;
    case ENVUP_STOPPED : 
      slammotStatusLabel->setText("stopped");
      break;
    case ENVUP_STARTED : 
      slammotStatusLabel->setText("started");
      break;
    }

  msgBox->setText("The moving object tracking process has been started.");
  return 1;
}

long MPC_HugrIfc::stopSLAMMOT(FXObject*,FXSelector,void*)
{
  if(rb_ == NULL)
    {
      FXMessageBox::error(this,MBOX_OK,"Cannot stop the environment perception process","No robotic system has been defined");
      msgBox->setText("Cannot stop the environment perception process. No robotic system has been defined.");
      return -1;
    }

  rb_->stopEnvPercept();

  switch(rb_->envpStatus())
    {
    case ENVUP_NULL : 
      slammotStatusLabel->setText("null");
      break;
    case ENVUP_DISCONNECTED : 
      slammotStatusLabel->setText("disconnected");
      break;
    case ENVUP_STOPPED : 
      slammotStatusLabel->setText("stopped");
      break;
    case ENVUP_STARTED : 
      slammotStatusLabel->setText("started");
      break;
    }

  msgBox->setText("The moving object tracking process has been stopped.");
  return 1;
}


/////////////////////////////////////////////////
////////////// Environment signals //////////////
/////////////////////////////////////////////////

long MPC_HugrIfc::updateDispRoadmap(FXObject*,FXSelector,void*)
{
  // update the roadmap displaying option
  if(dispRoadmap == false)
      dispRoadmap = true;
  else
    dispRoadmap = false;

  if(rb_ != NULL) rb_->setDispRM(dispRoadmap);
  return 1;
}

long MPC_HugrIfc::updateAutoBuildRoadmap(FXObject*,FXSelector,void*)
{
  // update the roadmap auto-building option
  if(autoBuildRoadmap == false)
    autoBuildRoadmap = true;
  else
    autoBuildRoadmap = false;

  if(rb_ != NULL) rb_->setAutobuildRM(autoBuildRoadmap);
  return 1;
}


/////////////////////////////////////////////////
/////////////// interface signals ///////////////
/////////////////////////////////////////////////

long MPC_HugrIfc::updateDynView(FXObject*,FXSelector,void*)
{
  // update the dynamic view option
  if(dynamicView == false)
    dynamicView = true;
  else
    dynamicView = false;

  return 1;
}

long MPC_HugrIfc::centerViewBtn(FXObject*,FXSelector,void*)
{
  this->centerView();
  return 1;
}

long MPC_HugrIfc::updateFixedTime(FXObject*,FXSelector,void*)
{
  // update the fixed final time
  if(fixed_T == false)
    fixed_T = true;
  else
    fixed_T = false;

  return 1;
}

long MPC_HugrIfc::updateDispFExt(FXObject*,FXSelector,void*)
{
  // update the display of dyn external forces
  if(dispFExt == false)
    dispFExt = true;
  else
    dispFExt = false;

  return 1;
}

long MPC_HugrIfc::updateDispStFExt(FXObject*,FXSelector,void*)
{
  // update the display of dyn external forces
  if(dispStFExt == false)
    dispStFExt = true;
  else
    dispStFExt = false;

  return 1;
}

long MPC_HugrIfc::updateDispNearDynObs(FXObject*,FXSelector,void*)
{
  // update the display of nearest dyn obstacles
  if(dispNearDynObs == false)
    dispNearDynObs = true;
  else
    dispNearDynObs = false;

  return 1;
}

long MPC_HugrIfc::updateDispNearStObs(FXObject*,FXSelector,void*)
{
  // update the display of nearest dyn obstacles
  if(dispNearStObs == false)
    dispNearStObs = true;
  else
    dispNearStObs = false;

  return 1;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////// Other functions //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// void MPC_HugrIfc::drawInitPos()
// {
//   // verify that the init position is available
//   if(map != NULL && x1 >= 0 && x1 < map->getWidth() && x1 >= 0 && x1 < map->getHeight())
//     {
//       // create the drawing window
//       FXDCWindow dc(imageview);
      
//       // compute the init position coord in the image frame
//       int xF=0, yF=0;
//       coordMapToCoordFrame((int)x1, (int)y1, xF, yF);
      
//       // draw the init position on the map
//       dc.setForeground(initColor);
//       dc.drawEllipse(xF - 2, yF - 2, 4, 4);

//       cout << "init position : (" << xF << "," << yF << ")" << endl;
//     }
//   else
//     {
//       cout << "Can't draw init position" << endl;
//     }
// }

// void MPC_HugrIfc::drawInitDir()
// {
//   // verify that the init position is available
//   if(map != NULL && x1 >= 0 && x1 < map->getWidth() && x1 >= 0 && x1 < map->getHeight())
//     {
//       // create the drawing window
//       FXDCWindow dc(imageview);
      
//       // compute the init position coord in the image frame
//       int x1F=0, y1F=0;
//       coordMapToCoordFrame((int)x1, (int)y1, x1F, y1F);
      
//       int xTgtM = (int)(x1 + 20.0f * cos(theta1));
//       int yTgtM = (int)(y1 + 20.0f * sin(theta1));
      
//       // compute the vector of the init position direction in the image frame
//       int xTgtF=0, yTgtF=0;
//       coordMapToCoordFrame(xTgtM, yTgtM, xTgtF, yTgtF);
      
//       // draw the init direction on the map
//       dc.setForeground(initColor);
//       dc.drawLine(x1F, y1F, xTgtF, yTgtF);

//       cout << "init direction : (" << x1F << "," << y1F << ") -> (" << xTgtF << "," << yTgtF << ")" << endl;
//     }
//   else
//     {
//       cout << "Can't draw init direction" << endl;
//     }
// }

// void MPC_HugrIfc::drawInitCfg()
// {
//   drawInitPos();
//   drawInitDir();
// }

// void MPC_HugrIfc::drawGoalPos()
// {
//   // verify that the goal position is available
//   if(map != NULL && x2 >= 0 && x2 < map->getWidth() && x2 >= 0 && x2 < map->getHeight())
//     {
//       // create the drawing window
//       FXDCWindow dc(imageview);
      
//       // compute the init position coord in the image frame
//       int xF=0, yF=0;
//       coordMapToCoordFrame((int)x2, (int)y2, xF, yF);
      
//       // draw the init position on the map
//       dc.setForeground(goalColor);
//       dc.drawEllipse(xF - 2, yF - 2, 4, 4);

//       cout << "goal position : (" << xF << "," << yF << ")" << endl;
//     }
//   else
//     {
//       cout << "Can't draw goal position" << endl;
//     }
// }

// void MPC_HugrIfc::drawGoalDir()
// {
//   // verify that the goal position is available
//   if(map != NULL && x2 >= 0 && x2 < map->getWidth() && x2 >= 0 && x2 < map->getHeight())
//     {
//       // create the drawing window
//       FXDCWindow dc(imageview);
      
//       // compute the init position coord in the image frame
//       int x2F=0, y2F=0;
//       coordMapToCoordFrame((int)x2, (int)y2, x2F, y2F);
      
//       int xTgtM = (int)(x2 + 20.0f * cos(theta2));
//       int yTgtM = (int)(y2 + 20.0f * sin(theta2));
      
//       // compute the vector of the init position direction in the image frame
//       int xTgtF=0, yTgtF=0;
//       coordMapToCoordFrame(xTgtM, yTgtM, xTgtF, yTgtF);
      
//       // draw the init direction on the map
//       dc.setForeground(goalColor);
//       dc.drawLine(x2F, y2F, xTgtF, yTgtF);

//       cout << "goal direction : (" << x2F << "," << y2F << ") -> (" << xTgtF << "," << yTgtF << ")" << endl;
//     }
//   else
//     {
//       cout << "Can't draw goal direction" << endl;
//     }
// }

// void MPC_HugrIfc::drawGoalCfg()
// {
//   drawGoalPos();
//   drawGoalDir();
// }


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////// Map treatment functions ///////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 *  \brief display the current map in the viewer
 */
void MPC_HugrIfc::loadMapInViewer()
{
  FXImage *img=NULL;
  FXImage *old;

  img=new FXPPMImage(getApp(),NULL,IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);

  // Load the image
  FXFileStream stream;
  if(stream.open(samplingStEnvFilename,FXStreamLoad)){
    getApp()->beginWaitCursor();
    img->loadPixels(stream);
    stream.close();
    
    img->create();
    
    // rescaling the image
    int wtF = imageview->getWidth();
    int htF = imageview->getHeight();

    wtI = img->getWidth();
    htI = img->getHeight();

    double ratioW = ((double) wtI) / ((double) wtF);
    double ratioH = ((double) htI) / ((double) htF);

    if(ratioW > ratioH)
      {
	img->scale((FXint)(((double)img->getWidth())/ratioW), (FXint)(((double)img->getHeight())/ratioW));
      }
    else
      {
	img->scale((FXint)(((double)img->getWidth())/ratioH), (FXint)(((double)img->getHeight())/ratioH));
      }

    old=imageview->getImage();
    imageview->setImage(img);
    delete old;
    getApp()->endWaitCursor();
  }
}

// Load image
FXbool MPC_HugrIfc::loadimage(const FXString& file){
  FXString ext=FXPath::extension(file);
  FXImage *img=NULL;
  FXImage *old;

  // delete the old map
  if(map != NULL) delete map;
  map = new Map();

  if(comparecase(ext,"pgm")==0){
    img=new FXPPMImage(getApp(),NULL,IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
  }

  if(img==NULL){
    FXMessageBox::error(this,MBOX_OK,"Error Loading Image","Unsupported type: %s",ext.text());
    return FALSE;
  }

  // Load the map
  map->load(file.text());

  // Load the image
  FXFileStream stream;
  if(stream.open(file,FXStreamLoad)){
    getApp()->beginWaitCursor();
    img->loadPixels(stream);
    stream.close();

    img->create();

    // rescaling the image
    int wtF = imageview->getWidth();
    int htF = imageview->getHeight();

    wtI = img->getWidth();
    htI = img->getHeight();

    double ratioW = ((double) wtI) / ((double) wtF);
    double ratioH = ((double) htI) / ((double) htF);

    if(ratioW > ratioH)
      {
	img->scale((FXint)(((double)img->getWidth())/ratioW), (FXint)(((double)img->getHeight())/ratioW));
      }
    else
      {
	img->scale((FXint)(((double)img->getWidth())/ratioH), (FXint)(((double)img->getHeight())/ratioH));
      }

    old=imageview->getImage();
    imageview->setImage(img);
    delete old;
    getApp()->endWaitCursor();

    // display both configuration if they are defined
//     drawInitCfg();
//     drawGoalCfg();
    }
  return TRUE;
}

// convert frame coordinates to map coordinates
void MPC_HugrIfc::coordFrameToCoordMap(const int& xF, const int& yF, double& xM, double& yM)
{
  // get back the dimensions of the frame
  int wtF = imageview->getWidth();
  int htF = imageview->getHeight();

  double minX = rb_->env()->minX();
  double maxX = rb_->env()->maxX();
  double minY = rb_->env()->minY();
  double maxY = rb_->env()->maxY();

  // compute the ratio between frame and map
//   double ratioW = ((double) wtI) / ((double) wtF);
//   double ratioH = ((double) htI) / ((double) htF);
  double ratioW = (maxX-minX) / ((double) wtF);
  double ratioH = (maxY-minY) / ((double) htF);

  if(ratioW > ratioH)
    {
      // compute the shift on y
//       int shift = (int)((htF - ((double)htI)/ratioW)/2.0f);
      double shift = ((htF - (maxY-minY)/ratioW)/2.0f);

      // compute the coordinates in the map
      xM = minX + (xF * ratioW);
      //      yM = htI -1 - (int)((yF-shift) * ratioW);
      yM = maxY - ((yF-shift) * ratioW);
      return;
    }
  else
    {
      // compute the shift on x
      //      int shift = (int)((wtF - ((double)wtI)/ratioH)/2.0f);
      double shift = ((wtF - (maxX-minX)/ratioH)/2.0f);

      // compute the coordinates in the map
//       xM = (int)((xF-shift) * ratioH);
//       yM = htI -1 - (int)(yF * ratioH);
      xM = minX + ((xF-shift) * ratioH);
      yM = maxY - (yF * ratioH);
      return;
    }
}

// convert map coordinates to frame coordinates
void MPC_HugrIfc::coordMapToCoordFrame(const double& xM, const double& yM, int& xF, int& yF)
{
  // get back the dimensions of the frame
  int wtF = imageview->getWidth();
  int htF = imageview->getHeight();

  double minX = rb_->env()->minX();
  double maxX = rb_->env()->maxX();
  double minY = rb_->env()->minY();
  double maxY = rb_->env()->maxY();

  // compute the ratio between frame and map
//   double ratioW = ((double) wtI) / ((double) wtF);
//   double ratioH = ((double) htI) / ((double) htF);
  double ratioW = (maxX-minX) / ((double) wtF);
  double ratioH = (maxY-minY) / ((double) htF);

  if(ratioW > ratioH)
    {
      // compute the shift on y
//       int shift = (int)((htF - ((double)htI)/ratioW)/2.0f);
      double shift = ((htF - (maxY-minY)/ratioW)/2.0f);

      // compute the coordinates in the frame
//       xF = 1 + (int)(xM / ratioW);
//       yF = 1 + (int)(((double)(htI -1 - yM) / ratioW) + shift);
      xF = 1 + (int)((xM - minX) / ratioW);
      yF = 1 + (int)(((maxY - yM) / ratioW) + shift);
      return;
    }
  else
    {
      // compute the shift on x
      int shift = (int)((wtF - ((double)wtI)/ratioH)/2.0f);

      // compute the coordinates in the frame
//       xF = 1 + (int)(xM / ratioH) + shift;
//       yF = 1 + (int)(((double)(htI -1 - yM)) / ratioH);
      xF = 1 + (int)(((xM - minX) / ratioH) + shift);
      yF = 1 + (int)(((maxY - yM)) / ratioH);
      return;
    }
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////// OpenGL Viewer Drawing functions ///////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// update the view of the dynamic viewer mode
void MPC_HugrIfc::updDynamicView()
{
  // retrieve the old parameters of the viewer
  double oldEX = glcanvas->getEyeX();
  double oldEY = glcanvas->getEyeY();
  double oldEZ = glcanvas->getEyeZ();
  double oldCX = glcanvas->getCenterX();
  double oldCY = glcanvas->getCenterY();
  double oldCZ = glcanvas->getCenterZ();
  double oldUX = glcanvas->getUpX();
  double oldUY = glcanvas->getUpY();
  double oldUZ = glcanvas->getUpZ();

  // get the current time
  double time;
  if(rb_ != NULL) time = rb_->timer()->totalElapsedTime();
  else time = 0.0;

  // compute the displacement needed on the time axis
  double dt = time - oldCZ;

  // set the new parameters of the viewer
  this->glcanvas->init(oldEX, oldEY, oldEZ + dt, oldCX, oldCY, oldCZ + dt, oldUX, oldUY, oldUZ);  

  //this->drawScene();
}

// draw the robotic system
void MPC_HugrIfc::drawRobot()
{
  // check if the robotic system has been defined
  if(rb_ != NULL)
    {
      // get back the current position-time of the robot
      PositionXYTh pos = rb_->statet()->posXYTh();
      
      glPushMatrix();
      glTranslated(pos.x(),pos.y(),rb_->statet()->t());
      glRotated(radToDeg(pos.th()),0,0,1);
      this->rb_->toGL();
      glPopMatrix();
    }
}

// draw the trajectory followed by the robot
void MPC_HugrIfc::drawTraj()
{
  // check if a trajectory has been defined
  if(rb_ != NULL && rb_->traj() != NULL)
    {
      rb_->trajGL();
    }
}

// draw the bounds of the field
void MPC_HugrIfc::drawFieldBounds()
{
  const GLfloat blueMaterial[]={0.,0.,1.,1.};
  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);

  glLineWidth(1.0);

  double t;
  if(rb_ != NULL) t = rb_->timer()->totalElapsedTime();
  else t = 0;

  glBegin(GL_LINE_LOOP);
  glVertex3f(minFieldX,minFieldY,t);
  glVertex3f(minFieldX,maxFieldY,t);
  glVertex3f(maxFieldX,maxFieldY,t);
  glVertex3f(maxFieldX,minFieldY,t);
  glEnd();
}

// draw the obstacles of the environment and their forecast models
void MPC_HugrIfc::drawObstacles()
{
  // check if a robot has been defined
  if(rb_ != NULL)
    {
      rb_->envGL();
    }
}

// draw the controller specifications if needed
void MPC_HugrIfc::drawControllerSpec()
{
  // check if a robot has been defined
  if(rb_ != NULL)
    {
//       cout << "deb rb_->ctrlerGL()" << endl;
      rb_->ctrlerGL();
//       cout << "fin rb_->ctrlerGL()" << endl;
    }
}

//  draw all elements of the scene
void MPC_HugrIfc::drawScene()
{
  //boost::mutex::scoped_lock lock(mutexDrawScene);
//   //mutexDrawScene.lock();

  const GLfloat lightPosition[]={-20.,-100.,100.,1.};
  const GLfloat lightAmbient[]={.3f,.3f,.3f,1.};
  const GLfloat lightDiffuse[]={.9f,.9f,.9f,1.};
//   const GLfloat redMaterial[]={1.,0.,0.,1.};
  const GLfloat blueMaterial[]={0.,0.,1.,1.};
//   const GLfloat greenMaterial[]={0.,1.,0.,1.};
//   const GLfloat otherMaterial[]={1.,0.,1.,1.};

  GLdouble width = glcanvas->getWidth();
  GLdouble height = glcanvas->getHeight();
  GLdouble aspect = height>0 ? width/height : 1.0;

  glcanvas->makeCurrent();

  glViewport(0,0,glcanvas->getWidth(),glcanvas->getHeight());

  glClearColor(1.0,1.0,1.0,1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glEnable(GL_DEPTH_TEST);

  glDisable(GL_DITHER);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  if(!orthoView) gluPerspective(30.,aspect,1.,-1);
  else glOrtho(-0.6*(maxFieldX-minFieldX), 0.6*(maxFieldX-minFieldX), -0.6*(maxFieldY-minFieldY), 0.6*(maxFieldY-minFieldY), -1000.0, 10000.0);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  if(dynamicView)
    {
      this->updDynamicView();
    }

//   cout << "EyeX() : " << glcanvas->getEyeX() << "  EyeY() : " << glcanvas->getEyeY() << "  EyeZ() : " << glcanvas->getEyeZ()
//        << "  CenterX() : " << glcanvas->getCenterX() << "  CenterY() : " << glcanvas->getCenterY() << "  CenterZ() : " << glcanvas->getCenterZ()
//        << "  UpX() : " << glcanvas->getUpX() << "  UpY() : " << glcanvas->getUpY() << "  UpZ() : " << glcanvas->getUpZ() << endl;

  gluLookAt(glcanvas->getEyeX(),glcanvas->getEyeY(),glcanvas->getEyeZ(),
	    glcanvas->getCenterX(),glcanvas->getCenterY(),glcanvas->getCenterZ(),
	    glcanvas->getUpX(),glcanvas->getUpY(),glcanvas->getUpZ());

  glShadeModel(GL_SMOOTH);
  glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
  glLightfv(GL_LIGHT0, GL_AMBIENT, lightAmbient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);

  glMaterialfv(GL_FRONT, GL_AMBIENT, blueMaterial);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, blueMaterial);

  //glScalef(1.0,1.0,10.0);
  glScalef(1.0,1.0,1.0);

  // draw the robotic system
  drawRobot();
  // draw the trajectory followed by the robot
  drawTraj();
  // draw the boundaries of the field
  drawFieldBounds();
  // draw the obstacles of the environment and their forecase models
  drawObstacles();
  // draw the controller specifications if needed
  drawControllerSpec();

  // Swap if it is double-buffered
  if(glvisual->isDoubleBuffer()){
    glcanvas->swapBuffers();
  }
  
  // Make context non-current
  glcanvas->makeNonCurrent();

  //mutexDrawScene.unlock();
}

/**
 *  \brief center the openGL view on the scene
 */
void MPC_HugrIfc::centerView()
{
  // center the viewer on the environment
  double CX = 0.5*(minFieldX + maxFieldX);
  double CY = 0.5*(minFieldY + maxFieldY);
  double CZ;
  if(rb_ != NULL) CZ = rb_->timer()->totalElapsedTime();
  else CZ = 0;
  double EX = CX; 
  double EY = CY; 
  double EZ = CZ + 50.0;
  this->glcanvas->init(EX, EY, EZ, CX, CY, CZ, 0, 1, 0);
  
  this->drawScene();
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////// Environment management ///////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
 *  \brief load a static environment from an input file
 *  \param nomfic : the input filename
 *  \return 0 if the process succeed
 *         -1 if the file has not been found
 *         -2 if the localization has not been started
 *         -3 if the environment perception has not been initialized
 *         -4 if an unknown error occured
 *         -5 if no system has been initilialized
 */
int MPC_HugrIfc::loadStaticEnv(const char* filename)
{
  DRoadNode::setMaxDistObs(2.5);

  // check if a system has already been initialized
  if(rb_ == NULL) return -5;

  // load the static environment
  char* rmfilename = NULL;

  switch(optRoadmap)
    {
    case 0 : 
      rmfilename = NULL;
      rb_->setAutobuildRM(false);
      break;
    case 1 : 
      rmfilename = NULL;
      rb_->setAutobuildRM(true);
      break;
    case 2 : 
      rmfilename = ((char*)(RMfilename.text()));
      rb_->setAutobuildRM(false);
      break;
    }

  int res = rb_->addStaticEnv(filename,rmfilename);
  return res;
}

/**
 *  \brief load a roadmap from an input file
 *  \param nomfic : the input filename
 *  \return 0 if the process succeed
 *         -1 if the file has not been found
 *         -2 if the localization has not been started
 *         -3 if the environment perception has not been initialized
 *         -4 if the loading process failed
 *         -5 if no system has been initilialized
 */
int MPC_HugrIfc::loadRoadmap(const char* filename)
{
  // check if a system has already been initialized
  if(rb_ == NULL) return -5;

  int res = rb_->loadRoadmap(filename);
  if(res == 0) return 0;
  else return -4;
}

/**
 *  \brief save a roadmap into an output file
 *  \param nomfic : the input filename
 *  \return 0 if the process succeed
 *         -1 if the file has not been found
 *         -2 if the localization has not been started
 *         -3 if the environment perception has not been initialized
 *         -4 if the loading process failed
 *         -5 if no system has been initilialized
 */
int MPC_HugrIfc::saveRoadmap(const char* filename)
{
  // check if a system has already been initialized
  if(rb_ == NULL) return -5;

  int res = rb_->saveRoadmap(filename);
  if(res == 0) return 0;
  else return -4;
}

/**
 *  \brief sample the static environment to select easily the goal configuration
 *  \return the sampled map if the process succeed, NULL else
 */
Map* MPC_HugrIfc::sampleStaticEnv()
{
  return rb_->env()->sample(sampleX,sampleY);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////// Recording tools ///////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 *  \brief reinit the trajectory deviation recording
 */
void MPC_HugrIfc::reinitTrajDevRec()
{
//   this->deleteTrajDevRec();

  // append the current record to the list of records
  if(vecTrajSt != NULL) recordsTrajSt.push_back(vecTrajSt);
  if(vecLoc != NULL) recordsLoc.push_back(vecLoc);
  if(vecDev != NULL) recordsDev.push_back(vecDev);
  if(vecCtrlT != NULL) recordsCtrlT.push_back(vecCtrlT);
  
  vecTrajSt = new std::vector<StateTime*>;
  vecLoc = new std::vector<StateTime*>;
  vecDev = new std::vector<StateTime*>;
  vecCtrlT = new std::vector<ControlTime*>;
}

/**
 *  \brief delete the current trajectory record
 */
void MPC_HugrIfc::deleteTrajDevRec()
{
  if(vecTrajSt != NULL)
    {
      if(vecTrajSt->size() != 0)
	{
	  std::vector<StateTime*>::iterator it1;
	  for(it1 = vecTrajSt->begin(); it1 != vecTrajSt->end(); it1++)
	    {
	      if(*it1 != NULL) delete *it1;
	    }
	  vecTrajSt->clear();
	}
      delete vecTrajSt;
    }
  vecTrajSt = NULL;

  if(vecLoc != NULL)
    {
      if(vecLoc->size() != 0)
	{
	  std::vector<StateTime*>::iterator it2;
	  for(it2 = vecLoc->begin(); it2 != vecLoc->end(); it2++)
	    {
	      if(*it2 != NULL) delete *it2;
	    }
	  vecLoc->clear();
	}
      delete vecLoc;
    }
  vecLoc = NULL;

  if(vecDev != NULL)
    {
      if(vecDev->size() != 0)
	{
	  std::vector<StateTime*>::iterator it3;
	  for(it3 = vecDev->begin(); it3 != vecDev->end(); it3++)
	    {
	      if(*it3 != NULL) delete *it3;
	    }
      vecDev->clear();
	}
      delete vecDev;
    }
  vecDev = NULL;

  if(vecCtrlT != NULL)
    {
      if(vecCtrlT->size() != 0)
	{
	  std::vector<ControlTime*>::iterator it4;
	  for(it4 = vecCtrlT->begin(); it4 != vecCtrlT->end(); it4++)
	    {
	      if(*it4 != NULL) delete *it4;
	    }
	  vecCtrlT->clear();
	}
      delete vecCtrlT;
    }
  vecCtrlT = NULL;
}

/**
 *  \brief delete all the trajectory records
 */
void MPC_HugrIfc::deleteAllTrajRec()
{
  if(recordsTrajSt.size() != 0)
    {
      std::vector<std::vector<StateTime*>*>::iterator it;
      for(it = recordsTrajSt.begin(); it != recordsTrajSt.end(); it++)
	{
	  std::vector<StateTime*>* curTrajSt = *it;
	  if(curTrajSt->size() != 0)
	    {
	      std::vector<StateTime*>::iterator it1;
	      for(it1 = curTrajSt->begin(); it1 != curTrajSt->end(); it1++)
		{
		  if(*it1 != NULL) delete *it1;
		}
	      curTrajSt->clear();
	    }
	  delete curTrajSt;
	}
      recordsTrajSt.clear();
    }

  if(recordsLoc.size() != 0)
    {
      std::vector<std::vector<StateTime*>*>::iterator it;
      for(it = recordsLoc.begin(); it != recordsLoc.end(); it++)
	{
	  std::vector<StateTime*>* curLoc = *it;
	  if(curLoc->size() != 0)
	    {
	      std::vector<StateTime*>::iterator it1;
	      for(it1 = curLoc->begin(); it1 != curLoc->end(); it1++)
		{
		  if(*it1 != NULL) delete *it1;
		}
	      curLoc->clear();
	    }
	  delete curLoc;
	}
      recordsLoc.clear();
    }

  if(recordsDev.size() != 0)
    {
      std::vector<std::vector<StateTime*>*>::iterator it;
      for(it = recordsDev.begin(); it != recordsDev.end(); it++)
	{
	  std::vector<StateTime*>* curDev = *it;
	  if(curDev->size() != 0)
	    {
	      std::vector<StateTime*>::iterator it1;
	      for(it1 = curDev->begin(); it1 != curDev->end(); it1++)
		{
		  if(*it1 != NULL) delete *it1;
		}
	      curDev->clear();
	    }
	  delete curDev;
	}
      recordsDev.clear();
    }

  if(recordsCtrlT.size() != 0)
    {
      std::vector<std::vector<ControlTime*>*>::iterator it;
      for(it = recordsCtrlT.begin(); it != recordsCtrlT.end(); it++)
	{
	  std::vector<ControlTime*>* curCtrlT = *it;
	  if(curCtrlT->size() != 0)
	    {
	      std::vector<ControlTime*>::iterator it1;
	      for(it1 = curCtrlT->begin(); it1 != curCtrlT->end(); it1++)
		{
		  if(*it1 != NULL) delete *it1;
		}
	      curCtrlT->clear();
	    }
	  delete curCtrlT;
	}
      recordsCtrlT.clear();
    }
}

/**
 *  \brief record the position deviation during a trajectory following process
 */
void MPC_HugrIfc::recordTrajDev()
{
  // check if the robot has been initialized
  if(rb_ != NULL)
    {
      // initialize the robots state-times
      StateTime* trajSt = NULL;
      StateTime* loc = NULL;
      StateTime* dev = NULL;
      ControlTime* ctrlt = NULL;
	  switch(typeSys_)
	    {
	    case WHEELCHAIR_SYS : 
	      {
		trajSt = new WCStateT();
		loc = new WCStateT();
		dev = new WCStateT();
		ctrlt = new WCControlT();
		break;
	      }
	    case CARLIKE_SYS : 
	      {
		trajSt = new CLStateT();
		loc = new CLStateT();
		dev = new CLStateT();
		ctrlt = new CCControlT();
		break;
	      }
	    }

      // get back its current localization and trajectory deviation
	  rb_->computeCurDeviation(trajSt,loc,dev,ctrlt);
      cout << "record -> trajSt : "; trajSt->display(); cout << "  loc : "; loc->display(); cout << "  dev : "; dev->display(); cout << endl;

      // add each state-time in the appropriate vector
      vecTrajSt->push_back(trajSt);
      vecLoc->push_back(loc);
      vecDev->push_back(dev);
      vecCtrlT->push_back(ctrlt);
    }
}

/**
 *  \brief display the trajectory deviation recording
 */
void MPC_HugrIfc::displayTrajDevRec()
{
  // check if the deviation vector is not empty
  if(vecDev->size() == 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Deviation record displaying","Cannot display the record. It is empty.");
      return;
    }

  // create the required vector to display the gnuplot graph
  std::vector<double> vecX;
  std::vector<double> vecY;
  std::vector<double> vecT;
  std::vector<double> vecDist;

  int N = vecDev->size();
  double initT = 0;
  if(N != 0) initT = vecDev->at(0)->posXYT().t();

  for(int i=0; i<N; i++)
    {
      // get back the current position-time
      cout << "vecDev->at(i) : "; vecDev->at(i)->display();
      PositionXYT posXYT = vecDev->at(i)->posXYT();
      cout << "  posXYT : "; posXYT.display(); cout << endl;

      // fill the appropriate vectors
//       double x = posXYT.x();
//       double y = posXYT.y();
//       double t = posXYT.t();
      double x = ((WCStateT*)(vecDev->at(i)))->x();
      double y = ((WCStateT*)(vecDev->at(i)))->y();
      double t = ((WCStateT*)(vecDev->at(i)))->t() - initT;
      double dist = 0.7*sqrt(x*x + y*y);

      vecX.push_back(x);
      vecY.push_back(y);
      vecT.push_back(t);
      vecDist.push_back(dist);
    }

  // create the required vector to display the gnuplot graph
  std::vector<double> vecLX;
  std::vector<double> vecLY;
  std::vector<double> vecLV;
  std::vector<double> vecLAV;
  std::vector<double> vecLT;

  int N2 = vecLoc->size();
  for(int i=0; i<N2; i++)
    {
      // get back the current position-time
      PositionXYT posXYT = vecLoc->at(i)->posXYT();

      // fill the appropriate vectors
      vecLX.push_back(posXYT.x());
      vecLY.push_back(posXYT.y());
      vecLV.push_back(((WCStateT*)(vecLoc->at(i)))->s());
      vecLAV.push_back(((WCStateT*)(vecLoc->at(i)))->om());
      vecLT.push_back(posXYT.t());
    }

  // create the required vector to display the gnuplot graph
  std::vector<double> vecTX;
  std::vector<double> vecTY;
  std::vector<double> vecTV;
  std::vector<double> vecTAV;
  std::vector<double> vecTT;

  int N3 = vecTrajSt->size();
  for(int i=0; i<N3; i++)
    {
      // get back the current position-time
      PositionXYT posXYT = vecTrajSt->at(i)->posXYT();
      cout << "posXYT : "; posXYT.display(); cout << endl;

      // fill the appropriate vectors
      vecTX.push_back(posXYT.x());
      vecTY.push_back(posXYT.y());
      vecTV.push_back(((WCStateT*)(vecTrajSt->at(i)))->s());
      vecTAV.push_back(((WCStateT*)(vecTrajSt->at(i)))->om());
      vecTT.push_back(posXYT.t());
    }

  // create the control vectors
  std::vector<double> vecCV;
  std::vector<double> vecCAV;

  int N4 = vecCtrlT->size();
  for(int i=0; i<N4; i++)
    {
      vecCV.push_back(((WCControlT*)(vecCtrlT->at(i)))->v());
      vecCAV.push_back(((WCControlT*)(vecCtrlT->at(i)))->av());
    }  

  Gnuplot gTraj;
  gTraj.reset_plot();
  gTraj.cmd("set linestyle 1");
  gTraj.set_style("lines").plot_xyz(vecLX,vecLY,vecLT,"Localization");
  gTraj.set_style("lines").plot_xyz(vecTX,vecTY,vecTT,"Traj followed");

  Gnuplot gDev;
  gDev.reset_plot();
  gDev.cmd("set linestyle 1");
  gDev.set_style("lines").plot_xy(vecT,vecDist,"Position deviation (in meters)");
//   gDev.set_style("lines").plot_xy(vecX,vecT,"Position X");
//   gDev.set_style("lines").plot_xy(vecY,vecT,"Position Y");

  Gnuplot gVel;
  gVel.reset_plot();
  gVel.cmd("set linestyle 1");
  gVel.set_style("lines").plot_xy(vecT,vecTV,"Trajectory's speed");
//   gVel.set_style("lines").plot_xy(vecT,vecLV,"Robot's speed");
  gVel.set_style("lines").plot_xy(vecT,vecCV,"Robot's speed");

  Gnuplot gAVel;
  gAVel.reset_plot();
  gAVel.cmd("set linestyle 1");
  gAVel.set_style("lines").plot_xy(vecT,vecTAV,"Trajectory's ang speed");
//   gAVel.set_style("lines").plot_xy(vecT,vecLAV,"Robot's ang speed");
  gAVel.set_style("lines").plot_xy(vecT,vecCAV,"Robot's ang speed");

  wait_for_key();
  return;
}

/**
 *  \brief display all the trajectory records
 */
void MPC_HugrIfc::displayAllTrajRec()
{
  cout << "-3" << endl;

  // check if the trajectory records are not empty
  if(recordsDev.size() == 0)
    {
      FXMessageBox::error(this,MBOX_OK,"Deviation records displaying","Cannot display the records. They are empty.");
      return;
    }

  cout << "-2" << endl;

  // create the required vector to display the gnuplot graph
  std::vector<std::vector<double>* > vecDevX;
  std::vector<std::vector<double>* > vecDevY;
  std::vector<std::vector<double>* > vecDevT;
  std::vector<std::vector<double>* > vecDevDist;

  std::vector<std::vector<StateTime*>*>::iterator it;

  cout << "-1" << endl;

  int M1 = recordsDev.size();
//   for(it = recordsDev.begin(); it != recordsDev.end(); it++)
  for(int j=0; j<M1; j++)
    {      
//       std::vector<StateTime*>* curDev = *it;
      std::vector<StateTime*>* curDev = recordsDev.at(j);
      int N = curDev->size();
      double initT = 0;
      if(N != 0) initT = curDev->at(0)->posXYT().t();
      
      std::vector<double>* curDevX = new std::vector<double>;
      std::vector<double>* curDevY = new std::vector<double>;
      std::vector<double>* curDevT = new std::vector<double>;
      std::vector<double>* curDevDist = new std::vector<double>;

      for(int i=0; i<N; i++)
	{
	  // get back the current position-time
	  PositionXYT posXYT = curDev->at(i)->posXYT();
	  
	  // fill the appropriate vectors
	  double x = ((WCStateT*)(curDev->at(i)))->x();
	  double y = ((WCStateT*)(curDev->at(i)))->y();
	  double t = ((WCStateT*)(curDev->at(i)))->t() - initT;
	  double dist = 1.0*sqrt(x*x + y*y);
	  
	  curDevX->push_back(x);
	  curDevY->push_back(y);
	  curDevT->push_back(t);
	  curDevDist->push_back(dist);
	}
      
      vecDevX.push_back(curDevX);
      vecDevY.push_back(curDevY);
      vecDevT.push_back(curDevT);
      vecDevDist.push_back(curDevDist);
    }

  // create the required vector to display the gnuplot graph
  std::vector<std::vector<double>*> vecLX;
  std::vector<std::vector<double>*> vecLY;
  std::vector<std::vector<double>*> vecLV;
  std::vector<std::vector<double>*> vecLAV;
  std::vector<std::vector<double>*> vecLT;

  std::vector<std::vector<StateTime*>*>::iterator it2;

//   cout << "recordsLoc.size() : " << recordsLoc.size() << endl;

  for(it2 = recordsLoc.begin(); it2 != recordsLoc.end(); it2++)
    {
      std::vector<StateTime*>* curLoc = *it2;
      int N2 = curLoc->size();

      std::vector<double>* curLX = new std::vector<double>;
      std::vector<double>* curLY = new std::vector<double>;
      std::vector<double>* curLV = new std::vector<double>;
      std::vector<double>* curLAV = new std::vector<double>;
      std::vector<double>* curLT = new std::vector<double>;

      for(int i=0; i<N2; i++)
	{
	  // get back the current position-time
	  PositionXYT posXYT = curLoc->at(i)->posXYT();
	  
	  // fill the appropriate vectors
	  curLX->push_back(posXYT.x());
	  curLY->push_back(posXYT.y());
	  curLV->push_back(((WCStateT*)(curLoc->at(i)))->s());
	  curLAV->push_back(((WCStateT*)(curLoc->at(i)))->om());
	  curLT->push_back(posXYT.t());
	}

      vecLX.push_back(curLX);
      vecLY.push_back(curLY);
      vecLV.push_back(curLV);
      vecLAV.push_back(curLAV);
      vecLT.push_back(curLT);
    }


  // create the required vector to display the gnuplot graph
  std::vector<std::vector<double>*> vecTX;
  std::vector<std::vector<double>*> vecTY;
  std::vector<std::vector<double>*> vecTV;
  std::vector<std::vector<double>*> vecTAV;
  std::vector<std::vector<double>*> vecTT;

  std::vector<std::vector<StateTime*>*>::iterator it3;

  for(it3 = recordsTrajSt.begin(); it3 != recordsTrajSt.end(); it3++)
    {
      std::vector<StateTime*>* curTrajSt = *it3;
      int N3 = curTrajSt->size();

      std::vector<double>* curTX = new std::vector<double>;
      std::vector<double>* curTY = new std::vector<double>;
      std::vector<double>* curTV = new std::vector<double>;
      std::vector<double>* curTAV = new std::vector<double>;
      std::vector<double>* curTT = new std::vector<double>;

      for(int i=0; i<N3; i++)
	{
	  // get back the current position-time
	  PositionXYT posXYT = curTrajSt->at(i)->posXYT();
	  
	  // fill the appropriate vectors
	  curTX->push_back(posXYT.x());
	  curTY->push_back(posXYT.y());
	  curTV->push_back(((WCStateT*)(curTrajSt->at(i)))->s());
	  curTAV->push_back(((WCStateT*)(curTrajSt->at(i)))->om());
	  curTT->push_back(posXYT.t());
	}

      vecTX.push_back(curTX);
      vecTY.push_back(curTY);
      vecTV.push_back(curTV);
      vecTAV.push_back(curTAV);
      vecTT.push_back(curTT);
    }


  // create the control vectors
  std::vector<std::vector<double>*> vecCV;
  std::vector<std::vector<double>*> vecCAV;

  std::vector<std::vector<ControlTime*>*>::iterator it4;
  
  for(it4 = recordsCtrlT.begin(); it4 != recordsCtrlT.end(); it4++)
    {
      std::vector<ControlTime*>* curCtrlT = *it4;
      int N4 = curCtrlT->size();

      std::vector<double>* curCV = new std::vector<double>;
      std::vector<double>* curCAV = new std::vector<double>;

      for(int i=0; i<N4; i++)
	{
	  curCV->push_back(((WCControlT*)(curCtrlT->at(i)))->v());
	  curCAV->push_back(((WCControlT*)(curCtrlT->at(i)))->av());
	}
      
      vecCV.push_back(curCV);
      vecCAV.push_back(curCAV);
    }

  Gnuplot gTraj;
  gTraj.reset_plot();
  gTraj.cmd("set linestyle 1");
  for(int i=0; i<vecLX.size(); i++)
    {
      std::vector<double>* curLX = vecLX.at(i);
      std::vector<double>* curLY = vecLY.at(i);
      std::vector<double>* curLT = vecLT.at(i);
      std::vector<double>* curTX = vecTX.at(i);
      std::vector<double>* curTY = vecTY.at(i);
      std::vector<double>* curTT = vecTT.at(i);

      if(curLX->size() != 0 && curLY->size() != 0 && curLT->size() != 0) gTraj.set_style("lines").plot_xyz(*curLX,*curLY,*curLT,"Localization");
      if(curTX->size() != 0 && curTY->size() != 0 && curTT->size() != 0) gTraj.set_style("lines").plot_xyz(*curTX,*curTY,*curTT,"Traj followed");
    }

  Gnuplot gDev;
  gDev.reset_plot();
  gDev.cmd("set linestyle 1");
  for(int i=0; i<vecDevDist.size(); i++)
    {
      std::vector<double>* curDevT = vecDevT.at(i);
      std::vector<double>* curDevDist = vecDevDist.at(i);

      if(curDevT->size() != 0 && curDevDist->size() != 0) gDev.set_style("lines").plot_xy(*curDevT,*curDevDist,"Position deviation (in meters)");
    }

  Gnuplot gVel;
  gVel.reset_plot();
  gVel.cmd("set linestyle 1");
  for(int i=0; i<vecCV.size(); i++)
    {
      std::vector<double>* curT = vecDevT.at(i);
      std::vector<double>* curTV = vecTV.at(i);

      if(curT->size() != 0 && curTV->size() != 0)
	{
	  if(i == 0) gVel.set_style("lines").plot_xy(*curT,*curTV,"Trajectory's speed");
	  else gVel.set_style("lines").plot_xy(*curT,*curTV,"");
	}
    }

  Gnuplot gVel2;
  gVel2.reset_plot();
  gVel2.cmd("set linestyle 1");
  for(int i=0; i<vecCV.size(); i++)
    {
      std::vector<double>* curT = vecDevT.at(i);
      std::vector<double>* curCV = vecCV.at(i);

      if(curT->size() != 0 && curCV->size() != 0)
	{
	  if(i == 0) gVel2.set_style("lines").plot_xy(*curT,*curCV,"Robot's speed");
	  else gVel2.set_style("lines").plot_xy(*curT,*curCV,"");
	}
    }


  Gnuplot gAVel;
  gAVel.reset_plot();
  gAVel.cmd("set linestyle 1");
  for(int i=0; i<vecCAV.size(); i++)
    {
      std::vector<double>* curT = vecDevT.at(i);
      std::vector<double>* curTAV = vecTAV.at(i);

      if(curT->size() != 0 && curTAV->size() != 0)
	{
	  if(i == 0) gAVel.set_style("lines").plot_xy(*curT,*curTAV,"Trajectory's ang speed");
	  else gAVel.set_style("lines").plot_xy(*curT,*curTAV,"");
	}
    }

  Gnuplot gAVel2;
  gAVel2.reset_plot();
  gAVel2.cmd("set linestyle 1");
  for(int i=0; i<vecCAV.size(); i++)
    {
      std::vector<double>* curT = vecDevT.at(i);
      std::vector<double>* curCAV = vecCAV.at(i);

      if(curT->size() != 0 && curCAV->size() != 0)
	{
	  if(i == 0) gAVel2.set_style("lines").plot_xy(*curT,*curCAV,"Robot's ang speed");
	  else gAVel2.set_style("lines").plot_xy(*curT,*curCAV,"");
	}
    }

  wait_for_key();

  //////////////////////////////////////////////////////////////////////////////////////////////
  ////////////// All displayed vectors must be deleted !!!!!!!! TODO!!!!!!!! ///////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////

  return;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////// GUI functions ////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// update the current connected system label
void MPC_HugrIfc::updateCurSystemLabel()
{
  switch(typeSys_)
    {
    case WHEELCHAIR_SYS : 
      currentSystemLabel->setText("System: Wheelchair");
      break;
    case CARLIKE_SYS : 
      currentSystemLabel->setText("System: Cycab");
      break;
    default : 
      currentSystemLabel->setText("System: No connected system");
      break;
    }
}

// update the current connected planner label
void MPC_HugrIfc::updateCurPlannerLabel()
{
  if(rb_ != NULL)
    {
      switch(rb_->plannerType())
	{
	case ROADMAP_TP : 
	  currentPlannerLabel->setText("Planner: Roadmap trajectory planner");
	  break;
	case PVD_TP : 
	  currentPlannerLabel->setText("Planner: PVD trajectory planner");
	  break;
	case SIMPLE_TP : 
	  currentPlannerLabel->setText("Planner: Simple trajectory planner");
	  break;
	case ND_TP : 
	default : 
	  currentPlannerLabel->setText("Planner: No connected planner");
	  break;
	}
    }
  else
    {
      currentPlannerLabel->setText("Planner: No connected planner");
    }
}

// update the current connected controller label
void MPC_HugrIfc::updateCurControllerLabel()
{
  if(rb_ != NULL)
    {
      switch(rb_->ctrlerType())
	{
	case SIMPLE_CTRLER : 
	  currentControllerLabel->setText("Controller: Manual controller");
	  break;
	case TRAJFOLLOW_CTRLER : 
	  currentControllerLabel->setText("Controller: Trajectory follower");
	  break;
	case TEDDY_CTRLER : 
	  currentControllerLabel->setText("Controller: Trajectory deformer");
	  break;
	case ND_CTRLER : 
	default : 
	  currentControllerLabel->setText("Controller: No connected controller");
	  break;
	}
    }
  else
    {
      currentControllerLabel->setText("Controller: No connected controller");
    }
}

/**
 *  \brief init the gui from an input configuration file
 *  \param filename : the input filename
 */
void MPC_HugrIfc::initHugrIfcFromFile(const char* filename)
{
  string WC_namePos_string;
  string WC_nameCmd_string;

  string plannerType_string;
  string ctrlerType_string;

  string keyboard_string;

  string stEnvFilename_string;
  string optRoadmap_string;
  string RMfilename_string;

  po::options_description config_files("Configuration Files"); // Configuration files
  config_files.add_options() 
    ("HugrIfc.wc_x0", po::value<double>(&WC_x), "initial X position of the wheelchair")
    ("HugrIfc.wc_y0", po::value<double>(&WC_y), "initial Y position of the wheelchair")
    ("HugrIfc.wc_th0", po::value<double>(&WC_th), "initial heading of the wheelchair")

    ("HugrIfc.cc_x0", po::value<double>(&CC_x), "initial X position of the cycab")
    ("HugrIfc.cc_y0", po::value<double>(&CC_y), "initial Y position of the cycab")
    ("HugrIfc.cc_th0", po::value<double>(&CC_th), "initial heading of the cycab")

    ("HugrIfc.wc_dflt_pose_name", po::value<string>(&WC_namePos_string), "default variable's name of the wheelchair position in the hugr store")
    ("HugrIfc.wc_dflt_cmd_name", po::value<string>(&WC_nameCmd_string), "default variable's name of the wheelchair command in the hugr store")

    ("HugrIfc.planner_dflt_mode", po::value<string>(&plannerType_string), "planner default mode (options: simple_planner, roadmap_planner)")
    ("HugrIfc.planner_dt", po::value<double>(&planner_DT), "planner time discretization")

    ("HugrIfc.splanner_fixed_time_opt", po::value<bool>((bool*)(&fixed_T)), "true if the final time of the simple planner must be fixed")
    ("HugrIfc.splanner_fixed_time_val", po::value<double>(&planner_goalT), "value of the fixed final time")

    ("HugrIfc.rmplanner_coefvel", po::value<double>(&planner_CoefSpd), "coef in [0;1] determining the ratio of the maximal speed used for the trajectory planning")
    ("HugrIfc.rmplanner_coefacc", po::value<double>(&planner_CoefAcc), "coef in [0;1] determining the ratio of the maximal acc used for the trajectory planning")

    ("HugrIfc.ctrler_dflt_mode", po::value<string>(&ctrlerType_string), "controller default mode (options: manual_ctrl, traj_follow, traj_deform)")
    ("HugrIfc.ctrler_dt", po::value<double>(&dtCtrl), "time interval between two control requests")


    ("HugrIfc.mctrl_coefvel", po::value<double>(&Coef_Ctrl_v), "velocity control coefficient of the wheelchair/cycab (in [0;1])")
    ("HugrIfc.mctrl_coefavel", po::value<double>(&Coef_Ctrl_avph), "angular velocity / steering control coefficient (in [0;1])")

    ("HugrIfc.mctrl_keyboard", po::value<string>(&keyboard_string), "keyboard used for the manual control (options : azerty, qwerty)")

    ("HugrIfc.k_int", po::value<double>(&k_int), "internal forces coefficient")
    ("HugrIfc.k_ext", po::value<double>(&k_ext), "external forces coefficient (dynamic obstacles)")
    ("HugrIfc.k_stext", po::value<double>(&k_stext), "external forces coefficient (static obstacles)")
    ("HugrIfc.dyn_med", po::value<double>(&dyn_med), "maximal effective distance (dynamic obstacles)")
    ("HugrIfc.st_med", po::value<double>(&st_med), "maximal effective distance (static obstacles)")
    ("HugrIfc.tmp_med", po::value<double>(&tmp_med), "maximal temporal effective distance")

    ("HugrIfc.static_env_filename", po::value<string>(&stEnvFilename_string), "filename of the static environment to load")
    ("HugrIfc.roadmap_build_mode", po::value<string>(&optRoadmap_string), "roadmap building mode (options: no_rm, autobuild_rm, loadfromfile_rm)")
    ("HugrIfc.roadmap_filename", po::value<string>(&RMfilename_string), "filename of the roadmap to load")
    ("HugrIfc.disp_roadmap", po::value<bool>((bool*)(&dispRoadmap)), "true if the roadmap must be displayed on the viewer")
//   //  bool                  autoBuildRoadmap;         // true if a roadmap must be automatically built when a static env is loaded  

    ("HugrIfc.dt_loc", po::value<double>(&dtLoc), "time interval between two localization requests")

    ("HugrIfc.filtering_dist", po::value<double>(&distFilter), "maximal distance from static obstacles to consider the dynamic obstacles perceived as noise")
    ("HugrIfc.security_dist", po::value<double>(&securDist), "security radius kept around dynamic obstacles")

    ("HugrIfc.dyn_view", po::value<bool>((bool*)(&dynamicView)), "true if the view move along the time")

    ("HugrIfc.record_traj", po::value<bool>((bool*)(&recordTraj)), "true if the trajectory profiles and deviation must be recorded")

    ("HugrIfc.ctrlerDebugLevel", po::value<int>(&ctrlerDebugLevel), "controller debug level. must be activated to display all forces (options: 0 -> disabled, 1 -> enabled)")

    ("HugrIfc.disp_fext", po::value<bool>((bool*)(&dispFExt)), "true if the external forces (due to dynamic obstacles) must be displayed")
    ("HugrIfc.disp_stfext", po::value<bool>((bool*)(&dispStFExt)), "true if the external forces (due to static obstacles) must be displayed")
    ("HugrIfc.disp_near_dynobs", po::value<bool>((bool*)(&dispNearDynObs)), "true if the nearest dynamic obstacle positions must be displayed")
    ("HugrIfc.disp_near_stobs", po::value<bool>((bool*)(&dispNearStObs)), "true if the nearest static obstacle positions must be displayed")
    ;

  po::variables_map vm_config;
  fs::path cfg_path(fs::initial_path());
  cfg_path = fs::system_complete(fs::path(filename,fs::native));
  string cfg_string = cfg_path.native_file_string();
  if ( !fs::exists( cfg_path ) )
    {
      cout << "Unable to load the hugr gui default parameters. Configuration file not found" << endl;
      return;
    }
  // Open configuration file
  ifstream ifs((const char *)cfg_string.c_str());
  po::store(parse_config_file(ifs, config_files), vm_config);
  notify(vm_config);

  // post initialization
  WC_namePos = FXString(WC_namePos_string.c_str());
  WC_nameCmd = FXString(WC_nameCmd_string.c_str());

  if(strcmp(plannerType_string.c_str(),"simple_planner") == 0) plannerType = SIMPLE_TP;
  else if(strcmp(plannerType_string.c_str(),"roadmap_planner") == 0) plannerType = ROADMAP_TP;
  else 
    {
      cerr << "MPC_HugrIfc::initHugrIfcFromFile : planner_dflt_mode -> Unknow option" << endl;
      exit(1);
    }

  if(strcmp(ctrlerType_string.c_str(),"manual_ctrl") == 0) ctrlerType = 0;
  else if(strcmp(ctrlerType_string.c_str(),"traj_follow") == 0) ctrlerType = 1;
  else if(strcmp(ctrlerType_string.c_str(),"traj_deform") == 0) ctrlerType = 2;
  else 
    {
      cerr << "MPC_HugrIfc::initHugrIfcFromFile : ctrler_dflt_mode -> Unknow option" << endl;
      exit(1);
    }

  if(strcmp(keyboard_string.c_str(),"azerty") == 0) { cerr << "default : azerty mode ! no switch available" << endl; }
  else if(strcmp(keyboard_string.c_str(),"qwerty") == 0) { cerr << "qwerty mode not implemented yet! default mode azerty" << endl; }
  else 
    {
      cerr << "MPC_HugrIfc::initHugrIfcFromFile : keyboard_string -> Unknow option" << endl;
      exit(1);
    }

  stEnvFilename = FXString(stEnvFilename_string.c_str());
  RMfilename= FXString(RMfilename_string.c_str());

  if(strcmp(optRoadmap_string.c_str(),"no_rm") == 0) optRoadmap = 0;
  else if(strcmp(optRoadmap_string.c_str(),"autobuild_rm") == 0) optRoadmap = 1;
  else if(strcmp(optRoadmap_string.c_str(),"loadfromfile_rm") == 0) optRoadmap = 2;
  else 
    {
      cerr << "MPC_HugrIfc::initHugrIfcFromFile : roadmap_build_mode -> Unknow option" << endl;
      exit(1);
    }
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////// Main ///////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Here we begin
int main(int argc,char *argv[])
{  
  // Make application
  FXApp application("Hugr MPC Ifc","Hugr MPC");

  // Open the display
  application.init(argc,argv);
  
  // Make window
  new MPC_HugrIfc(&application);
  
  // Create the application's windows
  application.create();
  
  // Run the application
  return application.run();
}
