#ifndef MPC_HUGR_CCTRAJFOLLOW_HPP
#define MPC_HUGR_CCTRAJFOLLOW_HPP

#include "hugr/hugrController/inc/Controller.hpp"
#include "trajFollow/inc/CCTrajFollow.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /************* Class Hugr_CCTrajFollow ************/
                                    /**************************************************/

  /** \class Hugr_CCTrajFollow
   * \brief Trajectory following process to control a cycab thanks to the hugr store
   */
  class Hugr_CCTrajFollow : public Controller, public CCTrajFollow
  {
  protected : 
    double curTimeOnTraj_;                /**< the current time of the supposed reached waypoint of the trajectory */
    CLStateT curRbST_;                    /**< last localization of the robot perceived (time relative to the trajectory followed) */
    CLStateT absRbST_;                    /**< last localization of the robot perceived (absolute time registered by the robot timer) */

    /********** Constructors and destructors **********/
  protected : 
    /**
     *  \brief default constructor
     */
    Hugr_CCTrajFollow();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the system constraints of the cycab
     *  \param traj : the trajectory to follow
     *  \param store : the hugr store on which to send the control data
     *  \param name : the name of the cycab controlled
     *  \param mutexLoc : the mutex used to prevent localization conflicts
     *  \param mutexCtrl : the mutex used to prevent control conflicts
     *  \param mutexTraj : the mutex used to prevent trajectory conflicts
     *  \param mutexEnv : the mutex used to prevent environment conflicts
     */
    Hugr_CCTrajFollow(CarLikeSys* sys,CLSTraj* traj,
		      hugr::Store* store,const std::string& name,
		      boost::mutex* mutexLoc,boost::mutex* mutexCtrl,
		      boost::mutex* mutexTraj,boost::mutex* mutexEnv);

    /**
     *  \brief destructor
     */
    virtual ~Hugr_CCTrajFollow();


    //   /***************** Initialization *****************/

    // public : 
    //   /**
    //    *  \brief initialize the controller execution
    //    */
    //   virtual void initExec();


    /************** Getters and setters ***************/

  public : 
    /**
     *  \brief getter : return the current control-time applied
     *  \return the current control-time applied
     */
    inline const CCControlT* controlt() const { return ((CCControlT*)(this->controlt_)); }

    //   /**
    //    *  \brief return the time of the last internal update
    //    *  \return time of the last internal update
    //    */
    //   virtual double lastTimeUp() const;

    /**
     *  \brief return the current time of the supposed reached waypoint of the trajectory
     *  \return the current time of the supposed reached waypoint of the trajectory
     */
    inline virtual double curTimeOnTraj() const { return this->curTimeOnTraj_; }

    /**
     *  \brief return the state of the trajectory followed (if exists) at absolute time t
     *  \param t : the input absolute time
     *  \param res : the resulting state of the trajectory followed at absolute time t if it exists, 0 else
     */
    virtual void getStateAtAbsTime(double t,StateTime& res) const;

    /******** hugr store connection functions *********/

  public : 
    /**
     *  \brief connect the controller to the hugr store
     *  \int return 0 if the connection to the hugr store has been correctly established, -1 else
     */
    virtual int connect();


    /********************* Actions ********************/

  protected : 
    /**
     *  \brief compute the control to apply from the time t during a time dt
     *  \param t : the initial time of application of the control
     *  \param dt : the duration of the application of the control
     *
     *  The resulting control is put in controlt_ to be send  
     */
    virtual void computeControl(double t,double dt);

    /**
     *  \brief send a control to the hugr store
     *  \return 0 if the control has been sent successfully, -1 else
     */
    virtual int sendControl() const;

    /**
     *  \brief update the localization of the trajectory follower
     *  \param curST : the current state-time of the robot (given at absolute time)
     *
     *  Modify the time of the input state to take into account to relative time to the trajectory followed
     */
    virtual void updateStateT(const StateTime& curST);

    /**
     *  \brief reinit the controller
     */
    virtual void reinit();
  };

};
#endif // MPC_HUGR_CCTRAJFOLLOW_HPP
