#ifndef MPC_HUGR_WC_TFD_HPP
#define MPC_HUGR_WC_TFD_HPP

#include "hugr/hugrController/inc/Hugr_WCTrajFollow.hpp"
#include "hugr/hugrDeformer/inc/Hugr_WCTrajDeform.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class Hugr_WC_TFD ***************/
                                    /**************************************************/

  /** \class Hugr_WC_TFD
   * \brief Trajectory following process to control a wheelchair coupled with a trajectory deformer
   */
  class Hugr_WC_TFD : public Hugr_WCTrajFollow
  {
  protected : 
    Hugr_WCTrajDeform* deformer_;                     /**< the trajectory deformer used */

    /********** Constructors and destructors **********/
  protected : 
    /**
     *  \brief default constructor
     */
    Hugr_WC_TFD();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the system constraints of the wheelchair
     *  \param env : the environment in which the system evolve
     *  \param traj : the trajectory to follow
     *  \param store : the hugr store on which to send the control data
     *  \param name : the name of the cycab controlled
     *  \param mutexLoc : the mutex used to prevent localization conflicts
     *  \param mutexCtrl : the mutex used to prevent control conflicts
     */
    Hugr_WC_TFD(WheelChairSys* sys,HEnvironment* env,WCSTraj* traj,hugr::Store* store,const std::string& name,
		boost::mutex* mutexLoc,boost::mutex* mutexCtrl,boost::mutex* mutexTraj,boost::mutex* mutexEnv);

    /**
     *  \brief destructor
     */
    virtual ~Hugr_WC_TFD();
  
    /************** Getters and setters ***************/

    /**
     *  \brief return the debug level used
     *  \return the debug level used
     */
    virtual void setDebug(int dbg);

    /**
     *  \brief set the deformation parameters
     *  \param k_int : internal forces coefficient
     *  \param k_ext : external forces coefficient (dynamic obstacles)
     *  \param k_stext : external forces coefficient (static obstacles)
     *  \param dyn_med : maximal effective distance (dynamic obstacles)
     *  \param st_med : maximal effective distance (static obstacles)
     *  \param dyn_med : maximal temporal effective distance
     */
    void setTeddyParams(double k_int,double k_ext,double k_stext,double dyn_med,double st_med,double tmp_med);

    /********************* Actions ********************/

  protected : 
    /**
     *  \brief execute the control loop
     */
    void loop();

  public : 
    /**
     *  \brief start the controller
     *  \return 0 if the starting process succeed, -1 else
     */
    int start();

    /**
     *  \brief stop the controller
     *  \return 0 if the stoping process succeed, -1 else
     */
    int stop();

  protected : 
    /**
     *  \brief update the trajectory to follow with the new deformed trajectory
     */
    void updateTrajToFollow();

    /*************** Display functions ****************/

  public : 
    /**
     *  \brief update the trajectory deformation process visualization
     *  \param dispFExt : true if the external forces (due to dynamic obstacles) must be displayed
     *  \param dispStFExt : true if the external forces (due to static obstacles) must be displayed
     *  \param dispNearDynObs : true if the position of the nearest dynamic obstacles must be displayed
     *  \param dispNearStObs : true if the position of the nearest static obstacles must be displayed
     */
    void updateTeddyView(bool dispFExt,bool dispStFExt,bool dispNearDynObs,bool dispNearStObs);

    /**
     *  \label get back informations about the deformation process
     *  \param nbDeform : the number of deformations proceeded
     *  \param avgDeformDur : the average deformation duration
     *  \param DPS : the average number of deformations per second
     */
    void getDeformInfo(int& nbDeform,double& avgDeformDur,double& DPS);

    /**
     *  \brief draw the controller features in an openGL viewer
     */
    virtual void toGL() const;

//     /**************** Signal functions ****************/

//     /**
//      *  \brief handle segmentation faults
//      */
//     void handleSegFault(int param);
  };

};
#endif // MPC_HUGR_WC_TFD_HPP
