#include "Controller.hpp"
using namespace mpc;

#include "WheelchairSimulator.hpp"
#include "CycabStructs.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <utime.h>

#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /**************** Class Controller ****************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
Controller::Controller()
{
  this->dt_ = 0.05;
  this->dtToWait_ = -1;
  this->controlt_ = new WCControlT();

  this->ctrlerSt_ = CTRLER_DISCONNECTED;
  this->ctrlerType_ = ND_CTRLER;

  this->store_ = NULL;
  this->controlId_ = -1;

  this->nameRb_ = std::string("");

  this->threadCtrl_ = NULL;
  this->mutexLoc_ = NULL;

  this->debug_ = 0;
}

/*
 *  \brief init constructor
 */
Controller::Controller(hugr::Store* store,const std::string& name,
		       boost::mutex* mutexLoc,boost::mutex* mutexCtrl,
		       boost::mutex* mutexTraj,boost::mutex* mutexEnv)
{
  this->dt_ = 0.05;
  this->dtToWait_ = -1;
  this->controlt_ = new WCControlT();

  this->ctrlerSt_ = CTRLER_DISCONNECTED;

  this->store_ = store;
  this->controlId_ = -1;

  this->nameRb_ = std::string(name);
  this->commandName_ = std::string(name);

  this->threadCtrl_ = NULL;
  this->mutexLoc_ = mutexLoc;
  this->mutexCtrl_ = mutexCtrl;
  this->mutexTraj_ = mutexTraj;
  this->mutexEnv_ = mutexEnv;

  this->debug_ = 0;
}

/*
 *  \brief destructor
 */
Controller::~Controller()
{
  if(this->threadCtrl_ != NULL) delete this->threadCtrl_;
}


/********************* Actions ********************/

/*
 *  \brief execute the control loop
 */
void Controller::loop()
{
  cout << "Controller::loop" << endl;
  while(ctrlerSt_ == CTRLER_STARTED)
    {
      // compute the control to apply from the time t during a time dt
      this->computeControl(timer_.totalElapsedTime(),dt_);
//       cout << "Controller::dt_ : " << dt_ << endl;

      // send the control to the hugr store
      this->sendControl();

      // compute the elapsed time since the last control sent
      double elapsedT = timer_.elapsedTime();

      if(dtToWait_ == -1)
	{
	  dtToWait_ = dt_ - elapsedT;
	  if(dtToWait_ > 0) usleep((int)(1000000.0*(dtToWait_)));
	  else usleep((int)(1000000.0*(dt_)));
	}
      else
	{
	  dtToWait_ = (dtToWait_ - dt_ + elapsedT);
	  if(dtToWait_ > 0) usleep((int)(1000000.0*(dtToWait_)));
	  else usleep((int)(1000000.0*(dt_)));
	}
    }
}

/*
 *  \brief start the timer
 *  \return 0 if the starting process succeed, -1 else
 */
int Controller::start()
{
  switch(ctrlerSt_)
    {
    case CTRLER_NULL : 
    case CTRLER_DISCONNECTED : 
      {
	cerr << "Controller::start : Warning. The controller is disconnected." << endl;
	return -1;
      }
    case CTRLER_STOPPED : 
      {
// 	cout << "start :-)" << endl;
	ctrlerSt_ = CTRLER_STARTED;
// 	cout << "time avant start : " << timer_.totalElapsedTime() << endl;
	timer_.start();
// 	cout << "time apres start : " << timer_.totalElapsedTime() << endl;
	threadCtrl_ = new boost::thread(boost::bind(&Controller::loop,this));
// 	cout << "time apres thread : " << timer_.totalElapsedTime() << endl;
	return 0;
      }
    case CTRLER_STARTED : 
      {
	// nothing to do
	cerr << "Controller::start : Controller already started" << endl;
	return 0;
      }
    default : 
      {
	cerr << "Controller::start : Unknown controller status" << endl;
	return -1;
      }
    }
}

/*
 *  \brief stop the timer
 *  \return 0 if the stoping process succeed, -1 else
 */
int Controller::stop()
{
  cout << "SSSSSTTTTTTTTTOOOOOOOOOOOOOOPPPPPPPP" << endl;
  cout << "ctrlerSt_ : " << ctrlerSt_ << endl;

  switch(ctrlerSt_)
    {
    case CTRLER_NULL : 
    case CTRLER_DISCONNECTED : 
      {
	cerr << "Controller::stop : Warning. The controller is disconnected." << endl;
	return -1;
      }
    case CTRLER_STOPPED : 
      {
	cout << "1" << endl;
	// nothing to do
	return 0;
      }
    case CTRLER_STARTED : 
      {
	cout << "2" << endl;
	ctrlerSt_ = CTRLER_STOPPED;
	dtToWait_ = -1;
	timer_.stop();
	if(threadCtrl_ != NULL) 
	  {
	    delete threadCtrl_;
	    threadCtrl_ = NULL;
	  }
	
	cout << "ctrlerSt_ : " << ctrlerSt_ << endl;
	return 0;
      }
    }
}


                                    /**************************************************/
                                    /*************** Class WCController ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
WCController::WCController() : Controller()
{
  this->ctrlerType_ = SIMPLE_CTRLER;
}

/*
 *  \brief init constructor
 */
WCController::WCController(hugr::Store* store,const std::string& name) : Controller(store,name,NULL,NULL,NULL,NULL)
{
  this->ctrlerType_ = SIMPLE_CTRLER;
}

/*
 *  \brief destructor
 */
WCController::~WCController()
{
}


/******** hugr store connection functions *********/

/*
 *  \brief connect the controller to the hugr store
 *  \return 0 if the connection to the hugr store has been correctly established, -1 else
 */
int WCController::connect()
{
  if(store_ == NULL)
    {
      cerr << "WCController::connect : Warning. The HUGR store hasn't been defined." << endl;
      ctrlerSt_ = CTRLER_DISCONNECTED;
      return -1;
    }
  
  // connect the hugr store if it has not been already done
  store_->connect();

  // get back the wheelchair command id
//   std::string str(nameRb_);
  try
    {
//       str += "_command";
//       controlId_ = store_->lookupVariable(str.c_str());
      controlId_ = store_->lookupVariable(commandName_.c_str());
    }
  catch(runtime_error e)
    {
      cerr << "WCController::connect : Warning. " << commandName_.c_str() << "cannot be found in the HUGR store." << endl;
      ctrlerSt_ = CTRLER_DISCONNECTED;
      return -1;
    }

  ctrlerSt_ = CTRLER_STOPPED;
  return 0;
}


/********************* Actions ********************/

/*
 *  \brief compute the control to apply from the time t during a time dt
 *  \param t : the initial time of application of the control
 *  \param dt : the duration of the application of the control
 *
 *  The resulting control is put in controlt_ to be send  
 */
void WCController::computeControl(double t,double dt)
{
  // nothing to do : the control given by the user is send
}

/*
 *  \brief send a control to the hugr store
 *  \return 0 if the control has been sent successfully, -1 else
 */
int WCController::sendControl() const
{
  if(store_ == NULL)
    {
      cerr << "WCController::sendControl : Warning. The HUGR store hasn't been defined." << endl;
      return -1;
    }
  
  WheelchairCommand WC_Command;
  WC_Command.v = this->controlt()->v();
  WC_Command.omega = this->controlt()->av();

  try
    {
      store_->writeVariable(controlId_, WC_Command);

      cout << "control : "; this->controlt()->display(); cout << endl;
    }
  catch(runtime_error e)
    {
      cerr << "WCController::sendControl : Warning. The control ID has not been found in the HUGR store." << endl;
      return -1;
    }
  return 0;
}

/*
 *  \brief reinit the controller
 */
void WCController::reinit()
{
  this->updateCtrl(WCControl(0,0));
  timer_.reinit();
}

/*
 *  \brief update the input control
 *  \param ctrl : the new input control
 *  \return 0 if the update process succeed, -1 else
 */
int WCController::updateCtrl(const Control& c)
{
  ((WCControlT*)(this->controlt_))->v() = ((WCControl*)(&c))->v();
  ((WCControlT*)(this->controlt_))->av() = ((WCControl*)(&c))->av();
  return 0;
}


                                    /**************************************************/
                                    /*************** Class WCController2 **************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
WCController2::WCController2() : Controller()
{
  this->ctrlerType_ = TF2_CTRLER;
}

/*
 *  \brief init constructor
 */
WCController2::WCController2(hugr::Store* store,const std::string& name) : Controller(store,name,NULL,NULL,NULL,NULL)
{
  this->ctrlerType_ = TF2_CTRLER;
}

/*
 *  \brief destructor
 */
WCController2::~WCController2()
{
}


/******** hugr store connection functions *********/

/*
 *  \brief connect the controller to the hugr store
 *  \return 0 if the connection to the hugr store has been correctly established, -1 else
 */
int WCController2::connect()
{
  if(store_ == NULL)
    {
      cout << "WCController::connect : Warning. The HUGR store hasn't been defined." << endl;
      ctrlerSt_ = CTRLER_DISCONNECTED;
      return -1;
    }
  
  // connect the hugr store if it has not been already done
  store_->connect();

  // get back the wheelchair command id
  std::string str(nameRb_);
  try
    {
      str += "_command";
      controlId_ = store_->lookupVariable(str.c_str());
    }
  catch(runtime_error e)
    {
      cerr << "WCController::connect : Warning. " << str.c_str() << "cannot be found in the HUGR store." << endl;
      ctrlerSt_ = CTRLER_DISCONNECTED;
      return -1;
    }

  ctrlerSt_ = CTRLER_STOPPED;
  return 0;
}


/********************* Actions ********************/

/*
 *  \brief compute the control to apply from the time t during a time dt
 *  \param t : the initial time of application of the control
 *  \param dt : the duration of the application of the control
 *
 *  The resulting control is put in controlt_ to be send  
 */
void WCController2::computeControl(double t,double dt)
{
  double elapsedT = timer_.totalElapsedTime();
  if(elapsedT >= 0.0 && elapsedT < 2.97)
    {
      ((WCControlT*)(this->controlt_))->v() = 0.5;
      ((WCControlT*)(this->controlt_))->av() = 0.0;
//       ((WCControlT*)(this->controlt_))->t() = 2.97;
    }
  else if(elapsedT >= 2.97 && elapsedT < 5.12)
    {
      ((WCControlT*)(this->controlt_))->v() = 0.8;
      ((WCControlT*)(this->controlt_))->av() = 0.03*M_PI;
//       ((WCControlT*)(this->controlt_))->t() = 2.15;
    }
  else if(elapsedT >= 5.12 && elapsedT < 6.35)
    {
      ((WCControlT*)(this->controlt_))->v() = 1.0;
      ((WCControlT*)(this->controlt_))->av() = 0.05*M_PI;
//       ((WCControlT*)(this->controlt_))->t() = 1.23;
    }
  else if(elapsedT >= 6.35 && elapsedT < 9.13)
    {
      ((WCControlT*)(this->controlt_))->v() = 1.0;
      ((WCControlT*)(this->controlt_))->av() = 0.05*M_PI;
//       ((WCControlT*)(this->controlt_))->t() = 2.78;
    }
  else if(elapsedT >= 9.13 && elapsedT < 12.14)
    {
      ((WCControlT*)(this->controlt_))->v() = 0.6;
      ((WCControlT*)(this->controlt_))->av() = -0.06*M_PI;
//       ((WCControlT*)(this->controlt_))->t() = 3.01;
    }
  else if(elapsedT >= 12.14 && elapsedT < 13.22)
    {
      ((WCControlT*)(this->controlt_))->v() = 0.2;
      ((WCControlT*)(this->controlt_))->av() = -0.02*M_PI;
//       ((WCControlT*)(this->controlt_))->t() = 1.08;
    }
  else if(elapsedT >= 13.22)
    {
      ((WCControlT*)(this->controlt_))->v() = 0.0;
      ((WCControlT*)(this->controlt_))->av() = 0.00*M_PI;
//       ((WCControlT*)(this->controlt_))->t() = 0.0;
    }
}

/*
 *  \brief send a control to the hugr store
 *  \return 0 if the control has been sent successfully, -1 else
 */
int WCController2::sendControl() const
{
  if(store_ == NULL)
    {
      cerr << "WCController::sendControl : Warning. The HUGR store hasn't been defined." << endl;
      return -1;
    }

  WheelchairCommand WC_Command;
  WC_Command.v = this->controlt()->v();
  WC_Command.omega = this->controlt()->av();

  try
    {
      store_->writeVariable(controlId_, WC_Command);

      cout << "control : "; this->controlt()->display(); cout << endl;
    }
  catch(runtime_error e)
    {
      cerr << "WCController::sendControl : Warning. The control ID has not been found in the HUGR store." << endl;
      return -1;
    }
  return 0;
}

/*
 *  \brief reinit the controller
 */
void WCController2::reinit()
{
  this->updateCtrl(WCControl(0,0));
  timer_.reinit();
}


                                    /**************************************************/
                                    /*************** Class CCController ***************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
CCController::CCController() : Controller()
{
  this->ctrlerType_ = SIMPLE_CTRLER;
}

/*
 *  \brief init constructor
 */
CCController::CCController(hugr::Store* store,const std::string& name) : Controller(store,name,NULL,NULL,NULL,NULL)
{
  this->ctrlerType_ = SIMPLE_CTRLER;
}

/*
 *  \brief destructor
 */
CCController::~CCController()
{
}


/******** hugr store connection functions *********/

/*
 *  \brief connect the controller to the hugr store
 *  \return 0 if the connection to the hugr store has been correctly established, -1 else
 */
int CCController::connect()
{
  if(store_ == NULL)
    {
      cerr << "CCController::connect : Warning. The HUGR store hasn't been defined." << endl;
      ctrlerSt_ = CTRLER_DISCONNECTED;
      return -1;
    }
  
  // connect the hugr store if it has not been already done
  store_->connect();

  // get back the wheelchair command id
  std::string str(nameRb_);
  try
    {
      str += "_cycabCarCommand";
      controlId_ = store_->lookupVariable(str.c_str());
    }
  catch(runtime_error e)
    {
      cerr << "CCController::connect : Warning. " << str.c_str() << "cannot be found in the HUGR store." << endl;
      ctrlerSt_ = CTRLER_DISCONNECTED;
      return -1;
    }

  ctrlerSt_ = CTRLER_STOPPED;
  return 0;
}


/********************* Actions ********************/

/*
 *  \brief compute the control to apply from the time t during a time dt
 *  \param t : the initial time of application of the control
 *  \param dt : the duration of the application of the control
 *
 *  The resulting control is put in controlt_ to be send  
 */
void CCController::computeControl(double t,double dt)
{
  // nothing to do : the control given by the user is send
}

/*
 *  \brief send a control to the hugr store
 *  \return 0 if the control has been sent successfully, -1 else
 */
int CCController::sendControl() const
{
  if(store_ == NULL)
    {
      cerr << "CCController::sendControl : Warning. The HUGR store hasn't been defined." << endl;
      return -1;
    }
  
  CycabCarCommand CC_Command;    
  CC_Command.v = this->controlt()->v();
  CC_Command.phi = 0.5*this->controlt()->ph();
  CC_Command.dualSteering = true;
  CC_Command.manualDriving = false;

  try
    {
      store_->writeVariable(controlId_, CC_Command);

      cout << "control : "; this->controlt()->display(); cout << endl;
    }
  catch(runtime_error e)
    {
      cerr << "CCController::sendControl : Warning. The control ID has not been found in the HUGR store." << endl;
      return -1;
    }
  return 0;
}

/*
 *  \brief reinit the controller
 */
void CCController::reinit()
{
  this->updateCtrl(CCControl(0,0));
  timer_.reinit();
}

/*
 *  \brief update the input control
 *  \param ctrl : the new input control
 *  \return 0 if the update process succeed
 */
int CCController::updateCtrl(const Control& c)
{
  ((CCControlT*)(this->controlt_))->v() = ((CCControl*)(&c))->v();
  ((CCControlT*)(this->controlt_))->ph() = ((CCControl*)(&c))->ph();
}
