#include "Hugr_CCTrajFollow.hpp"
using namespace mpc;

#include "CycabStructs.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <utime.h>

#include <iostream>
using namespace std;

                                    /**************************************************/
                                    /************* Class Hugr_CCTrajFollow ************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/*
 *  \brief default constructor
 */
Hugr_CCTrajFollow::Hugr_CCTrajFollow() : Controller(), CCTrajFollow()
{
  this->curTimeOnTraj_ = 0;
  this->ctrlerType_ = TRAJFOLLOW_CTRLER;

//   curRbST_ = NULL;
//   absRbST_ = NULL;
}

/*
 *  \brief init constructor
 *  \param sys : the system constraints of the cycab
 *  \param traj : the trajectory to follow
 *  \param store : the hugr store on which to send the control data
 *  \param name : the name of the cycab controlled
 *  \param mutexLoc : the mutex used to prevent localization conflicts
 *  \param mutexCtrl : the mutex used to prevent control conflicts
 *  \param mutexTraj : the mutex used to prevent trajectory conflicts
 *  \param mutexEnv : the mutex used to prevent environment conflicts
 */
Hugr_CCTrajFollow::Hugr_CCTrajFollow(CarLikeSys* sys,CLSTraj* traj,
				     hugr::Store* store,const std::string& name,
				     boost::mutex* mutexLoc,boost::mutex* mutexCtrl,
				     boost::mutex* mutexTraj,boost::mutex* mutexEnv)
  : Controller(store,name,mutexLoc,mutexCtrl,mutexTraj,mutexEnv), CCTrajFollow(sys,traj)
{
  cout << "Hugr_CCTrajFollow::Hugr_CCTrajFollow!!!!!!!!!!!!!!!!" << endl;
  this->curTimeOnTraj_ = 0;
  this->ctrlerType_ = TRAJFOLLOW_CTRLER;

//   curRbST_ = NULL;
//   absRbST_ = NULL;
}

/*
 *  \brief destructor
 */
Hugr_CCTrajFollow::~Hugr_CCTrajFollow()
{
}


// /***************** Initialization *****************/

// /**
//  *  \brief initialize the controller execution
//  */
// void Hugr_CCTrajFollow::initExec()
// {
// }


/************** Getters and setters ***************/

/*
 *  \brief return the state of the trajectory followed (if exists) at absolute time t
 *  \param t : the input absolute time
 *  \param res : the resulting state of the trajectory followed at absolute time t if it exists, 0 else
 */
void Hugr_CCTrajFollow::getStateAtAbsTime(double t,StateTime& res) const
{
}


/******** hugr store connection functions *********/

/*
 *  \brief connect the controller to the hugr store
 *  \int return 0 if the connection to the hugr store has been correctly established, -1 else
 */
int Hugr_CCTrajFollow::connect()
{
  if(store_ == NULL)
    {
      cout << "Hugr_CCTrajFollow::connect : Warning. The HUGR store hasn't been defined." << endl;
      ctrlerSt_ = CTRLER_DISCONNECTED;
      return -1;
    }
  
  // connect the hugr store if it has not been already done
  store_->connect();

  // get back the wheelchair command id
  std::string str(nameRb_);
  try
    {
      str += "_cycabCarCommand";
      controlId_ = store_->lookupVariable(str.c_str());
    }
  catch(runtime_error e)
    {
      cerr << "Hugr_CCTrajFollow::connect : Warning. " << str.c_str() << "cannot be found in the HUGR store." << endl;
      ctrlerSt_ = CTRLER_DISCONNECTED;
      return -1;
    }

  ctrlerSt_ = CTRLER_STOPPED;
  return 0;
}


/********************* Actions ********************/

/*
 *  \brief compute the control to apply from the time t during a time dt
 *  \param t : the initial time of application of the control
 *  \param dt : the duration of the application of the control
 *
 *  The resulting control is put in controlt_ to be send  
 */
void Hugr_CCTrajFollow::computeControl(double t,double dt)
{
  // lock the localization mutex to prevent loc conflicts
  if(MUTEX_DEBUG == 1) cout << "Hugr_CCTrajFollow::computeControl : waiting for mutex Loc" << endl;
  boost::mutex::scoped_lock lockLoc(*mutexLoc_);
  if(MUTEX_DEBUG == 1) cout << "Hugr_CCTrajFollow::computeControl : mutex Loc locked" << endl;

  cout << "///////////////////////////////////////////////////////" << endl;

//   // lock the control mutex to prevent control conflicts
//   boost::mutex::scoped_lock lockCtrl(*mutexCtrl_);

  // get the last saved state-time
  CLStateT curST(curRbST_);

  // compute the time elapsed from the previous control request
  double DT = timer_.elapsedTime();

  // set the current time as the time registered by the timer
  curTimeOnTraj_ = timer_.totalElapsedTime();
  //curST.t() = curTimeOnTraj_;

  CCControl ctrl;

//   cout << "curST : "; curST.display(); cout << "  curST.t() : " << curST.t() << "  lastDuration_ : " << lastDuration_ << endl;

//   int res = this->getControl(curST,
// 			     curST.t() + this->lastDuration_,
// 			     this->lastDuration_,
// 			     ctrl);
//   cout << "Hugr_CCTrajFollow::computeControl -> dt : " << dt << endl;

  cout << "dt : " << dt << "  DT : " << DT << "  CtrlTimerT : " << timer_.totalElapsedTime() << endl;

  int res = this->getControl(curST,
			     curTimeOnTraj_,
			     DT,
			     ctrl);
  //cout << "  res : " << res << endl;

  ((CCControlT*)(this->controlt_))->v() = ctrl.v();
  ((CCControlT*)(this->controlt_))->ph() = ctrl.ph();
  ((CCControlT*)(this->controlt_))->t() = dt;

  // modify the state of the robot to consider the actual velocities
  curRbST_.s() = ctrl.v();
  curRbST_.ph() = ctrl.ph();

  if(MUTEX_DEBUG == 1) cout << "Hugr_CCTrajFollow::computeControl : mutex Loc released" << endl;
  cout << "///////////////////////////////////////////////////////" << endl;
}

/*
 *  \brief send a control to the hugr store
 *  \return 0 if the control has been sent successfully, -1 else
 */
int Hugr_CCTrajFollow::sendControl() const
{
  if(store_ == NULL)
    {
      cerr << "CCController::sendControl : Warning. The HUGR store hasn't been defined." << endl;
      return -1;
    }

  CycabCarCommand CC_Command;    
  CC_Command.v = this->controlt()->v();
  CC_Command.phi = 0.5*this->controlt()->ph();
  CC_Command.dualSteering = true;
  CC_Command.manualDriving = false;

  try
    {
      store_->writeVariable(controlId_, CC_Command);

//       cout << "control : "; this->controlt()->display(); cout << endl;
    }
  catch(runtime_error e)
    {
      cerr << "CCController::sendControl : Warning. The control ID has not been found in the HUGR store." << endl;
      return -1;
    }
  return 0;
}

/*
 *  \brief update the localization of the trajectory follower
 *  \param curST : the current state-time of the robot (given at absolute time)
 *
 *  Modify the time of the input state to take into account to relative time to the trajectory followed
 */
void Hugr_CCTrajFollow::updateStateT(const StateTime& curST)
{
  // lock the localization mutex to prevent loc conflicts
  if(MUTEX_DEBUG == 1) cout << "Hugr_CCTrajFollow::updateStateT : waiting for mutex Loc" << endl;
  boost::mutex::scoped_lock lock(*mutexLoc_);
  if(MUTEX_DEBUG == 1) cout << "Hugr_CCTrajFollow::updateStateT : mutex Loc locked" << endl;

  // get the position of the input state
  PositionXYTh pos = curST.posXYTh();

  curRbST_.x() = pos.x();
  curRbST_.y() = pos.y();
  curRbST_.th() = normalizedAngle(pos.th());
  curRbST_.t() = timer_.totalElapsedTime();

  absRbST_.x() = pos.x();
  absRbST_.y() = pos.y();
  absRbST_.th() = normalizedAngle(pos.th());
  absRbST_.t() = curST.t();

  if(MUTEX_DEBUG == 1) cout << "Hugr_CCTrajFollow::updateStateT : mutex Loc released" << endl;
}

/*
 *  \brief reinit the controller
 */
void Hugr_CCTrajFollow::reinit()
{
  this->curTimeOnTraj_ = 0;
  timer_.reinit();
}
