#ifndef MPC_HENVIRONMENT_HPP
#define MPC_HENVIRONMENT_HPP

#include "environment/inc/Environment.hpp"
#include "environment/inc/RoadMap.hpp"
#include "common/inc/Timer.hpp"

#include "hugr.hpp"
#include <string>

#include "hugr/HCommon/inc/HCommon.hpp"

#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>

namespace mpc
{
  /**
   * \enum envUpStatus_t
   * \brief Enumeration of the different environment updater status types
   *
   * ENVUP_NULL : environment updater not allocated
   * ENVUP_DISCONNECTED : disconnected environment updater
   * ENVUP_STOPPED : stopped environment updater
   * ENVUP_STARTED : started environment updater
   */
  typedef enum {
    ENVUP_NULL          = -1,
    ENVUP_DISCONNECTED  =  0,
    ENVUP_STOPPED       =  1,
    ENVUP_STARTED       =  2
  } envUpStatus_t;

  /**
   * \enum OGrequest_t
   * \brief Enumeration of the different request for the occupancy grid types
   *
   * OGR_OFF : no request for the occupancy grid
   * OGR_ON : always request for the occupancy grid
   * OGR_ONESHOT : request for the occupancy grid only one time
   * OGR_TIMED : request for the occupancy grid each T sec
   */
  typedef enum {
    OGR_OFF         = 0,
    OGR_ON          = 1,
    OGR_ONESHOT     = 2,
    OGR_TIMED       = 3
  } OGrequest_t;

                                    /**************************************************/
                                    /*************** Class HEnvironment ***************/
                                    /**************************************************/

  /** \class HEnvironment
   * \brief Hugr environment
   */
  class HEnvironment : public Environment
  {
  protected : 
    DRoadMap* roadmap_;                          /**< roadmap built over the static environment if needed */
    bool autobuildRM_;                           /**< true if the roadmap must be autobuilt when a static environment is loaded */
    bool dispRM_;                                /**< true if the roadmap must be drawn during the environment display process */

    Timer *timer_;                               /**< global timer used by the robot to date the perception scans */
    double dt_;                                  /**< the interval of time to send controls (in sec.) */

    envUpStatus_t status_;                       /**< the environment update status (disconnected, connected and stopped, connected and started) */

    hugr::Store* store_;                         /**< the hugr store used to retrieve information about the environment */
    hugr::VariableId lObsId_;                    /**< ID of the hugr's variable of the list of dynamic obstacles */
    hugr::VariableId OGId_;                      /**< ID of the hugr's variable of the occupancy grid */
    hugr::VariableId SlammotStateId_;            /**< ID of the hugr's variable of the slammot state (frequency of request to the occupancy grid) */

    double distFilter_;                          /**< maximal distance from static obstacles to consider the dynamic obstacles perceived as noise */
    double securDist_;                           /**< radius of the perceived dynamic obstacles */

    OGrequest_t OGRequest_;                      /**< occupancy grid request type */
    double OGRdt_;                               /**< interval of time between two occupancy grid requests */
    double OGRelapsedT_;                         /**< elasped time since the last occupancy grid request */
    double OGRtotElapsedT_;                      /**< last total elasped time used for occupancy grid request */

    boost::thread* threadEnv_;                   /**< the thread used to execute the environment update */

    boost::mutex* mutexEnv_;                     /**< mutex used to prevent environment access conflicts */

    //   PositionXYTh pos_;                           /**< the location from which the perception process acts */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    HEnvironment();

  public : 
    /**
     *  \brief init constructor
     *  \param store : the hugr store from which is construct the environment
     *  \param timer : the timer used to date input information about the environment
     *  \param mutexEnv : the mutex used to prevent environment access conflicts
     */
    HEnvironment(hugr::Store* store,Timer* timer,boost::mutex* mutexEnv);

    /**
     *  \brief destructor
     */
    virtual ~HEnvironment();

    /************** Getters and setters ***************/

    /**
     *  \brief return the status of the environment perception process
     *  \return the status of the environment perception process
     */
    inline envUpStatus_t status() const { return this->status_; }

    /**
     *  \brief return the roadmap built over the static environment
     *  \return the roadmap built over the static environment
     */
    inline DRoadMap* roadmap() const { return this->roadmap_; }

    /**
     *  \brief return the autobuilding roadmap status
     *  \return true if the roadmap must be autobuilt when a static environment is loaded
     */
    inline bool autobuildRM() const { return this->autobuildRM_; }

    /**
     *  \brief set the autobuilding roadmap status
     *  \param the new autobuilding roadmap status
     */
    inline void setAutobuildRM(bool status) { this->autobuildRM_ = status; }

    /**
     *  \brief return the roadmap drawing status
     *  \return true if the roadmap must be drawn
     */
    inline bool dispRM() const { return this->dispRM_; }

    /**
     *  \brief set the roadmap drawing status
     *  \param the new roadmap drawing status
     */
    inline void setDispRM(bool status) { this->dispRM_ = status; }

    /**
     *  \brief return the current time of the environment perception process
     *  \return the current time of the environment perception process
     */
    inline double getTime() const { return timer_->totalElapsedTime(); }

    //   /**
    //    *  \brief return the frame of the environment
    //    *  \return the frame of the environment
    //    */
    //   inline const PositionXYTh& frame() const { return this->pos_; }

    /******** hugr store connection functions *********/
  
  public : 
    /**
     *  \brief connect the environment updater to the hugr store
     *  \return 0 if the connection to the hugr store has been correctly established, -1 else
     */
    virtual int connect();

    /********************* Actions ********************/

  protected : 
    /**
     *  \brief update the Detection and Tracking of Moving Objects
     *  \return 0 if the update process succeed, -1 else
     */
    virtual int updateDaTMO();

    /**
     *  \brief update the occupancy grid given by the SLAM process
     *  \return 0 if the update process succeed, -1 else
     */
    virtual int updateOG();

    /**
     *  \brief update the environment thanks to the SLAMMOT process
     *  \return 0 if the update process succeed, -1 else
     */
    virtual int updateEnv();

    /**
     *  \brief execute the control loop
     */
    void loop();

  public : 
    /**
     *  \brief init the detection and tracking of moving objects process
     *  \param distFilter : maximal distance from static obstacles to consider the dynamic obstacles perceived as noise
     *  \param securDist : minimal distance kept to dynamic obstacles
     *  \return 0 if the initialization succeed
     *  \return -1 if the DATMO has been already started
     *  \return -2 if the environment perception tool has not been defined
     *  \return -3 if an unknown error occured
     */
    int initDATMO(double distFilter,double securDist);

    /**
     *  \brief update the location from which the perception process acts
     */
    void updatePose(const PositionXYTh pos);

    /**
     *  \brief start the timer
     *  \return 0 if the starting process succeed, -1 else
     */
    int start();

    /**
     *  \brief stop the timer
     *  \return 0 if the stoping process succeed, -1 else
     */
    int stop();

    /**
     *  \brief provide knowledge of the static environment to the robot
     *  \param filename : filename of the input static environment
     *  \param size : the obstacles are incresed by this size
     *  \param RMfilename : filename of an input roadmap alredy built on the environment
     *  \return 0 if the process succeed
     *          -1 if the file has not been found
     *          -4 if an unknown error occured
     */
    int addStaticEnv(const char* filename,double size,const char* RMfilename = NULL);

    /**
     *  \brief clear the static environment
     *  \return 0 if the process succeed, -1 else
     *          -4 if an unknown error occured
     */
    int clearStaticEnv();

    /**
     *  \brief launch the roadmap building on the static environment
     *  \return 0 if the process succeed, -1 else
     */
    int buildRoadmap();

    /**
     *  \brief load the roadmap from an input file
     *  \param filename : the input filename
     *  \return 0 if the process succeed, -1 else
     */
    int loadRoadmap(const char* filename);

    /**
     *  \brief save the current roadmap
     *  \param filename : the output filename
     *  \return 0 if the process succeed, -1 else
     */
    int saveRoadmap(const char* filename);

    /************** Obstacles filtering ***************/

    /**
     *  \brief check if the input obstacles may be a false positive
     *  \param obs : the input obstacle
     *  \return true if the input obstacle is considered as a false positive
     */
    bool isFalsePositive(const DiskObs2D& obs);

    /*************** Display functions ****************/

    /**
     *  \brief draw the environment in an openGL viewer
     *  \param dispFM : true if the forecast model of the dynamic obstacles must be displayed
     */
    virtual void toGL(bool dispFM) const;
  };

};
#endif // MPC_HENVIRONMENT_HPP
