#ifndef MPC_CYCAB_HPP
#define MPC_CYCAB_HPP

#include "robot/inc/Robot.hpp"
#include "system/inc/CarLikeSys.hpp"

#include "CycabSimulator.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /******************* Class Cycab ******************/
                                    /**************************************************/

  /** \class Cycab
   * \brief robotic cycab
   */
  class Cycab : public Robot
  {
  protected : 
    hugr::VariableId      CC_poseId;                /**< store id for the pose of the wheelchair */
    hugr::VariableId      CC_commandId;             /**< store id for the command of the wheelchair */
    hugr::VariableId      CC_constantsId;           /**< store id for the constants of the wheelchair */

    PositionXYTh initPos_;                          /**< initial state at which is initialized */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    Cycab();

  public : 
    /**
     *  \brief default constructor
     *  \param name : the name of the robot
     *  \param store : the hugr store containing informations about the robot
     */
    Cycab(std::string name,hugr::Store* store);

    //   /**
    //    *  \brief default constructor
    //    *  \param name : the name of the robot
    //    *  \param sys : the robotic system configuration of the robot
    //    */
    //   Cycab(std::string name,CarLikeSys* sys);

    /**
     *  \brief destructor
     */
    virtual ~Cycab();

    /************** Getters and setters ***************/
  
    /**
     *  \brief return the system configuration of the robot
     *  \return the system configuration of the robot
     */
    inline CarLikeSys* sys() const { return ((CarLikeSys*)sys_); }

    /**
     *  \brief return the trajectory followed by the robot
     *  \return the trajectory followed by the robot
     */
    inline CLSTraj* traj() const { return ((CLSTraj*)traj_); }

    /**
     *  \brief return the state-time of the robot
     *  \return the state-time of the robot
     */
    inline CLStateT* statet() const { return ((CLStateT*)statet_); }

    /**
     *  \brief update the system configuration
     *  \param sys : the new system configuration
     */
    inline void setSys(CarLikeSys* sys) { this->sys_ = sys; }

#ifdef HUGR
  protected : 
    /**
     *  \brief return the controller of the wheelchair
     *  \return the controller of the wheelchair
     */
    inline const CCController* controller() const { return ((CCController*)(this->controller_)); }

    /**
     *  \brief return the current control applied on the wheelchair
     *  \return the current control applied on the wheelchair
     */
    const CCControlT* controlt();

    /**
     *  \brief return true if the robot is moving
     *  \return true if the robot is moving
     */
    virtual bool isMoving();

    /**
     *  \brief stop the robot moving
     */
    virtual void stopMoving();

    /*************** Hugr store reading ***************/
  public : 
    /**
     *  \brief update the system configuration from the HUGR store
     *  \return 0 if the update succeed, -1 else
     */
    int updateSys();

    /**
     *  \brief update the robot pose from the HUGR store
     *  \return 0 if the update succeed, -1 else
     */
    int updatePose();

    //   /**
    //    *  \brief send the control to apply on the robot to the HUGR store
    //    *  \return 0 if the update succeed, -1 else
    //    */
    //   int sendControl();

    /************* Localization functions *************/

    /**
     *  \brief compute the current deviation between the localization of the robot and the trajectory followed
     *  \param traj : the current state-time of the trajectory (set during the process)
     *  \param loc : the current localization state-time (set during the process)
     *  \param dev : the current deviation between the localization of the robot and the trajectory followed (set during the process)
     *  \param ctrlt : the current control applied on the robot
     *
     *  Every input parameters must be previously allocated
     *  Their value are set during the deviation computation
     */
    virtual void computeCurDeviation(StateTime* traj,StateTime* loc,StateTime* dev,ControlTime* ctrlt);

    /*************** Control functions ****************/

  public : 
    /**
     *  \brief change the controller type
     *  \param type : the new controller type
     *  \return 0 if the initialization succeed
     *  \return -1 if the current controller is running
     *  \return -2 if such a controller has already been initialized
     *  \return -3 if the input controller type is wrong or unknown
     *  \return -4 if an unknown problem occured
     *  \return -5 if the connection to the store cannot be established
     *  \return -6 if a required input of the controller is missing
     */
    virtual int initCtrler(controller_t type);
#endif

    /*************** Planner functions ****************/

  public : 
    /**
     *  \brief change the planner type
     *  \param type : the new planner type
     *  \return 0 if the initialization succeed
     *  \return -2 if such a planner has already been initialized
     *  \return -3 if the input planner type is wrong or unknown
     *  \return -4 if a roadmap is required
     *  \return -5 if the environment has not been initialized
     *  \return -6 if an unknown problem occured
     */
    int initPlanner(trajPlanner_t type);

    /**
     *  \brief reset the trajectory followed
     *  \param traj : the new trajectory
     */
    virtual void resetTraj(const STrajectory* traj);

    /**
     *  \brief start the control process of the robot
     *  \param goalS : the goal state
     *  \param t : the final time at which the goal state must be reached
     *  \return 0 if the controller starting process succeed
     *  \return -1 if no trajectory has been found
     *  \return -2 if the planner hasn't been initialized
     *  \return -3 if the planner type is not handled
     *  \return -4 if a roadmap is required
     *  \return -5 if the environment has not been initialized
     *  \return -6 if the robot is still moving
     */
    virtual int planTraj(const CLState& goalS,double t=0,double dt=0);

    /************* Deformation functions **************/

  public : 
    /**
     *  \brief set the deformation parameters
     *  \param k_int : internal forces coefficient
     *  \param k_ext : external forces coefficient (dynamic obstacles)
     *  \param k_stext : external forces coefficient (static obstacles)
     *  \param dyn_med : maximal effective distance (dynamic obstacles)
     *  \param st_med : maximal effective distance (static obstacles)
     *  \param dyn_med : maximal temporal effective distance
     */
    void setTeddyParams(double k_int,double k_ext,double k_stext,double dyn_med,double st_med,double tmp_med);

    /*************** Display functions ****************/

    /**
     *  \brief update the trajectory deformation process visualization
     *  \param dispFExt : true if the external forces (due to dynamic obstacles) must be displayed
     *  \param dispStFExt : true if the external forces (due to static obstacles) must be displayed
     *  \param dispNearDynObs : true if the position of the nearest dynamic obstacles must be displayed
     *  \param dispNearStObs : true if the position of the nearest static obstacles must be displayed
     */
    virtual void updateTeddyView(bool dispFExt,bool dispStFExt,bool dispNearDynObs,bool dispNearStObs);

    /**
     *  \brief display the trajectory followed by the robot
     */
    virtual void trajGL();
  };

};
#endif // MPC_CYCAB_HPP
