#ifndef MPC_ROBOT_HPP
#define MPC_ROBOT_HPP

#include "system/inc/System.hpp"
#include "common/inc/StateTime.hpp"
#include "common/inc/ControlTime.hpp"
#include "planner/inc/TrajPlanner.hpp"
#include "environment/inc/Environment.hpp"
#include "trajFollow/inc/TrajFollow.hpp"
#include "deformation/inc/TrajDeform.hpp"
#include "hugr/hugrController/inc/Hugr_CCTrajFollow.hpp"
//#include "Hugr_TrajDeform.hpp"
//using namespace mpc;

#ifdef HUGR
#include "hugr.hpp"
#include "hugr/hugrController/inc/Controller.hpp"
#include "hugr/HCommon/inc/HCommon.hpp"
#include "hugr/hugrEnvironment/inc/HEnvironment.hpp"
#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>
#endif

namespace mpc
{
  /**
   * \enum localizerState_t
   * \brief Enumeration of the different controller state types
   *
   * LOCZER_NULL : localizer not allocated
   * LOCZER_DISCONNECTED : disconnected localizer
   * LOCZER_STOPPED : stopped localizer
   * LOCZER_STARTED : started localizer
   */
  typedef enum {
    LOCZER_NULL          = -1,
    LOCZER_DISCONNECTED  =  0,
    LOCZER_STOPPED       =  1,
    LOCZER_STARTED       =  2
  } localizerStatus_t;

  /**
   * \enum envPerceptStatus_t
   * \brief Enumeration of the different environment perception status
   *
   * ENVP_NULL : environment perception process not allocated
   * ENVP_DISCONNECTED : disconnected environment perception process
   * ENVP_STOPPED : stopped environment perception process
   * ENVP_STARTED : started environment perception process
   */
  typedef enum {
    ENVP_NULL          = -1,
    ENVP_DISCONNECTED  =  0,
    ENVP_STOPPED       =  1,
    ENVP_STARTED       =  2
  } envPerceptStatus_t;

                                    /**************************************************/
                                    /****************** Class Robot *******************/
                                    /**************************************************/

  /** \class Robot
   * \brief Generic robot class
   *
   *  Each robot shall derivate from this class
   */
  class Robot
  {
  protected : 
    System* sys_;                         /**< the robotic system configuration of the robot */
    StateTime* statet_;                   /**< the state of the robot at a given time */
    std::string name_;                    /**< the name of the robot */

    TrajPlanner* planner_;                /**< the trajectory planner of the robot */
    STrajectory* traj_;                  /**< the trajectory followed by the robot */

#ifdef HUGR
    hugr::Store* store_;                  /**< the hugr store containing informations about the robot */

    Controller* controller_;              /**< the controller of the robot */
    //   Hugr_TrajDeform* deformer_;           /**< the trajectory deformer used to modify the motion of the robot */

    localizerStatus_t loczerStatus_;      /**< the localizer status */
    double dtLoczer_;                     /**< the interval of time to check the localization (in sec.) */
    boost::thread* threadLoc_;            /**< the thread used to execute the localization process */  

    HEnvironment* env_;                   /**< Representation of the perception of the robot's environment */
#endif

    Timer* timer_;                        /**< the global timer to date all informations received by the robot */

    boost::mutex mutexEnv_;               /**< mutex used to prevent environment access conflicts */
    boost::mutex mutexLoc_;               /**< mutex used to prevent localization access conflicts */
    boost::mutex mutexCtrl_;              /**< mutex used to prevent controller access conflicts */
    boost::mutex mutexTraj_;              /**< mutex used to prevent trajectory access conflicts */

    int debug_;                           /**< debug level */

  public : 
    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    Robot();

  public : 
    /**
     *  \brief default constructor
     *  \param name : the name of the robot
     *  \param store : the hugr store containing informations about the robot
     */
    Robot(std::string name,hugr::Store* store);

    //   /**
    //    *  \brief default constructor
    //    *  \param name : the name of the robot
    //    *  \param sys : the robotic system configuration of the robot
    //    */
    //   Robot(std::string name,System* sys);

    /**
     *  \brief destructor
     */
    virtual ~Robot();

    /************** Getters and setters ***************/
  
  public : 
    /**
     *  \brief return the system configuration of the robot
     *  \return the system configuration of the robot
     */
    inline System* sys() const { return this->sys_; }

    /**
     *  \brief return the trajectory followed by the robot
     *  \return the trajectory followed by the robot
     */
    virtual STrajectory* traj() const = 0;

    /**
     *  \brief return the environment in which the system evolve
     *  \return the environment in which the system evolve
     */
    inline HEnvironment* env() const { return this->env_; }

    /**
     *  \brief return the state-time of the robot
     *  \return the state-time of the robot
     */
    inline StateTime* statet() const { return this->statet_; }

    //   /**
    //    *  \brief return the control applied on the robot
    //    *  \return the control applied on the robot
    //    */
    //   inline ControlTime* ctrl() const { return this->ctrl_; }

    /**
     *  \brief update the system configuration
     *  \param sys : the new system configuration
     */
    inline void setSys(System* sys) { this->sys_ = sys; }

    /**
     *  \brief return a pointer to the timer used
     *  \return a pointer to the timer used
     */
    inline const Timer* timer() const { return this->timer_; }

    /**
     *  \brief return the current time of the supposed reached waypoint of the trajectory
     *  \return the current time of the supposed reached waypoint of the trajectory
     */
    virtual double curTimeOnTraj() const;

    /**
     *  \brief return true if the robot is moving
     *  \return true if the robot is moving
     */
    virtual bool isMoving() = 0;

    /**
     *  \brief stop the robot moving
     */
    virtual void stopMoving() = 0;

#ifdef HUGR
    /**
     *  \brief update the interval of time to check the localization 
     *  \param dt : the new interval of time (in sec.)
     */
    inline void setDtLoczer(double dt) { this->dtLoczer_ = dt; }

    /**
     *  \brief update the interval of time to check the localization 
     *  \param dt : the new interval of time (in sec.)
     */
    inline void setDtCtrler(double dt) { if(controller_ != NULL) controller_->dt() = dt; }

    /**
     *  \brief return the controller status
     *  \return the controller status
     */
    controllerStatus_t ctrlerStatus() const;

    /**
     *  \brief return the localizer status
     *  \return the localizer status
     */
    localizerStatus_t loczerStatus() const;

    /**
     *  \brief return the localizer status
     *  \return the localizer status
     */
    envUpStatus_t envpStatus() const;

    //   /**
    //    *  \brief return the deformer status
    //    *  \return the deformer status
    //    */
    //   virtual deformerStatus_t deformerStatus() const = 0;

    /**
     *  \brief set the controller debug level
     *  \param dbg : the new controller debug level
     */
    void setCtrlerDebugLevel(int dbg);

    /**
     *  \brief return the type of controller used
     *  \return the type of controller used
     */
    controller_t ctrlerType() const;

    /**
     *  \brief return the type of controller used
     *  \return the type of controller used
     */
    trajPlanner_t plannerType() const;

    /**
     *  \brief set the debugging level
     *  \param dbg : the new debugging level
     */
    void setDebug(int dbg);

    /**
     *  \brief return the roadmap built over the static environment
     *  \return the roadmap built over the static environment
     */
    DRoadMap* roadmap() const;

    /**
     *  \brief set the autobuilding roadmap status
     *  \param the new autobuilding roadmap status
     */
    void setAutobuildRM(bool status);

    /**
     *  \brief set the roadmap drawing status
     *  \param the new roadmap drawing status
     */
    void setDispRM(bool status);

    /******** hugr store connection functions *********/

  protected : 
    /**
     *  \brief update the system configuration from the HUGR store
     *  \return 0 if the update succeed, -1 else
     */
    virtual int updateSys() = 0;

    /**
     *  \brief update the robot pose from the HUGR store
     *  \return 0 if the update succeed, -1 else
     */
    virtual int updatePose() = 0;

  public : 
    //   /**
    //    *  \brief send the control to apply on the robot to the HUGR store
    //    *  \return 0 if the update succeed, -1 else
    //    */
    //   virtual int sendControl() = 0;

    /************* Localization functions *************/

  protected : 
    /**
     *  \brief execute the localization loop
     */
    virtual void execLoc();

  public : 
    /**
     *  \brief start the localization process of the robot
     *  \return 0 if the localization starting process succeed, -1 else
     */
    virtual int startLoc();

    /**
     *  \brief stop the localization process of the robot
     *  \return 0 if the localization stoping process succeed, -1 else
     */
    virtual int stopLoc();

    /**
     *  \brief compute the current deviation between the localization of the robot and the trajectory followed
     *  \param traj : the current state-time of the trajectory (set during the process)
     *  \param loc : the current localization state-time (set during the process)
     *  \param dev : the current deviation between the localization of the robot and the trajectory followed (set during the process)
     *  \param ctrlt : the current control applied on the robot
     *
     *  Every input parameters must be previously allocated
     *  Their value are set during the deviation computation
     */
    virtual void computeCurDeviation(StateTime* traj,StateTime* loc,StateTime* dev,ControlTime* ctrlt) = 0;

    /*************** Control functions ****************/

  public : 
    /**
     *  \brief change the controller type
     *  \param type : the new controller type
     *  \return 0 if the initialization succeed
     *  \return -1 if the current controller is running
     *  \return -2 if such a controller has already been initialized
     *  \return -3 if the input controller type is wrong or unknown
     *  \return -4 if an unknown problem occured
     *  \return -5 if the connection to the store cannot be established
     *  \return -6 if a required input of the controller is missing
     */
    virtual int initCtrler(controller_t type) = 0;

    /**
     *  \brief start the control process of the robot
     *  \return 0 if the controller starting process succeed, -1 else
     */
    virtual int startCtrler();

    /**
     *  \brief stop the control process of the robot
     *  \return 0 if the controller stoping process succeed, -1 else
     */
    virtual int stopCtrler();

    /**
     *  \brief update the input control
     *  \param ctrl : the new input control
     *  \return 0 if the update process succeed
     */
    virtual int updateCtrl(const Control& c);

  protected : 
    /**
     *  \brief reset the controller
     */
    virtual void resetCtrler();


#endif

    /*************** Planner functions ****************/

  public : 
    /**
     *  \brief change the planner type
     *  \param type : the new planner type
     *  \return 0 if the initialization succeed
     *  \return -2 if such a planner has already been initialized
     *  \return -3 if the input planner type is wrong or unknown
     *  \return -4 if an unknown problem occured
     */
    virtual int initPlanner(trajPlanner_t type) = 0;

    //   /**
    //    *  \brief start the control process of the robot
    //    *  \return 0 if the controller starting process succeed, -1 else
    //    */
    //   virtual int planTraj(const State& goalS);

  protected : 
    /**
     *  \brief reset the trajectory followed
     *  \param traj : the new trajectory
     */
    virtual void resetTraj(const STrajectory* traj) = 0;

    /******** Environment perception functions ********/

  public : 
    /**
     *  \brief change the environment perception process type
     *  \return 0 if the initialization succeed
     *  \return -2 if such an environment perception process has already been initialized
     *  \return -3 if the input environment perception process type is wrong or unknown
     *  \return -4 if an unknown problem occured
     *  \return -5 if the connection to the hugr store cannot be established
     */
    virtual int initEnvPercept();

    /**
     *  \brief init the detection and tracking of moving objects process
     *  \param distFilter : maximal distance from static obstacles to consider the dynamic obstacles perceived as noise
     *  \param securDist : minimal distance kept to dynamic obstacles
     *  \return 0 if the initialization succeed
     *  \return -1 if the DATMO has been already started
     *  \return -2 if the environment perception tool has not been defined
     *  \return -3 if an unknown error occured
     */
    int initDATMO(double distFilter,double securDist);

    /**
     *  \brief start the robot's environment perception process
     */
    virtual void startEnvPercept();

    /**
     *  \brief stop the robot's environment perception process
     */
    virtual void stopEnvPercept();

    /**
     *  \brief provide knowledge of the static environment to the robot
     *  \param filename : filename of the input static environment
     *  \param RMfilename : filename of an input roadmap alredy built on the environment
     *  \return 0 if the process succeed
     *          -1 if the file has not been found
     *          -2 if the localization has not been started
     *          -3 if the environment perception has not been initialized
     *          -4 if an unknown error occured
     *
     *  The static env coordinates must be expressed in the global frame
     */
    int addStaticEnv(const char* filename,const char* RMfilename = NULL);

    /**
     *  \brief clear the static environment
     *  \return 0 if the process succeed
     *          -3 if the environment perception has not been initialized
     *          -4 if an unknown error occured
     */
    int clearStaticEnv();

    /**
     *  \brief launch the roadmap building on the static environment
     *  \return 0 if the process succeed, -1 else
     */
    int buildRoadmap();

    /**
     *  \brief load a new roadmap on the environment
     *  \param filename : the input filename
     *  \return 0 if the process succeed, -1 else
     */
    int loadRoadmap(const char* filename);

    /**
     *  \brief save the current roadmap
     *  \param filename : the output filename
     *  \return 0 if the process succeed, -1 else
     */
    int saveRoadmap(const char* filename);

    /************* Deformation functions **************/

  public : 
    /**
     *  \brief set the deformation parameters
     *  \param k_int : internal forces coefficient
     *  \param k_ext : external forces coefficient (dynamic obstacles)
     *  \param k_stext : external forces coefficient (static obstacles)
     *  \param dyn_med : maximal effective distance (dynamic obstacles)
     *  \param st_med : maximal effective distance (static obstacles)
     *  \param dyn_med : maximal temporal effective distance
     */
    virtual void setTeddyParams(double k_int,double k_ext,double k_stext,double dyn_med,double st_med,double tmp_med) = 0;

    /**
     *  \brief start the deformation process of the trajectory of the robot
     *  \return 0 if the deformation starting process succeed, -1 else
     */
    virtual int startDeformer();

    /**
     *  \brief stop the deformation process of the trajectory of the robot
     *  \return 0 if the deformation stoping process succeed, -1 else
     */
    virtual int stopDeformer();

    /**
     *  \label get back informations about the deformation process
     *  \param nbDeform : the number of deformations proceeded
     *  \param avgDeformDur : the average deformation duration
     *  \param DPS : the average number of deformations per second
     */
    virtual void getDeformInfo(int& nbDeform,double& avgDeformDur,double& DPS) 
    {
      nbDeform = 0;
      avgDeformDur = 0;
      DPS = 0;
    }

    /*************** Display functions ****************/

  public : 
    /**
     *  \brief update the trajectory deformation process visualization
     *  \param dispFExt : true if the external forces (due to dynamic obstacles) must be displayed
     *  \param dispStFExt : true if the external forces (due to static obstacles) must be displayed
     *  \param dispNearDynObs : true if the position of the nearest dynamic obstacles must be displayed
     *  \param dispNearStObs : true if the position of the nearest static obstacles must be displayed
     */
    virtual void updateTeddyView(bool dispFExt,bool dispStFExt,bool dispNearDynObs,bool dispNearStObs) = 0;

    /**
     *  \brief draw the robotic system in an openGL viewer
     */
    void toGL() const;

    /**
     *  \brief display the trajectory followed by the robot
     */
    virtual void trajGL() = 0;

    /**
     *  \brief display the environment perception of the robot
     */
    virtual void envGL();

    /**
     *  \brief display the controller specifications of the robot
     */
    virtual void ctrlerGL();
  };

};
#endif // MPC_ROBOT_HPP
