#include "Cycab.hpp"
using namespace mpc;

#include "CycabConstantsINRIA-RA-grey.hpp"
#include "SimpleTrajPlanner.hpp"
#include "Hugr_CCTrajFollow.hpp"
#include "Hugr_CC_TFD.hpp"

                                    /**************************************************/
                                    /******************* Class Cycab ******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
Cycab::Cycab() : Robot()
{
}

/**
 *  \brief default constructor
 *  \param name : the name of the robot
 *  \param store : the hugr store containing informations about the robot
 */
Cycab::Cycab(std::string name,hugr::Store* store) : Robot(name,store)
{
  this->sys_ = new CarLikeSys();
  this->statet_ = new CLStateT();

  // init a simple controller
  this->initCtrler(SIMPLE_CTRLER);
  //this->controller_ = new CCController(store,name);

  // init a simple planner
  this->initPlanner(SIMPLE_TP);

  // update the system from the hugr store
  int resUp = this->updateSys();
  if(resUp != 0)
    {
      cerr << "Cycab::Cycab : Warning. A problem occured during the system update." << endl;
    }
}

// /**
//  *  \brief default constructor
//  *  \param name : the name of the robot
//  *  \param sys : the robotic system configuration of the robot
//  */
// Cycab::Cycab(std::string name,CarLikeSys* sys) : Robot(name,sys)
// {
//   this->statet_ = new CLStateT();
// }

/**
 *  \brief destructor
 */
Cycab::~Cycab()
{
}


/************** Getters and setters ***************/

/**
 *  \brief return the current control applied on the wheelchair
 *  \return the current control applied on the wheelchair
 */
const CCControlT* Cycab::controlt()
{ 
  if(MUTEX_DEBUG == 1) cout << "Cycab::controlt : waiting for mutex Ctrl" << endl;
  boost::mutex::scoped_lock lock(mutexCtrl_);
  if(MUTEX_DEBUG == 1) cout << "Cycab::controlt : mutex Ctrl locked" << endl;

  return this->controller()->controlt();

  if(MUTEX_DEBUG == 1) cout << "Cycab::controlt : mutex Traj released" << endl;
}

/**
 *  \brief return true if the robot is moving
 *  \return true if the robot is moving
 */
bool Cycab::isMoving()
{
  const CCControlT* ctrl = this->controlt();
  if(ctrl != NULL) { cout << "ctrl : "; ctrl->display(); cout << endl; }
  else cout << "ctrl NULL" << endl;
  return (ctrl != NULL && (ctrl->v() != 0 || ctrl->ph() != 0));
}

/**
 *  \brief stop the robot moving
 */
void Cycab::stopMoving()
{
  // check if the robot is moving
  if(this->isMoving())
    {     
      // stop the controller
      this->stopCtrler();

      // stop the robot motion
      this->initCtrler(SIMPLE_CTRLER);
      this->updateCtrl(CCControl(0,0));
      this->startCtrler();
    }
}


/*************** Hugr store reading ***************/

#ifdef HUGR
/**
 *  \brief update the system configuration from the HUGR store
 *  \return 0 if the update succeed, -1 else
 */
int Cycab::updateSys()
{
  cout << "1" << endl;

  if(store_ == NULL)
    {
      cerr << "Cycab::updateSys : Warning. The HUGR store hasn't been defined." << endl;
      return -1;
    }

  // get back the cycab constants
  // actually given in static variables from struct CycabConstants

//   std::string str(name_);
//   try
//     {
//       str += "_constants";
//       CC_constantsId = store_->lookupVariable(str.c_str());
//     }
//   catch(runtime_error e)
//     {
//       cerr << "Cycab::updateSys : Warning. " << str.c_str() << "cannot be found in the hugr store" << endl;
//       return -1;
//     }
  
//   CycabConstants CL_constants;
//   store_->readVariable(CL_constantsId, CL_constants, false);

  cout << "2" << endl;

  // update the system configuration of the cycab
  this->sys()->setLength(CycabConstants::L);
  this->sys()->setRadius(CycabConstants::robot_length*0.5);
  this->sys()->setVMax(CycabConstants::max_allowed_speed);
  this->sys()->setAMax(CycabConstants::max_allowed_speed);
  this->sys()->setMaxSteering(CycabConstants::max_allowed_steering);
  this->sys()->setMaxSteerSpd(CycabConstants::max_allowed_steering);

  cout << "3" << endl;

  // display the system configuration
  this->sys()->display();

  cout << "4" << endl;

  // connect the controller to the hugr store
  int resConnecCtrler = this->controller_->connect();
  if(resConnecCtrler != 0)
    {
      cerr << "Cycab::updateSys : Warning. The cycab cannot be controlled." << endl;
      return -1;
    }

  cout << "5" << endl;

  // connect the localizer to the hugr store
  // just check here if the pose is defined
  int resConnecLoczer = this->updatePose();
  if(resConnecLoczer != 0)
    {
      cerr << "Cycab::updateSys : Warning. The cycab cannot be localized." << endl;
      this->loczerStatus_ = LOCZER_DISCONNECTED;
      return -1;
    }
  this->loczerStatus_ = LOCZER_STOPPED;
  
  cout << "6" << endl;

  return 0;
}

/**
 *  \brief update the robot pose from the HUGR store
 *  \return 0 if the update succeed, -1 else
 */
int Cycab::updatePose()
{
  if(store_ == NULL)
    {
      cerr << "Cycab::updatePose : Warning. The HUGR store hasn't been defined." << endl;
      return -1;
    }
  
  // get back the cycab pose id
  std::string str(name_);
  try
    {
      str += "_simuState";
      CC_poseId = store_->lookupVariable(str.c_str());
    }
  catch(runtime_error e)
    {
      cerr << "Cycab::updatePose : Warning. " << str.c_str() << " cannot be found in the hugr store" << endl;
      return -1;
    }
  
  SimulatorState CC_pose;
  store_->readVariable(CC_poseId, CC_pose, false);

  // update the state of the system with the pose found
  this->statet()->x() = CC_pose.x;
  this->statet()->y() = CC_pose.y;
  this->statet()->th() = normalizedAngle(CC_pose.theta);
  this->statet()->t() = timer_->totalElapsedTime();

//   cout << "state : "; this->statet()->display(); cout << endl;

  // update the state registered in the controller
  if(controller_ != NULL)
    this->controller_->updateStateT(*(statet()));

  // update the state registered in the environment perception process
  if(env_ != NULL)
    this->env_->updatePose(PositionXYTh(CC_pose.x,CC_pose.y,normalizedAngle(CC_pose.theta)));

  return 0;
}


/************* Localization functions *************/

/*
 *  \brief compute the current deviation between the localization of the robot and the trajectory followed
 *  \param trajSt : the current state-time of the trajectory (set during the process)
 *  \param loc : the current localization state-time (set during the process)
 *  \param dev : the current deviation between the localization of the robot and the trajectory followed (set during the process)
 *  \param ctrlt : the current control applied on the robot
 *
 *  Every input parameters must be previously allocated
 *  Their value are set during the deviation computation
 */
void Cycab::computeCurDeviation(StateTime* trajSt,StateTime* loc,StateTime* dev,ControlTime* ctrlt)
{
  CLStateT* CLtrajSt = ((CLStateT*) trajSt);
  CLStateT* CLloc = ((CLStateT*) loc);
  CLStateT* CLdev = ((CLStateT*) dev);
  CCControlT* CCctrlt = ((CCControlT*) ctrlt);

  double time = this->statet()->t();

  CCctrlt->v() = this->controlt()->v();
  CCctrlt->ph() = this->controlt()->ph();
  CCctrlt->t() = this->controlt()->t();

  CLloc->x()  = this->statet()->x();
  CLloc->y()  = this->statet()->y();
  CLloc->th() = this->statet()->th();
  CLloc->ph() = this->statet()->ph();
  CLloc->s()  = this->statet()->s();
  CLloc->t()  = this->statet()->t();

  // get back the state of teh trajectry followed at the localization time
  if(this->traj() != NULL) 
    {
      this->traj()->getStateAtTime(time,*((CLStateT*)trajSt));

      CLdev->x()  = CLloc->x() - CLtrajSt->x();
      CLdev->y()  = CLloc->y() - CLtrajSt->y();
      CLdev->th() = mpc::computeMinDiffAngle(CLloc->th(),CLtrajSt->th());
      CLdev->ph() = CLloc->ph() - CLtrajSt->ph();
      CLdev->s()  = CLloc->s() - CLtrajSt->s();
      CLdev->t()  = CLloc->t() - CLtrajSt->t();
    }
  else
    {
      CLtrajSt->x() = 0;
      CLtrajSt->y() = 0;
      CLtrajSt->th() = 0;
      CLtrajSt->ph() = 0;
      CLtrajSt->s() = 0;
      CLtrajSt->t() = 0;

      CLdev->x() = 0;
      CLdev->y() = 0;
      CLdev->th() = 0;
      CLdev->ph() = 0;
      CLdev->s() = 0;
      CLdev->t() = 0;
    }
}


/*************** Control functions ****************/

/**
 *  \brief change the controller type
 *  \param type : the new controller type
 *  \return 0 if the initialization succeed
 *  \return -1 if the current controller is running
 *  \return -2 if such a controller has already been initialized
 *  \return -3 if the input controller type is wrong or unknown
 *  \return -4 if an unknown problem occured
 *  \return -5 if the connection to the store cannot be established
 *  \return -6 if a required input of the controller is missing
 */
int Cycab::initCtrler(controller_t type)
{
  // check if the current controller (if it exists) can be replaced
  if(this->controller_ != NULL)
    {
      if(this->controller_->ctrlerType() == type)
	{
	  // nothing t do : such a controller has been already initialized
	  return -2;
	}
      else switch(this->controller_->status())
	{
	case CTRLER_STARTED : 
	  // cannot change the controller : the current controller is running
	  return -1;
	case CTRLER_DISCONNECTED : 
	case CTRLER_STOPPED : 
	  // delete the current controller
	  delete this->controller_;
	  break;
	default : 
	  return -4;
	}
    }

  // initialize a new controller
  switch(type)
    {
    case SIMPLE_CTRLER : 
      this->controller_ = new CCController(this->store_,this->name_);
      break;
    case TRAJFOLLOW_CTRLER : 
      this->controller_ = new Hugr_CCTrajFollow(sys(),traj(),store_,name_,&mutexLoc_,&mutexCtrl_,&mutexTraj_,&mutexEnv_);
      break;
    case TEDDY_CTRLER : 
      if(traj() == NULL)
	{
	  cerr << "Cycab::initCtrler : traj() null" << endl;
	  this->controller_ = NULL;
	  return -6;
	}
      this->controller_ = new Hugr_CC_TFD(sys(),env(),traj(),store_,name_,&mutexLoc_,&mutexCtrl_,&mutexTraj_,&mutexEnv_);
      break;
    default : 
      cerr << "Cycab::initCtrler : Warning. Wrong or unknown controller type (" << type <<"). Such a controller cannot be defined." << endl;
      return -3;
    }

  // connect it to the hugr store
  int resConnec = this->controller_->connect();
  if(resConnec != 0)
    {
      cerr << "Cycab::initCtrler : Warning. The Connection of the controller to the store cannot be established." << endl;
      return -5;
    }
  return 0;
}
#endif


/*************** Planner functions ****************/

/**
 *  \brief change the planner type
 *  \param type : the new planner type
 *  \return 0 if the initialization succeed
 *  \return -2 if such a planner has already been initialized
 *  \return -3 if the input planner type is wrong or unknown
 *  \return -4 if a roadmap is required
 *  \return -5 if the environment has not been initialized
 *  \return -6 if an unknown problem occured
 */
int Cycab::initPlanner(trajPlanner_t type)
{
  // check if the current controller (if it exists) can be replaced
  if(this->planner_ != NULL)
    {
      if(this->planner_->type() == type)
	{
	  // nothing t do : such a planner has been already initialized
	  return -2;
	}
      else
	{
	  // delete the current planner
	  delete this->planner_;
	}
    }

  // initialize a new planner
  switch(type)
    {
    case SIMPLE_TP : 
      this->planner_ = new CLSimpleTrajPlanner(sys(),NULL);
      break;
//     case PVD_TP : 
//       this->planner_ = new 
//       break;
//     case ROADMAP_TP : 
//       this->planner_ = new 
//       break;
    default : 
      cerr << "Cycab::initPlanner : Warning. Wrong or unknown planner type. Such a planner cannot be defined." << endl;
      return -3;
    }
  return 0;
}

/**
 *  \brief reset the trajectory followed
 *  \param traj : the new trajectory
 */
void Cycab::resetTraj(const STrajectory* traj)
{
  // stop the motion of the robot
  this->stopMoving();
  
  // stop the controller
  this->stopCtrler();
  
  // reset the new trajectory followed
  //if(traj_ != NULL) delete traj_;                       ////////////////////////////////// NON??????????????? A TESTER
  traj_ = new CLSTraj(*((CLSTraj*)traj));
  
  // init the current time of the supposed reached waypoint of the trajectory
  //   this->curTimeOnTraj_ = 0;
}

/**
 *  \brief start the control process of the robot
 *  \param goalS : the goal state
 *  \param t : the final time at which the goal state must be reached
 *  \return 0 if the controller starting process succeed
 *  \return -1 if no trajectory has been found
 *  \return -2 if the planner hasn't been initialized
 *  \return -3 if the planner type is not handled
 *  \return -4 if a roadmap is required
 *  \return -5 if the environment has not been initialized
 *  \return -6 if the robot is still moving
 */
int Cycab::planTraj(const CLState& goalS,double t,double dt)
{
  // check if the planner has been initialized
  if(this->planner_ == NULL)
    {
      cerr << "Cycab::planTraj : The planner hasn't been initialized..." << endl;
      return -2;
    }

  // check if the the robot is still in motion
  if(this->isMoving())
    {
      cerr << "WheelChair::planTraj : The robot is still moving. Cannot plan a new path." << endl;
      return -6;
    }

  // plan the trajectory between the current state of the robot and the goal state
  int res;
  switch(this->planner_->type())
    {
    case SIMPLE_TP : 
      res = ((CLSimpleTrajPlanner*)planner_)->computeTraj(statet()->toState(),goalS,t,dt);
      break;
//     case PVD_TP : 
//       break;
//     case ROADMAP_TP : 
//       break;
    default : 
      cerr << "Cycab::planTraj : Wrong or unknown planner type." << endl;
      return -3;
    }

  // check if the trajectory has been correctly computed
  if(res == TRAJ_CPTD)
    {
      // set the computed trajectory as the current trajectory followed
      traj_ = new CLSTraj(*((CLSTraj*)(planner_->getTraj())));
//       // init the current time of the supposed reached waypoint of the trajectory
//       this->curTimeOnTraj_ = 0;
      return 0;
    }
  else
    {
      cerr << "Cycab::planTraj : No trajectory has been found" << endl;
      return -1;
    }
}


/************* Deformation functions **************/

/**
 *  \brief set the deformation parameters
 *  \param k_int : internal forces coefficient
 *  \param k_ext : external forces coefficient (dynamic obstacles)
 *  \param k_stext : external forces coefficient (static obstacles)
 *  \param dyn_med : maximal effective distance (dynamic obstacles)
 *  \param st_med : maximal effective distance (static obstacles)
 *  \param dyn_med : maximal temporal effective distance
 */
void Cycab::setTeddyParams(double k_int,double k_ext,double k_stext,double dyn_med,double st_med,double tmp_med)
{
  // check if a deformation process has been initialized
  if(this->ctrlerType() == TEDDY_CTRLER) ((Hugr_CC_TFD*)controller_)->setTeddyParams(k_int,k_ext,k_stext,dyn_med,st_med,tmp_med);
}


/*************** Display functions ****************/

/**
 *  \brief update the trajectory deformation process visualization
 *  \param dispFExt : true if the external forces (due to dynamic obstacles) must be displayed
 *  \param dispStFExt : true if the external forces (due to static obstacles) must be displayed
 *  \param dispNearDynObs : true if the position of the nearest dynamic obstacles must be displayed
 *  \param dispNearStObs : true if the position of the nearest static obstacles must be displayed
 */
void Cycab::updateTeddyView(bool dispFExt,bool dispStFExt,bool dispNearDynObs,bool dispNearStObs)
{
  // check if a deformation process has been initialized
  if(this->ctrlerType() == TEDDY_CTRLER) ((Hugr_CC_TFD*)controller_)->updateTeddyView(dispFExt,dispStFExt,dispNearDynObs,dispNearStObs);
}

/**
 *  \brief display the trajectory followed by the robot
 */
void Cycab::trajGL()
{
  // check if the trajectory has been defined
  if(traj_ != NULL)
    {
      // copy the current trajectory
      CLSTraj traj(*(this->traj()));

      // get back the current time of the robot
      double curT = this->statet()->t();

      // get back the time reached on the trajectory
      double trajT = this->curTimeOnTraj();

//       cout << "curT : " << curT << "  trajT : " << trajT << endl;

      // shift the trajectory to the current time
      traj.shift(curT-trajT);

//       traj.display();

      // display the trajectory in an openGL viewer
      traj.toGL();
    }
}
