#include "Robot.hpp"
using namespace mpc;

                                    /**************************************************/
                                    /****************** Class Robot *******************/
                                    /**************************************************/

/********** Constructors and destructors **********/

/**
 *  \brief default constructor
 *
 *  Only used by subclasses
 */
Robot::Robot()
{
  this->sys_           = NULL;
  this->statet_        = NULL;
//   this->ctrl_         = NULL;
  this->name_          = string("");
  this->planner_       = NULL;
  //this->traj_          = new STrajectory*;
  this->traj_          = NULL;
#ifdef HUGR
  this->store_         = NULL;
  this->controller_    = NULL;
  this->loczerStatus_  = LOCZER_DISCONNECTED;
  this->dtLoczer_      = 0.05;
  this->threadLoc_     = NULL;
  this->env_           = NULL;
#endif

  this->debug_         = 0;
}

#ifdef HUGR
/**
 *  \brief default constructor
 *  \param name : the name of the robot
 *  \param store : the hugr store containing informations about the robot
 */
Robot::Robot(std::string name,hugr::Store* store)
{
  this->sys_           = NULL;
  this->statet_        = NULL;
  this->name_          = name;
  this->planner_       = NULL;
  //this->traj_          = new STrajectory*;
  this->traj_          = NULL;
  this->store_         = store;
  this->controller_    = NULL;
  this->loczerStatus_  = LOCZER_DISCONNECTED;
  this->dtLoczer_      = 0.05;
  this->threadLoc_     = NULL;

  // initialize the timer of the robot
  timer_ = new Timer();
  timer_->start();

  // init the environment perception
  this->env_ = new HEnvironment(store_,timer_,&mutexEnv_);
  this->initEnvPercept();

  this->debug_         = 0;
}
#endif

// /**
//  *  \brief default constructor
//  *  \param name : the name of the robot
//  *  \param sys : the robotic system configuration of the robot
//  */
// Robot::Robot(std::string name,System* sys)
// {
//   this->sys_           = sys;
//   this->statet_        = NULL;
//   this->name_          = name;
//   this->planner_       = NULL;
//   //this->traj_          = new STrajectory*;
//   this->traj_          = NULL;
// #ifdef HUGR
//   this->store_         = NULL;
//   this->controller_    = NULL;
//   this->loczerStatus_  = LOCZER_DISCONNECTED;
//   this->dtLoczer_      = 0.05;
//   this->threadLoc_     = NULL;
//   this->env_           = NULL;
// #endif

//   // initialize the timer of the robot
//   timer_ = new Timer();
//   timer_->start();

//   this->debug_         = 0;
// }

/**
 *  \brief destructor
 */
Robot::~Robot()
{
  if(threadLoc_ != NULL) delete threadLoc_;
  if(timer_ != NULL) delete timer_;
}


/************** Getters and setters ***************/

/**
 *  \brief return the current time of the supposed reached waypoint of the trajectory
 *  \return the current time of the supposed reached waypoint of the trajectory
 */
double Robot::curTimeOnTraj() const
{
  if(controller_ != NULL) return controller_->curTimeOnTraj();
  else return 0;
}

#ifdef HUGR

/**
 *  \brief return the controller status
 *  \return the controller status
 */
controllerStatus_t Robot::ctrlerStatus() const
{
  if(this->controller_ == NULL) return CTRLER_NULL;
  else return this->controller_->status();
}

/**
 *  \brief return the localizer status
 *  \return the localizer status
 */
localizerStatus_t Robot::loczerStatus() const
{
  return this->loczerStatus_;
}

/**
 *  \brief return the localizer status
 *  \return the localizer status
 */
envUpStatus_t Robot::envpStatus() const
{
  if(this->env_ == NULL) return ENVUP_NULL;
  else return this->env_->status();
}

// /**
//  *  \brief return the deformer status
//  *  \return the deformer status
//  */
// deformerStatus_t Robot::deformerStatus() const
// {
//   if(this->env_ == NULL) return DEFORMER_NULL;
//   else return this->deformer_->status();
// }

/**
 *  \brief set the controller debug level
 *  \param dbg : the new controller debug level
 */
void Robot::setCtrlerDebugLevel(int dbg)
{
  if(controller_ != NULL) controller_->setDebug(dbg);
}

/**
 *  \brief return the type of controller used
 *  \return the type of controller used
 */
controller_t Robot::ctrlerType() const
{
  if(controller_ == NULL) return ND_CTRLER;
  else return controller_->ctrlerType();
}

/**
 *  \brief return the type of controller used
 *  \return the type of controller used
 */
trajPlanner_t Robot::plannerType() const
{
  if(planner_ == NULL) return ND_TP;
  else return planner_->type();
}

/**
 *  \brief set the debugging level
 *  \param dbg : the new debugging level
 */
void Robot::setDebug(int dbg)
{
  // update the local debug level
  this->debug_ = dbg;

  // update the controller debug level
  if(controller_ != NULL) 
    {
      controller_->setDebug(dbg);
      if(dbg == 0) this->updateTeddyView(false,false,false,false);
      else // if(dbg == 1)
	this->updateTeddyView(true,true,true,true);
    }
}

/**
 *  \brief return the roadmap built over the static environment
 *  \return the roadmap built over the static environment
 */
DRoadMap* Robot::roadmap() const
{
  // check if the environment perception has been initialized
  if(this->env_ == NULL) return NULL;
  else return env_->roadmap();
}

/**
 *  \brief set the autobuilding roadmap status
 *  \param the new autobuilding roadmap status
 */
void Robot::setAutobuildRM(bool status)
{
  // check if the environment perception has been initialized
  if(this->env_ != NULL) env_->setAutobuildRM(status);
}

/**
 *  \brief set the roadmap drawing status
 *  \param the new roadmap drawing status
 */
void Robot::setDispRM(bool status)
{
  // check if the environment perception has been initialized
  if(this->env_ != NULL) env_->setDispRM(status);
}


/************* Localization functions *************/

/**
 *  \brief execute the localization loop
 */
void Robot::execLoc()
{
  while(this->loczerStatus_ == LOCZER_STARTED)
    {
      int resUp = this->updatePose();
      if(resUp == 0) usleep((int)(1000000.0*(dtLoczer_)));
      else this->loczerStatus_ = LOCZER_STOPPED;
    }
}

/**
 *  \brief start the localization process of the robot
 *  \return 0 if the localization starting process succeed, -1 else
 */
int Robot::startLoc()
{
  switch(loczerStatus_)
    {
    case LOCZER_NULL : 
    case LOCZER_DISCONNECTED : 
      {
	cerr << "Robot::startLoc : Warning. The localizer is disconnected." << endl;
	return -1;
      }
    case LOCZER_STOPPED : 
      {
	loczerStatus_ = LOCZER_STARTED;
	threadLoc_ = new boost::thread(boost::bind(&Robot::execLoc,this));
	return 0;
      }
    case LOCZER_STARTED : 
      {
	// nothing to do : the localizer is already started
	return 0;
      }
    default : 
      cout << "Robot::startLoc : " << endl;
      exit(1);
    }
}

/**
 *  \brief stop the localization process of the robot
 *  \return 0 if the localization stoping process succeed, -1 else
 */
int Robot::stopLoc()
{
  switch(loczerStatus_)
    {
    case LOCZER_NULL : 
    case LOCZER_DISCONNECTED : 
      {
	cerr << "Robot::stopLoc : Warning. The localizer is disconnected." << endl;
	return -1;
      }
    case LOCZER_STOPPED : 
      {
	// nothing to do
	return 0;
      }
    case LOCZER_STARTED : 
      {
	this->loczerStatus_ = LOCZER_STOPPED;
	if(this->threadLoc_ != NULL) delete this->threadLoc_;
	return 0;
      }
    default : 
      cout << "Robot::stopLoc : " << endl;
      exit(1);
    }
}


/*************** Control functions ****************/

/**
 *  \brief start the control process of the robot
 *  \return 0 if the controller starting process succeed, -1 else
 */
int Robot::startCtrler()
{
  if(this->controller_ == NULL) return -1;
  else return this->controller_->start();
}

/**
 *  \brief stop the control process of the robot
 *  \return 0 if the controller stoping process succeed, -1 else
 */
int Robot::stopCtrler()
{
  if(this->controller_ == NULL) return -1;
  else return this->controller_->stop();
}

/**
 *  \brief update the input control
 *  \param ctrl : the new input control
 *  \return 0 if the update process succeed
 */
int Robot::updateCtrl(const Control& c)
{
  if(this->controller_ == NULL) return -1;
  else return this->controller_->updateCtrl(c);
}

/**
 *  \brief reset the controller
 */
void Robot::resetCtrler()
{
  if(this->controller_ != NULL) controller_->reinit();
}

#endif


/******** Environment perception functions ********/

/**
 *  \brief change the environment perception process type
 *  \return 0 if the initialization succeed
 *  \return -2 if such an environment perception process has already been initialized
 *  \return -3 if the input environment perception process type is wrong or unknown
 *  \return -4 if an unknown problem occured
 *  \return -5 if the connection to the hugr store cannot be established
 */
int Robot::initEnvPercept()
{
//   // reinit the environment if needed
//   if(env_ != NULL) delete this->env_;
//   env_ = new HEnvironment(store_,timer_,&mutexEnv_);

  // connect the environment perception process to the hugr store
  int resConnec = env_->connect();
  if(resConnec != 0)
    {
      cerr << "Robot::initEnvPercept : Warning. The connection to the hugr store cannot be established." << endl;      
      return -5;
    }
  return 0;
}

/**
 *  \brief init the detection and tracking of moving objects process
 *  \param distFilter : maximal distance from static obstacles to consider the dynamic obstacles perceived as noise
 *  \param securDist : minimal distance kept to dynamic obstacles
 *  \return 0 if the initialization succeed
 *  \return -1 if the DATMO has been already started
 *  \return -2 if the environment perception tool has not been defined
 *  \return -3 if an unknown error occured
 */
int Robot::initDATMO(double distFilter,double securDist)
{
  if(env_ != NULL) 
    {
      return env_->initDATMO(distFilter,securDist);
    }
  else return -2;
}

/**
 *  \brief start the robot's environment perception process
 */
void Robot::startEnvPercept()
{
  if(env_ != NULL) 
    {
      // check if the environment perception process is already connected
      if(env_->status() == ENVUP_DISCONNECTED) this->initEnvPercept();

      env_->start();
    }
}

/**
 *  \brief stop the robot's environment perception process
 */
void Robot::stopEnvPercept()
{
  if(env_ != NULL) env_->stop();
}

/**
 *  \brief provide knowledge of the static environment to the robot
 *  \param filename : filename of the input static environment
 *  \param RMfilename : filename of an input roadmap alredy built on the environment
 *  \return 0 if the process succeed
 *          -1 if the file has not been found
 *          -2 if the localization has not been started
 *          -3 if the environment perception has not been initialized
 *          -4 if an unknown error occured
 */
int Robot::addStaticEnv(const char* filename,const char* RMfilename)
{
  // check if the localization has been started
  if(this->loczerStatus() != LOCZER_STARTED) return -2;

  // check if the environment perception has been initialized
  if(this->env_ == NULL) return -3;

  // add the static environment to the perception of the robot
  return env_->addStaticEnv(filename, 1.2*sys()->getRadius(),RMfilename);
}

/**
 *  \brief clear the static environment
 *  \return 0 if the process succeed
 *          -3 if the environment perception has not been initialized
 *          -4 if an unknown error occured
 */
int Robot::clearStaticEnv()
{
  // check if the environment perception has been initialized
  if(this->env_ == NULL) return -3;

  // clear the a prioiri defined static environment
  return env_->clearStaticEnv();
}

/**
 *  \brief launch the roadmap building on the static environment
 *  \return 0 if the process succeed 
 *         -1 if the environment perception has not been initialized
 */
int Robot::buildRoadmap()
{
  // check if the environment perception has been initialized
  if(this->env_ == NULL) return -1;

  // build the roadmap
  return env_->buildRoadmap();
}

/**
 *  \brief load a new roadmap on the environment
 *  \param filename : the input filename
 *  \return 0 if the process succeed, -1 else
 */
int Robot::loadRoadmap(const char* filename)
{
  // check if the environment perception has been initialized
  if(this->env_ == NULL) return -1;

  // load the roadmap
  return env_->loadRoadmap(filename);
}

/**
 *  \brief save the current roadmap
 *  \param filename : the output filename
 *  \return 0 if the process succeed, -1 else
 */
int Robot::saveRoadmap(const char* filename)
{
  // check if the environment perception has been initialized
  if(this->env_ == NULL) return -1;

  // save the roadmap
  return env_->saveRoadmap(filename);
}


/************* Deformation functions **************/

/**
 *  \brief start the deformation process of the trajectory of the robot
 *  \return 0 if the deformation starting process succeed, -1 else
 */
int Robot::startDeformer()
{
  cout << "Robot::startDeformer : ?????????????????" << endl;
  exit(1);
}

/**
 *  \brief stop the deformation process of the trajectory of the robot
 *  \return 0 if the deformation stoping process succeed, -1 else
 */
int Robot::stopDeformer()
{
  cout << "Robot::stopDeformer : ?????????????????" << endl;
  exit(1);
}


/*************** Display functions ****************/

/**
 *  \brief draw the robotic system in an openGL viewer
 */
void Robot::toGL() const
{
  this->sys()->toGL();
}

/**
 *  \brief display the environment perception of the robot
 */
void Robot::envGL()
{
//   cout << "Robot::envGL" << endl;
  if(MUTEX_DEBUG == 1) cout << "Robot::envGL : waiting for mutex Env" << endl;
  boost::mutex::scoped_lock lock(mutexEnv_);
  if(MUTEX_DEBUG == 1) cout << "Robot::envGL : mutex Env locked" << endl;

  if(env_ != NULL) 
    {
      // display the environment
      env_->toGL(true);
    }

  if(MUTEX_DEBUG == 1) cout << "Robot::envGL : mutex Env released" << endl;
}

/**
 *  \brief display the controller specifications of the robot
 */
void Robot::ctrlerGL()
{
  if(MUTEX_DEBUG == 1) cout << "Robot::ctrlerGL : waiting for mutex Ctrl" << endl;
  boost::mutex::scoped_lock lock(mutexCtrl_);
  if(MUTEX_DEBUG == 1) cout << "Robot::ctrlerGL : mutex Ctrl locked" << endl;

//   cout << "this->ctrlerType() : " << this->ctrlerType() << endl;

  // check if the controller has been already started
  if(ctrlerStatus() != CTRLER_STARTED)
    {
      if(MUTEX_DEBUG == 1) cout << "Robot::ctrlerGL : mutex Ctrl released2" << endl;
      return;
    }

  switch(this->ctrlerType())
    {
    case ND_CTRLER : 
      break;
    case SIMPLE_CTRLER : 
      break;
    case TRAJFOLLOW_CTRLER : 
//       // draw the state the robot try to reach
//       ((Hugr_CCTrajFollow*)controller_)->lastStateToReach().toGL();
//       cout << "state to reach : "; ((Hugr_CCTrajFollow*)controller_)->lastStateToReach().display(); cout << endl;
//       // draw the trajectory generated
//       ((Hugr_CCTrajFollow*)controller_)->displayTrajGen();
      break;
    case TEDDY_CTRLER : 
      controller_->toGL();
      break;
    }

  if(MUTEX_DEBUG == 1) cout << "Robot::ctrlerGL : mutex Ctrl released" << endl;
}
